#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

#include "Comparison.h"


Comparison::Comparison()
{
}


Comparison::Comparison(const Comparison &copy_me)
{
	operand1 = copy_me.operand1;
	whichAtt1 = copy_me.whichAtt1;

	operand2 = copy_me.operand2;
	whichAtt2 = copy_me.whichAtt2;

	attType = copy_me.attType;

	op = copy_me.op;
}


void Comparison :: Print () {

	cout << "Att " << whichAtt1 << " from ";

	if (operand1 == Left)
		cout << "left record ";
	else if (operand1 == Right)
		cout << "right record ";
	else 
		cout << "literal record ";


	if (op == LessThan)
		cout << "< ";
	else if (op == GreaterThan)
		cout << "> ";
	else
		cout << "= ";


	cout << "Att " << whichAtt2 << " from ";

	if (operand2 == Left)
		cout << "left record ";
	else if (operand2 == Right)
		cout << "right record ";
	else
		cout << "literal record ";
 

	if (attType == Int)
		cout << "(Int)";
	else if (attType == Double)
		cout << "(Double)";
	else
		cout << "(String)";
}




OrderMaker :: OrderMaker() {
	numAtts = 0;
}

OrderMaker :: OrderMaker(Schema *schema) {
	numAtts = 0;

	int n = schema->GetNumAtts();
	Attribute *atts = schema->GetAtts();

	//first add the Integer attributes
	for (int i = 0; i < n; i++) {
		if (atts[i].myType == Int) {
			whichAtts[numAtts] = i;
			whichTypes[numAtts] = Int;
			numAtts++;
		}
	}

	// now add in the doubles
	for (int i = 0; i < n; i++) {
                if (atts[i].myType == Double) {
                        whichAtts[numAtts] = i;
                        whichTypes[numAtts] = Double;
                        numAtts++;
                }
        }

	// and finally the strings
        for (int i = 0; i < n; i++) {
                if (atts[i].myType == String) {
                        whichAtts[numAtts] = i;
                        whichTypes[numAtts] = String;
                        numAtts++;
                }
        }
}


OrderMaker :: OrderMaker(Schema *schema, int *join_atts, int join_atts_no) {
	numAtts = join_atts_no;

	Attribute *atts = schema->GetAtts();

	//only the specified attributes are inserted into the sorting order
	for (int i = 0; i < numAtts; i++) {
		whichAtts[i] = join_atts[i];
		whichTypes[i] = atts[join_atts[i]].myType;
	}
}


OrderMaker::OrderMaker(const OrderMaker &om1) {
	numAtts = om1.numAtts;

	for (int i = 0; i < om1.numAtts; i++)
	{
		whichAtts[i] = om1.whichAtts[i];
		whichTypes[i] = om1.whichTypes[i];
	}
}


void OrderMaker::ANDMerge(OrderMaker *om1, OrderMaker *om2) {
	numAtts = om1->numAtts;

	//copy the attributes from om1
	for (int i = 0; i < om1->numAtts; i++)
	{
		whichAtts[i] = om1->whichAtts[i];
		whichTypes[i] = om1->whichTypes[i];
	}

	//copy the attributes from om2 that do not appear in om1
	for (int i = 0; i < om2->numAtts; i++)
	{
		int found = 0;
		for (int j = 0; j < numAtts; j++)
		{
			if (whichAtts[j] == om2->whichAtts[i])
			{
				//attribute already appears in om1; discard it
				found = 1;
				break;
			}
		}

		//attribute is not appearing in om1
		if (found == 0)
		{
			whichAtts[numAtts] = om2->whichAtts[i];
			whichTypes[numAtts] = om2->whichTypes[i];
			numAtts += 1;
		}
	}
}


void OrderMaker :: Print () {
	printf("NumAtts = %5d\n", numAtts);
	for (int i = 0; i < numAtts; i++)
	{
		printf("%3d: %5d ", i, whichAtts[i]);
		if (whichTypes[i] == Int)
			printf("Int\n");
		else if (whichTypes[i] == Double)
			printf("Double\n");
		else
			printf("String\n");
	}
}



int CNF :: GetSortOrders (OrderMaker &left, OrderMaker &right) {

	// initialize the size of the OrderMakers
	left.numAtts = 0;
	right.numAtts = 0;

	// loop through all of the disjunctions in the CNF and find those
	// that are acceptable for use in a sort ordering
	for (int i = 0; i < numAnds; i++) {
		
		// if we don't have a disjunction of length one, then it
		// can't be acceptable for use with a sort ordering
		if (orLens[i] != 1) {
			continue;
		}

		// made it this far, so first verify that it is an equality check
		if (orList[i][0].op != Equals) {
			continue;
		}

		// now verify that it operates over atts from both tables
		if (!((orList[i][0].operand1 == Left && orList[i][0].operand2 == Right) ||
		      (orList[i][0].operand2 == Left && orList[i][0].operand1 == Right))) {
			continue;		
		}

		// since we are here, we have found a join attribute!!!
		// so all we need to do is add the new comparison info into the
		// relevant structures
		if (orList[i][0].operand1 == Left) {
			left.whichAtts[left.numAtts] = orList[i][0].whichAtt1;
			left.whichTypes[left.numAtts] = orList[i][0].attType;	
		}	

		if (orList[i][0].operand1 == Right) {
                        right.whichAtts[right.numAtts] = orList[i][0].whichAtt1;
                        right.whichTypes[right.numAtts] = orList[i][0].attType;
                }

		if (orList[i][0].operand2 == Left) {
                        left.whichAtts[left.numAtts] = orList[i][0].whichAtt2;
                        left.whichTypes[left.numAtts] = orList[i][0].attType;
                }

		if (orList[i][0].operand2 == Right) {
                        right.whichAtts[right.numAtts] = orList[i][0].whichAtt2;
                        right.whichTypes[right.numAtts] = orList[i][0].attType;
                }

		// note that we have found two new attributes
		left.numAtts++;
		right.numAtts++;
	}

	return left.numAtts;
}


int CNF :: EliminateEquiJoinCond (CNF *from_me) {
	//practically the number of orLists will decrease with the number of equi-join conditions
	numAnds = 0;

	// loop through all of the disjunctions in the CNF and find those
	// that are acceptable for use in a sort ordering
	for (int i = 0; i < from_me->numAnds; i++) {
		
		// if any of the non-acceptable conditions for GetSortOrders is satisfied, then we have to keep it
		if ((from_me->orLens[i] != 1) || (from_me->orList[i][0].op != Equals) || (!((from_me->orList[i][0].operand1 == Left && from_me->orList[i][0].operand2 == Right) || (from_me->orList[i][0].operand2 == Left && from_me->orList[i][0].operand1 == Right)))) {
			for (int j = 0; j < from_me->orLens[i]; j++)
				orList[numAnds][j] = from_me->orList[i][j];
			orLens[numAnds] = from_me->orLens[i];
			
			numAnds += 1;
		}
	}

	return numAnds;
}


void CNF :: AddCNFAtts(int *read_atts, int &read_atts_no) {
	for (int i = 0; i < numAnds; i++) {
		for (int j = 0; j < orLens[i]; j++) {
			if (orList[i][j].operand1 == Left)
			{
				int att_ind = orList[i][j].whichAtt1;
				int already_in = 0;
				for (int k = 0; k < read_atts_no; k++)
				{
					//check if the attribute is already in (output schema)
					if (read_atts[k] == att_ind)
					{
						orList[i][j].whichAtt1 = k;
						already_in = 1;
						break;
					}
				}

				if (already_in == 0)
				{
					//attribute not in, so add it
					//but before, modify the CNF to reflect the change
					orList[i][j].whichAtt1 = read_atts_no;

					read_atts[read_atts_no] = att_ind;
					read_atts_no += 1;
				}
			}

			if (orList[i][j].operand2 == Left)
			{
				int att_ind = orList[i][j].whichAtt2;
				int already_in = 0;
				for (int k = 0; k < read_atts_no; k++)
				{
					//check if the attribute is already in (output schema)
					if (read_atts[k] == att_ind)
					{
						orList[i][j].whichAtt2 = k;
						already_in = 1;
						break;
					}
				}

				if (already_in == 0)
				{
					//attribute not in, so add it
					//but before, modify the CNF to reflect the change
					orList[i][j].whichAtt2 = read_atts_no;

					read_atts[read_atts_no] = att_ind;
					read_atts_no += 1;
				}
			}
		}
	}
}


void CNF :: Print () {

	for (int i = 0; i < numAnds; i++) {
		
		cout << "( ";
		for (int j = 0; j < orLens[i]; j++) {
			orList[i][j].Print ();
			if (j < orLens[i] - 1)
				cout << " OR ";
		}
		cout << ") ";
		if (i < numAnds - 1)
			cout << " AND\n";
		else
			cout << "\n";
	}
}
