#include "DEFS.h"
#include "Pipe.h"
#include "Compiler.h"

#include <iostream>
#include <string>
#include <set>
#include <vector>
#include <map>
#include <algorithm>





using namespace std;


//this is the head of the abstrac syntax tree
extern struct OperationList *finalope;



OperationItem::OperationItem(char *schema_file)
{
	flag=0;
	mylevel=-1;
	type=-1;

	InputPipe1=-1;
	InputPipe2=-1;
	OutputPipe=-1;

	AttstoProj=NULL;
	NumAttstoProj=0;

	CurrentLevelAtts.clear();
	InMemoryJoinBufferSize = 0;

	this->schema_file = strdup(schema_file);
	InputSchema1 = InputSchema2 = OutputSchema = NULL;

	parent=child1=child2=NULL;
}


OperationItem::~OperationItem()
{
	flag=0;
	mylevel=-1;
	type=-1;

	InputPipe1=-1;
	InputPipe2=-1;
	OutputPipe=-1;

	delete [] AttstoProj;
	AttstoProj=NULL;
	NumAttstoProj=0;

	CurrentLevelAtts.clear();
	InMemoryJoinBufferSize = 0;

	free(schema_file);
	schema_file = NULL;

	delete InputSchema1;
	delete InputSchema2;
	delete OutputSchema;
	InputSchema1 = InputSchema2 = OutputSchema = NULL;

	parent=child1=child2=NULL;
}


void OperationItem::CheckCircle()
{
	if(flag!=0)
	{
		cerr << "Find a circle." << "\n";
		exit(1);
	}

	++flag;

	if(child1!=NULL)
		child1->CheckCircle();

	if(child2!=NULL)
		child2->CheckCircle();
}


int OperationItem::Count()
{
	if((child1==NULL) && (child2==NULL))
		return 1;

	if(IsUnary())
	{
		return child1->Count()+1;
	}
	else
	{
		return child1->Count()+child2->Count()+1;
	}
}


void OperationItem::SetLevel(int i)
{
	if(child1!=NULL)
		child1->SetLevel(i+1);

	if(child2!=NULL)
		child2->SetLevel(i+1);

	mylevel=i;
}




TableScanOperation::TableScanOperation(char *schema_file)
: OperationItem(schema_file)
{
	CurrentAndList = NULL;

	myCNF = NULL;

	RecLit = NULL;
	RecLit_Schema = NULL;
}


TableScanOperation::~TableScanOperation()
{
	CurrentAndList = NULL;

	delete myCNF;
	myCNF = NULL;

	delete RecLit;
	RecLit = NULL;

	delete RecLit_Schema;
	RecLit_Schema = NULL;
}


bool TableScanOperation::IsUnary()
{
	return true;
}



void TableScanOperation :: Initialization(TableScanYac *CurrentOperation)
{
	type=TYPETABLESCAN;

	AttrRenameList *RenameList = CurrentOperation->RenameList;

	CurrentAndList= CurrentOperation->Condition;

	InMemoryJoinBufferSize= CurrentOperation->InMemoryJoinBufferSize;

	InputSchema1 = new (std::nothrow) Schema(schema_file, CurrentOperation->RelationName);
	if (InputSchema1 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	//rename everything needs renamed
	AttrRenameList* current=RenameList;
	for(;current != NULL; current = current->NextItem)
		InputSchema1->RenameAtt(current->CurrentItem->OldName,current->CurrentItem->NewName);

	OutputPipe=CurrentOperation->OutPutPipe;

	struct AndList *parseTree=CurrentAndList;

	//clear current atts;
	CurrentLevelAtts.clear();
	for (; parseTree!=NULL; parseTree = parseTree->rightAnd)
	{
		OrList *myOr = parseTree->left;
		for (; myOr != NULL; myOr = myOr->rightOr)
		{
			if (myOr->left->left->code == NAME)
			{
				CurrentLevelAtts.insert(myOr->left->left->value);
			}

			if (myOr->left->right->code == NAME)
			{
				CurrentLevelAtts.insert(myOr->left->right->value);
			}
		}
	}

}



void TableScanOperation :: BuildSchema(set<string> AttsFromAbove)
{
	set<string> tempAtts(AttsFromAbove.begin(), AttsFromAbove.end());

	vector<int> tempAttstoProj;

	int numAtts = InputSchema1->GetNumAtts();
	Attribute* AttList = InputSchema1->GetAtts();

	//for every attribute in the list, search whether this is in 
	//the set of attributes that are required to keep;
	// if it is, push the index to the vector tempAttstoProj 
	for(int i = 0; i < numAtts; i++)
	{
		if (tempAtts.find(AttList[i].name) != tempAtts.end())
			tempAttstoProj.push_back(i);
	}

	NumAttstoProj = tempAttstoProj.size();
	AttstoProj = new (std::nothrow) int[NumAttstoProj];
	if (AttstoProj == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	
	for(int i = 0; i < NumAttstoProj; i++)
	{
		AttstoProj[i] = tempAttstoProj[i];
	}

	OutputSchema = new (std::nothrow) Schema();
	if (OutputSchema == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	OutputSchema->CopyFrom(InputSchema1);
	OutputSchema->Project(AttstoProj, NumAttstoProj);

	myCNF = new (std::nothrow) CNF();
	if (myCNF == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	RecLit = new (std::nothrow) Record();
	if (RecLit == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	RecLit->GrowFromParseTree(CurrentAndList, InputSchema1, *myCNF, RecLit_Schema);
}



void TableScanOperation :: PrintParameters(int recursive)
{
	cout << "Begin" << "\n";
	cout << "Operation Type: TableScan" << "\n";
	cout << "InputFile: " << InputSchema1->GetName() << "\n";
	cout << "Output Pipe: " << OutputPipe << "\n";

	cout << "Input Schema:" << "\n";
	InputSchema1->Print();

	cout << "Output Schema:" << "\n";
	OutputSchema->Print();

	cout << "CNF:" <<"\n";
	myCNF->Print();

	cout << "Literal Record:" <<"\n";
	RecLit->Print(RecLit_Schema);

	cout << "InMemoryJoin Buffer Size: " << InMemoryJoinBufferSize << "\n";

	cout << "End" << "\n" << "\n";
}





SortJoinOperation::SortJoinOperation(char *schema_file)
: OperationItem(schema_file)
{
	MemorySize = 0;

	CurrentAndList = NULL;

	myCNF = NULL;

	RecLit = NULL;
	RecLit_Schema = NULL;
}


SortJoinOperation::~SortJoinOperation()
{
	MemorySize = 0;

	CurrentAndList = NULL;

	delete myCNF;
	myCNF = NULL;

	delete RecLit;
	RecLit = NULL;

	delete RecLit_Schema;
	RecLit_Schema = NULL;
}


bool SortJoinOperation::IsUnary()
{
	return false;
}



void SortJoinOperation :: Initialization(SortJoinYac *CurrentOperation)
{
	type = TYPESORTJOIN;

	InputPipe1 = CurrentOperation->InPutPipe1;
	InputPipe2 = CurrentOperation->InPutPipe2;
	OutputPipe = CurrentOperation->OutPutPipe;
	
	InMemoryJoinBufferSize= CurrentOperation->InMemoryJoinBufferSize;
	MemorySize = CurrentOperation->MemorySize;
	CurrentAndList = CurrentOperation->Condition;
	AndList *parseTree = CurrentAndList;

	CurrentLevelAtts.clear();
	for (; parseTree != NULL; parseTree = parseTree->rightAnd)
	{
		OrList *myOr = parseTree->left;
		for (; myOr != NULL; myOr = myOr->rightOr)
		{
			if (myOr->left->left->code == NAME)
			{
				CurrentLevelAtts.insert(myOr->left->left->value);
			}

			if (myOr->left->right->code == NAME)
			{
				CurrentLevelAtts.insert(myOr->left->right->value);	
			}
		}
	}
}



void SortJoinOperation :: BuildSchema(set<string> AttsFromAbove)
{
	set<string> tempAtts;
	set_union(CurrentLevelAtts.begin(), CurrentLevelAtts.end(), AttsFromAbove.begin(), AttsFromAbove.end(), inserter(tempAtts, tempAtts.begin()));

	child1->BuildSchema(tempAtts);
	child2->BuildSchema(tempAtts);

	InputSchema1 = new (std::nothrow) Schema();
	if (InputSchema1 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	InputSchema1->CopyFrom(child1->OutputSchema);

	InputSchema2 = new (std::nothrow) Schema();
	if (InputSchema2 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	InputSchema2->CopyFrom(child2->OutputSchema);

	OutputSchema = new (std::nothrow) Schema();
	if (OutputSchema == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	OutputSchema->CopyFrom(InputSchema1, InputSchema2);

	int numAtts = OutputSchema->GetNumAtts();
	Attribute* AttList = OutputSchema->GetAtts();

	vector<int> tempAttstoProj;
	for(int i = 0; i < numAtts; i++)
	{
		if (AttsFromAbove.find(AttList[i].name) != AttsFromAbove.end())
			tempAttstoProj.push_back(i);
	}

	NumAttstoProj = tempAttstoProj.size();
	AttstoProj = new (std::nothrow) int[NumAttstoProj];
	if (AttstoProj == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for(int i = 0; i < NumAttstoProj; i++)
	{
		AttstoProj[i] = tempAttstoProj[i];
	}
	
	OutputSchema->Project(AttstoProj, NumAttstoProj);

	myCNF = new (std::nothrow) CNF();
	if (myCNF == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	RecLit = new (std::nothrow) Record();
	if (RecLit == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	RecLit->GrowFromParseTree(CurrentAndList, InputSchema1, InputSchema2, *myCNF, RecLit_Schema);
}


void SortJoinOperation :: PrintParameters(int recursive)
{
	cout << "Begin" << "\n";
	cout << "Operation Type: SortJoin" << "\n";

	cout << "LeftInputPipe: " << InputPipe1 << " RightInputPipe: " << InputPipe2 << " OutputPipe: " << OutputPipe << "\n";

	cout << "LeftInputSchema:" << "\n";
	InputSchema1->Print();

	cout << "RightInputSchema:" << "\n";
	InputSchema2->Print();

	cout << "Output Schema:" << "\n";
	OutputSchema->Print();

	cout << "CNF:" << "\n";
	myCNF->Print();

	cout << "Literal Record:" <<"\n";
	RecLit->Print(RecLit_Schema);

	cout << "Buffer Size: " << MemorySize << "\n";

	cout << "InMemoryJoin Buffer Size: " << InMemoryJoinBufferSize << "\n";

	cout << "End" << "\n" << "\n";
	

	if (recursive == 1)
	{
		cout << "My Left Child:" << "\n";
		child1->PrintParameters(recursive);
	
		cout << "My Right Child:" << "\n";
		child2->PrintParameters(recursive);
	}
}




RehashingOperation :: RehashingOperation(char *schema_file)
: OperationItem(schema_file)
{
	MemorySize = 0;

}


RehashingOperation :: ~RehashingOperation()
{
	MemorySize = 0;

}


bool RehashingOperation :: IsUnary()
{
	return true;
}

void RehashingOperation :: Initialization(RehashingYac *CurrentOperation)
{
	type = TYPEREHASHING;

	InputPipe1 = CurrentOperation->InPutPipe;
	OutputPipe = CurrentOperation->OutPutPipe;
	MemorySize = CurrentOperation->MemorySize;
	InMemoryJoinBufferSize= CurrentOperation->InMemoryJoinBufferSize;

}


void RehashingOperation :: BuildSchema(set<string> AttsFromAbove)
{
	
	child1->BuildSchema(AttsFromAbove);
	
	InputSchema1 = new (std::nothrow) Schema();
	if (InputSchema1 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	InputSchema1->CopyFrom(child1->OutputSchema);

	OutputSchema = new (std::nothrow) Schema();
	if (OutputSchema == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	OutputSchema->CopyFrom(InputSchema1);
}

void RehashingOperation :: PrintParameters(int recursive)
{
	cout << "Begin" << "\n";

	cout << "Operation Type: Rehashing" << "\n";
	cout << "InputPipe: " << InputPipe1 << " OutputPipe: " << OutputPipe << "\n";

	cout << "InputSchema/OutputSchema:" << "\n";
	InputSchema1->Print();

	cout << "Buffer Size: " << MemorySize << "\n";

	cout << "InMemoryJoin Buffer Size: " << InMemoryJoinBufferSize << "\n";

	cout << "End" << "\n" << "\n";

	if (recursive == 1)
	{
		cout << "My Child:" << "\n";
		child1->PrintParameters(recursive);
	}
}


AggregateOperation::AggregateOperation(char *schema_file)
: OperationItem(schema_file)
{
	CurrentOperator = NULL;
	myFunc = NULL;

	JoinAttIndex = NULL;
	numJoinAtts = 0;

	JoinAtts.clear();
}


AggregateOperation::~AggregateOperation()
{
	CurrentOperator = NULL;

	delete myFunc;
	myFunc = NULL;

	delete [] JoinAttIndex;
	JoinAttIndex = NULL;
}


bool AggregateOperation::IsUnary()
{
	return true;
}



void AggregateOperation :: PrepareAtts(FuncOperator *Operator)
{
	if (Operator->leftOperator != NULL)
		PrepareAtts(Operator->leftOperator);

	if (Operator->right != NULL)
		PrepareAtts(Operator->right);
	
	if (Operator->leftOperand != NULL)
	{
		if (Operator->leftOperand->code == NAME)
			CurrentLevelAtts.insert(Operator->leftOperand->value);
	}
}



void AggregateOperation :: Initialization(struct AggregateStep * CurrentOperation)
{
	type = TYPEAGGREGATE; 

	InputPipe1 = CurrentOperation->InPutPipe;

	CurrentOperator = CurrentOperation->Operator;
	CurrentLevelAtts.clear();
	PrepareAtts(CurrentOperator);
	
	

}


void AggregateOperation :: BuildSchema(set<string> AttsFromAbove)
{
	JoinAtts.clear();
	if(child1->type == TYPESORTJOIN)
	{
		SortJoinOperation *p= (SortJoinOperation *)child1;
		for(struct AndList * tmp= p->CurrentAndList;tmp!=NULL;tmp=tmp->rightAnd)
		{
			if(tmp->left->rightOr==NULL&&tmp->left->left->code==EQUALS&&tmp->left->left->left->code==NAME&&tmp->left->left->right->code==NAME)
			CurrentLevelAtts.insert(tmp->left->left->left->value);
			JoinAtts.insert(tmp->left->left->left->value);
		}
	}

	child1->BuildSchema(CurrentLevelAtts);
	
	InputSchema1 = new (std::nothrow) Schema();
	if (InputSchema1 == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	InputSchema1->CopyFrom(child1->OutputSchema);

	OutputSchema = new (std::nothrow) Schema();
	if (OutputSchema == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	OutputSchema->CopyFrom(InputSchema1);
	
	set<string>::iterator iend = CurrentLevelAtts.end(), ibegin;

	for (ibegin = CurrentLevelAtts.begin(); ibegin != iend; ++ibegin)
	{
		char* temp = (char*)( (*ibegin).c_str());
		if (OutputSchema->Find(temp) == -1)
		{
			cerr << "I could not find this attribute in the output schema: " << temp << "\n";
			exit(1);
		}
	}

	myFunc = new (std::nothrow) Function();
	if (myFunc == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	myFunc->GrowFromParseTree(CurrentOperator, InputSchema1);


	if(child1->type == TYPESORTJOIN)
	{
		int numAtts = OutputSchema->GetNumAtts();
		Attribute* AttList = OutputSchema->GetAtts();
	
		vector<int> tempJoinAtts;
		for(int i = 0; i < numAtts; i++)
		{
			if (JoinAtts.find(AttList[i].name) != JoinAtts.end())
				tempJoinAtts.push_back(i);
		}
	
		numJoinAtts = tempJoinAtts.size();
		JoinAttIndex = new (std::nothrow) int[numJoinAtts];
		if (JoinAttIndex == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	
		for(int i = 0; i < numJoinAtts; i++)
		{
			JoinAttIndex[i] = tempJoinAtts[i];
		}
	}
}


void AggregateOperation :: PrintParameters(int recursive)
{
	cout << "Begin" << "\n";

	cout << "Operation Type: Output" << "\n";
	cout << "InputPipe " << InputPipe1 << "\n";

	cout << "InputSchema: " << "\n";
	InputSchema1->Print();

	cout << "Function: " << "\n";
	myFunc->Print(InputSchema1);

	cout << "Join Attributes: " << "\t";
	for (int i = 0; i < numJoinAtts; i++)
		cout << JoinAttIndex[i] << " ";
	cout << "\n";

	cout << "End" << "\n" << "\n";

	if (recursive == 1)
	{
		cout << "My child" << "\n";
		child1->PrintParameters(recursive);
	}
}





LevelWiseStepCompiler::LevelWiseStepCompiler()
{
	LevelWiseNumber = -1;
	LevelChecked = false;

	Operations.clear();

	InMemoryJoinMask = NULL;

	InMemoryCNFs=NULL;
	OtherConditions=NULL;

	InMemoryFunc=NULL;
	func_schema = NULL;

	RecLit=NULL;
	RecLit_Schema=NULL;
}


LevelWiseStepCompiler::~LevelWiseStepCompiler()
{
	LevelWiseNumber = -1;
	LevelChecked = false;

	int size=Operations.size();

	if(InMemoryJoinMask!=NULL)
	{
		for(int i=0;i<size;++i)
		{	
			delete [] InMemoryJoinMask[i];
			delete [] InMemoryCNFs[i];
		}
	
		delete [] InMemoryJoinMask;
		delete [] InMemoryCNFs;
	}

	Operations.clear();

	delete InMemoryFunc;
	delete func_schema;

	delete OtherConditions;

	delete RecLit;
	delete RecLit_Schema;
}


void LevelWiseStepCompiler::Add(OperationItem* Node)
{
	Operations.push_back(Node);
}


int LevelWiseStepCompiler::GetTreeLevel()
{
	if (LevelChecked == false)
	{
		cerr << "Please call CheckTreeLevel before calling this function." << "\n";
		exit(1);
	}

	return (Operations[0])->mylevel;
}


void LevelWiseStepCompiler ::  CheckTreeLevel()
{
	int size = Operations.size();
	int level = (Operations[0])->mylevel;

	for (int i = 1; i < size; i++)
	{
		if (level != (Operations[i])->mylevel)
		{
			cerr << "Nodes claimed in the same levelwise step are not in the same level of the query tree." << "\n";
			exit(1);
		}
	}

	LevelChecked = true;
}



void LevelWiseStepCompiler::PrintParameters(int recursive)
{
	int n = Operations.size(), i, j;

	cout << "Level: " << LevelWiseNumber << "\n";

	cout << "Operations:\n";
	for (i = 0; i < n; i++)
		Operations[i]->PrintParameters(recursive);

	cout << "Join Mask:\n";
	for (i = 0; i < n; i++)
	{
		for (j = 0; j < n; j++)
		{
			cout << InMemoryJoinMask[i][j] << "\t";
		}
		cout << "\n";
	}

	cout << "\n";

	cout << "CNFs:\n";
	for (i = 0; i < n; i++)
	{
		for (j = 0; j < n; j++)
		{
			cout << "(" << i << "," << j << "):\n";
			InMemoryCNFs[i][j].Print();
		}
	}

	cout << "Additional CNF:\n";
	OtherConditions->Print();

	cout << "Literal Record:" <<"\n";
	RecLit->Print(RecLit_Schema);

	cout << "Function:\n";
	InMemoryFunc->Print(func_schema);

	cout << "\n\n";
}






OperationManager::OperationManager(char *schema_file)
{
	this->schema_file = strdup(schema_file);
	root = NULL;

	join_level = NULL;
	aggregate_level = NULL;

	RunningOrder = NULL;

	CurrentOperations.clear();

	PipeIdtoIndex.clear();
	CurrentPipes.clear();
}


OperationManager::~OperationManager()
{
	free(schema_file);

	root = NULL;

	delete [] RunningOrder;
	RunningOrder = NULL;

	int i, n;

	n = ActualJoinLevel + 1;
	for (i = 0; i < n; i++)
	{
		delete join_level[i];
		join_level[i] = NULL;

		if (i < n - 1)
		{
			delete imj_join_level[i];
			imj_join_level[i] = NULL;
		}
	}

	delete [] join_level;
	join_level = NULL;

	delete [] imj_join_level;
	imj_join_level = NULL;

	ActualJoinLevel = 0;


	delete aggregate_level;
	aggregate_level = NULL;


	n = CurrentOperations.size();
	for (i = 0; i < n; i++)
	{
		OperationItem *e = CurrentOperations[i];
		delete e;
	}
	CurrentOperations.clear();


	PipeIdtoIndex.clear();

	n = CurrentPipes.size();
	for (i = 0; i < n; i++)
	{
		Pipe *e = CurrentPipes[i];
		e = NULL;
	}
	CurrentPipes.clear();

	delete [] relation_size;
}



void OperationManager::InitCurrentOperations()
{
	//three steps;
	//build the tablescan list
	//build each joinlevel
	//build the aggregation part

	//parse tablescanlist
	TableScanOperation *ptablescan;
	for (TableScanList *currentTableScanList = finalope->CurrentTableScanList; currentTableScanList != NULL; currentTableScanList = currentTableScanList->NextItem)
	{
		ptablescan = new (std::nothrow) TableScanOperation(schema_file);
		if (ptablescan == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		ptablescan->Initialization(currentTableScanList->CurrentItem);
		TableScanLevel.Add((OperationItem*)ptablescan);
		CurrentOperations.push_back((OperationItem*)ptablescan);
	}


	//now scan for each join level ,and parse
	ActualJoinLevel = 0;
	SortJoinOperation *psortjoin;
	RehashingOperation *prehashing;
	for (LevelWiseList *currentLevelWise = finalope->CurrentLevelWiseList; currentLevelWise != NULL; currentLevelWise = currentLevelWise->NextLevel)
	{
		JoinLevel[ActualJoinLevel].LevelWiseNumber = currentLevelWise->CurrentLevelNumber;
		for (SortJoinList *currentJoinList = currentLevelWise->CurrentJoinList; currentJoinList != NULL; currentJoinList = currentJoinList->NextItem)
		{
			psortjoin = new (std::nothrow) SortJoinOperation(schema_file);
			if (psortjoin == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			psortjoin->Initialization(currentJoinList->CurrentItem);
			JoinLevel[ActualJoinLevel].Add((OperationItem*)psortjoin);
			CurrentOperations.push_back((OperationItem*)psortjoin);
		}
		for(RehashingList *currentRehashingList = currentLevelWise->CurrentRehashingList; currentRehashingList !=NULL; currentRehashingList = currentRehashingList->NextItem)
		{
			prehashing = new (std::nothrow) RehashingOperation(schema_file);
			if (prehashing == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			prehashing->Initialization(currentRehashingList->CurrentItem);
			JoinLevel[ActualJoinLevel].Add((OperationItem*)prehashing);
			CurrentOperations.push_back((OperationItem*)prehashing);
		}
		++ActualJoinLevel;
		if(ActualJoinLevel>=MAXLEVEL)
		{
			cerr << "execeed max levelwise join" << "\n";
			exit(1);
		}
	}


	//now deal with the aggregation part
	AggregateOperation *pAggregate = new (std::nothrow) AggregateOperation(schema_file);
	if (pAggregate == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	pAggregate->Initialization(finalope->CurrentAggregationStep);
	root = (OperationItem*)pAggregate;
	CurrentOperations.push_back(root);
}



//setup the pipe and index relation
void OperationManager :: SetupPipeIndex()
{
	PipeIdtoIndex.clear();

	int size = CurrentOperations.size();
	int index = 0;

	for(int i = 0; i < size; i++)
	{
		if (CurrentOperations[i]->OutputPipe != -1)
		{
			if (PipeIdtoIndex.find(CurrentOperations[i]->OutputPipe) != PipeIdtoIndex.end())
			{
				cerr << "find two different operations have the same output pipe" << CurrentOperations[i]->OutputPipe << "\n";
				exit(1);
			}
			else
			{
				PipeIdtoIndex[CurrentOperations[i]->OutputPipe] = index;
				++index;
			}
		}
	}

	//create all pipes
	size = PipeIdtoIndex.size();
	for (int i = 0; i < size; i++)
	{
		Pipe *in_pipe = new (std::nothrow) Pipe(PIPE_BUFFERSIZE);
		if (in_pipe == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		CurrentPipes.push_back(in_pipe);
	}
}



void OperationManager :: BuiltQueryTree()
{
	//for every operation in reverse order(later operations
	//are poteintially parents of previous operations)
	//use the relationship specified in inputpipe /outputpipe
	//to build the tree
	int NumOperations = CurrentOperations.size();
	
	
	for(int i = NumOperations - 1; i >= 0; i--)
	{
		
		//if the operation is TableScan, we do nothing
		//because they are leaf nodes.
		if (CurrentOperations[i]->type != TYPETABLESCAN)
		{
			if (CurrentOperations[i]->IsUnary())
			{
				int input = CurrentOperations[i]->InputPipe1;
				int find = 0;
				for(int j = NumOperations- 2; j >= 0; j--)
				{
					
					//we find the child
					if (input == CurrentOperations[j]->OutputPipe)
					{
						find = 1;
						CurrentOperations[i]->child1 = CurrentOperations[j];
						CurrentOperations[j]->parent = CurrentOperations[i];
						break;	
					}
				}

				//fail to construct the tree
				if (find == 0)
				{
					cerr << "Could not find the chile for node number " << i << "\n";
					exit(0);
				}
			}

			else
			{
				int input1 = CurrentOperations[i]->InputPipe1;
				int input2 = CurrentOperations[i]->InputPipe2;

				int find1 = 0, find2 = 0;
				
				for(int j = NumOperations- 2; j >= 0; j--)
				{
					
					if (input1 == CurrentOperations[j]->OutputPipe)
					{
						find1 = 1;
						CurrentOperations[i]->child1 = CurrentOperations[j];
						CurrentOperations[j]->parent = CurrentOperations[i];
					}

					if (input2 == CurrentOperations[j]->OutputPipe)
					{
						find2 = 1;
						CurrentOperations[i]->child2 = CurrentOperations[j];
						CurrentOperations[j]->parent = CurrentOperations[i];
					}

					if ((find1 == 1) && (find2 == 1))
						break;
				}

				if ((find1 == 0) || (find2 == 0))
				{
					cerr << "Could not find the chile for node number " << i << "\n";
					exit(0);
				}
			}
		}
	}
}



//add check codes here
void OperationManager::CheckTree()
{
	root->CheckCircle();
	if (root->Count() != CurrentOperations.size())
	{
		cout << "Warning!!! Some nodes are not in the tree." << "\n";
		exit(1);
	}

	root->SetLevel(1);
}



//check tree nodes defined in the same level are actually in the same level
void OperationManager :: CheckLevel()
{
	for(int i = 0; i < ActualJoinLevel; ++i)
	{
		JoinLevel[i].CheckTreeLevel();
	}
}


//prepare an order of join levels defined in the JoinLevel array
//remember when defining a query plan, the levelwise step could be in arbitrary order
//when actually executing this we need to follow an order from bottom to the top along the tree
void OperationManager :: PrepareRunningOrder()
{
	RunningOrder = new (std::nothrow) int[ActualJoinLevel];
	if (RunningOrder == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	
	for (int i = 0; i < ActualJoinLevel; ++i)
		RunningOrder[i] = i;

	int temp;
	for (int i = 0; i < ActualJoinLevel; ++i)
	{
		temp = i;
		for (int j = i+1; j < ActualJoinLevel; ++j)
		{
			if (JoinLevel[temp].GetTreeLevel() < JoinLevel[j].GetTreeLevel())
				temp = j;
		} 
		
		if(temp != i)
		{
			int abc = RunningOrder[i];
			RunningOrder[i] = RunningOrder[temp];
			RunningOrder[temp] = abc;
		}
	}


/*	for (int i = 0; i < ActualJoinLevel; ++i)
		cout << RunningOrder[i] << '\t';
	cout << "\n";
*/
}
//This function prepares all parameters needed for in-memory-join
//remember: since we read from output, the tablescan level needs to be handled
void OperationManager :: BuildInMemoryJoinParameters()
{
	//the root of the join level is not needed to be set up since there is only
	//one relation;

	//only scan operations in the query plan, so no InMemoryJoins
	if (ActualJoinLevel == 0)
		return;


	int count=0;
	Record temprec;
	Schema *temp_schema;

	AndList *pfirst=NULL,*pCurrent=NULL, *ptemp=NULL;

	// this set up all join conditions for each level
	for(int i= ActualJoinLevel-2;i>=0;--i)
	{
		
		if(i==ActualJoinLevel-2)
		{
			if(JoinLevel[RunningOrder[i+1]].Operations[0]->IsUnary())
			{
				cout<<"TopLevel should only have Join Operations \n";
				exit(1);
			}
			pfirst=((SortJoinOperation*)(JoinLevel[RunningOrder[i+1]].Operations[0]))->CurrentAndList;	
		}
		//build the cnfs and masks;
		int size = JoinLevel[RunningOrder[i]].Operations.size();
		JoinLevel[RunningOrder[i]].InMemoryJoinMask = new (std::nothrow) int* [size];
		if (JoinLevel[RunningOrder[i]].InMemoryJoinMask == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		JoinLevel[RunningOrder[i]].InMemoryCNFs	= new (std::nothrow) CNF* [size];
		if (JoinLevel[RunningOrder[i]].InMemoryCNFs == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for(int j = 0;j < size; ++j)
		{
			JoinLevel[RunningOrder[i]].InMemoryJoinMask[j]= new (std::nothrow) int[size];
			if (JoinLevel[RunningOrder[i]].InMemoryJoinMask[j] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			JoinLevel[RunningOrder[i]].InMemoryCNFs[j]= new (std::nothrow) CNF[size];
			if (JoinLevel[RunningOrder[i]].InMemoryCNFs[j] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

		}
		for(int j=0; j< size;++j)
			for(int k=0;k< size;++k)
			{
				JoinLevel[RunningOrder[i]].InMemoryJoinMask[j][k]=0;
			}
		
		

		//search for each condition the list pointed by pfirst
		//for each CNF,find two relations involved, setup the cnf and mask
		//each parse tree should only contain one condition
		//and this is a join condition
		
		count=0;
		for(pCurrent=pfirst;pCurrent!=NULL;pCurrent=pCurrent->rightAnd)
		{
			count++;
			//we find a qualified entry
			// a singleton OR list with equality check
			if(pCurrent->left->rightOr==NULL&&pCurrent->left->left->code==EQUALS
			&&pCurrent->left->left->left->code==NAME&&pCurrent->left->left->right->code==NAME)
			{
				char *left=pCurrent->left->left->left->value;
				char *right=pCurrent->left->left->right->value;
				int leftIndex=-1, rightIndex=-1, temp=-1;
				for(int k=0;k<size;++k)
				{
					
					temp = JoinLevel[RunningOrder[i]].Operations[k]->OutputSchema->Find(left);
					if(temp != -1)
						leftIndex =k;
					temp = JoinLevel[RunningOrder[i]].Operations[k]->OutputSchema->Find(right);
					if(temp != -1)
						rightIndex =k;
					
				}
				if(leftIndex == -1)
					cout<<"Error, could not find "<<left<<" in any relation.\n";
				if(rightIndex == -1)
					cout<<"Error, could not find "<<right<<" in any relation.\n";
				//disconnect the right And of pCurrent so that we can use this node only
				//to build the CNF
				if(leftIndex>rightIndex)
				{
					int tempind=leftIndex;
					leftIndex=rightIndex;
					rightIndex=tempind;
				}
				ptemp=pCurrent->rightAnd;
				pCurrent->rightAnd=NULL;
				JoinLevel[RunningOrder[i]].InMemoryJoinMask[leftIndex][rightIndex] = JoinLevel[RunningOrder[i]].InMemoryJoinMask[rightIndex][leftIndex]=1;


				temprec.GrowFromParseTree(pCurrent, JoinLevel[RunningOrder[i]].Operations[leftIndex]->OutputSchema, JoinLevel[RunningOrder[i]].Operations[rightIndex]->OutputSchema, JoinLevel[RunningOrder[i]].InMemoryCNFs[rightIndex][leftIndex], temp_schema);

				delete temp_schema;
				temp_schema = NULL;


				temprec.GrowFromParseTree(pCurrent, JoinLevel[RunningOrder[i]].Operations[leftIndex]->OutputSchema, JoinLevel[RunningOrder[i]].Operations[rightIndex]->OutputSchema, JoinLevel[RunningOrder[i]].InMemoryCNFs[leftIndex][rightIndex], temp_schema);

				delete temp_schema;
				temp_schema = NULL;



/*
				cout<<"i="<<i<<endl;
				cout<<"leftIndex="<<leftIndex<<" rightIndex="<<rightIndex<<endl;
				JoinLevel[RunningOrder[i]].Operations[leftIndex]->OutputSchema->Print();
				JoinLevel[RunningOrder[i]].Operations[rightIndex]->OutputSchema->Print();
				JoinLevel[RunningOrder[i]].InMemoryCNFs[leftIndex][rightIndex].Print();
*/

				//getchar();
				//after building the CNF, put the connection back
				pCurrent->rightAnd=ptemp;
			}
// 			
  		}

		
		//build the function object;
		Schema *temp1= new (std::nothrow) Schema, *temp2= new (std::nothrow) Schema;
		if ((temp1 == NULL) || (temp2 == NULL))
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		JoinLevel[RunningOrder[i]].InMemoryFunc = new (std::nothrow) Function;
		if (JoinLevel[RunningOrder[i]].InMemoryFunc == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		temp2->CopyFrom(JoinLevel[RunningOrder[i]].Operations[0]->OutputSchema,JoinLevel[RunningOrder[i]].Operations[1]->OutputSchema);
		for(int j = 2 ; j < size ; ++j)
		{
			temp1->CopyFrom(temp2, JoinLevel[RunningOrder[i]].Operations[j]->OutputSchema);
			temp2->CopyFrom(temp1);
		}
		
		JoinLevel[RunningOrder[i]].func_schema = temp2;
		
		JoinLevel[RunningOrder[i]].InMemoryFunc ->GrowFromParseTree(((AggregateOperation*)root)->CurrentOperator, temp2);

		//temp2->Print();
		//getchar();
		AndList* qfirst=CopyAndList(pfirst);
		AndList *d_qfirst=qfirst;
		AndList *qcurrent=NULL;
		for(;qfirst!=NULL&&(qfirst->left->rightOr==NULL&&qfirst->left->left->code==EQUALS&&qfirst->left->left->left->code==NAME&&qfirst->left->left->right->code==NAME);qfirst=qfirst->rightAnd);
		qcurrent=qfirst;
		if(qfirst!=NULL)
		for(qcurrent=qfirst;qcurrent->rightAnd!=NULL;qcurrent=qcurrent->rightAnd)
		{
			if(qcurrent->rightAnd->left->rightOr==NULL&&qcurrent->rightAnd->left->left->code==EQUALS
			&&qcurrent->rightAnd->left->left->left->code==NAME&&qcurrent->rightAnd->left->left->right->code==NAME)
			//we need to delete this node;
			qcurrent->rightAnd=qcurrent->rightAnd->rightAnd;
		}
		JoinLevel[RunningOrder[i]].OtherConditions= new (std::nothrow) CNF;
		if (JoinLevel[RunningOrder[i]].OtherConditions == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		JoinLevel[RunningOrder[i]].RecLit=new (std::nothrow) Record;
		if (JoinLevel[RunningOrder[i]].RecLit == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		JoinLevel[RunningOrder[i]].RecLit->GrowFromParseTree(qfirst,temp2,*(JoinLevel[RunningOrder[i]].OtherConditions), JoinLevel[RunningOrder[i]].RecLit_Schema);

		CleanAndList(d_qfirst);

//		JoinLevel[RunningOrder[i]].OtherConditions->Print();


		delete temp1;
//		delete temp2;


		//add in current level,used for next level's generation
		//re-set pCurrent for safety;
		pCurrent=pfirst;
		for(int j = 0 ; j < size ; ++j)
		{
			for(;pCurrent->rightAnd!=NULL;pCurrent=pCurrent->rightAnd);
			if(!JoinLevel[RunningOrder[i]].Operations[j]->IsUnary())
			pCurrent->rightAnd=((SortJoinOperation*)(JoinLevel[RunningOrder[i]].Operations[j]))->CurrentAndList;
		}
	}
	
	//add in tablescan level
	
	if(ActualJoinLevel ==1)
		pfirst=((SortJoinOperation*)(JoinLevel[0].Operations[0]))->CurrentAndList;
	count=0;
	
	int size=TableScanLevel.Operations.size();
	TableScanLevel.InMemoryJoinMask = new (std::nothrow) int* [size];
	if (TableScanLevel.InMemoryJoinMask == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	TableScanLevel.InMemoryCNFs = new (std::nothrow) CNF* [size];
	if (TableScanLevel.InMemoryCNFs == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for(int j = 0;j < size; ++j)
	{
		TableScanLevel.InMemoryJoinMask[j]= new (std::nothrow) int[size];
		if (TableScanLevel.InMemoryJoinMask[j] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		TableScanLevel.InMemoryCNFs[j]= new (std::nothrow) CNF[size];
		if (TableScanLevel.InMemoryCNFs[j] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

	}
	for(int j=0; j< size;++j)
		for(int k=0;k< size;++k)
		{
			TableScanLevel.InMemoryJoinMask[j][k]=0;
		}
	//search for each condition in the AllJoinConditions
	//for each CNF,find two relations involved, setup the cnf and mask
	//each parse tree should only contain one condition
	//and this is a join condition
	
	
	count=0;
	
	for(pCurrent=pfirst;pCurrent!=NULL;pCurrent=pCurrent->rightAnd)
	{
		count++;
		if(pCurrent->left->rightOr==NULL&&pCurrent->left->left->code==EQUALS
		&&pCurrent->left->left->left->code==NAME&&pCurrent->left->left->right->code==NAME)
		{
			char *left=pCurrent->left->left->left->value;
			char *right=pCurrent->left->left->right->value;
			int leftIndex=-1, rightIndex=-1, temp=-1;
			for(int k=0;k<size;++k)
			{
				
				temp = TableScanLevel.Operations[k]->OutputSchema->Find(left);
				if(temp != -1)
					leftIndex = k;
				temp = TableScanLevel.Operations[k]->OutputSchema->Find(right);
				if(temp != -1)
					rightIndex = k;
				
			}
			if(leftIndex == -1)
				cout<<"Error, could not find "<<left<<" in any relation.\n";
			if(rightIndex == -1)
				cout<<"Error, could not find "<<right<<" in any relation.\n";
			if(leftIndex>rightIndex)
			{
			
				int tempind=leftIndex;
				leftIndex=rightIndex;
				rightIndex=tempind;
			}
			ptemp=pCurrent->rightAnd;
			pCurrent->rightAnd=NULL;
			TableScanLevel.InMemoryJoinMask[leftIndex][rightIndex] = TableScanLevel.InMemoryJoinMask[rightIndex][leftIndex]=1;

			temprec.GrowFromParseTree(pCurrent, TableScanLevel.Operations[leftIndex]->OutputSchema, TableScanLevel.Operations[rightIndex]->OutputSchema, TableScanLevel.InMemoryCNFs[rightIndex][leftIndex], temp_schema);

			delete temp_schema;
			temp_schema = NULL;


			temprec.GrowFromParseTree(pCurrent, TableScanLevel.Operations[leftIndex]->OutputSchema, TableScanLevel.Operations[rightIndex]->OutputSchema, TableScanLevel.InMemoryCNFs[leftIndex][rightIndex], temp_schema);

			delete temp_schema;
			temp_schema = NULL;



//			TableScanLevel.InMemoryCNFs[leftIndex][rightIndex].Print();


			pCurrent->rightAnd=ptemp;
		}
		
	}
	Schema *temp1= new (std::nothrow) Schema, *temp2= new (std::nothrow) Schema;
	if ((temp1 == NULL) || (temp2 == NULL))
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	TableScanLevel.InMemoryFunc = new (std::nothrow) Function;
	if (TableScanLevel.InMemoryFunc == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	temp2->CopyFrom(TableScanLevel.Operations[0]->OutputSchema,TableScanLevel.Operations[1]->OutputSchema);
	for(int j = 2 ; j < size ; ++j)
	{
		temp1->CopyFrom(temp2, TableScanLevel.Operations[j]->OutputSchema);
		temp2->CopyFrom(temp1);
	}

	//temp2->Print();

	TableScanLevel.func_schema = temp2;

	TableScanLevel.InMemoryFunc ->GrowFromParseTree(((AggregateOperation*)root)->CurrentOperator, temp2);
	AndList* qfirst=CopyAndList(pfirst);
	AndList *d_qfirst=qfirst;
	AndList *qcurrent=NULL;
	for(;qfirst!=NULL&&(qfirst->left->rightOr==NULL&&qfirst->left->left->code==EQUALS&&qfirst->left->left->left->code==NAME&&qfirst->left->left->right->code==NAME);qfirst=qfirst->rightAnd);
	qcurrent=qfirst;
	if(qfirst!=NULL)
	for(qcurrent=qfirst;qcurrent->rightAnd!=NULL;qcurrent=qcurrent->rightAnd)
	{
		if(qcurrent->rightAnd->left->rightOr==NULL&&qcurrent->rightAnd->left->left->code==EQUALS
		&&qcurrent->rightAnd->left->left->left->code==NAME&&qcurrent->rightAnd->left->left->right->code==NAME)
		//we need to delete this node;
		qcurrent->rightAnd=qcurrent->rightAnd->rightAnd;
	}
	TableScanLevel.OtherConditions= new (std::nothrow) CNF;
	if (TableScanLevel.OtherConditions == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	TableScanLevel.RecLit=new (std::nothrow) Record;
	if (TableScanLevel.RecLit == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	if(pfirst!=NULL)
	TableScanLevel.RecLit->GrowFromParseTree(qfirst,temp2,*(TableScanLevel.OtherConditions), TableScanLevel.RecLit_Schema);

	CleanAndList(d_qfirst);

//	TableScanLevel.OtherConditions->Print();

	delete temp1;
//	delete temp2;

}

//put all codes together
void OperationManager :: RunAll()
{
	InitCurrentOperations();

	BuiltQueryTree();

	CheckTree();

	SetupPipeIndex();

	set<string> abc;
	abc.clear();
	root->BuildSchema(abc);


	CheckLevel();

	PrepareRunningOrder();


//build the parameters that are passed for query execution
	BuildRelationSizes();

	BuildInMemoryJoinParameters();

	BuildAggregateLevel();

	BuildJoinLevels();

//print all the structures sent to the execution engine

/*
	root->PrintParameters(1);

	if (ActualJoinLevel > 0)
		TableScanLevel.PrintParameters(0);

	for (int i = 0; i < ActualJoinLevel - 1; i++)
		JoinLevel[i].PrintParameters(0);
*/

}



void OperationManager :: BuildRelationSizes()
{
	relations_no = TableScanLevel.Operations.size();
	relation_size = new (std::nothrow) int[relations_no];
	if (relation_size == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	for (int i = 0; i < relations_no; i++)
	{
		relation_size[i] = TableScanLevel.Operations[i]->InputSchema1->GetRecsNo();
	}
}



void OperationManager :: BuildAggregateLevel()
{
	aggregate_level = new (std::nothrow) LevelwiseStep(relations_no, relation_size);
	if (aggregate_level == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	AggregateOperation *ag = (AggregateOperation*)root;
	aggregate_level->Append(CurrentPipes[PipeIdtoIndex[ag->InputPipe1]], ag->InputPipe1, ag->InputSchema1, ag->myFunc, ag->JoinAttIndex, ag->numJoinAtts);
}


void OperationManager :: BuildJoinLevels()
{
	join_level = new (std::nothrow) LevelwiseStep*[ActualJoinLevel + 1];
	if (join_level == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	imj_join_level = new (std::nothrow) InMemoryJoin*[ActualJoinLevel];
	if (imj_join_level == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int i, j, n;

	//first build the level corresponding to scans
	n = TableScanLevel.Operations.size();

	Schema **sch_array = new (std::nothrow) Schema*[n];
	if (sch_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (i = 0; i < n; i++)
		sch_array[i] = TableScanLevel.Operations[i]->OutputSchema;

	imj_join_level[0] = new (std::nothrow) InMemoryJoin(n, TableScanLevel.InMemoryCNFs, TableScanLevel.OtherConditions, TableScanLevel.RecLit, TableScanLevel.InMemoryJoinMask, sch_array, TableScanLevel.InMemoryFunc);
	if (imj_join_level[0] == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	join_level[0] = new (std::nothrow) LevelwiseStep(relations_no, relation_size);
	if (join_level[0] == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (i = 0; i < n; i++)
	{
		TableScanOperation *ts = (TableScanOperation*)TableScanLevel.Operations[i];
		join_level[0]->Append(ts->InputSchema1->GetName(), CurrentPipes[PipeIdtoIndex[ts->OutputPipe]], ts->InMemoryJoinBufferSize, ts->myCNF, ts->RecLit, ts->InputSchema1, ts->AttstoProj, ts->NumAttstoProj, ts->InputSchema1->GetRecsNo());
	}

	delete [] sch_array;



	//then build the levels for the joins
	for (i = 0; i < ActualJoinLevel; i++)
	{
		n = JoinLevel[RunningOrder[i]].Operations.size();

		if (i < ActualJoinLevel - 1)
		{
			Schema **sch_array = new (std::nothrow) Schema*[n];
			if (sch_array == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			for (j = 0; j < n; j++)
				sch_array[j] = JoinLevel[RunningOrder[i]].Operations[j]->OutputSchema;

			imj_join_level[i + 1] = new (std::nothrow) InMemoryJoin(n, JoinLevel[RunningOrder[i]].InMemoryCNFs, JoinLevel[RunningOrder[i]].OtherConditions, JoinLevel[RunningOrder[i]].RecLit, JoinLevel[RunningOrder[i]].InMemoryJoinMask, sch_array, JoinLevel[RunningOrder[i]].InMemoryFunc);
			if (imj_join_level[i+1] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}


			delete [] sch_array;
		}

		join_level[i + 1] = new (std::nothrow) LevelwiseStep(relations_no, relation_size);
		if (join_level[i+1] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (j = 0; j < n; j++)
		{
			//build Rehashing
			if(JoinLevel[RunningOrder[i]].Operations[j]->IsUnary())
			{
				RehashingOperation *sj=(RehashingOperation *)JoinLevel[RunningOrder[i]].Operations[j];
				join_level[i+1]->Append(CurrentPipes[PipeIdtoIndex[sj->InputPipe1]], sj->InputPipe1,CurrentPipes[PipeIdtoIndex[sj->OutputPipe]],sj->InMemoryJoinBufferSize, sj->InputSchema1,sj->MemorySize);
			}
			else
			{
				SortJoinOperation *sj = (SortJoinOperation*)JoinLevel[RunningOrder[i]].Operations[j];
				join_level[i+1]->Append(CurrentPipes[PipeIdtoIndex[sj->InputPipe1]], sj->InputPipe1, CurrentPipes[PipeIdtoIndex[sj->InputPipe2]], sj->InputPipe2, CurrentPipes[PipeIdtoIndex[sj->OutputPipe]], sj->InMemoryJoinBufferSize, sj->myCNF, sj->RecLit, sj->InputSchema1, sj->InputSchema2, sj->AttstoProj, sj->NumAttstoProj, sj->MemorySize);
			}
		}
	}
}

/*//This function prepares all parameters needed for in-memory-join
//remember: since we read from output, the tablescan level needs to be handled
void OperationManager :: BuildInMemoryJoinParameters()
{
	//the root of the join level is not needed to be set up since there is only
	//one relation;

	//only scan operations in the query plan, so no InMemoryJoins
	if (ActualJoinLevel == 0)
		return;

	Record temprec;
	// this set up all join conditions for each level
	for(int i= ActualJoinLevel-2;i>=0;--i)
	{
		
		if(i==ActualJoinLevel-2)
		{
			if(JoinLevel[RunningOrder[i+1]].Operations[0]->IsUnary())
			{
				cout<<"TopLevel should only have Join Operations \n";
				exit(1);
			}
			JoinLevel[RunningOrder[i]].AllJoinConditions.push_back(((SortJoinOperation*)(JoinLevel[RunningOrder[i+1]].Operations[0]))->CurrentAndList);	
		
		}
		else
			//first copy from the up level
			copy(JoinLevel[RunningOrder[i+1]].AllJoinConditions.begin(),JoinLevel[RunningOrder[i+1]].AllJoinConditions.end(),back_inserter(JoinLevel[RunningOrder[i]].AllJoinConditions));
		//build the cnfs and masks;
		int size=JoinLevel[RunningOrder[i]].Operations.size();
		JoinLevel[RunningOrder[i]].InMemoryJoinMask = new (std::nothrow) int* [size];
		JoinLevel[RunningOrder[i]].InMemoryCNFs	= new (std::nothrow) CNF* [size];
		for(int j = 0;j < size; ++j)
		{
			JoinLevel[RunningOrder[i]].InMemoryJoinMask[j]= new (std::nothrow) int[size];
			JoinLevel[RunningOrder[i]].InMemoryCNFs[j]= new (std::nothrow) CNF[size];
		}
		for(int j=0; j< size;++j)
			for(int k=0;k< size;++k)
			{
				JoinLevel[RunningOrder[i]].InMemoryJoinMask[j][k]=0;
			}
		
		

		//search for each condition in the AllJoinConditions
		//for each CNF,find two relations involved, setup the cnf and mask
		//each parse tree should only contain one condition
		//and this is a join condition
		
		int condsize=JoinLevel[RunningOrder[i]].AllJoinConditions.size();
		for(int j=0; j<condsize;++j)
		{
			char *left=JoinLevel[RunningOrder[i]].AllJoinConditions[j]->left->left->left->value;
			char *right=JoinLevel[RunningOrder[i]].AllJoinConditions[j]->left->left->right->value;
			int leftIndex=-1, rightIndex=-1, temp=-1;
			for(int k=0;k<size;++k)
			{
				
				temp = JoinLevel[RunningOrder[i]].Operations[k]->OutputSchema->Find(left);
				if(temp != -1)
					leftIndex =k;
				temp = JoinLevel[RunningOrder[i]].Operations[k]->OutputSchema->Find(right);
				if(temp != -1)
					rightIndex =k;
				
			}
			if(leftIndex == -1)
				cout<<"Error, could not find "<<left<<" in any relation.\n";
			if(rightIndex == -1)
				cout<<"Error, could not find "<<right<<" in any relation.\n";
			
			JoinLevel[RunningOrder[i]].InMemoryJoinMask[leftIndex][rightIndex] = JoinLevel[RunningOrder[i]].InMemoryJoinMask[rightIndex][leftIndex]=1;
			temprec.GrowFromParseTree(JoinLevel[RunningOrder[i]].AllJoinConditions[j], JoinLevel[RunningOrder[i]].Operations[leftIndex]->OutputSchema, JoinLevel[RunningOrder[i]].Operations[rightIndex]->OutputSchema, JoinLevel[RunningOrder[i]].InMemoryCNFs[rightIndex][leftIndex]);
			temprec.GrowFromParseTree(JoinLevel[RunningOrder[i]].AllJoinConditions[j], JoinLevel[RunningOrder[i]].Operations[leftIndex]->OutputSchema, JoinLevel[RunningOrder[i]].Operations[rightIndex]->OutputSchema, JoinLevel[RunningOrder[i]].InMemoryCNFs[leftIndex][rightIndex]);
  		}

		//add in current level,used for next level's generation
		
		for(int j = 0 ; j < size ; ++j)
		{
			if(!JoinLevel[RunningOrder[i]].Operations[j]->IsUnary())
			JoinLevel[RunningOrder[i]].AllJoinConditions.push_back(((SortJoinOperation*)(JoinLevel[RunningOrder[i]].Operations[j]))->CurrentAndList);
		}
		//build the function object;
		Schema *temp1= new (std::nothrow) Schema, *temp2= new (std::nothrow) Schema;
		JoinLevel[RunningOrder[i]].InMemoryFunc = new (std::nothrow) Function;
		temp2->CopyFrom(JoinLevel[RunningOrder[i]].Operations[0]->OutputSchema,JoinLevel[RunningOrder[i]].Operations[1]->OutputSchema);
		for(int j = 2 ; j < size ; ++j)
		{
			temp1->CopyFrom(temp2, JoinLevel[RunningOrder[i]].Operations[j]->OutputSchema);
			temp2->CopyFrom(temp1);
		}
		
		
		
		JoinLevel[RunningOrder[i]].InMemoryFunc ->GrowFromParseTree(((AggregateOperation*)root)->CurrentOperator, temp2);
		delete temp1;
		delete temp2;
	}


	//add in tablescan level
	if(ActualJoinLevel ==1)
		TableScanLevel.AllJoinConditions.push_back(((SortJoinOperation*)(JoinLevel[0].Operations[0]))->CurrentAndList);
	else
		copy(JoinLevel[RunningOrder[0]].AllJoinConditions.begin(),JoinLevel[RunningOrder[0]].AllJoinConditions.end(),back_inserter(TableScanLevel.AllJoinConditions));

	int size=TableScanLevel.Operations.size();
	TableScanLevel.InMemoryJoinMask = new (std::nothrow) int* [size];
	TableScanLevel.InMemoryCNFs	= new (std::nothrow) CNF* [size];
	for(int j = 0;j < size; ++j)
	{
		TableScanLevel.InMemoryJoinMask[j]= new (std::nothrow) int[size];
		TableScanLevel.InMemoryCNFs[j]= new (std::nothrow) CNF[size];
	}
	for(int j=0; j< size;++j)
		for(int k=0;k< size;++k)
		{
			TableScanLevel.InMemoryJoinMask[j][k]=0;
		}
	//search for each condition in the AllJoinConditions
	//for each CNF,find two relations involved, setup the cnf and mask
	//each parse tree should only contain one condition
	//and this is a join condition
	
	int condsize=TableScanLevel.AllJoinConditions.size();
	
	for(int j=0; j<condsize;++j)
	{
		char *left=TableScanLevel.AllJoinConditions[j]->left->left->left->value;
		char *right=TableScanLevel.AllJoinConditions[j]->left->left->right->value;
		int leftIndex=-1, rightIndex=-1, temp=-1;
		for(int k=0;k<size;++k)
		{
			
			temp = TableScanLevel.Operations[k]->OutputSchema->Find(left);
			if(temp != -1)
				leftIndex = k;
			temp = TableScanLevel.Operations[k]->OutputSchema->Find(right);
			if(temp != -1)
				rightIndex = k;
			
		}
		if(leftIndex == -1)
			cout<<"Error, could not find "<<left<<" in any relation.\n";
		if(rightIndex == -1)
			cout<<"Error, could not find "<<right<<" in any relation.\n";
		
		TableScanLevel.InMemoryJoinMask[leftIndex][rightIndex] = TableScanLevel.InMemoryJoinMask[rightIndex][leftIndex]=1;
		temprec.GrowFromParseTree(TableScanLevel.AllJoinConditions[j], TableScanLevel.Operations[leftIndex]->OutputSchema, TableScanLevel.Operations[rightIndex]->OutputSchema, TableScanLevel.InMemoryCNFs[rightIndex][leftIndex]);
		temprec.GrowFromParseTree(TableScanLevel.AllJoinConditions[j], TableScanLevel.Operations[leftIndex]->OutputSchema, TableScanLevel.Operations[rightIndex]->OutputSchema, TableScanLevel.InMemoryCNFs[leftIndex][rightIndex]);
	}
	Schema *temp1= new (std::nothrow) Schema, *temp2= new (std::nothrow) Schema;
	TableScanLevel.InMemoryFunc = new (std::nothrow) Function;
	temp2->CopyFrom(TableScanLevel.Operations[0]->OutputSchema,TableScanLevel.Operations[1]->OutputSchema);
	for(int j = 2 ; j < size ; ++j)
	{
		temp1->CopyFrom(temp2, TableScanLevel.Operations[j]->OutputSchema);
		temp2->CopyFrom(temp1);
	}
	TableScanLevel.InMemoryFunc ->GrowFromParseTree(((AggregateOperation*)root)->CurrentOperator, temp2);
	delete temp1;
	delete temp2;

}*/


