#include "Controller.h"
#include "Statistics.h"
#include "SamplingEstimate.h"

#include <iostream>



Controller::Controller(int join_levels_no, LevelwiseStep **join_level, InMemoryJoin **imj_join_level, LevelwiseStep *aggregate_level, double max_err, double conf_level)
{
	this->join_level = join_level;
	this->join_levels_no = join_levels_no;
	this->imj_join_level = imj_join_level;

	this->aggregate_level = aggregate_level;

	tuples_no = new (std::nothrow) double*[join_levels_no];
	if (tuples_no == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	stat_module = new (std::nothrow) Statistics(max_err);
	if (stat_module == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->conf_level = conf_level;

	are_estimates_available = 0;
}


Controller::~Controller()
{
	join_level = NULL;
	imj_join_level = NULL;

	aggregate_level = NULL;

	for (int i = 0; i < join_levels_no; i++)
		delete [] tuples_no[i];

	delete [] tuples_no;

	delete stat_module;
	stat_module = NULL;

	join_levels_no = 0;
}



void Controller::Control()
{
	Control_StartUp();

	Control_JoinLevel();

	Control_UpJoinLevel();

	Control_AggregateLevel();


	for (int i = 0; i < join_levels_no; i++)
	{
		for (int j = 0; j < join_level[i]->GetSize(); j++)
		{
			printf("Level %3d, Join %3d : %20.2f tuples\n", i, j, tuples_no[i][j]);
		}
	}
}



void Controller::Control_StartUp()
{
	for (int i = 0; i < join_levels_no; i++)
		join_level[i]->StartUp();

	aggregate_level->StartUp();
}



void Controller::Control_JoinLevel()
{
	int i, j, k, n;

	int output_size;

	for (j = 0; j < join_levels_no - 1; j++)
	{
		n = join_level[j]->GetSize();

		tuples_no[j] = new (std::nothrow) double[n];
		if (tuples_no[j] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		double *p = new (std::nothrow) double[n];
		if (p == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		double *old_p = new (std::nothrow) double[n];
		if (old_p == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		double *cur_tot = new (std::nothrow) double[n];
		if (cur_tot == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (i = 0; i < n; i++)
		{
			p[i] = -1.0;
			tuples_no[j][i] = 0;
		}


		//add this level to the statistics module
		stat_module->AddLevel(n, 0);


		int first_time = 1;
		while (1)
		{
			int still_running = 0;

			if (first_time == 1)
			{
				for (i = 0; i < n; i++)
				{
					int buffer_size = join_level[j]->out_pipe_sizes[i];

					Record *match_list = new (std::nothrow) Record[buffer_size];
					if (match_list == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					//do this trick in order to obtain a stable p
					buffer_size = (int)((double)buffer_size * 0.9);

					double in_l = join_level[j]->HowMuchIsLeft(i);

					int ret = join_level[j]->GetNextBatch(i, p[i], match_list, buffer_size, output_size);
					tuples_no[j][i] += output_size;

					if (ret == 0)
					{
						delete [] match_list;
						continue;
					}
					else if (ret == 1)
					{
						still_running = 1;
					}
					else if (ret == -1)
					{
						if (output_size == 0)
						{
							printf("Level %3d, Join %3d: Skew too large for a %10d buffer !!!\n", j, i, buffer_size);

							exit(1);
						}

						printf("Level %3d, Join %3d : p = %20.18f is too large !!!\n", j, i, p[i]);
	
						double out_l = join_level[j]->HowMuchIsLeft(i);

						printf("Level %3d, Join %3d : new p = %20.18f !!!\n", j, i, in_l - out_l);

						still_running = 1;
					}
	
					Record *out_imj_list = NULL;
					imj_join_level[j]->Swap(match_list, out_imj_list, i, output_size);
	
					delete [] out_imj_list;

					cur_tot[i] = join_level[j]->HowMuchIsLeft(i);
					printf("Level %3d, Join %3d : left %20.18f !!!\n", j, i, cur_tot[i]);

					p[i] = in_l - cur_tot[i];
					old_p[i] = p[i];
				}
	
				int **join_res = NULL;
				double *aggregate = NULL;
	
				int tup_no = imj_join_level[j]->Join(join_res, aggregate);

				//choose what king of collapsing: sort-based or hash-based
				//tup_no = Collaps_Hash(join_res, aggregate, tup_no, imj_join_level[j]->relations_no);
				//tup_no = Collaps_Sort(join_res, aggregate, tup_no, imj_join_level[j]->relations_no);



//statistics stuff gets in here
				SamplingData *sampling_prop = NULL;
				if (j == 0)
				{
					int *tuples_now = new (std::nothrow) int[n];
					if (tuples_now == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					for (k = 0; k < n; k++)
						tuples_now[k] = (int)(p[k] * join_level[0]->relation_size[k]);

					sampling_prop = new (std::nothrow) SamplingData(n, join_level[j]->relation_size, tuples_now);
					if (sampling_prop == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					delete [] tuples_now;
				}
				else
				{
					sampling_prop = new (std::nothrow) SamplingData(n, p);
					if (sampling_prop == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
				}

				int stat_ret = stat_module->ProcessTuples(join_res, aggregate, tup_no, sampling_prop);

				are_estimates_available |= stat_ret;
				if (are_estimates_available)
				{
					double estimate, variance, high, low;
					stat_ret = stat_module->GetEst(estimate, variance, low, high, conf_level);
	
					printf("STATS: %30.4f\t %30.4f\t %30.4f\t %40.10f\n", low, estimate, high, variance);
				}
				else
				{
					printf("Estimates are unavailable at this moment !!!\n");
				}

				delete sampling_prop;

//statistics stuff end

				for (int k = 0; k < tup_no; k++)
					delete [] join_res[k];
	
				delete [] join_res;
				delete [] aggregate;

				first_time = 0;
			}
	
			else
			{
				for (i = 0; i < n; i++)
				{
					int buffer_size = join_level[j]->out_pipe_sizes[i];

					Record *match_list = new (std::nothrow) Record[buffer_size];
					if (match_list == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					double in_l = join_level[j]->HowMuchIsLeft(i);

					int ret = join_level[j]->GetNextBatch(i, p[i], match_list, buffer_size, output_size);
					tuples_no[j][i] += output_size;

					if (ret == 0)
					{
						delete [] match_list;
						continue;
					}
					else if (ret == 1)
					{
						still_running = 1;
					}
					else if (ret == -1)
					{
						if (output_size == 0)
						{
							printf("Level %3d, Join %3d: Skew too large for a %10d buffer !!!\n", j, i, buffer_size);

							exit(1);
						}

						printf("Level %3d, Join %3d : p = %20.18f is too large !!!\n", j, i, p[i]);
	
						double out_l = join_level[j]->HowMuchIsLeft(i);
	
						printf("Level %3d, Join %3d : new p = %20.18f !!!\n", j, i, in_l - out_l);

						still_running = 1;
					}
		
					Record *out_imj_list = NULL;
					imj_join_level[j]->Swap(match_list, out_imj_list, i, output_size);
	
					cur_tot[i] = join_level[j]->HowMuchIsLeft(i);
					printf("Level %3d, Join %3d : left %20.18f !!!\n", j, i, cur_tot[i]);

					p[i] = in_l - cur_tot[i];


					int **join_res = NULL;
					double *aggregate = NULL;
	

//statistics stuff gets in here
					SamplingData *sampling_prop = NULL;
					if (j == 0)
					{
						int *tuples_now = new (std::nothrow) int[n];
						if (tuples_now == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
	
						for (k = 0; k < n; k++)
							tuples_now[k] = (int)(p[k] * join_level[0]->relation_size[k]);

						sampling_prop = new (std::nothrow) SamplingData(n, join_level[j]->relation_size, tuples_now);
						if (sampling_prop == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}

						delete [] tuples_now;
					}
					else
					{
						sampling_prop = new (std::nothrow) SamplingData(n, p);
						if (sampling_prop == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
					}

					int tup_no = 0;
					int stat_ret = 0;

					tup_no = imj_join_level[j]->Join(join_res, aggregate);

					if ((!are_estimates_available) || (old_p[i] != p[i]))
					{
						//choose what king of collapsing: sort-based or hash-based
						//tup_no = Collaps_Hash(join_res, aggregate, tup_no, imj_join_level[j]->relations_no);
						//tup_no = Collaps_Sort(join_res, aggregate, tup_no, imj_join_level[j]->relations_no);

						stat_ret = stat_module->ProcessTuples(join_res, aggregate, tup_no, sampling_prop);
					}
					else
					{
						stat_module->ProcessTuples(aggregate, tup_no, sampling_prop);
					}
	

					are_estimates_available |= stat_ret;
					if (are_estimates_available)
					{
						double estimate, variance, high, low;
						stat_ret = stat_module->GetEst(estimate, variance, low, high, conf_level);
		
						printf("STATS: %30.4f\t %30.4f\t %30.4f\t %40.10f\n", low, estimate, high, variance);
					}
					else
					{
						printf("Estimates are unavailable at this moment !!!\n");
					}

					delete sampling_prop;

//statistics stuff end

					old_p[i] = p[i];

					for (int k = 0; k < tup_no; k++)
						delete [] join_res[k];
	
					delete [] join_res;
					delete [] aggregate;
	
	
					delete [] out_imj_list;
				}
			}
	
			if (still_running == 0)
				break;
		}

		delete [] p;
		delete [] old_p;
		delete [] cur_tot;
	}
}


void Controller::Control_UpJoinLevel()
{
	int i, n;

	int output_size;

	n = join_level[join_levels_no - 1]->GetSize();

	double *p = new (std::nothrow) double[n];
	if (p == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	tuples_no[join_levels_no - 1] = new (std::nothrow) double[n];
	if (tuples_no[join_levels_no - 1] == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *cur_tot = new (std::nothrow) double[n];
	if (cur_tot == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (i = 0; i < n; i++)
	{
		p[i] = -1.0;
		tuples_no[join_levels_no - 1][i] = 0;
	}



	//add this level to the statistics module
	stat_module->AddLevel(n, 1);

	while (1)
	{
		int still_running = 0;


		for (i = 0; i < n; i++)
		{
			int buffer_size = join_level[join_levels_no - 1]->out_pipe_sizes[i];

			Record *match_list = new (std::nothrow) Record[buffer_size];
			if (match_list == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			double in_l = join_level[join_levels_no - 1]->HowMuchIsLeft(i);

			int ret = join_level[join_levels_no - 1]->GetNextBatch(i, p[i], match_list, buffer_size, output_size);
			tuples_no[join_levels_no - 1][i] += output_size;

			if (ret == 0)
			{
				delete [] match_list;
				continue;
			}
			else
			{
				still_running = 1;
			}


			//build the aggregate function values for the returned tuples
			double *aggregate = new (std::nothrow) double[output_size];
			if (aggregate == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			Function *func = aggregate_level->GetFunction(0);
			for (int j = 0; j < output_size; j++)
			{
				int intResult;
				double doubleResult;
		
				Type t = func->Apply(&match_list[j], intResult, doubleResult);
		
				if (t == Int)
					aggregate[j] = (double)intResult;
				else if (t == Double)
					aggregate[j] = doubleResult;
			}


			//collapsing goes in here
			Schema *agg_schema = aggregate_level->GetSchema(0);

			int *join_atts;
			int join_atts_no;
			aggregate_level->GetJoinAtts(0, join_atts, join_atts_no);

			OrderMaker *join_order = new (std::nothrow) OrderMaker(agg_schema, join_atts, join_atts_no);
			if (join_order == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}


			int tup_no = Collaps_Aggregate_Level(match_list, aggregate, output_size, join_order);



			cur_tot[i] = 1.0 - join_level[join_levels_no - 1]->HowMuchIsLeft(i);
			printf("UpLevel, Join %3d : left %20.18f !!!\n", i, 1.0 - cur_tot[i]);


//statistics stuff gets in here
			SamplingData *sampling_prop = new (std::nothrow) SamplingData(n, cur_tot);
			if (sampling_prop == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			stat_module->ProcessTuples(aggregate, tup_no, sampling_prop);

			double estimate, variance, high, low;
			int stat_ret = stat_module->GetEst(estimate, variance, low, high, conf_level);

			printf("STATS: %30.4f\t %30.4f\t %30.4f\t %40.10f\n", low, estimate, high, variance);

			delete sampling_prop;
//statistics stuff end


			delete join_order;

			delete [] match_list;
			delete [] aggregate;
		}

		if (still_running == 0)
			break;
	}

	delete [] p;
	delete [] cur_tot;
}


void Controller::Control_AggregateLevel()
{
	int i, n;

	n = aggregate_level->GetSize();
	for (i = 0; i < n; i++)
	{
		//fake call to GetNextBatch for synchronization purposes
		int output_size;
		int ret = aggregate_level->GetNextBatch(i, 0.0, NULL, -1, output_size);

		double aggregate = aggregate_level->GetAggregate(i);
		double tup_no = aggregate_level->GetTuplesNo(i);

		printf("Result:%30.4f\n", aggregate);
		printf("Tuples:%30.1f\n\n", tup_no);
	}
}





//collapsing method for the aggregate level; it aggregates the function values for each of the distinct groups
int Collaps_Aggregate_Level(Record *join_res, double *&output, int v_size, OrderMaker *join_order)
{
	Hash *eq_func = new (std::nothrow) Hash();
	if (eq_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *new_output = new (std::nothrow) double[v_size];
	if (new_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		new_output[colaps_count] = output[c_i];

		c_j = c_i + 1;


		while ((c_j < v_size) && (eq_func->Compare(&join_res[c_i], &join_res[c_j], *join_order) == 0))
		{
			new_output[colaps_count] += output[c_j];

			c_j += 1;
		}

		c_i = c_j;
		colaps_count += 1;
	}


	delete [] output;
	output = new_output;

	delete eq_func;

	return colaps_count;
}




//collapsing method based on linear hashing; it aggregates the function values for each of the distinct groups
int Collaps_Hash(int **&join_res, double *&output, int v_size, int relations_no)
{

	//THIS IMPLEMENTATION IS BASED ON HASHING//

	int *colaps_array = new (std::nothrow) int[v_size];
	if (colaps_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < v_size; i++)
	{
		colaps_array[i] = -1;
	}

	int colaps_count = 0;
	for (int i = 0; i < v_size; i++)
	{
		//compute the index in the hash array of each record in the output

		unsigned int index = 0;
		for (int j = 0; j < relations_no; j++)
		{
			index += join_res[i][j];

		}
		index = index % v_size;

		if (colaps_array[index] == -1)
		{
			//in this case the position is free, so just add the index to the array
			colaps_array[index] = i;
			colaps_count += 1;

		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = memcmp(join_res[i], join_res[colaps_array[index]], relations_no * sizeof(int));
			if (are_identical == 0)
			{
				//the records are identical, so collapse the aggregate values
				output[colaps_array[index]] += output[i];

			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % v_size; (colaps_array[j] != -1) && (j != index); j = (j + 1) % v_size)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = memcmp(join_res[i], join_res[colaps_array[j]], relations_no * sizeof(int));
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					output[colaps_array[j]] += output[i];
				}
				else
				{
					if (colaps_array[j] == -1)
					{
						colaps_array[j] = i;
						colaps_count += 1;
					}
					else
					{
						cout << "This should be an error. No place was found in the hash array for record with index " << i << " in the result.";
						exit(1);
					}
				}
			}
		}
	}

	int **p_join_res = new (std::nothrow) int*[colaps_count];
	if (p_join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *p_output = new (std::nothrow) double[colaps_count];
	if (p_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

//	double tot_res = 0.0;

	int c_i = 0;
	for (int i = 0; i < v_size; i++)
	{
		if (colaps_array[i] != -1)
		{
			p_join_res[c_i] = new (std::nothrow) int[relations_no];
			if (p_join_res[c_i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			memcpy(p_join_res[c_i], join_res[colaps_array[i]], relations_no * sizeof(int));
	
			p_output[c_i] = output[colaps_array[i]];

//			tot_res += p_output[c_i];

			c_i += 1;
			if (c_i > colaps_count)
			{
				cout << "Error with putting the collapsed records together !";
				exit(1);
			}
		}

		delete [] join_res[i];
	}

	delete [] join_res;
	join_res = p_join_res;

	delete [] output;
	output = p_output;


	delete [] colaps_array;

	//end of collapsing


// 	printf("InMemoryJoinResult: %30.5f\n\n", tot_res);


	return colaps_count;
}



//collapsing method based on sorting; it aggregates the function values for each of the distinct groups
int Collaps_Sort(int **&join_res, double *&output, int v_size, int relations_no)
{
	//THIS IMPLEMENTATION IS BASED ON SORTING//

	int *colaps_array = new (std::nothrow) int[v_size * (relations_no + 2)];
	if (colaps_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < v_size; i++)
	{
		colaps_array[i * (relations_no + 2)] = relations_no;

		memcpy(colaps_array + i * (relations_no + 2) + 1, join_res[i], relations_no * sizeof(int));

		colaps_array[i * (relations_no + 2) + relations_no + 1] = i;

		delete [] join_res[i];
	}

	delete [] join_res;

	qsort(colaps_array, v_size, sizeof(int) * (relations_no + 2), Compare_Direct_IMJ);

	int *colaps_index = new (std::nothrow) int[v_size];
	if (colaps_index == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int colaps_count = 0;

	int c_i, c_j;

	c_i = 0;
	while (c_i < v_size)
	{
		colaps_index[colaps_count] = c_i;
		colaps_count += 1;

		c_j = c_i + 1;


		while ((c_j < v_size) && (memcmp(colaps_array + c_i * (relations_no + 2), colaps_array + c_j * (relations_no + 2), (relations_no + 1) * sizeof(int)) == 0))
		{
			output[colaps_array[c_i * (relations_no + 2) + relations_no + 1]] += output[colaps_array[c_j * (relations_no + 2) + relations_no + 1]];

			c_j += 1;
		}

		c_i = c_j;
	}


	//here we have in colaps_index colaps_count records that have survived

	join_res = new (std::nothrow) int*[colaps_count];
	if (join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *p_output = new (std::nothrow) double[colaps_count];
	if (p_output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


//	double tot_res = 0.0;


	for (int i = 0; i < colaps_count; i++)
	{
		join_res[i] = new (std::nothrow) int[relations_no];
		if (join_res[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		memcpy(join_res[i], colaps_array + colaps_index[i] * (relations_no + 2) + 1, relations_no * sizeof(int));

		p_output[i] = output[colaps_array[colaps_index[i] * (relations_no + 2) + relations_no + 1]];

//		tot_res += p_output[i];
	}
	//done with putting the results in the output


	delete [] colaps_index;

	delete [] colaps_array;

	delete [] output;
	output = p_output;

	//end of collapsing

// 	printf("InMemoryJoinResult: %30.5f\n\n", tot_res);


	return colaps_count;
}



int Compare_IMJ(const void *r1, const void *r2)
{
	int *a, *b;

	a = (int*)*(int*)r1;
	b = (int*)*(int*)r2;

	for (int i = 1; i <= *a; i++)
	{
		if (*(a + i) > *(b + i))
			return 1;
		else if (*(a + i) < *(b + i))
			return -1;
	}

	return 0;
}


int Compare_Direct_IMJ(const void *r1, const void *r2)
{
	int *a, *b;

	a = (int*)r1;
	b = (int*)r2;

	int res = memcmp(a + 1, b + 1, *a * sizeof(int));
	return res;

/*
	for (int i = 1; i <= *a; i++)
	{
		if (*(a + i) > *(b + i))
			return 1;
		else if (*(a + i) < *(b + i))
			return -1;
	}

	return 0;
*/

}



