#include "ControllerIQ.h"
#include "Pipe.h"
#include "RelOpLevel.h"
#include "Record.h"
#include "DEFS.h"

#include <iostream>
#include <stdio.h>


using namespace std;



ControllerInputQueue::ControllerInputQueue(RelationalOpLevel *q_operator, int imj_buffer_size, int queue_size)
{
	this->q_operator = q_operator;
	this->imj_buffer_size = imj_buffer_size;

	// Alin: q_size keeps track of how many things are in the queue so it should be initialized to 0
	q_size = 0;
	// Alin: fixed argumetn to Pipe
	queue = new (std::nothrow) Pipe(queue_size);
	if (queue == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	is_all_done = 0;

	//initialize the mutex
	pthread_mutex_init (&q_mutex, NULL);
	pthread_cond_init (&q_cond, NULL);

	// and let the thread go
	pthread_attr_t t_attr;
	int ret;

	ret = pthread_attr_init(&t_attr);
	ret = pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_DETACHED);
	ret = pthread_create(&thr, &t_attr, ControllerInputQueue::InputQueue_Thread, (void *) this);
	if (ret) {
		cerr << "ERROR; return code from pthread_create() is " << ret << "\n";
		exit(-1);
	}
}


ControllerInputQueue::~ControllerInputQueue()
{
	delete queue;
	queue = NULL;

	pthread_mutex_destroy (&q_mutex);
	pthread_cond_destroy (&q_cond);
}


int ControllerInputQueue::IsAllDone ()
{
	int ret = 0;

//	pthread_mutex_lock(&q_mutex);

	if ((is_all_done == 1) && (q_size <= 0))
		ret = 1;

//	pthread_mutex_unlock(&q_mutex);

	return ret;
}


int ControllerInputQueue::IsConstant ()
{
	int ret = 0;

//	pthread_mutex_lock(&q_mutex);

	if (is_all_done == 1)
		ret = 1;

//	pthread_mutex_unlock(&q_mutex);

	return ret;
}


int ControllerInputQueue::GetNumTuples ()
{
//	pthread_mutex_lock(&q_mutex);

	int ret = q_size;

//	pthread_mutex_unlock(&q_mutex);

	return ret;
}


int ControllerInputQueue::GetNextTuple (Record *nextOne, int &timeStamp)
{
	if (IsAllDone() == 1)
		return 0;

	pthread_mutex_lock(&q_mutex);

	while ((q_size == 0) && (is_all_done == 0))
		pthread_cond_wait (&q_cond, &q_mutex);

	if (q_size > 0)
	{
		queue->Remove(nextOne);
		timeStamp = nextOne->hash;
	
		q_size -= 1;

		pthread_mutex_unlock(&q_mutex);
	
		return 1;
	}

	if (is_all_done == 1)
	{
		pthread_mutex_unlock(&q_mutex);
	
		return 0;
	}
}


void *ControllerInputQueue::InputQueue_Thread (void *arg)
{
	ControllerInputQueue *ciq = (ControllerInputQueue *)arg;

	Record *rec = new (std::nothrow) Record[ciq->imj_buffer_size];
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	while (1)
	{
		int output_size;
		int ret = ciq->q_operator->GetNextBatch (-1.0, rec, ciq->imj_buffer_size, output_size);

		for (int i = 0; i < output_size; i++)
		{
			ciq->queue->Insert(&rec[i]);

			pthread_mutex_lock(&ciq->q_mutex);
		
			ciq->q_size += 1;
		
			pthread_mutex_unlock(&ciq->q_mutex);
			pthread_cond_signal(&ciq->q_cond);
		}

		if (ret == 1)
		{
			ciq->queue->ShutDown();

			pthread_mutex_lock(&ciq->q_mutex);
		
			ciq->is_all_done = 1;
		
			pthread_mutex_unlock(&ciq->q_mutex);
			pthread_cond_signal(&ciq->q_cond);

			break;
		}

		if ((ret == -1) && (output_size == 0))
		{
			cout << "ERROR : Memory to extract tuples from operator is not sufficient !!!\n";
			exit(1);
		}
	}

	delete [] rec;
}


