#include "Estimation.h"
#include "Statistics.h"

#include <iostream>
#include <stdlib.h>
#include <vector>



Estimation::Estimation()
{
}


Estimation::~Estimation()
{
}




Estimation_SPJ::Estimation_SPJ(int tuples_stable, int var_estimates_needed)
{
	stat_module = new (std::nothrow) Statistics(tuples_stable, var_estimates_needed);
	if (stat_module == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}


Estimation_SPJ::~Estimation_SPJ()
{
	delete stat_module;
	stat_module = NULL;
}



void Estimation_SPJ::AddLevel(int numRels, int isLastLevel)
{
	stat_module->AddLevel(numRels, isLastLevel);
}


void Estimation_SPJ::AddGeneralLevel(int numRels, int isLastLevel)
{
	stat_module->AddGeneralLevel(numRels, isLastLevel);
}


int Estimation_SPJ::GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel, double p)
{
	estimates_no = 1;

	estimate = new (std::nothrow) double;
	if (estimate == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	variance = new (std::nothrow) double;
	if (variance == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	low = new (std::nothrow) double;
	if (low == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	high = new (std::nothrow) double;
	if (high == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	groups_id = NULL;

	int res = stat_module->GetEst(*estimate, *variance, *low, *high, confLevel, p);
	return res;
}


int Estimation_SPJ::GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel)
{
	estimates_no = 1;

	estimate = new (std::nothrow) double;
	if (estimate == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	variance = new (std::nothrow) double;
	if (variance == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	low = new (std::nothrow) double;
	if (low == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	high = new (std::nothrow) double;
	if (high == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	groups_id = NULL;

	int res = stat_module->GetEst(*estimate, *variance, *low, *high, confLevel);
	return res;
}


double Estimation_SPJ::WhatIfVariance(double *proposedSamplingRatios)
{
	double res = stat_module->WhatIfVariance(proposedSamplingRatios);
	return res;
}


void Estimation_SPJ::ProcessTuples(SearchResults *newResults, double p)
{
	stat_module->ProcessTuples(newResults, p);
}


int Estimation_SPJ::ProcessTuples(SearchResults *newResults, SimpleSamplingData *data)
{
	int res = stat_module->ProcessTuples(newResults->resTuples, newResults->funcRes, newResults->numRes, data);
	return res;
}


int Estimation_SPJ::ProcessTuples(double *funcVals, Record *groups_id, int numTuples, SimpleSamplingData *data)
{
	int res = stat_module->ProcessTuples(funcVals, numTuples, data);
	return res;
}


void Estimation_SPJ::Print()
{
	stat_module->Print();
}


void Estimation_SPJ::Print_Level()
{
	stat_module->Print_Level();
}



Estimation_Group::Estimation_Group(int tuples_stable, int var_estimates_needed, int levels_no, int groups_no, Schema *group_schema)
{
	this->tuples_stable = tuples_stable;
	this->var_estimates_needed = var_estimates_needed;

	this->levels_no = levels_no;

	this->groups_no = groups_no;

	this->group_schema = group_schema;


	this->group_order = new (std::nothrow) OrderMaker(this->group_schema);
	if (group_order == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->eq_func = new (std::nothrow) Hash();
	if (this->eq_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	level_groups_no = new (std::nothrow) int[this->levels_no];
	if (level_groups_no == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	stat_module = new (std::nothrow) vector<Statistics*>[this->levels_no];
	if (stat_module == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	group_ids = new (std::nothrow) vector<Record*>[this->levels_no];
	if (group_ids == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < this->levels_no; i++)
	{
		level_groups_no[i] = 0;

		stat_module[i].clear();
		group_ids[i].clear();
	}

	current_level = -1;
	cur_level_rels_no = -1;
	is_last_level = -1;
}


Estimation_Group::~Estimation_Group()
{
	delete group_order;
	group_order = NULL;

	delete eq_func;
	eq_func = NULL;

	for (int i = 0; i < levels_no; i++)
	{
		for (int j = 0; j < level_groups_no[i]; j++)
		{
			delete (stat_module[i])[j];
			(stat_module[i])[j] = NULL;

			delete (group_ids[i])[j];
			(group_ids[i])[j] = NULL;
		}

		stat_module[i].clear();
		group_ids[i].clear();
	}

	delete [] stat_module;
	delete [] group_ids;
}



void Estimation_Group::AddLevel(int numRels, int isLastLevel)
{
	cur_level_rels_no = numRels;
	is_last_level = isLastLevel;

	current_level += 1;
}


void Estimation_Group::AddGeneralLevel(int numRels, int isLastLevel)
{
	//first we need to freeze all the estimators from the current level
	if (current_level >= 0)
	{
		//but only if this is not the first level
		for (int i = 0; i < level_groups_no[current_level]; i++)
		{
			(stat_module[current_level])[i]->FreezeCoefficients();
		}
	}

	cur_level_rels_no = numRels;
	is_last_level = isLastLevel;

	current_level += 1;
}


int Estimation_Group::GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel, double p)
{
	estimates_no = level_groups_no[current_level];

	estimate = new (std::nothrow) double[estimates_no];
	if (estimate == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	variance = new (std::nothrow) double[estimates_no];
	if (variance == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	low = new (std::nothrow) double[estimates_no];
	if (low == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	high = new (std::nothrow) double[estimates_no];
	if (high == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	groups_id = new (std::nothrow) Record[estimates_no];
	if (groups_id == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int res = 0;

	//get the estimates from each stat_module in the current level
	for (int i = 0; i < level_groups_no[current_level]; i++)
	{
		res = (stat_module[current_level])[i]->GetEst(estimate[i], variance[i], low[i], high[i], confLevel, p);

		groups_id[i].Copy((group_ids[current_level])[i]);
	}

	return res;
}


int Estimation_Group::GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel)
{
	estimates_no = level_groups_no[current_level];

	estimate = new (std::nothrow) double[estimates_no];
	if (estimate == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	variance = new (std::nothrow) double[estimates_no];
	if (variance == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	low = new (std::nothrow) double[estimates_no];
	if (low == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	high = new (std::nothrow) double[estimates_no];
	if (high == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	groups_id = new (std::nothrow) Record[estimates_no];
	if (groups_id == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int res = 0;

	//get the estimates from each stat_module in the current level
	for (int i = 0; i < level_groups_no[current_level]; i++)
	{
		res = (stat_module[current_level])[i]->GetEst(estimate[i], variance[i], low[i], high[i], confLevel);

		groups_id[i].Copy((group_ids[current_level])[i]);
	}

	return res;
}


double Estimation_Group::WhatIfVariance(double *proposedSamplingRatios)
{
	double ret_var = 0.0;

	double *est_var = new (std::nothrow) double[level_groups_no[current_level]];
	if (est_var == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	//get a WhatIfVariance estimate from each stat_module in the current level
	for (int i = 0; i < level_groups_no[current_level]; i++)
	{
		est_var[i] = (stat_module[current_level])[i]->WhatIfVariance(proposedSamplingRatios);
		ret_var += est_var[i];
	}

	//determine the returned estimate from all available
	//in this case, the average is used
	ret_var = ret_var / level_groups_no[current_level];

	delete [] est_var;

	return ret_var;
}


void Estimation_Group::ProcessTuples(SearchResults *newResults, double p)
{
	Record **group_part;

	vector<int> *group_add;
	vector<int> *group_del;

	group_part = new (std::nothrow) Record*[this->groups_no];
	if (group_part == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	group_add = new (std::nothrow) vector<int>[this->groups_no];
	if (group_add == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	group_del = new (std::nothrow) vector<int>[this->groups_no];
	if (group_del == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < this->groups_no; i++)
	{
		group_part[i] = NULL;
		group_add[i].clear();
		group_del[i].clear();
	}

	Hash *hash_func = new (std::nothrow) CWHash(this->group_order->numAtts + 1, this->groups_no);
	if (hash_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	//first partition the tuples given by IMJ into groups
	for (int i = 0; i < newResults->numRes; i++)
	{
		//compute the position in the hash table
		unsigned int index = hash_func->Compute_Hash(&newResults->resGroups[i], this->group_order);

		//check to see if there is space in the hash table
		if (group_part[index] == NULL)
		{
			//in this case the position is free, so just add everything in the hash tables
			group_part[index] = new (std::nothrow) Record();
			if (group_part[index] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
			group_part[index]->Consume(&newResults->resGroups[i]);
			group_add[index].push_back(i);
		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = this->eq_func->Compare(&newResults->resGroups[i], group_part[index], this->group_order);
			if (are_identical == 0)
			{
				//the records are identical, so add the index
				group_add[index].push_back(i);
			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % this->groups_no; (group_part[j] != NULL) && (j != index); j = (j + 1) % this->groups_no)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = this->eq_func->Compare(&newResults->resGroups[i], group_part[j], this->group_order);
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					group_add[j].push_back(i);
				}
				else
				{
					if (group_part[j] == NULL)
					{
						//in this case the position is free, so just add everything in the hash tables
						group_part[j] = new (std::nothrow) Record();
						if (group_part[j] == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						group_part[j]->Consume(&newResults->resGroups[i]);
						group_add[j].push_back(i);
					}
					else
					{
						//in this case the hash is full; we simply ignore the extra groups
					}
				}
			}
		}
	}

	for (int i = 0; i < newResults->numDel; i++)
	{
		//compute the position in the hash table
		unsigned int index = hash_func->Compute_Hash(&newResults->delGroups[i], this->group_order);

		//check to see if there is space in the hash table
		if (group_part[index] == NULL)
		{
			//in this case the position is free, so just add everything in the hash tables
			group_part[index] = new (std::nothrow) Record();
			if (group_part[index] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
			group_part[index]->Consume(&newResults->delGroups[i]);
			group_del[index].push_back(i);
		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = this->eq_func->Compare(&newResults->delGroups[i], group_part[index], this->group_order);
			if (are_identical == 0)
			{
				//the records are identical, so add the index
				group_del[index].push_back(i);
			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % this->groups_no; (group_part[j] != NULL) && (j != index); j = (j + 1) % this->groups_no)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = this->eq_func->Compare(&newResults->delGroups[i], group_part[j], this->group_order);
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					group_del[j].push_back(i);
				}
				else
				{
					if (group_part[j] == NULL)
					{
						//in this case the position is free, so just add everything in the hash tables
						group_part[j] = new (std::nothrow) Record();
						if (group_part[j] == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						group_part[j]->Consume(&newResults->delGroups[i]);
						group_del[j].push_back(i);
					}
					else
					{
						//in this case the hash is full; we simply ignore the extra groups
					}
				}
			}
		}
	}


	//the partitions are created at this point
	//next step is to decide for which ones we need estimators
	for (int i = 0; i < this->groups_no; i++)
	{
		if (group_part[i] != NULL)
		{
			//first check if we have an estimator for the group at the current level
			int found = 0;
			int index;
	
			for (int j = 0; j < level_groups_no[current_level]; j++)
			{
				if (eq_func->Compare(group_part[i], (group_ids[current_level])[j], group_order) == 0)
				{
					//group found
					found = 1;
					index = j;
	
					break;
				}
			}
	
			if (found == 1)
			{
				//group exists; do the updates
				//create a new SearchResults structure for this group
				SearchResults *group_res = new (std::nothrow) SearchResults;
				if (group_res == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				group_res->numConst = newResults->numConst;
	
				group_res->constRels = new (std::nothrow) int[group_res->numConst];
				if (group_res->constRels == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
				memcpy(group_res->constRels, newResults->constRels, group_res->numConst * sizeof(int));
	
				group_res->ratios = new (std::nothrow) double[group_res->numConst];
				if (group_res->ratios == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
				memcpy(group_res->ratios, newResults->ratios, group_res->numConst * sizeof(double));
	
	
				group_res->numRes = group_add[i].size();
	
				group_res->funcRes = new (std::nothrow) double[group_res->numRes];
				if (group_res->funcRes == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				group_res->resTuples = new (std::nothrow) int*[group_res->numRes];
				if (group_res->resTuples == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				for (int j = 0; j < group_res->numRes; j++)
				{
					group_res->funcRes[j] = newResults->funcRes[(group_add[i])[j]];
	
					group_res->resTuples[j] = new (std::nothrow) int[group_res->numConst];
					if (group_res->resTuples[j] == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
					memcpy(group_res->resTuples[j], newResults->resTuples[(group_add[i])[j]], group_res->numConst * sizeof(int));
				}
	
	
				group_res->numDel = group_del[i].size();
	
				group_res->funcDel = new (std::nothrow) double[group_res->numDel];
				if (group_res->funcDel == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				group_res->delTuples = new (std::nothrow) int*[group_res->numDel];
				if (group_res->delTuples == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				for (int j = 0; j < group_res->numDel; j++)
				{
					group_res->funcDel[j] = newResults->funcDel[(group_del[i])[j]];
	
					group_res->delTuples[j] = new (std::nothrow) int[group_res->numConst];
					if (group_res->delTuples[j] == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
					memcpy(group_res->delTuples[j], newResults->delTuples[(group_del[i])[j]], group_res->numConst * sizeof(int));
				}
	
				(stat_module[current_level])[index]->ProcessTuples(group_res, p);
	
				delete group_res;
			}
	
			else if (found == 0)
			{
				//check to see if there is still space for new groups at the current level
				if (level_groups_no[current_level] < groups_no)
				{
					//in this case we have to check it this group already exists at a previous level
					//if it exists, we have to move it at the current level
					//if it does not exist, we have to create it at the current level
					int group_found = 0;

					for (int j = current_level - 1; j >= 0; j--)
					{
						for (int k = 0; k < level_groups_no[j]; k++)
						{
							if (((group_ids[j])[k] != NULL) && (eq_func->Compare(group_part[i], (group_ids[j])[k], group_order) == 0))
							{
								//group found
								group_found = 1;
				
								(group_ids[current_level]).push_back((group_ids[j])[k]);
								(group_ids[j])[k] = NULL;

								(stat_module[j])[k]->AddGeneralLevel(cur_level_rels_no, is_last_level);

								//create a new SearchResults structure for this group
								SearchResults *group_res = new (std::nothrow) SearchResults;
								if (group_res == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								group_res->numConst = newResults->numConst;
					
								group_res->constRels = new (std::nothrow) int[group_res->numConst];
								if (group_res->constRels == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
								memcpy(group_res->constRels, newResults->constRels, group_res->numConst * sizeof(int));
					
								group_res->ratios = new (std::nothrow) double[group_res->numConst];
								if (group_res->ratios == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
								memcpy(group_res->ratios, newResults->ratios, group_res->numConst * sizeof(double));
					
					
								group_res->numRes = group_add[i].size();
					
								group_res->funcRes = new (std::nothrow) double[group_res->numRes];
								if (group_res->funcRes == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								group_res->resTuples = new (std::nothrow) int*[group_res->numRes];
								if (group_res->resTuples == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								for (int l = 0; l < group_res->numRes; l++)
								{
									group_res->funcRes[l] = newResults->funcRes[(group_add[i])[l]];
					
									group_res->resTuples[l] = new (std::nothrow) int[group_res->numConst];
									if (group_res->resTuples[l] == NULL)
									{
										cout << "ERROR : Not enough memory. EXIT !!!\n";
										exit(1);
									}
									memcpy(group_res->resTuples[l], newResults->resTuples[(group_add[i])[l]], group_res->numConst * sizeof(int));
								}
					
					
								group_res->numDel = group_del[i].size();
					
								group_res->funcDel = new (std::nothrow) double[group_res->numDel];
								if (group_res->funcDel == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								group_res->delTuples = new (std::nothrow) int*[group_res->numDel];
								if (group_res->delTuples == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								for (int l = 0; l < group_res->numDel; l++)
								{
									group_res->funcDel[l] = newResults->funcDel[(group_del[i])[l]];
					
									group_res->delTuples[j] = new (std::nothrow) int[group_res->numConst];
									if (group_res->delTuples[l] == NULL)
									{
										cout << "ERROR : Not enough memory. EXIT !!!\n";
										exit(1);
									}
									memcpy(group_res->delTuples[l], newResults->delTuples[(group_del[i])[l]], group_res->numConst * sizeof(int));
								}
					
								(stat_module[j])[k]->ProcessTuples(group_res, p);
					
								delete group_res;

								(stat_module[current_level]).push_back((stat_module[j])[k]);
								(stat_module[j])[k] = NULL;

								level_groups_no[current_level] += 1;

								break;
							}
						}

						if (group_found == 1)
							break;
					}

					if (group_found == 0)
					{
						//groups has to be added
						Record *new_group_rec = new (std::nothrow) Record();
						if (new_group_rec == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						new_group_rec->Consume(group_part[i]);
						(group_ids[current_level]).push_back(new_group_rec);

						Statistics *new_stat_group = new (std::nothrow) Statistics(tuples_stable, var_estimates_needed);
						if (new_stat_group == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						new_stat_group->AddGeneralLevel(cur_level_rels_no, is_last_level);


						//create a new SearchResults structure for this group
						SearchResults *group_res = new (std::nothrow) SearchResults;
						if (group_res == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						group_res->numConst = newResults->numConst;
			
						group_res->constRels = new (std::nothrow) int[group_res->numConst];
						if (group_res->constRels == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						memcpy(group_res->constRels, newResults->constRels, group_res->numConst * sizeof(int));
			
						group_res->ratios = new (std::nothrow) double[group_res->numConst];
						if (group_res->ratios == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						memcpy(group_res->ratios, newResults->ratios, group_res->numConst * sizeof(double));
			
			
						group_res->numRes = group_add[i].size();
			
						group_res->funcRes = new (std::nothrow) double[group_res->numRes];
						if (group_res->funcRes == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						group_res->resTuples = new (std::nothrow) int*[group_res->numRes];
						if (group_res->resTuples == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						for (int j = 0; j < group_res->numRes; j++)
						{
							group_res->funcRes[j] = newResults->funcRes[(group_add[i])[j]];
			
							group_res->resTuples[j] = new (std::nothrow) int[group_res->numConst];
							if (group_res->resTuples[j] == NULL)
							{
								cout << "ERROR : Not enough memory. EXIT !!!\n";
								exit(1);
							}
							memcpy(group_res->resTuples[j], newResults->resTuples[(group_add[i])[j]], group_res->numConst * sizeof(int));
						}
			
			
						group_res->numDel = group_del[i].size();
			
						group_res->funcDel = new (std::nothrow) double[group_res->numDel];
						if (group_res->funcDel == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						group_res->delTuples = new (std::nothrow) int*[group_res->numDel];
						if (group_res->delTuples == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						for (int j = 0; j < group_res->numDel; j++)
						{
							group_res->funcDel[j] = newResults->funcDel[(group_del[i])[j]];
			
							group_res->delTuples[j] = new (std::nothrow) int[group_res->numConst];
							if (group_res->delTuples[j] == NULL)
							{
								cout << "ERROR : Not enough memory. EXIT !!!\n";
								exit(1);
							}
							memcpy(group_res->delTuples[j], newResults->delTuples[(group_del[i])[j]], group_res->numConst * sizeof(int));
						}
			
						new_stat_group->ProcessTuples(group_res, p);
			
						delete group_res;

						(stat_module[current_level]).push_back(new_stat_group);

						level_groups_no[current_level] += 1;
					}
				}
			}
		}
	}


	for (int i = 0; i < this->groups_no; i++)
	{
		delete group_part[i];
		group_part[i] = NULL;

		group_add[i].clear();
		group_del[i].clear();
	}

	delete [] group_part;
	delete [] group_add;
	delete [] group_del;

	delete hash_func;
}


int Estimation_Group::ProcessTuples(SearchResults *newResults, SimpleSamplingData *data)
{
	int res = 0;

	Record **group_part;

	vector<int> *group_add;
	vector<int> *group_del;

	group_part = new (std::nothrow) Record*[this->groups_no];
	if (group_part == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	group_add = new (std::nothrow) vector<int>[this->groups_no];
	if (group_add == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	group_del = new (std::nothrow) vector<int>[this->groups_no];
	if (group_del == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < this->groups_no; i++)
	{
		group_part[i] = NULL;
		group_add[i].clear();
		group_del[i].clear();
	}

	Hash *hash_func = new (std::nothrow) CWHash(this->group_order->numAtts + 1, this->groups_no);
	if (hash_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	//first partition the tuples given by IMJ into groups
	for (int i = 0; i < newResults->numRes; i++)
	{
		//compute the position in the hash table
		unsigned int index = hash_func->Compute_Hash(&newResults->resGroups[i], this->group_order);

		//check to see if there is space in the hash table
		if (group_part[index] == NULL)
		{
			//in this case the position is free, so just add everything in the hash tables
			group_part[index] = new (std::nothrow) Record();
			if (group_part[index] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
			group_part[index]->Consume(&newResults->resGroups[i]);
			group_add[index].push_back(i);
		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = this->eq_func->Compare(&newResults->resGroups[i], group_part[index], this->group_order);
			if (are_identical == 0)
			{
				//the records are identical, so add the index
				group_add[index].push_back(i);
			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % this->groups_no; (group_part[j] != NULL) && (j != index); j = (j + 1) % this->groups_no)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = this->eq_func->Compare(&newResults->resGroups[i], group_part[j], this->group_order);
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					group_add[j].push_back(i);
				}
				else
				{
					if (group_part[j] == NULL)
					{
						//in this case the position is free, so just add everything in the hash tables
						group_part[j] = new (std::nothrow) Record();
						if (group_part[j] == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						group_part[j]->Consume(&newResults->resGroups[i]);
						group_add[j].push_back(i);
					}
					else
					{
						//in this case the hash is full; we simply ignore the extra groups
					}
				}
			}
		}
	}

	for (int i = 0; i < newResults->numDel; i++)
	{
		//compute the position in the hash table
		unsigned int index = hash_func->Compute_Hash(&newResults->delGroups[i], this->group_order);

		//check to see if there is space in the hash table
		if (group_part[index] == NULL)
		{
			//in this case the position is free, so just add everything in the hash tables
			group_part[index] = new (std::nothrow) Record();
			if (group_part[index] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
			group_part[index]->Consume(&newResults->delGroups[i]);
			group_del[index].push_back(i);
		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = this->eq_func->Compare(&newResults->delGroups[i], group_part[index], this->group_order);
			if (are_identical == 0)
			{
				//the records are identical, so add the index
				group_del[index].push_back(i);
			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % this->groups_no; (group_part[j] != NULL) && (j != index); j = (j + 1) % this->groups_no)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = this->eq_func->Compare(&newResults->delGroups[i], group_part[j], this->group_order);
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					group_del[j].push_back(i);
				}
				else
				{
					if (group_part[j] == NULL)
					{
						//in this case the position is free, so just add everything in the hash tables
						group_part[j] = new (std::nothrow) Record();
						if (group_part[j] == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						group_part[j]->Consume(&newResults->delGroups[i]);
						group_del[j].push_back(i);
					}
					else
					{
						//in this case the hash is full; we simply ignore the extra groups
					}
				}
			}
		}
	}


	//the partitions are created at this point
	//next step is to decide for which ones we need estimators
	int modified_no = level_groups_no[current_level];
	int *modified = new (std::nothrow) int[modified_no];
	if (modified == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	for (int j = 0; j < modified_no; j++)
	{
		modified[j] = 0;
	}

	for (int i = 0; i < this->groups_no; i++)
	{
		if (group_part[i] != NULL)
		{
			//first check if we have an estimator for the group at the current level
			int found = 0;
			int index;
	
			for (int j = 0; j < level_groups_no[current_level]; j++)
			{
				if (eq_func->Compare(group_part[i], (group_ids[current_level])[j], group_order) == 0)
				{
					//group found
					found = 1;
					index = j;
	
					break;
				}
			}
	
			if (found == 1)
			{
				//group exists; do the updates
				//create a new SearchResults structure for this group
				SearchResults *group_res = new (std::nothrow) SearchResults;
				if (group_res == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				group_res->numConst = newResults->numConst;
	
				group_res->numRes = group_add[i].size();
	
				group_res->funcRes = new (std::nothrow) double[group_res->numRes];
				if (group_res->funcRes == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				group_res->resTuples = new (std::nothrow) int*[group_res->numRes];
				if (group_res->resTuples == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				for (int j = 0; j < group_res->numRes; j++)
				{
					group_res->funcRes[j] = newResults->funcRes[(group_add[i])[j]];
	
					group_res->resTuples[j] = new (std::nothrow) int[group_res->numConst];
					if (group_res->resTuples[j] == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
					memcpy(group_res->resTuples[j], newResults->resTuples[(group_add[i])[j]], group_res->numConst * sizeof(int));
				}
	
				(stat_module[current_level])[index]->ProcessTuples(group_res->resTuples, group_res->funcRes, group_res->numRes, data);
	
				delete group_res;

				modified[index] = 1;
			}
	
			else if (found == 0)
			{
				//check to see if there is still space for new groups at the current level
				if (level_groups_no[current_level] < groups_no)
				{
					//in this case we have to check it this group already exists at a previous level
					//if it exists, we have to move it at the current level
					//if it does not exist, we have to create it at the current level
					int group_found = 0;

					for (int j = current_level - 1; j >= 0; j--)
					{
						for (int k = 0; k < level_groups_no[j]; k++)
						{
							if (((group_ids[j])[k] != NULL) && (eq_func->Compare(group_part[i], (group_ids[j])[k], group_order) == 0))
							{
								//group found
								group_found = 1;
				
								(group_ids[current_level]).push_back((group_ids[j])[k]);
								(group_ids[j])[k] = NULL;

								(stat_module[j])[k]->AddGeneralLevel(cur_level_rels_no, is_last_level);

								//create a new SearchResults structure for this group
								SearchResults *group_res = new (std::nothrow) SearchResults;
								if (group_res == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								group_res->numConst = newResults->numConst;
										
								group_res->numRes = group_add[i].size();
					
								group_res->funcRes = new (std::nothrow) double[group_res->numRes];
								if (group_res->funcRes == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								group_res->resTuples = new (std::nothrow) int*[group_res->numRes];
								if (group_res->resTuples == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								for (int l = 0; l < group_res->numRes; l++)
								{
									group_res->funcRes[l] = newResults->funcRes[(group_add[i])[l]];
					
									group_res->resTuples[l] = new (std::nothrow) int[group_res->numConst];
									if (group_res->resTuples[l] == NULL)
									{
										cout << "ERROR : Not enough memory. EXIT !!!\n";
										exit(1);
									}
									memcpy(group_res->resTuples[l], newResults->resTuples[(group_add[i])[l]], group_res->numConst * sizeof(int));
								}
					
					
								(stat_module[j])[k]->ProcessTuples(group_res->resTuples, group_res->funcRes, group_res->numRes, data);
					
								delete group_res;

								(stat_module[current_level]).push_back((stat_module[j])[k]);
								(stat_module[j])[k] = NULL;

								level_groups_no[current_level] += 1;

								break;
							}
						}

						if (group_found == 1)
							break;
					}

					if (group_found == 0)
					{
						//groups has to be added
						Record *new_group_rec = new (std::nothrow) Record();
						if (new_group_rec == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						new_group_rec->Consume(group_part[i]);
						(group_ids[current_level]).push_back(new_group_rec);

						Statistics *new_stat_group = new (std::nothrow) Statistics(tuples_stable, var_estimates_needed);
						if (new_stat_group == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						new_stat_group->AddLevel(cur_level_rels_no, is_last_level);


						//create a new SearchResults structure for this group
						SearchResults *group_res = new (std::nothrow) SearchResults;
						if (group_res == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						group_res->numConst = newResults->numConst;
			
						group_res->numRes = group_add[i].size();
			
						group_res->funcRes = new (std::nothrow) double[group_res->numRes];
						if (group_res->funcRes == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						group_res->resTuples = new (std::nothrow) int*[group_res->numRes];
						if (group_res->resTuples == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						for (int j = 0; j < group_res->numRes; j++)
						{
							group_res->funcRes[j] = newResults->funcRes[(group_add[i])[j]];
			
							group_res->resTuples[j] = new (std::nothrow) int[group_res->numConst];
							if (group_res->resTuples[j] == NULL)
							{
								cout << "ERROR : Not enough memory. EXIT !!!\n";
								exit(1);
							}
							memcpy(group_res->resTuples[j], newResults->resTuples[(group_add[i])[j]], group_res->numConst * sizeof(int));
						}
			
			
						new_stat_group->ProcessTuples(group_res->resTuples, group_res->funcRes, group_res->numRes, data);
			
						delete group_res;

						(stat_module[current_level]).push_back(new_stat_group);

						level_groups_no[current_level] += 1;
					}
				}
			}
		}
	}


	for (int j = 0; j < modified_no; j++)
	{
		if (modified[j]== 0)
		{
			(stat_module[current_level])[j]->ProcessTuples(NULL, NULL, 0, data);
		}
	}

	for (int i = 0; i < this->groups_no; i++)
	{
		delete group_part[i];
		group_part[i] = NULL;

		group_add[i].clear();
		group_del[i].clear();
	}

	delete [] modified;

	delete [] group_part;
	delete [] group_add;
	delete [] group_del;

	delete hash_func;

	return res;
}


int Estimation_Group::ProcessTuples(double *funcVals, Record *groups_id, int numTuples, SimpleSamplingData *data)
{
	int res = 0;

	Record **group_part;

	vector<int> *group_add;

	group_part = new (std::nothrow) Record*[this->groups_no];
	if (group_part == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	group_add = new (std::nothrow) vector<int>[this->groups_no];
	if (group_add == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < this->groups_no; i++)
	{
		group_part[i] = NULL;
		group_add[i].clear();
	}

	Hash *hash_func = new (std::nothrow) CWHash(this->group_order->numAtts + 1, this->groups_no);
	if (hash_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	//first partition the tuples given by IMJ into groups
	for (int i = 0; i < numTuples; i++)
	{
		//compute the position in the hash table
		unsigned int index = hash_func->Compute_Hash(&groups_id[i], this->group_order);

		//check to see if there is space in the hash table
		if (group_part[index] == NULL)
		{
			//in this case the position is free, so just add everything in the hash tables
			group_part[index] = new (std::nothrow) Record();
			if (group_part[index] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
			group_part[index]->Consume(&groups_id[i]);
			group_add[index].push_back(i);
		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = this->eq_func->Compare(&groups_id[i], group_part[index], this->group_order);
			if (are_identical == 0)
			{
				//the records are identical, so add the index
				group_add[index].push_back(i);
			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % this->groups_no; (group_part[j] != NULL) && (j != index); j = (j + 1) % this->groups_no)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = this->eq_func->Compare(&groups_id[i], group_part[j], this->group_order);
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					group_add[j].push_back(i);
				}
				else
				{
					if (group_part[j] == NULL)
					{
						//in this case the position is free, so just add everything in the hash tables
						group_part[j] = new (std::nothrow) Record();
						if (group_part[j] == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						group_part[j]->Consume(&groups_id[i]);
						group_add[j].push_back(i);
					}
					else
					{
						//in this case the hash is full; we simply ignore the extra groups
					}
				}
			}
		}
	}

	//the partitions are created at this point
	//next step is to decide for which ones we need estimators
	for (int i = 0; i < this->groups_no; i++)
	{
		if (group_part[i] != NULL)
		{
			//first check if we have an estimator for the group at the current level
			int found = 0;
			int index;
	
			for (int j = 0; j < level_groups_no[current_level]; j++)
			{
				if (eq_func->Compare(group_part[i], (group_ids[current_level])[j], group_order) == 0)
				{
					//group found
					found = 1;
					index = j;
	
					break;
				}
			}
	
			if (found == 1)
			{
				//group exists; do the updates
				//create a new group aggregate structure for this group
				int group_size = group_add[i].size();

				double *group_aggregate = new (std::nothrow) double[group_size];
				if (group_aggregate == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
	
				for (int j = 0; j < group_size; j++)
				{
					group_aggregate[j] = funcVals[(group_add[i])[j]];
				}
	
				(stat_module[current_level])[index]->ProcessTuples(group_aggregate, group_size, data);
	
				delete group_aggregate;
			}
	
			else if (found == 0)
			{
				//check to see if there is still space for new groups at the current level
				if (level_groups_no[current_level] < groups_no)
				{
					//in this case we have to check it this group already exists at a previous level
					//if it exists, we have to move it at the current level
					//if it does not exist, we have to create it at the current level
					int group_found = 0;

					for (int j = current_level - 1; j >= 0; j--)
					{
						for (int k = 0; k < level_groups_no[j]; k++)
						{
							if (((group_ids[j])[k] != NULL) && (eq_func->Compare(group_part[i], (group_ids[j])[k], group_order) == 0))
							{
								//group found
								group_found = 1;
				
								(group_ids[current_level]).push_back((group_ids[j])[k]);
								(group_ids[j])[k] = NULL;

								(stat_module[j])[k]->AddGeneralLevel(cur_level_rels_no, is_last_level);

								//create a new SearchResults structure for this group
								int group_size = group_add[i].size();
				
								double *group_aggregate = new (std::nothrow) double[group_size];
								if (group_aggregate == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}
					
								for (int l = 0; l < group_size; l++)
								{
									group_aggregate[l] = funcVals[(group_add[i])[l]];
								}
					
								(stat_module[j])[k]->ProcessTuples(group_aggregate, group_size, data);
					
								delete group_aggregate;

								(stat_module[current_level]).push_back((stat_module[j])[k]);
								(stat_module[j])[k] = NULL;

								level_groups_no[current_level] += 1;

								break;
							}
						}

						if (group_found == 1)
							break;
					}

					if (group_found == 0)
					{
						//groups has to be added
						Record *new_group_rec = new (std::nothrow) Record();
						if (new_group_rec == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						new_group_rec->Consume(group_part[i]);
						(group_ids[current_level]).push_back(new_group_rec);

						Statistics *new_stat_group = new (std::nothrow) Statistics(tuples_stable, var_estimates_needed);
						if (new_stat_group == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						new_stat_group->AddGeneralLevel(cur_level_rels_no, is_last_level);


						//create a new SearchResults structure for this group
						int group_size = group_add[i].size();
		
						double *group_aggregate = new (std::nothrow) double[group_size];
						if (group_aggregate == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
			
						for (int j = 0; j < group_size; j++)
						{
							group_aggregate[j] = funcVals[(group_add[i])[j]];
						}
			
						new_stat_group->ProcessTuples(group_aggregate, group_size, data);
			
						delete group_aggregate;

						(stat_module[current_level]).push_back(new_stat_group);

						level_groups_no[current_level] += 1;
					}
				}
			}
		}
	}


	for (int i = 0; i < this->groups_no; i++)
	{
		delete group_part[i];
		group_part[i] = NULL;

		group_add[i].clear();
	}

	delete [] group_part;
	delete [] group_add;

	delete hash_func;

	return res;
}


void Estimation_Group::Print()
{
}


void Estimation_Group::Print_Level()
{
}

