#ifndef _ESTIMATION
#define _ESTIMATION

#include "Statistics.h"
#include "Record.h"
#include "Comparison.h"
#include "Hashes.h"

#include <vector>


using namespace std;


class Estimation {
public:
	Estimation();
	virtual ~Estimation();

	virtual void AddLevel(int numRels, int isLastLevel) = 0;
	virtual void AddGeneralLevel(int numRels, int isLastLevel) = 0;

	virtual int GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel, double p) = 0;
	virtual int GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel) = 0;

	virtual double WhatIfVariance(double *proposedSamplingRatios) = 0;

	virtual void ProcessTuples(SearchResults *newResults, double p) = 0;
	virtual int ProcessTuples(SearchResults *newResults, SimpleSamplingData *data) = 0;
	// this version works for the top-most level, where no tuples need to be passed
	virtual int ProcessTuples(double *funcVals, Record *groups_id, int numTuples, SimpleSamplingData *data) = 0;

	virtual void Print() = 0;
	virtual void Print_Level() = 0;
};



class Estimation_SPJ : public Estimation {
public:
	Estimation_SPJ(int tuples_stable, int var_estimates_needed);
	virtual ~Estimation_SPJ();

	virtual void AddLevel(int numRels, int isLastLevel);
	virtual void AddGeneralLevel(int numRels, int isLastLevel);

	virtual int GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel, double p);
	virtual int GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel);

	virtual double WhatIfVariance(double *proposedSamplingRatios);

	virtual void ProcessTuples(SearchResults *newResults, double p);
	virtual int ProcessTuples(SearchResults *newResults, SimpleSamplingData *data);
	virtual int ProcessTuples(double *funcVals, Record *groups_id, int numTuples, SimpleSamplingData *data);

	virtual void Print();
	virtual void Print_Level();

private:
	Statistics *stat_module;

};



class Estimation_Group : public Estimation {
public:
	Estimation_Group(int tuples_stable, int var_estimates_needed, int levels_no, int groups_no, Schema *group_schema);
	virtual ~Estimation_Group();

	virtual void AddLevel(int numRels, int isLastLevel);
	virtual void AddGeneralLevel(int numRels, int isLastLevel);

	virtual int GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel, double p);
	virtual int GetEst(double *&estimate, double *&variance, double *&low, double *&high, Record *&groups_id, int &estimates_no, double confLevel);

	virtual double WhatIfVariance(double *proposedSamplingRatios);

	virtual void ProcessTuples(SearchResults *newResults, double p);
	virtual int ProcessTuples(SearchResults *newResults, SimpleSamplingData *data);
	virtual int ProcessTuples(double *funcVals, Record *groups_id, int numTuples, SimpleSamplingData *data);

	virtual void Print();
	virtual void Print_Level();

private:
	int tuples_stable;
	int var_estimates_needed;

	int levels_no;

	int groups_no;

	int *level_groups_no;

	int current_level;
	int cur_level_rels_no;
	int is_last_level;

	Schema *group_schema;

	OrderMaker *group_order;
	Hash *eq_func;

	vector<Statistics*> *stat_module;
	vector<Record*> *group_ids;
};


#endif


