#ifndef EXTENDEDHASH_H
#define EXTENDEDHASH_H

#include "DEFS.h"
#include "Record.h"
#include "Comparison.h"
#include "Schema.h"
#include "Hashes.h"
#include <vector>
#include "TwoWayList.h"

#define MAXBUCKET 2147483647 //2^31
#define EXTENDEDHASHCONST  100

using namespace std;

//this is the class defined for record array
//the class maintains an array of record points 
//along with their time stamp(timestamp array if they are join results)
//some extra information might be added later

class RecordArray{

public:
	int timeStampSize;//size of timestamp array
	


    vector <double *> ratios;

    vector <int *> timeStamps;//store all timestamps

    vector <Record *> allRecords;//store all records

    RecordArray (int size);

    RecordArray (RecordArray& myArray);	

    RecordArray ();	

    ~RecordArray();

    // in the following functions, return the index when success -1 otherwise

    void SetTimeStampSize(int size);

    //this returns the number of records in the array

    int GetNumOfRecs();

	//the following functions add a record and possibly a set of timestamps with the record
	//into the array the first parameter is the record, the following are timestamps
    void AddRecord(Record& addMe);
	void AddRecord(Record& addMe, int *times, double *ratios);

    //this is provided for convience of merging records
	void AddRecord(Record& addMe, int * times, int extraTime, double *ratios, double extraRatio);
	void Print(int index, Schema& mySchema);

    //obtain the number of bytes occupied.
	double GetNumOfBytes();
};

//This is an int wrapper used to store indices for extended hashing
class IntWrapper{

public:

	int value;
	
	IntWrapper (int inputVal);

	IntWrapper ();
	
	void Consume (IntWrapper* fromMe);

	void Copy (IntWrapper* fromMe);

};

class ExtendedHash{

private:

    struct HashPair{

       unsigned int hashValue;

       TwoWayList<IntWrapper> Indices;	
	};
	

	vector<HashPair*> ** AllIndices;	
 

    // we first take a look at the last 6 bits, initial to 64;
	unsigned int size;
	
   //the total number of distinct keys so far
	int totalDistinctKeys;

    // for hashing	
	CWHash *myHashFunc;	

    OrderMaker myOrderMaker;

    void Expand();//increase the looking of number of bits;
 
   
public :
	
    ExtendedHash(OrderMaker & fromMe, CWHash* theHashFunc);

    ExtendedHash(OrderMaker & fromMe, CWHash* theHashFUnc, int theSize);	

    ExtendedHash(ExtendedHash &myHashes);

    ~ExtendedHash();

    void AddRecIndex(Record& fromMe, int myIndex);

    //this function returns the pointer to the list that
       //contains all hashed indices. return 0 if succeed and -1
	//if no key is found
	int Retrieve(unsigned int key, TwoWayList<IntWrapper> ** pIndex);
   void DeleteIndex(Record& fromMe, int myIndex);	
    void Print();
};
#endif
