#include "File.h"
#include "TwoWayList.cc"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <iostream.h>



//path prefixes for dealing with stripped files
char *prefixes[NUM_DISKS] = {"data1/", "data2/", "data3/", "data4/"};



Page :: Page () {
	curSizeInBytes = sizeof (int);
	numRecs = 0;

	myRecs = new (std::nothrow) TwoWayList<Record>;
	if (myRecs == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}

Page :: ~Page () {
	delete myRecs;
}


void Page :: EmptyItOut () {
	// get rid of all of the records
	while (1) {
		Record temp;
		if (!GetFirst (&temp))
		{
			break;
		}
	}

	// reset the page size
	curSizeInBytes = sizeof (int);
	numRecs = 0;
}


int Page :: GetFirst (Record *firstOne) {
	// move to the first record
	myRecs->MoveToStart ();

	// make sure there is data
	if (!myRecs->RightLength ()) {
		return 0;
	}

	// and remove it
	myRecs->Remove (firstOne);
	numRecs--;

	char *b = firstOne->GetBits();
	curSizeInBytes -= ((int *) b)[0];

	return 1;
}


int Page :: Append (Record *addMe) {
	char *b = addMe->GetBits();

	// first see if we can fit the record
	if (curSizeInBytes + ((int *) b)[0] > PAGE_SIZE) {
		return 0;
	}

	// move to the last record
	myRecs->MoveToFinish ();

	// and add it
	curSizeInBytes += ((int *) b)[0];
	myRecs->Insert(addMe);
	numRecs++;

	return 1;
}


void Page :: ToBinary (char *bits) {
	// first write the number of records on the page
	((int *) bits)[0] = numRecs;

	char *curPos = bits + sizeof (int);

	// and copy the records one-by-one
	myRecs->MoveToStart ();
	for (int i = 0; i < numRecs; i++) {	
		char *b = myRecs->Current(0)->GetBits();
		
		// copy over the bits of the current record
		memcpy (curPos, b, ((int *) b)[0]);
		curPos += ((int *) b)[0];

		// and traverse the list
		myRecs->Advance ();
	}
}


void Page :: FromBinary (char *bits) {
	// first read the number of records on the page
	numRecs = ((int *) bits)[0];

	// sanity check
	if (numRecs < 0) {
		cerr << "This is probably an error.  Found " << numRecs << " records on a page.\n";
		exit (1);
	}

	// first, empty out the list of current records
	// now loop through and re-populate it
	myRecs->MoveToStart ();
	while (myRecs->RightLength ()) {
		Record *temp = new (std::nothrow) Record();
		if (temp == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		myRecs->Remove(temp);

		delete temp;
	}

	// and now get the binary representations of each
	char *curPos = bits + sizeof (int);
	curSizeInBytes = sizeof (int);

	for (int i = 0; i < numRecs; i++) {
		Record *temp = new (std::nothrow) Record();
		if (temp == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		// get the length of the current record
		int len = ((int *) curPos)[0];
		curSizeInBytes += len;

		// create the record
		temp->CopyBits(curPos, len);

		// add it
		myRecs->Insert(temp);

		// and move along
		myRecs->Advance ();
		curPos += len;

		delete temp;
	}
}




C_Page :: C_Page () {
}

C_Page :: ~C_Page () {
}




C_Page_INT :: C_Page_INT () {
	numItems = 0;
	curSizeInBytes = sizeof(int);

	elem = new (std::nothrow) deque<char*>;
	if (elem == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}

C_Page_INT :: ~C_Page_INT () {
	int n = elem->size();
	for (int i = 0; i < n; i++)
	{
		char *p = elem->front();
		delete [] p;

		(*elem)[0] = NULL;
		elem->pop_front();
	}

	elem->clear();
	delete elem;
}


void C_Page_INT :: EmptyItOut () {
	// reset the page size
	numItems = 0;
	curSizeInBytes = sizeof (int);

	int n = elem->size();
	for (int i = 0; i < n; i++)
	{
		char *p = elem->front();
		delete [] p;

		(*elem)[0] = NULL;
		elem->pop_front();
	}
	elem->clear();
}


int C_Page_INT :: GetFirst (char *&first) {
	// make sure there is data
	if (!(numItems > 0)) {
		return 0;
	}

	first = new (std::nothrow) char[3 * sizeof(int)];
	if (first == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	// and remove it
	char *p = elem->front();

	((int*)first)[0] = 3 * sizeof(int);
	((int*)first)[1] = sizeof(int) << 1;
	((int*)first)[2] = *(int*)p;

	delete [] p;
	(*elem)[0] = NULL;
	elem->pop_front();

	numItems -= 1;
	curSizeInBytes -= sizeof(int);

	return 1;
}


int C_Page_INT :: Append (char *&addMe) {
	// first see if we can fit the record
	if (curSizeInBytes + sizeof(int) > PAGE_SIZE) {
		return 0;
	}

	// and add it
	char *n_int = new (std::nothrow) char[sizeof(int)];
	if (n_int == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	
	memcpy(n_int, addMe + (sizeof(int) << 1), sizeof(int));
	elem->push_back(n_int);

	delete [] addMe;
	addMe = NULL;

	curSizeInBytes += sizeof(int);
	numItems += 1;

	return 1;
}


void C_Page_INT :: ToBinary (char *bits) {
	// first write the number of records on the page
	*(int*)bits = numItems;

	char *curPos = bits + sizeof (int);

	// and copy the records one-by-one
	for (int i = 0; i < numItems; i++) {
		int p = *(int*)((*elem)[i]);
		*(int*)curPos = p;
		
		curPos += sizeof(int);
	}
}


void C_Page_INT :: FromBinary (char *bits) {
	EmptyItOut();

	// first read the number of records on the page
	numItems = *(int*)bits;
	if (numItems < 0) {
		cerr << "This is probably an error.  Found " << numItems << " records on a page.\n";
		exit (1);
	}

	// and now get the binary representations of each
	char *curPos = bits + sizeof (int);
	for (int i = 0; i < numItems; i++) {
		char *first = new (std::nothrow) char[sizeof(int)];
		if (first == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		memcpy(first, curPos, sizeof(int));
		elem->push_back(first);

		curSizeInBytes += sizeof(int);

		curPos += sizeof(int);
	}
}




C_Page_DOUBLE :: C_Page_DOUBLE () {
	numItems = 0;
	curSizeInBytes = sizeof(int);

	elem = new (std::nothrow) deque<char*>;
	if (elem == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}

C_Page_DOUBLE :: ~C_Page_DOUBLE () {
	int n = elem->size();
	for (int i = 0; i < n; i++)
	{
		char *p = elem->front();
		delete [] p;

		(*elem)[0] = NULL;
		elem->pop_front();
	}

	elem->clear();
	delete elem;
}


void C_Page_DOUBLE :: EmptyItOut () {
	// reset the page size
	numItems = 0;
	curSizeInBytes = sizeof (int);

	int n = elem->size();
	for (int i = 0; i < n; i++)
	{
		char *p = elem->front();
		delete [] p;

		(*elem)[0] = NULL;
		elem->pop_front();
	}
	elem->clear();
}


int C_Page_DOUBLE :: GetFirst (char *&first) {
	// make sure there is data
	if (!(numItems > 0)) {
		return 0;
	}

	first = new (std::nothrow) char[(sizeof(int) << 1) + sizeof(double)];
	if (first == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	// and remove it
	char *p = elem->front();

	((int*)first)[0] = (sizeof(int) << 1) + sizeof(double);
	((int*)first)[1] = sizeof(int) << 1;
	memcpy(first + (sizeof(int) << 1), p, sizeof(double));

	delete [] p;
	(*elem)[0] = NULL;
	elem->pop_front();

	numItems -= 1;
	curSizeInBytes -= sizeof(double);

	return 1;
}


int C_Page_DOUBLE :: Append (char *&addMe) {
	// first see if we can fit the record
	if (curSizeInBytes + sizeof(double) > PAGE_SIZE) {
		return 0;
	}

	// and add it
	char *n_int = new (std::nothrow) char[sizeof(double)];
	if (n_int == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	
	memcpy(n_int, addMe + (sizeof(int) << 1), sizeof(double));
	elem->push_back(n_int);

	delete [] addMe;
	addMe = NULL;

	curSizeInBytes += sizeof(double);
	numItems += 1;

	return 1;
}


void C_Page_DOUBLE :: ToBinary (char *bits) {
	// first write the number of records on the page
	*(int*)bits = numItems;

	char *curPos = bits + sizeof (int);

	// and copy the records one-by-one
	for (int i = 0; i < numItems; i++) {
		double p = *(double*)((*elem)[i]);
		*(double*)curPos = p;
		
		curPos += sizeof(double);
	}
}


void C_Page_DOUBLE :: FromBinary (char *bits) {
	EmptyItOut();

	// first read the number of records on the page
	numItems = *(int*)bits;
	if (numItems < 0) {
		cerr << "This is probably an error.  Found " << numItems << " records on a page.\n";
		exit (1);
	}

	// and now get the binary representations of each
	char *curPos = bits + sizeof (int);
	for (int i = 0; i < numItems; i++) {
		char *first = new (std::nothrow) char[sizeof(double)];
		if (first == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		memcpy(first, curPos, sizeof(double));
		elem->push_back(first);

		curSizeInBytes += sizeof(double);

		curPos += sizeof(double);
	}
}




C_Page_STRING :: C_Page_STRING () {
	numItems = 0;
	curSizeInBytes = sizeof(int) << 1;

	elem = new (std::nothrow) deque<char*>;
	if (elem == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	e_size = new (std::nothrow) deque<int>;
	if (e_size == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}

C_Page_STRING :: ~C_Page_STRING () {
	int n = elem->size();
	for (int i = 0; i < n; i++)
	{
		char *p = elem->front();
		delete [] p;

		(*elem)[0] = NULL;
		elem->pop_front();
	}

	elem->clear();
	delete elem;

	e_size->clear();
	delete e_size;
}


void C_Page_STRING :: EmptyItOut () {
	// reset the page size
	numItems = 0;
	curSizeInBytes = sizeof (int) << 1;

	int n = elem->size();
	for (int i = 0; i < n; i++)
	{
		char *p = elem->front();
		delete [] p;

		(*elem)[0] = NULL;
		elem->pop_front();
	}
	elem->clear();

	e_size->clear();
}


int C_Page_STRING :: GetFirst (char *&first) {
	// make sure there is data
	if (!(numItems > 0)) {
		return 0;
	}

	int s = e_size->front();
	first = new (std::nothrow) char[(sizeof(int) << 1) + s];
	if (first == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	// and remove it
	((int*)first)[0] = (sizeof(int) << 1) + s;
	((int*)first)[1] = sizeof(int) << 1;

	memcpy((char*)((int*)first + 2), (*elem)[0], s);

	delete [] (*elem)[0];
	(*elem)[0] = NULL;
	elem->pop_front();

	e_size->pop_front();

	numItems -= 1;
	curSizeInBytes -= s;

	return 1;
}


int C_Page_STRING :: Append (char *&addMe) {
	// first see if we can fit the record
	// + 1 for the extra '\0' NULL character
	int s = *(int*)addMe - (sizeof(int) << 1);
	if (curSizeInBytes + s + sizeof(int) > PAGE_SIZE) {
		return 0;
	}

	// and add it
	char *n_int = new (std::nothrow) char[s];
	if (n_int == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	
	memcpy(n_int, addMe + (sizeof(int) << 1), s);
	elem->push_back(n_int);
	e_size->push_back(s);

	delete [] addMe;
	addMe = NULL;

	curSizeInBytes += s + sizeof(int);
	numItems += 1;

	return 1;
}


void C_Page_STRING :: ToBinary (char *bits) {
	// first write the number of records on the page
	*(int*)bits = numItems;

	char *curPos = bits + sizeof(int);
	*(int*)curPos = curSizeInBytes;

	curPos += sizeof(int);
	int c_start = sizeof(int) * (numItems + 2);

	// first copy the sizes one-by-one
	for (int i = 0; i < numItems; i++) {
		*(int*)curPos = c_start;

		int p = (*e_size)[i];
		c_start += p;

		curPos += sizeof(int);
	}

	// second copy the actual strings
	for (int i = 0; i < numItems; i++) {
		int s = (*e_size)[i];

		char *p = (*elem)[i];
		memcpy(curPos, p, s);

		curPos += s;
	}
}


void C_Page_STRING :: FromBinary (char *bits) {
	EmptyItOut();

	// first read the number of records on the page
	numItems = *(int*)bits;
	if (numItems < 0) {
		cerr << "This is probably an error.  Found " << numItems << " records on a page.\n";
		exit (1);
	}

	curSizeInBytes = ((int*)bits)[1];

	// first get the sizes
	char *curPos = bits + (sizeof (int) << 1);
	for (int i = 0; i < numItems; i++) {
		int p = *(int*)curPos;
		e_size->push_back(p);

		curPos += sizeof(int);
	}

	// second get the actual data and modify the sizes accordingly
	for (int i = 0; i < numItems - 1; i++) {
		int len = (*e_size)[i+1] - (*e_size)[i];

		char *p = new (std::nothrow) char[len];
		if (p == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		memcpy(p, curPos, len);
		elem->push_back(p);

		(*e_size)[i] = len;

		curPos += len;
	}

	int len = curSizeInBytes - (*e_size)[numItems - 1];
	char *p = new (std::nothrow) char[len];
	if (p == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	memcpy(p, curPos, len);
	elem->push_back(p);

	(*e_size)[numItems - 1] = len;
}




C_Page_RECORD :: C_Page_RECORD () {
	numItems = 0;
	curSizeInBytes = sizeof(int) << 1;

	elem = new (std::nothrow) deque<char*>;
	if (elem == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	e_size = new (std::nothrow) deque<int>;
	if (e_size == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}

C_Page_RECORD :: ~C_Page_RECORD () {
	int n = elem->size();
	for (int i = 0; i < n; i++)
	{
		char *p = elem->front();
		delete [] p;

		(*elem)[0] = NULL;
		elem->pop_front();
	}

	elem->clear();
	delete elem;

	e_size->clear();
	delete e_size;
}


void C_Page_RECORD :: EmptyItOut () {
	// reset the page size
	numItems = 0;
	curSizeInBytes = sizeof (int);

	int n = elem->size();
	for (int i = 0; i < n; i++)
	{
		char *p = elem->front();
		delete [] p;

		(*elem)[0] = NULL;
		elem->pop_front();
	}
	elem->clear();

	e_size->clear();
}


int C_Page_RECORD :: GetFirst (char *&first) {
	// make sure there is data
	if (!(numItems > 0)) {
		return 0;
	}

	int s = (*e_size)[0];

	// and remove it
	first = (*elem)[0];

	elem->pop_front();

	e_size->pop_front();

	numItems -= 1;
	curSizeInBytes -= s;

	return 1;
}


int C_Page_RECORD :: Append (char *&addMe) {
	// first see if we can fit the record
	int s = *(int*)addMe;
	if (curSizeInBytes + s > PAGE_SIZE) {
		return 0;
	}

	// and add it
	elem->push_back(addMe);
	e_size->push_back(s);

	curSizeInBytes += s;
	numItems += 1;

	return 1;
}


void C_Page_RECORD :: ToBinary (char *bits) {
	// first write the number of records on the page
	*(int*)bits = numItems;

	char *curPos = bits + sizeof(int);
	for (int i = 0; i < numItems; i++) {
		int s = (*e_size)[i];

		char *p = (*elem)[i];
		memcpy(curPos, p, s);

		curPos += s;
	}
}


void C_Page_RECORD :: FromBinary (char *bits) {
	EmptyItOut();

	// first read the number of records on the page
	numItems = *(int*)bits;
	if (numItems < 0) {
		cerr << "This is probably an error.  Found " << numItems << " records on a page.\n";
		exit (1);
	}

	// get the sizes and the records
	char *curPos = bits + sizeof (int);
	for (int i = 0; i < numItems; i++) {
		int len = *(int*)curPos;
		e_size->push_back(len);
		curSizeInBytes += len;

		char *p = new (std::nothrow) char[len];
		if (p == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		memcpy(p, curPos, len);
		elem->push_back(p);

		curPos += len;
	}
}






File_Interface :: File_Interface () {
	myBuff = NULL;
}

File_Interface :: ~File_Interface () {
}


off_t File_Interface :: GetLength () {
	return curLength;
}


void File_Interface :: MovePointer (off_t whichPage) {
	// this is done because the very first page does not have any data
	whichPage++;
	
	// see if someone wants to move past the end of the file
	if (whichPage >= curLength) {
		cerr << "Error: you tried to move past the end of the file.\n";
		exit (1);
	}
	
	// write out the current page, if needed
	char *bits = new (std::nothrow) char[PAGE_SIZE];
	if (bits == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	if (changed) {
		off_t tmp = curPos;
		tmp *= PAGE_SIZE;
		Lseek (tmp);
	
		myBuff->ToBinary (bits);
		Write (bits);
		changed = 0;
	}
	
	// read the bits
	off_t tmp = whichPage;
	tmp *= PAGE_SIZE;	
	Lseek (tmp);

	Read (bits);
	// and set up the buffer
	myBuff->FromBinary (bits);
	delete [] bits;

	// set up the current page
	curPos = whichPage;
}



int File_Interface :: NextRec (Record *getMe) {
	// try to remove the record from the buffer
	char *gm_bits = NULL;
	if (myBuff->GetFirst (gm_bits)) {
		getMe->Consume(gm_bits);

		delete [] gm_bits;
		gm_bits = NULL;

		return 1;
	}

	// can't do it, so load the next page
	while (1) {

		// try to move to the next page
		if (curPos + 1 < curLength) {
			MovePointer (curPos);

		} else {
			return 0;
		}

		// now that we are at the next page, try
		// to return the next record
		if (myBuff->GetFirst (gm_bits)) {
			getMe->Consume(gm_bits);

			delete [] gm_bits;
			gm_bits = NULL;

	                return 1;
		}
        }
}


void File_Interface :: Append (Record *addMe) {
	// make sure that we are on the last page
	if (curPos != curLength - 1) {
		MovePointer (curLength - 2);
	}
	
	// try to add the record
	char *am_bits = addMe->GetBits();
	if (myBuff->Append (am_bits)) {
		addMe->Nullify();

		changed = 1;
		return;
	}
	
	// getting here means that we cannot fit the record, so first
	// write the current page to disk
	if (changed) {
		char *bits = new (std::nothrow) char[PAGE_SIZE];
		if (bits == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		myBuff->ToBinary (bits);

		off_t tmp = curPos;
		tmp *= PAGE_SIZE;
		Lseek (tmp);

		Write (bits);

		delete [] bits;
	}

	// note that we have moved along
	curPos = curLength;
	curLength++;
	
	// then do the actual write; first empty out the buffer
	myBuff->EmptyItOut ();
	
	// and add the record
	if (myBuff->Append (am_bits)) {
		addMe->Nullify();

		changed = 1;
		return;
	}

	// error?
	cerr << "You tried to add a record that was too big for the page!\n";
	exit (1);
}



void File_Interface :: GetPage (Page *putItHere, off_t whichPage) {
	// this is because the first page has no data
	whichPage++;
	
	if (whichPage >= curLength) {
		cerr << "BAD: you tried to read past the end of the file\n";
		return;
	}
	
	// read in the specified page
	char *bits = new (std::nothrow) char[PAGE_SIZE];
	if (bits == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	off_t tmp = whichPage;
	tmp *= PAGE_SIZE;
	Lseek (tmp);

	Read (bits);
	putItHere->FromBinary (bits);

	delete [] bits;
}


void File_Interface :: AddPage (Page *addMe, off_t whichPage) {
	// this is because the first page has no data
	whichPage++;
	
	// if we are trying to add past the end of the file, then
	// zero all of the pages out
	// if we are trying to add past the end of the file, then
	// zero all of the pages out
	if (whichPage >= curLength) {
		
		// do the zeroing
		for (off_t i = curLength; i < whichPage; i++) {
			char *bits = new (std::nothrow) char[PAGE_SIZE];
			if (bits == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			off_t tmp = i;
			tmp *= PAGE_SIZE;
			Lseek (tmp);
		
			Write (bits);

			delete [] bits;
		}

		// set the size
		curLength = whichPage + 1;	
	}

	// now write the page
	char *bits = new (std::nothrow) char[PAGE_SIZE];
	if (bits == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	addMe->ToBinary (bits);

	off_t tmp = whichPage;
	tmp *= PAGE_SIZE;
	Lseek (tmp);

	Write (bits);

	delete [] bits;
}






File :: File () {
	path = NULL;

	myFilDes = new (std::nothrow) int;
	if (myFilDes == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	file_type = FT_NONE;
}

File :: ~File () {
	free(path);
	path = NULL;

	delete myFilDes;

	delete myBuff;
}


void File :: Lseek (off_t howFar) {
	lseek (*myFilDes, howFar, SEEK_SET);
}


void File :: Write (char *writeMe) {
	write (*myFilDes, writeMe, PAGE_SIZE);
}


void File :: Read (char *readMe) {
	read (*myFilDes, readMe, PAGE_SIZE);
}



void File :: Open (int fileLen, char *fName, int file_type) {

	// figure out the flags for the system open call
        int mode;
        if (fileLen == 0)
                mode = O_TRUNC | O_RDWR | O_CREAT;
        else
                mode = O_RDWR;

	// actually do the open
        *myFilDes = open (fName, mode, S_IRUSR | S_IWUSR);

	// see if there was an error
	if (*myFilDes < 0) {
		cerr << "BAD!  Open did not work for " << fName << "\n";
		exit (1);
	}

	// read in the buffer if needed
	if (fileLen != 0) {
		char *bits = new (std::nothrow) char[PAGE_SIZE];
		if (bits == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		// read in the first few bits, which is the page size
		Lseek (0);

		Read (bits);
		curLength = *((off_t *) bits);

		if (file_type == FT_NONE)
			this->file_type = *(int *) (bits + sizeof(off_t));
		else
			this->file_type = file_type;

		switch (this->file_type)
		{
			case FT_INT:
				myBuff = new (std::nothrow) C_Page_INT();
				break;
			case FT_REAL:
				myBuff = new (std::nothrow) C_Page_DOUBLE();
				break;
			case FT_STRING:
				myBuff = new (std::nothrow) C_Page_STRING();
				break;
			case FT_RECORD:
				myBuff = new (std::nothrow) C_Page_RECORD();
				break;
			default:
				myBuff = new (std::nothrow) C_Page_RECORD();
				break;
		}

		if (myBuff == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}


		// read in the first page
		Lseek (PAGE_SIZE);

		Read (bits);
		myBuff->FromBinary (bits);

		delete [] bits;
		
		// the buffer is not dirty!
		changed = 0;
	}
	else {
		this->file_type = file_type;

		switch (this->file_type)
		{
			case FT_INT:
				myBuff = new (std::nothrow) C_Page_INT();
				break;
			case FT_REAL:
				myBuff = new (std::nothrow) C_Page_DOUBLE();
				break;
			case FT_STRING:
				myBuff = new (std::nothrow) C_Page_STRING();
				break;
			case FT_RECORD:
				myBuff = new (std::nothrow) C_Page_RECORD();
				break;
			default:
				myBuff = new (std::nothrow) C_Page_RECORD();
				break;
		}

		if (myBuff == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		// otherwise, empty it out
		myBuff->EmptyItOut ();
		curLength = 2;
		
		// the buffer is dirty, since nothing is there!
		changed = 1;
	}

	// and note that we are at the first pos!
	curPos = 1;

	// set up the path
	free(path);
	path = strdup (fName);
}


int File :: Close () {
	// write out the buffer if needed
	char *bits = new (std::nothrow) char[PAGE_SIZE];
	if (bits == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	// write out the current length in pages
	*((off_t *) bits) = curLength;
	*(int *) (bits + sizeof(off_t)) = file_type;


	Lseek (0);

	Write (bits);

	if (changed) {
		off_t tmp = curPos;
		tmp *= PAGE_SIZE;
		Lseek (tmp);

		myBuff->ToBinary (bits);
		Write (bits);
	}
	
	delete [] bits;

	// close the file
	close (*myFilDes);

	// and return the size
	return curLength;
}


void File :: Kill () {
	close (*myFilDes);
	unlink (path);
}






Stripe_File :: Stripe_File () {
	path = NULL;
	
	pthread_mutex_init(&mainMutex, NULL);
	pthread_cond_init(&mainSignal, NULL);
	
	for (int i = 0; i < NUM_DISKS; i++) {
		// initialize this guy's mutex and cond var
		pthread_mutex_init(&(mutexes[i]), NULL);
		pthread_cond_init(&(mySignals[i]), NULL);
	}

	status = new (std::nothrow) int[NUM_DISKS];
	if (status == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	execute = new (std::nothrow) int[NUM_DISKS];
	if (execute == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	myFilDes = new (std::nothrow) int[NUM_DISKS];
	if (myFilDes == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	counter = new (std::nothrow) int;
	if (counter == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	bitPointers = new (std::nothrow) char*[NUM_DISKS];
	if (bitPointers == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	myThreadParam = new (std::nothrow) ThreadParam[NUM_DISKS];
	if (myThreadParam == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}


Stripe_File :: ~Stripe_File () {
	free(path);
	path = NULL;

	delete myBuff;

	pthread_mutex_destroy(&mainMutex);
	pthread_cond_destroy(&mainSignal);
	
	for (int i = 0; i < NUM_DISKS; i++) {
		// initialize this guy's mutex and cond var
		pthread_mutex_destroy(&(mutexes[i]));
		pthread_cond_destroy(&(mySignals[i]));
	}

	delete [] status;
	delete [] execute;
	delete [] myFilDes;
	delete counter;

	delete [] bitPointers;

	delete [] myThreadParam;
}


void Stripe_File :: Lseek (off_t howFar) {
	// this just moves all of the underlying pointers
	for (int i = 0; i < NUM_DISKS; i++) {
		lseek (myFilDes[i], (howFar / NUM_DISKS), SEEK_SET);
	}
}


void Stripe_File :: Write (char *writeMe) {
	pthread_mutex_lock(&mainMutex);

	*counter = 0;
	
	// ask all of the disks to write
	for (int i = 0; i < NUM_DISKS; i++) {
		pthread_mutex_lock (&(mutexes[i]));

		bitPointers[i] = writeMe + i * (PAGE_SIZE / NUM_DISKS);
		status[i] = WRITE;

		execute[i] = 1;

		pthread_mutex_unlock (&(mutexes[i]));
		pthread_cond_signal(&(mySignals[i]));
	}
	
	while (*counter != NUM_DISKS)
		pthread_cond_wait (&mainSignal, &mainMutex);

	pthread_mutex_unlock(&mainMutex);
}


void Stripe_File :: Read (char *readMe) {
	pthread_mutex_lock(&mainMutex);

	*counter = 0;
	
	for (int i = 0; i < NUM_DISKS; i++) {
		pthread_mutex_lock (&(mutexes[i]));

		bitPointers[i] = readMe + i * (PAGE_SIZE / NUM_DISKS);
		status[i] = READ;

		execute[i] = 1;

		pthread_mutex_unlock (&(mutexes[i]));
		pthread_cond_signal(&(mySignals[i]));
	}
	
	// wait for all of them to be done
	while (*counter != NUM_DISKS)
		pthread_cond_wait (&mainSignal, &mainMutex);

	pthread_mutex_unlock(&mainMutex);
}




void Stripe_File :: Open (int fileLen, char *fName, int file_type) {
	// figure out the flags for the system open call
	int mode;
	if (fileLen == 0)
		mode = O_TRUNC | O_RDWR | O_CREAT;
	else
		mode = O_RDWR;

	// actually do the open
	for (int i = 0; i < NUM_DISKS; i++) {
		char path[200];
		sprintf (path, "%s%s", prefixes[i], fName);

		myFilDes[i] = open (path, mode, S_IRUSR | S_IWUSR);
		if (myFilDes[i] < 0) {
			cerr << "BAD!  Open did not work.\n";
			exit (1);
		}
	}

	
	// OK, we first set up the read/write threads
	for (int i = 0; i < NUM_DISKS; i++) {
		// set up the parameters to this thread
		myThreadParam[i].bits = &bitPointers[i];

		myThreadParam[i].status = status + i;

		execute[i] = 0;
		myThreadParam[i].execute = execute + i;

		myThreadParam[i].myCond = &(mySignals[i]);
		myThreadParam[i].myMutex = &(mutexes[i]);

		myThreadParam[i].mainMutex = &(mainMutex);
		myThreadParam[i].mainSignal = &(mainSignal);

		myThreadParam[i].myFD = myFilDes + i;
		myThreadParam[i].counter = counter;
		
		// and let him go
		pthread_attr_t t_attr;
		int ret;
	
		ret = pthread_attr_init(&t_attr);
		ret = pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_DETACHED);
		ret = pthread_create(&(myThreads[i]), &t_attr, Stripe_Thread, (void *) (myThreadParam + i));
		if (ret) {
			cerr << "ERROR; return code from pthread_create() is " << ret << "\n";
			exit(-1);
		}
	}

	// read in the buffer if needed
	if (fileLen != 0) {
		char *bits = new (std::nothrow) char[PAGE_SIZE];
		if (bits == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		// read in the first few bits, which is the page size
		Lseek (0);

		Read (bits);
		curLength = *((off_t *) bits);
		
		if (file_type == FT_NONE)
			this->file_type = *(int *) (bits + sizeof(off_t));
		else
			this->file_type = file_type;

		switch (this->file_type)
		{
			case FT_INT:
				myBuff = new (std::nothrow) C_Page_INT();
				break;
			case FT_REAL:
				myBuff = new (std::nothrow) C_Page_DOUBLE();
				break;
			case FT_STRING:
				myBuff = new (std::nothrow) C_Page_STRING();
				break;
			case FT_RECORD:
				myBuff = new (std::nothrow) C_Page_RECORD();
				break;
			default:
				myBuff = new (std::nothrow) C_Page_RECORD();
				break;
		}

		if (myBuff == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		// read in the first page
		Lseek (PAGE_SIZE);

		Read (bits);
		myBuff->FromBinary (bits);

		delete [] bits;
		bits = NULL;
		
		// the buffer is not dirty!
		changed = 0;
	}
	else {
		this->file_type = file_type;

		switch (this->file_type)
		{
			case FT_INT:
				myBuff = new (std::nothrow) C_Page_INT();
				break;
			case FT_REAL:
				myBuff = new (std::nothrow) C_Page_DOUBLE();
				break;
			case FT_STRING:
				myBuff = new (std::nothrow) C_Page_STRING();
				break;
			case FT_RECORD:
				myBuff = new (std::nothrow) C_Page_RECORD();
				break;
			default:
				myBuff = new (std::nothrow) C_Page_RECORD();
				break;
		}

		if (myBuff == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		// otherwise, empty it out
		myBuff->EmptyItOut ();
		curLength = 2;
		
		// the buffer is dirty, since nothing is there!
		changed = 1;
	}

	// and note that we are at the first pos!
	curPos = 1;
	
	// set up the path
	free(path);
	path = strdup(fName);
}


int Stripe_File :: Close () {
	// write out the buffer if needed
	char *bits = new (std::nothrow) char[PAGE_SIZE];
	if (bits == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	// write out the current length in pages
	*((off_t *) bits) = curLength;
	*(int *) (bits + sizeof(off_t)) = file_type;

	Lseek (0);

	Write (bits);

	if (changed) {
		off_t tmp = curPos;
		tmp *= PAGE_SIZE;
		Lseek (tmp);

		myBuff->ToBinary (bits);
		Write (bits);
	}
	
	delete [] bits;

	// close the file
	for (int i = 0; i < NUM_DISKS; i++) {
		close (myFilDes[i]);
	}

	// now, let everyone know that we are done
	pthread_mutex_lock(&mainMutex);
	*counter = 0;

	for (int i = 0; i < NUM_DISKS; i++) {
		pthread_mutex_lock (&(mutexes[i]));

		status[i] = DONE;
		execute[i] = 1;

		pthread_mutex_unlock (&(mutexes[i]));
		pthread_cond_signal (&(mySignals[i]));
	}

	while (*counter != NUM_DISKS)
		pthread_cond_wait (&mainSignal, &mainMutex);

	pthread_mutex_unlock(&mainMutex);

	// and return the size
	return curLength;
}


void Stripe_File :: Kill () {
	// close the file
	for (int i = 0; i < NUM_DISKS; i++) {
		close (myFilDes[i]);

		char str[200];
		sprintf (str, "%s%s", prefixes[i], path);

		unlink (str);
	}
}







Run_File_Interface :: Run_File_Interface () {
}

Run_File_Interface :: ~Run_File_Interface () {
}



void Run_File_Interface :: MovePointer (int run_no) {
	off_t whichPage = run_start_page[run_no] + cur_page_pos[run_no];

	// see if someone wants to move past the end of the file
	if (whichPage >= current_length) {

		cerr << "Error: you tried to move past the end of the file " << path << ".\n";
		exit (1);
	}
	
	char *bits = new (std::nothrow) char[PAGE_SIZE];
	if (bits == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Lseek (whichPage * PAGE_SIZE);
	Read (bits);
	myBuff[run_no]->FromBinary (bits);

	delete [] bits;
}


int Run_File_Interface :: NextRec (Record *getMe, int run_no) {
	//if we returned all the records in this run, we simply return
	if (cur_pos[run_no] >= run_length[run_no])
		return 0;

	// try to remove the record from the buffer
	if (myBuff[run_no]->GetFirst (getMe)) {
		cur_pos[run_no] += 1;
		return 1;
	}

	//move to the next page of this run
	cur_page_pos[run_no] += 1;
	MovePointer (run_no);

	// now that we are at the next page, try to return the next record
	if (myBuff[run_no]->GetFirst (getMe)) {
		cur_pos[run_no] += 1;
		return 1;
	}
	else
	{
		cout << "Could not get a record from run " << run_no << " although it should exist since " << cur_pos[run_no] << " and " << run_length[run_no] << "\n";
		exit(1);
	}
}



// append a new run to the file; the records of the run are stored sorted in SortRecord which has the size buffer_size; return the new size of the file on success, or -1 on failure
off_t Run_File_Interface::Append_Run(SortRecord *buffer, int buffer_size)
{
	Page *page = new (std::nothrow) Page();
	if (page == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int recs_in = 0;

	run_start_page.push_back(current_length);
	cur_page_pos.push_back(current_length);

	for(int i = 0; i < buffer_size; i++) {
		// try to add the record
		if (page->Append (buffer[i].record) == 0) {
			//the page is full, so write it to the disk
			char *bits = new (std::nothrow) char[PAGE_SIZE];
			if (bits == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			page->ToBinary (bits);

			Lseek (PAGE_SIZE * current_length);

			Write (bits);

			delete [] bits;

			current_length += 1;

			page->EmptyItOut ();

			// add the record; now it should be free space for sure
			if (page->Append (buffer[i].record) != 0) {
				recs_in += 1;
			}
			else {
				// error?
				cerr << "You tried to add a record that was too big for the page!\n";
				exit (1);
			}
		}
		else {
			recs_in += 1;
		}
	}

	//write the last page to the disk
	char *bits = new (std::nothrow) char[PAGE_SIZE];
	if (bits == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	page->ToBinary (bits);

	Lseek (PAGE_SIZE * current_length);

	Write (bits);

	delete [] bits;

	current_length += 1;

	page->EmptyItOut ();


	run_length.push_back(recs_in);
	cur_pos.push_back(0);

	runs_no += 1;

	delete page;

	return current_length;
}




Run_File :: Run_File () {
	path = NULL;
	myBuff = NULL;

	runs_no = 0;

	//this is for the master record (the first page with metadata)
	current_length = 1;

	run_length.clear();
	run_start_page.clear();
	cur_pos.clear();
	cur_page_pos.clear();

	myFilDes = new (std::nothrow) int;
	if (myFilDes == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}

Run_File :: ~Run_File () {
	free(path);
	path = NULL;

	delete myFilDes;

	if (myBuff != NULL)
	{
		for (int i = 0; i < runs_no; i++)
			delete myBuff[i];

		delete [] myBuff;
	}

	runs_no = 0;

	run_length.clear();
	run_start_page.clear();
	cur_pos.clear();
	cur_page_pos.clear();
}



void Run_File :: Lseek (off_t howFar) {
	lseek (*myFilDes, howFar, SEEK_SET);
}


void Run_File :: Write (char *writeMe) {
	write (*myFilDes, writeMe, PAGE_SIZE);
}


void Run_File :: Read (char *readMe) {
	read (*myFilDes, readMe, PAGE_SIZE);
}



void Run_File :: Open (int fileLen, char *fName) {

	// figure out the flags for the system open call
        int mode;
        if (fileLen == 0)
                mode = O_TRUNC | O_RDWR | O_CREAT;
        else
                mode = O_RDWR;

	// actually do the open
        *myFilDes = open (fName, mode, S_IRUSR | S_IWUSR);

	// see if there was an error
	if (myFilDes < 0) {
		cerr << "BAD!  Open did not work for " << fName << "\n";
		exit (1);
	}


	//do the initializations in constructor again (for allowing multiple opens)
	if (myBuff != NULL)
	{
		for (int i = 0; i < runs_no; i++)
			delete myBuff[i];

		delete [] myBuff;
	}
	myBuff = NULL;

	runs_no = 0;

	//this is for the master record (the first page with metadata)
	current_length = 1;

	run_length.clear();
	run_start_page.clear();
	cur_pos.clear();
	cur_page_pos.clear();



	// read in the buffer if needed
	if (fileLen != 0) {
		int i;

		char *bits = new (std::nothrow) char[PAGE_SIZE];
		if (bits == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Lseek (0);
		Read (bits);

		// read in the first few bits, which is the number of pages in the file
		current_length = *(off_t*)bits;

		//read the number of runs in the file
		runs_no = *(off_t*)(bits + sizeof(off_t));

		//fill up the structures and the buffers with the first page for each of the runs
		myBuff = new (std::nothrow) Page*[runs_no];
		if (myBuff == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (i = 0; i < runs_no; i++)
		{
			off_t aux;

			//first the size of the run in records
			aux = *(off_t*)(bits + 2 * (i + 1) * sizeof(off_t));
			run_length.push_back(aux);

			//second the starting page of the current run
			aux = *(off_t*)(bits + (2 * (i + 1) + 1) * sizeof(off_t));
			run_start_page.push_back(aux);

			cur_pos.push_back(0);
			cur_page_pos.push_back(0);

			myBuff[i] = new (std::nothrow) Page();
			if (myBuff[i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			// read in the first page for each run
			char *run_bits = new (std::nothrow) char[PAGE_SIZE];
			if (run_bits == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			off_t real_offset = run_start_page[i];
			real_offset = real_offset * PAGE_SIZE;
			Lseek (real_offset);

			Read (run_bits);
			myBuff[i]->FromBinary (run_bits);
			delete [] run_bits;
		}

		delete [] bits;

	// otherwise, empty it out
	} else {
		//the file is open for inserting the runs; everything is empty and exactly as it was initialized in the constructor
	}

	// set up the path
	free(path);
	path = strdup (fName);
}


int Run_File :: Close (int write_needed) {
	int i;

	//write out the buffer if needed
	if (write_needed != 0) {
		char *bits = new (std::nothrow) char[PAGE_SIZE];
		if (bits == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	

		//write out the metadata in the first page of the file
		*(off_t*)bits = current_length;
	
		*(off_t*)(bits + sizeof(off_t)) = runs_no;
	
		for (i = 0; i < runs_no; i++)
		{
			off_t aux;
	
			//first the size of the run in records
			aux = run_length[i];
			*(off_t*)(bits + 2 * (i + 1) * sizeof(off_t)) = aux;
	
			//second the starting page of the current run
			aux = run_start_page[i];
			*(off_t*)(bits + (2 * (i + 1) + 1) * sizeof(off_t)) = aux;
		}

		Lseek (0);
	
		Write (bits);

		delete [] bits;
        }

	//close the file
	close (*myFilDes);

	// and return the size
	return current_length;	
}


void Run_File :: Kill () {
	close (*myFilDes);
	unlink (path);
}





Run_Stripe_File :: Run_Stripe_File () {
	path = NULL;
	myBuff = NULL;

	runs_no = 0;

	//this is for the master record (the first page with metadata)
	current_length = 1;

	run_length.clear();
	run_start_page.clear();
	cur_pos.clear();
	cur_page_pos.clear();

	pthread_mutex_init(&mainMutex, NULL);
	pthread_cond_init(&mainSignal, NULL);
	
	for (int i = 0; i < NUM_DISKS; i++) {
		// initialize this guy's mutex and cond var
		pthread_mutex_init(&(mutexes[i]), NULL);
		pthread_cond_init(&(mySignals[i]), NULL);
	}

	counter = new (std::nothrow) int;
	if (counter == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	myFilDes = new (std::nothrow) int[NUM_DISKS];
	if (myFilDes == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}

Run_Stripe_File :: ~Run_Stripe_File () {
	free(path);
	path = NULL;

	if (myBuff != NULL)
	{
		for (int i = 0; i < runs_no; i++)
			delete myBuff[i];

		delete [] myBuff;
	}

	runs_no = 0;

	run_length.clear();
	run_start_page.clear();
	cur_pos.clear();
	cur_page_pos.clear();

	pthread_mutex_destroy(&mainMutex);
	pthread_cond_destroy(&mainSignal);
	
	for (int i = 0; i < NUM_DISKS; i++) {
		// initialize this guy's mutex and cond var
		pthread_mutex_destroy(&(mutexes[i]));
		pthread_cond_destroy(&(mySignals[i]));
	}

	delete counter;

	delete [] myFilDes;
}



void Run_Stripe_File :: Lseek (off_t howFar) {
	// this just moves all of the underlying pointers
	for (int i = 0; i < NUM_DISKS; i++) {
		lseek (myFilDes[i], (howFar / NUM_DISKS), SEEK_SET);
	}
}


void Run_Stripe_File :: Write (char *writeMe) {
	pthread_mutex_lock(&mainMutex);

	*counter = 0;
	
	// ask all of the disks to write
	for (int i = 0; i < NUM_DISKS; i++) {
		pthread_mutex_lock (&(mutexes[i]));

		bitPointers[i] = writeMe + i * (PAGE_SIZE / NUM_DISKS);
		status[i] = WRITE;

		execute[i] = 1;

		pthread_mutex_unlock (&(mutexes[i]));
		pthread_cond_signal(&(mySignals[i]));
	}
	
	while (*counter != NUM_DISKS)
		pthread_cond_wait (&mainSignal, &mainMutex);

	pthread_mutex_unlock(&mainMutex);
}


void Run_Stripe_File :: Read (char *readMe) {
	pthread_mutex_lock(&mainMutex);

	*counter = 0;
	
	for (int i = 0; i < NUM_DISKS; i++) {
		pthread_mutex_lock (&(mutexes[i]));

		bitPointers[i] = readMe + i * (PAGE_SIZE / NUM_DISKS);
		status[i] = READ;

		execute[i] = 1;

		pthread_mutex_unlock (&(mutexes[i]));
		pthread_cond_signal(&(mySignals[i]));
	}
	
	// wait for all of them to be done
	while (*counter != NUM_DISKS)
		pthread_cond_wait (&mainSignal, &mainMutex);

	pthread_mutex_unlock(&mainMutex);
}



void Run_Stripe_File :: Open (int fileLen, char *fName) {

	// figure out the flags for the system open call
        int mode;
        if (fileLen == 0)
                mode = O_TRUNC | O_RDWR | O_CREAT;
        else
                mode = O_RDWR;

	// actually do the open
	for (int i = 0; i < NUM_DISKS; i++) {
		char path[200];
		sprintf (path, "%s%s", prefixes[i], fName);

		myFilDes[i] = open (path, mode, S_IRUSR | S_IWUSR);
		if (myFilDes[i] < 0) {
			cerr << "BAD!  Open did not work.\n";
			exit (1);
		}
	}


	//do the initializations in constructor again (for allowing multiple opens)
	if (myBuff != NULL)
	{
		for (int i = 0; i < runs_no; i++)
			delete myBuff[i];

		delete [] myBuff;
	}
	myBuff = NULL;

	runs_no = 0;

	//this is for the master record (the first page with metadata)
	current_length = 1;

	run_length.clear();
	run_start_page.clear();
	cur_pos.clear();
	cur_page_pos.clear();



	// OK, we first set up the read/write threads
	for (int i = 0; i < NUM_DISKS; i++) {
		// set up the parameters to this thread
		myThreadParam[i].bits = &bitPointers[i];

		myThreadParam[i].status = &(status[i]);

		execute[i] = 0;
		myThreadParam[i].execute = &(execute[i]);

		myThreadParam[i].myCond = &(mySignals[i]);
		myThreadParam[i].myMutex = &(mutexes[i]);

		myThreadParam[i].mainMutex = &(mainMutex);
		myThreadParam[i].mainSignal = &(mainSignal);

		myThreadParam[i].myFD = &(myFilDes[i]);
		myThreadParam[i].counter = counter;
		
		// and let him go
		pthread_attr_t t_attr;
		int ret;
	
		ret = pthread_attr_init(&t_attr);
		ret = pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_DETACHED);
		ret = pthread_create(&(myThreads[i]), &t_attr, Stripe_Thread, (void *) (&myThreadParam[i]));
		if (ret) {
			cerr << "ERROR; return code from pthread_create() is " << ret << "\n";
			exit(-1);
		}
	}


	// read in the buffer if needed
	if (fileLen != 0) {
		int i;

		char *bits = new (std::nothrow) char[PAGE_SIZE];
		if (bits == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Lseek (0);
		Read (bits);

		// read in the first few bits, which is the number of pages in the file
		current_length = *(off_t*)bits;

		//read the number of runs in the file
		runs_no = *(off_t*)(bits + sizeof(off_t));

		//fill up the structures and the buffers with the first page for each of the runs
		myBuff = new (std::nothrow) Page*[runs_no];
		if (myBuff == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (i = 0; i < runs_no; i++)
		{
			off_t aux;

			//first the size of the run in records
			aux = *(off_t*)(bits + 2 * (i + 1) * sizeof(off_t));
			run_length.push_back(aux);

			//second the starting page of the current run
			aux = *(off_t*)(bits + (2 * (i + 1) + 1) * sizeof(off_t));
			run_start_page.push_back(aux);

			cur_pos.push_back(0);
			cur_page_pos.push_back(0);

			myBuff[i] = new (std::nothrow) Page();
			if (myBuff[i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			// read in the first page for each run
			char *run_bits = new (std::nothrow) char[PAGE_SIZE];
			if (run_bits == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			off_t real_offset = run_start_page[i];
			real_offset = real_offset * PAGE_SIZE;
			Lseek (real_offset);

			Read (run_bits);
			myBuff[i]->FromBinary (run_bits);
			delete [] run_bits;
		}

		delete [] bits;

	// otherwise, empty it out
	} else {
		//the file is open for inserting the runs; everything is empty and exactly as it was initialized in the constructor
	}

	// set up the path
	free(path);
	path = strdup (fName);
}


int Run_Stripe_File :: Close (int write_needed) {
	int i;

	//write out the buffer if needed
	if (write_needed != 0) {
		char *bits = new (std::nothrow) char[PAGE_SIZE];
		if (bits == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	

		//write out the metadata in the first page of the file
		*(off_t*)bits = current_length;
	
		*(off_t*)(bits + sizeof(off_t)) = runs_no;
	
		for (i = 0; i < runs_no; i++)
		{
			off_t aux;
	
			//first the size of the run in records
			aux = run_length[i];
			*(off_t*)(bits + 2 * (i + 1) * sizeof(off_t)) = aux;
	
			//second the starting page of the current run
			aux = run_start_page[i];
			*(off_t*)(bits + (2 * (i + 1) + 1) * sizeof(off_t)) = aux;
		}

		Lseek (0);
	
		Write (bits);

		delete [] bits;
        }


	// now, let everyone know that we are done
	pthread_mutex_lock(&mainMutex);
	*counter = 0;

	for (int i = 0; i < NUM_DISKS; i++) {
		pthread_mutex_lock (&(mutexes[i]));

		status[i] = DONE;
		execute[i] = 1;

		pthread_mutex_unlock (&(mutexes[i]));
		pthread_cond_signal (&(mySignals[i]));
	}

	while (*counter != NUM_DISKS)
		pthread_cond_wait (&mainSignal, &mainMutex);

	pthread_mutex_unlock(&mainMutex);

	//close the file
	for (int i = 0; i < NUM_DISKS; i++) {
		close (myFilDes[i]);
	}

	// and return the size
	return current_length;
}


void Run_Stripe_File :: Kill () {
	// close the file
	for (int i = 0; i < NUM_DISKS; i++) {
		close (myFilDes[i]);

		char str[200];
		sprintf (str, "%s%s", prefixes[i], path);

		unlink (str);
	}
}







// this asks the individual thread to write or to read
void *Stripe_Thread (void *param) {
	int endIt = 0;
	
	// remember all of our params
	char **bitPointer = ((ThreadParam *) param)->bits;

	pthread_mutex_t *mainMutex = ((ThreadParam *) param)->mainMutex;
	pthread_cond_t *mainSignal = ((ThreadParam *) param)->mainSignal;

	pthread_mutex_t *myMutex = ((ThreadParam *) param)->myMutex;
	pthread_cond_t *myCond = ((ThreadParam *) param)->myCond;

	int *status = ((ThreadParam *) param)->status;
	int *execute = ((ThreadParam *) param)->execute;

	int *mainCounter = ((ThreadParam *) param)->counter;
	int *myFD = ((ThreadParam *) param)->myFD;
	
	while (1) {
		// now wait for him to signal us
		pthread_mutex_lock (myMutex);

		while (*execute == 0)
			pthread_cond_wait(myCond, myMutex);

		// see if we have been told to die
		if (*status == DONE) {
			endIt = 1;
		}
		else if (*status == WRITE) {
			write (*myFD, *bitPointer, (PAGE_SIZE / NUM_DISKS));
		}
		else if (*status == READ) {
			read (*myFD, *bitPointer, (PAGE_SIZE / NUM_DISKS));
		}
		else {
			cerr << "Weird command given to read/write thread.\n";
			exit (1);
		}

		*execute = 0;

		pthread_mutex_unlock(myMutex);


		// tell the main program that another thread is done
		pthread_mutex_lock (mainMutex);
		(*mainCounter)++;

		// if we are the last one, let the main program go
		if (*mainCounter == NUM_DISKS) {
			pthread_mutex_unlock (mainMutex);
			pthread_cond_signal (mainSignal);
		}
		else {
			pthread_mutex_unlock (mainMutex);
		}

		if (endIt)
			return NULL;
	}

}
