#include "File.h"
#include "Record.h"
#include "Comparison.h"
#include "Schema.h"
#include "Hashes.h"

#include <iostream>
#include <stdio.h>


using namespace std;


int main(int argc, char **argv)
{
	if (argc == 4)
	{
		int i;

		//first stripe
		char f_name[200];
		sprintf (f_name, "../Tables/%s", argv[1]);

		Schema *schema = new (std::nothrow) Schema(argv[2], argv[3]);
		if (schema == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		int sch_atts_no = schema->GetNumAtts();
		Attribute *sch_atts = schema->GetAtts();


		Schema **schemas = new (std::nothrow) Schema*[sch_atts_no];
		if (schemas == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}



		File_Interface *f_in = new (std::nothrow) File();
		if (f_in == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
		f_in->Open(1, f_name, FT_RECORD);


		File_Interface **f_out = new (std::nothrow) File_Interface*[sch_atts_no];
		if (f_out == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (i = 0; i < sch_atts_no; i++)
		{
			f_out[i] = new (std::nothrow) File();
			if (f_out[i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			int file_type;
			switch(sch_atts[i].myType)
			{
				case Int:
					file_type = FT_INT;
					break;
				case Double:
					file_type = FT_REAL;
					break;
				case String:
					file_type = FT_STRING;
					break;
			}

			char fc_name[200];
			sprintf(fc_name, "data/%s", sch_atts[i].name);

			f_out[i]->Open(0, fc_name, file_type);


			schemas[i] = new (std::nothrow) Schema();
			if (schemas[i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			schemas[i]->CopyFrom(schema);
			schemas[i]->Project(&i, 1);
		}



		Record *in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Record *out_rec = new (std::nothrow) Record();
		if (out_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}


		while (f_in->NextRec(in_rec))
		{
//			in_rec->Print(schema);

			for (i = 0; i < sch_atts_no; i++)
			{
				out_rec->Copy(in_rec);
				out_rec->Project(&i, 1, sch_atts_no);

//				out_rec->Print(schemas[i]);

				f_out[i]->Append(out_rec);
			}
		}

		delete in_rec;
		delete out_rec;


/*

		f_in->Close();
		delete f_in;

		for (i = 0; i < sch_atts_no; i++)
		{
			f_out[i]->Close();

			delete f_out[i];
			f_out[i] = NULL;
		}

		delete [] f_out;








		f_in = new (std::nothrow) File();
		if (f_in == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
		f_in->Open(1, f_name, FT_RECORD);


		f_out = new (std::nothrow) File_Interface*[sch_atts_no];
		if (f_out == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (i = 0; i < sch_atts_no; i++)
		{
			f_out[i] = new (std::nothrow) Stripe_File();
			if (f_out[i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			f_out[i]->Open(1, sch_atts[i].name, FT_NONE);
		}


		in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Record *in_rec1 = new (std::nothrow) Record();
		if (in_rec1 == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Record *in_rec2 = new (std::nothrow) Record();
		if (in_rec2 == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		out_rec = new (std::nothrow) Record();
		if (out_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}


		OrderMaker *order = new OrderMaker(schema);

		Hash *hash = new Hash();


		while (f_in->NextRec(in_rec))
		{
//			in_rec->Print(schema);

			f_out[0]->NextRec(out_rec);

//			out_rec->Print(schemas[0]);

			for (i = 1; i < sch_atts_no; i++)
			{
				f_out[i]->NextRec(in_rec1);

//				in_rec1->Print(schemas[i]);

				in_rec2->AppendRecords(out_rec, in_rec1, i, 1);

				out_rec->Consume(in_rec2);
			}

			if (hash->Compare(in_rec, out_rec, order) != 0)
			{
				cout << "ERROR!!!\n";
			}
		}

		delete hash;

		delete schema;
		delete order;

		delete in_rec;
		delete in_rec1;
		delete in_rec2;
		delete out_rec;

*/


		f_in->Close();
		delete f_in;

		for (i = 0; i < sch_atts_no; i++)
		{
			f_out[i]->Close();

			delete f_out[i];
			f_out[i] = NULL;

			delete schemas[i];
		}

		delete [] f_out;

		delete [] schemas;

	}
	else
		cerr << "Usage: " << argv[0] << " [input file] [schema file] [schema name]\n";

	return(0);
}

