#include "DEFS.h"
#include "Hashes.h"

#include <string.h>
#include <stdlib.h>
#include <iostream.h>



Hash::Hash() {
}


Hash::~Hash() {
}


//this is just an implementation of the Jenkins one-at-a-time hash function (http://www.burtleburtle.net/bob/hash/doobs.html)
unsigned int Hash::Compute_Hash(Record *rec, OrderMaker *order) {
	unsigned int ret = 0;

	for (int i = 0; i < order->numAtts; i++)
	{
		char *ptr = rec->GetBits() + ((int *)rec->GetBits())[order->whichAtts[i] + 1];
		int len = 0;

		if (order->whichTypes[i] == Int)
			len = sizeof(int);
		else if (order->whichTypes[i] == Double)
			len = sizeof(double);
		else if (order->whichTypes[i] == String)
		{
			char *t_ptr = ptr;

			while (*t_ptr != '\0')
			{
				len += 1;
				t_ptr += 1;
			}
		}

		for (int j = 0; j < len; j++)
		{
			ret += ptr[j];
			ret += (ret << 10);
			ret ^= (ret >> 6);
		}
	}

	ret += (ret << 3);
	ret ^= (ret >> 11);
	ret += (ret << 15);

	return ret;
}


unsigned int Hash::Compute_Hash(int *v, int n) {
	unsigned int ret = 0;

	for (int i = 0; i < n; i++)
	{
		ret += v[i];
		ret += (ret << 10);
		ret ^= (ret >> 6);
	}

	ret += (ret << 3);
	ret ^= (ret >> 11);
	ret += (ret << 15);

	return ret;
}



int Hash::Compare(Record *left, Record *right, OrderMaker *orderUs) {
	char *val1, *val2;

	char *left_bits = left->GetBits();
	char *right_bits = right->GetBits();

	for (int i = 0; i < orderUs->numAtts; i++) {
		val1 = left_bits + ((int *) left_bits)[orderUs->whichAtts[i] + 1];
		val2 = right_bits + ((int *) right_bits)[orderUs->whichAtts[i] + 1];
	
		// these are used to store the two operands, depending on their type
		int val1Int, val2Int;
		double val1Double, val2Double;
		
		// now check the type and do the comparison
		switch (orderUs->whichTypes[i]) {
	
			// first case: we are dealing with integers
			case Int:
	
			// cast the two bit strings to ints
			val1Int = *((int *) val1);
			val2Int = *((int *) val2);
	
			// and do the comparison
			if (val1Int < val2Int) 
				return -1;
			else if (val1Int > val2Int)
				return 1;
	
			break;
	
	
			// second case: dealing with doubles
			case Double:
	
			// cast the two bit strings to doubles
			val1Double = *((double *) val1);
			val2Double = *((double *) val2);
	
			// and do the comparison
			if (val1Double < val2Double)
				return -1;
			else if (val1Double > val2Double)
				return 1;
	
			break;
	
	
			// last case: dealing with strings
			default:
			int sc = strcmp (val1, val2);
			if (sc != 0)
				return sc;

			break;
	
		}
	}

	return 0;
}



int Hash::Compare(Record *left, OrderMaker *order_left, Record *right, OrderMaker *order_right) {
	char *val1, *val2;

	char *left_bits = left->GetBits();
	char *right_bits = right->GetBits();

	for (int i = 0; i < order_left->numAtts; i++) {
		val1 = left_bits + ((int *) left_bits)[order_left->whichAtts[i] + 1];
		val2 = right_bits + ((int *) right_bits)[order_right->whichAtts[i] + 1];
	
		// these are used to store the two operands, depending on their type
		int val1Int, val2Int;
		double val1Double, val2Double;
		
		// now check the type and do the comparison
		switch (order_left->whichTypes[i]) {
	
			// first case: we are dealing with integers
			case Int:
	
			// cast the two bit strings to ints
			val1Int = *((int *) val1);
			val2Int = *((int *) val2);
	
			// and do the comparison
			if (val1Int < val2Int) 
				return -1;
			else if (val1Int > val2Int)
				return 1;
	
			break;
	
	
			// second case: dealing with doubles
			case Double:
	
			// cast the two bit strings to doubles
			val1Double = *((double *) val1);
			val2Double = *((double *) val2);
	
			// and do the comparison
			if (val1Double < val2Double)
				return -1;
			else if (val1Double > val2Double)
				return 1;
	
			break;
	
	
			// last case: dealing with strings
			default:
			int sc = strcmp (val1, val2);
			if (sc != 0)
				return sc;

			break;
	
		}
	}

	return 0;
}



// this is the version of the record compare that works on a single record
int Hash::Compare (Record *left, Record *literal, CNF *myComparison) {

	for (int i = 0; i < myComparison->numAnds; i++) {

		for (int j = 0; j < myComparison->orLens[i]; j++) {

			// this returns a 0 if the comparison did not eval to true
			int result = Run(left, literal, &myComparison->orList[i][j]);
			
			if (result != 0) {
				break;	
			}
			
			// if we made it through all of the comparisons without a hit, return a 0
			if (j == myComparison->orLens[i] - 1) {
				return 0;
			}
		}

		// this ends the for-loop for the OR-list... if we made it this far, then we'll go on to the
		// next disjunction in the overall CNF expression
	}

	// this ends the for-loop for the AND-list... if we made it this far, then we know that we
	// have satisfied the overall CNF expression
	return 1;	
}



// this is the version of the record compare that works on a pair of records
int Hash::Compare (Record *left, Record *right, Record *literal, CNF *myComparison) {

	for (int i = 0; i < myComparison->numAnds; i++) {

		for (int j = 0; j < myComparison->orLens[i]; j++) {

			// this returns a 0 if the comparison did not eval to true
			int result = Run (left, right, literal, &myComparison->orList[i][j]);
			
			if (result != 0) {
				break;	
			}

			// if we made it through all of the comparisons without a hit, return a 0
			if (j == myComparison->orLens[i] - 1) {
				return 0;
			}
		}

		// this ends the for-loop for the OR-list... if we made it this far, then we'll go on to the
		// next disjunction in the overall CNF expression
	}

	// this ends the for-loop for the AND-list... if we made it this far, then we know that we
	// have satisfied the overall CNF expression
	return 1;	
}



int Hash::Run (Record *left, Record *literal, Comparison *c) {

	char *val1, *val2;

	char *left_bits = left->GetBits();
	char *lit_bits = literal->GetBits();

	// first get a pointer to the first value to compare
	if (c->operand1 == Left) {
		val1 = left_bits + ((int *) left_bits)[c->whichAtt1 + 1];
	} else {
		val1 = lit_bits + ((int *) lit_bits)[c->whichAtt1 + 1];
	}

	// next get a pointer to the second value to compare
	if (c->operand2 == Left) {
		val2 = left_bits + ((int *) left_bits)[c->whichAtt2 + 1];
	} else {
		val2 = lit_bits + ((int *) lit_bits)[c->whichAtt2 + 1];
	}


	int val1Int, val2Int, tempResult;
	double val1Double, val2Double;

	// now check the type and the comparison operation
	switch (c->attType) {

		// first case: we are dealing with integers
		case Int:

		val1Int = *((int *) val1);
		val2Int = *((int *) val2);

		// and check the operation type in order to actually do the comparison
		switch (c->op) {
				
			case LessThan:
			return (val1Int < val2Int);
			break;

			case GreaterThan:
			return (val1Int > val2Int);
			break;
					
			default:
			return (val1Int == val2Int);
			break;	
		}	
		break;

		// second case: dealing with doubles
		case Double:
		val1Double = *((double *) val1);
		val2Double = *((double *) val2);

		// and check the operation type in order to actually do the comparison
		switch (c->op) {
						
			case LessThan:
			return (val1Double < val2Double);
			break;

			case GreaterThan:
			return (val1Double > val2Double);
			break;
					
			default:
			return (val1Double == val2Double);
			break;	
		}	
		break;

		// final case: dealing with strings
		default:

		// so check the operation type in order to actually do the comparison
		tempResult = strcmp (val1, val2);
		switch (c->op) {
						
			case LessThan:
			return tempResult < 0;
			break;

			case GreaterThan:
			return tempResult > 0;
			break;
					
			default:
			return tempResult == 0;
			break;	
		}	
		break;
	}

}


int Hash::Run (Record *left, Record *right, Record *literal, Comparison *c) {

	char *val1, *val2;

	char *left_bits = left->GetBits();
	char *right_bits = right->GetBits();
	char *lit_bits = literal->GetBits();

	// first get a pointer to the first value to compare
	switch (c->operand1) {

		case Left:
		val1 = left_bits + ((int *) left_bits)[c->whichAtt1 + 1];
		break;
	
		case Right:
		val1 = right_bits + ((int *) right_bits)[c->whichAtt1 + 1];
		break;

		default:
		val1 = lit_bits + ((int *) lit_bits)[c->whichAtt1 + 1];
		break;	
	}

	// next get a pointer to the second value to compare
	switch (c->operand2) {

		case Left:
		val2 = left_bits + ((int *) left_bits)[c->whichAtt2 + 1];
		break;

		case Right:
		val2 = right_bits + ((int *) right_bits)[c->whichAtt2 + 1];
		break;

		default:
		val2 = lit_bits + ((int *) lit_bits)[c->whichAtt2 + 1];
		break;	

	}
			
	int val1Int, val2Int, tempResult;
	double val1Double, val2Double;

	// now check the type and the comparison operation
	switch (c->attType) {

		// first case: we are dealing with integers
		case Int:

		val1Int = *((int *) val1);
		val2Int = *((int *) val2);

		// and check the operation type in order to actually do the comparison
		switch (c->op) {
				
			case LessThan:
			return (val1Int < val2Int);
			break;

			case GreaterThan:
			return (val1Int > val2Int);
			break;
					
			default:
			return (val1Int == val2Int);
			break;	
		}	
		break;

		// second case: dealing with doubles
		case Double:
		val1Double = *((double *) val1);
		val2Double = *((double *) val2);

		// and check the operation type in order to actually do the comparison
		switch (c->op) {
						
			case LessThan:
			return (val1Double < val2Double);
			break;

			case GreaterThan:
			return (val1Double > val2Double);
			break;
					
			default:
			return (val1Double == val2Double);
			break;	
		}	
		break;

		// final case: dealing with strings
		default:

		// so check the operation type in order to actually do the comparison
		tempResult = strcmp (val1, val2);
		switch (c->op) {
						
			case LessThan:
			return tempResult < 0;
			break;

			case GreaterThan:
			return tempResult > 0;
			break;
					
			default:
			return tempResult == 0;
			break;	
		}	
		break;
	}

}


void Hash::Print_Parameters()
{
	cout << "Basic hash. No parameters.\n";
}



CWHash::CWHash() {
	params_no = 0;
	params = NULL;

	buckets = 1;
}


CWHash::CWHash(int params_no, unsigned int buckets) {
	this->params_no = params_no;
	
	params = new (std::nothrow) unsigned long long[params_no];
	if (params == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < params_no; i++)
		params[i] = Random_Generate();

	this->buckets = buckets;
}


CWHash::~CWHash() {
	params_no = 0;

	delete [] params;
	params = NULL;

	buckets = 1;
}


int CWHash::Compare(Record *left, Record *right, OrderMaker *order) {
	
	
	if (left->hash == -1)
		left->hash = Compute_Hash(left, order);

	if (right->hash == -1)
		right->hash = Compute_Hash(right, order);

	if (left->hash > right->hash)
		return 1;
	else if (left->hash < right->hash)
		return -1;
	else
		return 0;
}


int CWHash::Compare(Record *left, OrderMaker *order_left, Record *right, OrderMaker *order_right) {

	
	if (left->hash == -1)
		left->hash = Compute_Hash(left, order_left);

	if (right->hash == -1)
		right->hash = Compute_Hash(right, order_right);

	if (left->hash > right->hash)
		return 1;
	else if (left->hash < right->hash)
		return -1;
	else
		return 0;
}




unsigned int CWHash::Compute_Hash(Record *rec, OrderMaker *order) {
	unsigned int res = 0;


	
	char *rec_bits = rec->GetBits();

	for (int i = 0; i < order->numAtts; i++)
	{
 		char *val;
		val = rec_bits + ((int *) rec_bits)[order->whichAtts[i] + 1];

		unsigned long long a = params[i];
		unsigned long long b = (i == 0) ? params[order->numAtts] : res;

		Type attType = order->whichTypes[i];
		switch (attType) {
	
			case Int:
			{
				int valInt = *((int *) val);
				res = hash31(a, b, (unsigned long long)valInt);
	
				break;
			}
	
			case Double:
			{
				int sInt = sizeof(int);
				int sDouble = sizeof(double);
				
				int q = sDouble / sInt;
				int r = sDouble % sInt;
				
				for (int j = 0; j < q; j++)
				{
					int valInt = *((int *) (val + j * sInt));
					res = hash31(a, b, (unsigned long long)valInt);

					b = res;
				}

				if (r != 0)
				{
					char *c = new (std::nothrow) char[sInt];
					if (c == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					for (int j = 0; j < sInt; j++)
						c[j] = 0;
					for (int j = 0; j < r; j++)
						c[j] = *(val + q * sInt + r);

					int valInt = *((int *) c);
					res = hash31(a, b, (unsigned long long)valInt);

					delete [] c;
				}

				break;
			}
	
			default:
			{
				int sInt = sizeof(int);
				int sLen = strlen(val);
				
				int q = sLen / sInt;
				int r = sLen % sInt;
				
				for (int j = 0; j < q; j++)
				{
					int valInt = *((int *) (val + j * sInt));
					res = hash31(a, b, (unsigned long long)valInt);

					b = res;
				}

				if (r != 0)
				{
					char *c = new (std::nothrow) char[sInt];
					if (c == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					for (int j = 0; j < sInt; j++)
						c[j] = 0;
					for (int j = 0; j < r; j++)
						c[j] = *(val + q * sInt + r);

					int valInt = *((int *) c);
					res = hash31(a, b, (unsigned long long)valInt);

					delete [] c;
				}

				break;
			}
		}
	}

	res = res % buckets;
	return (int)res;


/*
	unsigned int res = 0;

	char *rec_bits = rec->GetBits();

	for (int i = 0; i < order->numAtts; i++)
	{
		char *val;
		val = rec_bits + ((int *) rec_bits)[order->whichAtts[i] + 1];

		unsigned long long a = (i == 0) ? params[order->numAtts] : res;
		unsigned long long b = params[i];

		Type attType = order->whichTypes[i];
		switch (attType) {
	
			case Int:
			{
				int valInt = *((int *) val);
				res = hash31(a, b, (unsigned long long)valInt);
	
				break;
			}
	
			case Double:
			{
				int sInt = sizeof(int);
				int sDouble = sizeof(double);
				
				int q = sDouble / sInt;
				int r = sDouble % sInt;
				
				for (int j = 0; j < q; j++)
				{
					int valInt = *((int *) (val + j * sInt));
					res = hash31(a, b, (unsigned long long)valInt);

					a = res;
				}

				if (r != 0)
				{
					char *c = new (std::nothrow) char[sInt];
					if (c == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					for (int j = 0; j < sInt; j++)
						c[j] = 0;
					for (int j = 0; j < r; j++)
						c[j] = *(val + q * sInt + r);

					int valInt = *((int *) c);
					res = hash31(a, b, (unsigned long long)valInt);

					delete [] c;
				}

				break;
			}
	
			default:
			{
				int sInt = sizeof(int);
				int sLen = strlen(val);
				
				int q = sLen / sInt;
				int r = sLen % sInt;
				
				for (int j = 0; j < q; j++)
				{
					int valInt = *((int *) (val + j * sInt));
					res = hash31(a, b, (unsigned long long)valInt);

					a = res;
				}

				if (r != 0)
				{
					char *c = new (std::nothrow) char[sInt];
					if (c == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					for (int j = 0; j < sInt; j++)
						c[j] = 0;
					for (int j = 0; j < r; j++)
						c[j] = *(val + q * sInt + r);

					int valInt = *((int *) c);
					res = hash31(a, b, (unsigned long long)valInt);

					delete [] c;
				}

				break;
			}
		}
	}

	res = res % buckets;
	return (int)res;
*/


}


unsigned int CWHash::Compute_Hash(int *v, int n) {
	unsigned int res = 0;

	for (int i = 0; i < n; i++)
	{

		unsigned long long a = params[i];
		unsigned long long b = (i == 0) ? params[n] : res;

/*
		unsigned long long a = (i == 0) ? params[n] : res;
		unsigned long long b = params[i];
*/

		res = hash31(a, b, (unsigned long long)v[i]);
	}

	res = res % buckets;
	return (int)res;
}


void CWHash::Print_Parameters()
{
	cout << "Buckets: " << buckets << "\n";
	cout << "No of Parameters: " << params_no << "\n";
	for (int i = 0; i < params_no; i++)
	{
		cout << params[i] << "\n";
	}
}



inline unsigned int hash31(unsigned long long a, unsigned long long b, unsigned long long x) {
	unsigned long long result;
	unsigned int lresult;

//	cout << a << " " << b << " " << x << " -> ";

	result = (a * x) + b;
	result = ((result >> (unsigned int)HL) + result) & (unsigned long long)MOD;
	lresult = (unsigned int)result;

//	cout << lresult << "\n";

	return lresult;
}
