#include "InMemoryJoin.h"

#include <iostream>
#include <stack>
#include <stdlib.h>


using namespace std;



InMemoryJoin::InMemoryJoin(int relations_no, CNF **cnf, CNF *extra_cond, Record *extra_lit, int **mask, Schema **relSchemas, Function *toApply)
{
	this->relations_no = relations_no;

	this->schemaSizes = new (std::nothrow) int[relations_no];
	if (this->schemaSizes == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->schemas = new (std::nothrow) Schema*[relations_no];
	if (this->schemas == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->extra_cond = extra_cond;
	this->extra_lit = extra_lit;

	this->cnf = new (std::nothrow) OrderMaker* [this->relations_no];
	if (this->cnf == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->mask = new (std::nothrow) int* [this->relations_no];
	if (this->mask == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	recs_no = new (std::nothrow) int[this->relations_no];
	if (recs_no == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	records = new (std::nothrow) Record* [this->relations_no];
	if (records == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	// create the hash table
	// 
	// LUIS - ALRIGHT, THIS TIME THE SIZE SHOULD BE ((this->relations_no)*2) - 2 
	hash_table = new (std::nothrow) HashChain*** [this->relations_no];
	if (hash_table == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

        // LUIS -- SAME SIZE HERE.
	hash_func = new (std::nothrow) Hash**[this->relations_no];
	if (hash_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	exact_hash = new (std::nothrow) Hash();
	if (exact_hash == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	int i, j;
	for (i = 0; i < this->relations_no; i++)
	{
		this->schemaSizes[i] = relSchemas[i]->GetNumAtts();
		this->schemas[i] = relSchemas[i];

		this->cnf[i] = new (std::nothrow) OrderMaker[this->relations_no];
		if (this->cnf[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		this->mask[i] = new (std::nothrow) int[this->relations_no];
		if (this->mask[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		recs_no[i] = 0;
		records[i] = NULL;
	   
		hash_table[i] = new (std::nothrow) HashChain** [this->relations_no];
		if (hash_table[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		hash_func[i] = new (std::nothrow) Hash* [this->relations_no];
		if (hash_func[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}


	        // LUIS: WE WILL HAVE TO CHANGE THIS LOOP.
		for (j = 0; j < this->relations_no; j++)
		{
			hash_table[i][j] = NULL;
			hash_func[i][j] = NULL;
		}
	}

	for (i = 0; i < this->relations_no; i++)
	{
		for (j = 0; j < this->relations_no; j++)
		{
			this->mask[i][j] = 0;

			if (j > i)
			{
				cnf[i][j].GetSortOrders(this->cnf[i][j], this->cnf[j][i]);
				this->mask[i][j] = mask[i][j];
			}

/*
			printf("%d %d\n", i, j);
			this->cnf[i][j].Print();
			printf("\n");
*/

		}
	}


	// get the search order
	int starting_node = -1;

	for (i = 0; i < this->relations_no; i++)
	{
		int all_zero = 1;
		for (j = 0; j < this->relations_no; j++)
		{
			if (this->mask[i][j] == 1)
			{
				all_zero = 0;
				break;
			}
		}

		if (all_zero == 1)
		{
			starting_node = i;
			break;
		}

		all_zero = 1;
		for (j = 0; j < this->relations_no; j++)
		{
			if (this->mask[j][i] == 1)
			{
				all_zero = 0;
				break;
			}
		}

		if (all_zero == 1)
		{
			starting_node = i;
			break;
		}
	}

	if (starting_node == -1)
	{
		printf("Error : Join condition in InMemoryJoin is not acceptable !!!\n");
		exit(1);
	}

	pairs_no = GetDFSOrder(starting_node, dfs_order);
	for (i = 0; i < pairs_no - 1; i++)
	{
		if (dfs_order[i].c2 != dfs_order[i+1].c1)
		{
			printf("Error : Join condition in InMemoryJoin after building is not acceptable !!!\n");
			exit(1);
		}
	}

//	cout << "Pairs No = " << pairs_no << "\n";


	colaps_om = new (std::nothrow) OrderMaker[relations_no];
	if (colaps_om == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

        // LUIS: I KNOW THIS IS FOR THE COLLAPSING PART, BUT I DON'T KNOW IF WE HAVE TO CHANGE THIS TO.
	
	colaps_hash = new (std::nothrow) Hash*[relations_no];
	if (colaps_hash == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


        // LUIS: SAME WITH THIS, I DON'T KNOW HOW DOES THIS AFFECT OUR NEW HASHING SCHEME.
	for (i = 0; i < this->relations_no; i++)
	{
		colaps_om[i] = this->cnf[i][0];
		for (j = 1; j < this->relations_no; j++)
		{
			OrderMaker om_old(colaps_om[i]);
			colaps_om[i].ANDMerge(&om_old, &this->cnf[i][j]);
		}

		colaps_hash[i] = new (std::nothrow) CWHash(colaps_om[i].numAtts + 1, HASH_DOMAIN);
		if (colaps_hash[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	}

	this->func = toApply;
}


InMemoryJoin::~InMemoryJoin()
{
	int i, j;

	for (i = 0; i < relations_no; i++)
	{
	        // LUIS: LOOP SIZE CHANGES, TOO.
		for (j = 0; j < relations_no; j++)
		{
			DestroyHashTable(i, j);
		}

		delete [] cnf[i];
		cnf[i] = NULL;

		delete [] mask[i];
		mask[i] = NULL;

		delete [] records[i];
		records[i] = NULL;

		delete [] hash_table[i];
		hash_table[i] = NULL;

		delete [] hash_func[i];
		hash_func[i] = NULL;

		delete colaps_hash[i];
		colaps_hash[i] = NULL;
	}

	delete [] colaps_hash;
	colaps_hash = NULL;

	delete [] colaps_om;
	colaps_om = NULL;

	extra_cond = NULL;
	extra_lit = NULL;

	delete [] cnf;
	cnf = NULL;

	delete [] mask;
	mask = NULL;

	delete [] records;
	records = NULL;

	delete [] hash_table;
	hash_table = NULL;

	delete [] hash_func;
	hash_func = NULL;

	delete [] recs_no;
	recs_no = NULL;

	delete exact_hash;
	exact_hash = NULL;

	delete [] schemaSizes;
	schemaSizes = NULL;

	delete [] schemas;
	schemas = NULL;

	relations_no = 0;

	delete [] dfs_order;
	dfs_order = NULL;

	pairs_no = 0;
}




void InMemoryJoin::Swap(Record *&input, Record *&output, int which_relation, int &output_no)
{
	if (records[which_relation] != NULL)
	{
		// move previous stuff to the output array.
		output = new (std::nothrow) Record[recs_no[which_relation]];
		if (output == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (int i = 0; i < recs_no[which_relation]; i++)
			output[i].Consume(&records[which_relation][i]);

		// destroy the old hash arrays.
		// 
		// LUIS: NOW, IT SEEMS WE WILL ONLY HAVE TO KILL --ONE-- OF THEM. NOW, WHICH ONE?
		for (int j = 0; j < relations_no; j++)
		{
			DestroyHashTable(which_relation,j);
			DestroyHashTable(j,which_relation);
		}

		// destroy the old array.
		delete [] records[which_relation];
		records[which_relation] = NULL;
	}

        // LUIS: THIS IS ALRIGHT.
	// we can put a 0 as a parameter if we only want to get the last records out of the join
	if (output_no > 0)
	{
		// build new array
		records[which_relation] = new (std::nothrow) Record[output_no];
		if (records[which_relation] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		// move everything in the structure.
		for (int x = 0; x < output_no; x++) 
			records[which_relation][x].Consume(&input[x]);

		delete [] input;
		input = NULL;
	}

	// exchange array sizes and stuff
	int size = output_no;
	output_no = recs_no[which_relation];
	recs_no[which_relation] = size;
}



// destroys a given hash_table.
// LUIS: I DON'T THINK IT IS NECESSARY TO CHANGE THIS PART.

void InMemoryJoin::DestroyHashTable(int i, int j)
{
	if (hash_table[i][j] == NULL)
		return;

	for (int k = 0; k < recs_no[i]; k++)
	{
		HashChain *mx = hash_table[i][j][k];

		while (mx != NULL)
		{
			HashChain *tempmx = mx;
			mx = mx->next;
			delete tempmx;
		}
	}

	delete [] hash_table[i][j];
	hash_table[i][j] = NULL;

	delete hash_func[i][j];
	hash_func[i][j] = NULL;
}



// join whatever is in memory
int InMemoryJoin::Join(int **&join_res, double *&output)
{
/*
	int max_res = recs_no[0];
	for (int i = 1; i < relations_no; i++)
		max_res *= recs_no[i];
*/

	int max_res = recs_no[0];
	int is_zero = 0;
	for (int i = 0; i < relations_no; i++)
	{
		if (recs_no[i] == 0)
			is_zero = 1;

		if (recs_no[i] > max_res)
			max_res = recs_no[i];
	}
	max_res = max_res << 2;


	//extreme case : no tuples are in the result
	if (is_zero == 1)
	{
		return 0;
	}


	// build all the necessary hash tables
	// LUIS: WE WILL CHANGE THIS LOOP.
	for (int i = 0; i < pairs_no; i++)
	{
	        // LUIS: ARE THESE STEPS BEING CORRECTLY DONE? IT SEEMS SO!
		//       LOOKS LIKE THE STRUCTURE WE HAVE RIGHT NOW CAN
		//       HANDLE THE NEW (0,1), (1,2), ..., (n-2, n-1) SCHEME
		if (hash_table[dfs_order[i].c1][dfs_order[i].c2] == NULL)
		{
			// build new hash table and initialize.
			hash_table[dfs_order[i].c1][dfs_order[i].c2] = new (std::nothrow) HashChain*[recs_no[dfs_order[i].c1]];
			if (hash_table[dfs_order[i].c1][dfs_order[i].c2] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			hash_func[dfs_order[i].c1][dfs_order[i].c2] = new (std::nothrow) CWHash(cnf[dfs_order[i].c1][dfs_order[i].c2].numAtts + 1, recs_no[dfs_order[i].c1]);
			if (hash_func[dfs_order[i].c1][dfs_order[i].c2] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}


//			hash_func[dfs_order[i].c1][dfs_order[i].c2]->Print_Parameters();


			for (int x = 0; x < recs_no[dfs_order[i].c1]; x++)
				hash_table[dfs_order[i].c1][dfs_order[i].c2][x] = NULL;

//			cout << "lols: " << dfs_order[i].c1 << ", " << dfs_order[i].c2 << "\n";
//			cout << "size: " << recs_no[dfs_order[i].c1] << "," << recs_no[dfs_order[i].c2] << "\n";

		        // LUIS: IT SEEMS THIS PART IS GOING TO BE KIND OF SIMILAR,
			//       BUUUUUUUT... WILL WE BE KEEPING THE HashChain OR
			//       REMOVING IT IN FAVOR OF A stl::vector ?
			
			for (int x = 0; x < recs_no[dfs_order[i].c2]; x++)
			{
//				cout << "1111111111111111\n";

//				records[dfs_order[i].c2][x].Print(schemas[dfs_order[i].c2]);

//				cout << "2222222222222222\n";


				unsigned int k = hash_func[dfs_order[i].c1][dfs_order[i].c2]->Compute_Hash(&records[dfs_order[i].c2][x], &cnf[dfs_order[i].c2][dfs_order[i].c1]);

//				cout << "hash: " << x << " -> " << k << "\n";

				if (hash_table[dfs_order[i].c1][dfs_order[i].c2][k] == NULL)
				{
//					cout << "333333333333333333\n";

					hash_table[dfs_order[i].c1][dfs_order[i].c2][k] = new (std::nothrow) HashChain;
					if (hash_table[dfs_order[i].c1][dfs_order[i].c2][k] == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

				        // LUIS: WE SHOULD ALSO SET THE BOOLEAN FLAG TO FALSE HERE
					hash_table[dfs_order[i].c1][dfs_order[i].c2][k]->rec_index = x;
					hash_table[dfs_order[i].c1][dfs_order[i].c2][k]->next = NULL;
				}
				else
				{
//					cout << "44444444444444444444\n";

					HashChain *mx = new (std::nothrow) HashChain;
					if (mx == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					mx->rec_index = x;
					mx->next = hash_table[dfs_order[i].c1][dfs_order[i].c2][k];
					hash_table[dfs_order[i].c1][dfs_order[i].c2][k] = mx;
				}

//				cout << "55555555555555555555555\n";
			}

//			fflush(NULL);
		}
	}


        // LUIS: SOOO... ARE WE KEEPING THIS AS A GLOBAL?
	int ***results = new (std::nothrow) int**[pairs_no + 1];
	if (results == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

        // LUIS: ALRIGHT, IT SEEMS LIKE WE _WILL_ NOT CHANGE
	//       THE STRUCTURE OF THE results ARRAY ITSELF
	//       BUT RATHER MESS DIRECTLY WITH THE HASHING
	//       PART.
	for (int i = 0; i < pairs_no + 1; i++)
	{
		results[i] = new (std::nothrow) int*[max_res];
		if (results[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (int j = 0; j < max_res; j++)
		{
			results[i][j] = new (std::nothrow) int[relations_no];
			if (results[i][j] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
		}
	}

	int count = recs_no[dfs_order[0].c1];
	for (int i = 0; i < count; i++)
		results[0][i][dfs_order[0].c1] = i;

        // LUIS: OKAY, CURRENTLY WE ARE DOING A _COMPLETE_
        //       TRAVERSAL OF THE PATH, BUT I BELIVE WE CAN
        //       AVOID IT BY HAVING A BOOLEAN VALUE IN EACH
	//       ENTRY OF THE HashChain THAT "TELLS" YOU TO
	//       NOT GO FURTHER IN THE PATH SINCE A Compare()
	//       WAS DONE BEFORE AND WE AREADY KNOW THE RESULT.
	//    
	//       THE PROBLEM IS THAT IT SEEMS WE CANNOT DO THAT
	//       STRAIGHTFORWARDLY WITH THE CURRENT STRUCTURE OF
	//       THIS LOOP. WE WILL PROBABLY NEED TO MAKE "j" THE
	//       EXTERNAL LOOP AND "i" THE INTERNAL, SO THAT WE CAN
	//       STOP IT (THE INTERNAL ONE) WHENEVER WE REACH A
	//       PATH WE ALREADY KNOW OF.

	for (int i = 0; i < pairs_no; i++)
	{
		int new_count = 0;

 	        for (int j = 0; j < count; j++)
		{
//			cout << "testing (" << i << "," << j << ")" << dfs_order[i].c1 << "." << results[i][j][dfs_order[i].c1] << "\n";

			Record *ax = &records[dfs_order[i].c1][results[i][j][dfs_order[i].c1]];
			unsigned int k = hash_func[dfs_order[i].c1][dfs_order[i].c2]->Compute_Hash(ax, &cnf[dfs_order[i].c1][dfs_order[i].c2]);

			HashChain *hc = hash_table[dfs_order[i].c1][dfs_order[i].c2][k];
			while (hc != NULL)
			{
				Record *bx = &records[dfs_order[i].c2][hc->rec_index];
			   
			        // LUIS: WHAT FOLLOWS SEEMS TO BE THE MOST "INTENSIVE" PART OF THE
				//       ALGORITHM, WHICH WE WILL TRY TO AVOID BY HAVING A BOOLEAN
				//       VALUE THAT MEANS WE ALREADY Compared/MOVED/WHATEVER AND
				//       WILL CUT OFF WHAT IS INSIDE, MOVING TO SOME "EASIER" LOOP.
				// 
				//       MOREOVER, IT SEEMS LIKE HashChain IS A VERY MESSY AND
				//       TIME-CONSUMING THING, WHICH WE SHOULD REPLACE ASAP BY 
				//       A std::vector.
			       
				if (exact_hash->Compare(bx, &cnf[dfs_order[i].c2][dfs_order[i].c1], ax, &cnf[dfs_order[i].c1][dfs_order[i].c2]) == 0)
				{
//					cout << "join " << results[i][j][dfs_order[i].c1] << " and " << hc->rec_index << "/" << new_count << "/" << i+1 << "\n";
 
//					cout << "i=" << i << " j=" << j << " newc=" << new_count << "\n";

					for (int z = 0; z < i; z++)
					{
						results[i+1][new_count][dfs_order[z].c1] = results[i][j][dfs_order[z].c1];
						results[i+1][new_count][dfs_order[z].c2] = results[i][j][dfs_order[z].c2];
					}

					results[i+1][new_count][dfs_order[i].c1] = results[i][j][dfs_order[i].c1];
					results[i+1][new_count][dfs_order[i].c2] = hc->rec_index;

					new_count += 1;
					if (new_count == max_res)
					{
						//the size of the InMemoryJoin buffer is too small; it needs to be increased (doubled)
						max_res = max_res << 1;

						int ***aux_results = new (std::nothrow) int**[pairs_no + 1];
						if (aux_results == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
					
						for (int ii = 0; ii < pairs_no + 1; ii++)
						{
							aux_results[ii] = new (std::nothrow) int*[max_res];
							if (aux_results[ii] == NULL)
							{
								cout << "ERROR : Not enough memory. EXIT !!!\n";
								exit(1);
							}
					
							for (int jj = 0; jj < max_res; jj++)
							{
								aux_results[ii][jj] = new (std::nothrow) int[relations_no];
								if (aux_results[ii][jj] == NULL)
								{
									cout << "ERROR : Not enough memory. EXIT !!!\n";
									exit(1);
								}

								if (jj < new_count)
								{
									memcpy(aux_results[ii][jj], results[ii][jj], relations_no * sizeof(int));

									delete [] results[ii][jj];
								}
							}

							delete [] results[ii];
						}

						delete [] results;

						results = aux_results;
					}

				}

				hc = hc->next;
			}
		}

		count = new_count;
	}


/*
	printf("Count=%10d\n", count);
	for (int i = 0; i < count; i++)
	{
		printf("%10d:", i);
		for (int j = 0; j < relations_no; j++)
		{
			printf("%10d", results[pairs_no][i][j]);
		}
		printf("\n");

		for (int j = 0; j < relations_no; j++)
		{
			Record *rec = &records[j][results[pairs_no][i][j]];
			rec->Print(schemas[j]);
		}
		printf("\n");
	}
*/

        // LUIS: WELL, FROM NOW ON IT IS THE COLLAPSING PART, WHICH IS
	//       BEYOND WHAT WE ARE TRYING TO ACHIEVE.

	if (count == 0)
	{
		for (int i = 0; i < pairs_no + 1; i++)
		{
			for (int j = 0; j < max_res; j++)
				delete [] results[i][j];
			delete [] results[i];
		}
		delete [] results;

		return 0;
	}


	join_res = new (std::nothrow) int*[count];
	if (join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	output = new (std::nothrow) double[count];
	if (output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int new_count = 0;
	for (int i = 0; i < count; i++)
	{
		Record *toMerge = NULL;

//		printf("%10d:\n", i);

		int acc_schema_size = 0;
//		Schema cur_schema;

		for (int j = 0; j < relations_no; j++)
		{
//			records[j][p_join_res[i][j]].Print(schemas[j]);

			if (toMerge == NULL)
			{
				toMerge = new (std::nothrow) Record();
				if (toMerge == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				toMerge->Copy(&records[j][results[pairs_no][i][j]]);
				acc_schema_size = schemaSizes[j];
			}
			else
			{
				// merge left (toMerge) with right (from records[][]) into newRec.
				Record *newRec = new (std::nothrow) Record();
				if (newRec == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				int *toKeep = new (std::nothrow) int[acc_schema_size + schemaSizes[j]];
				if (toKeep == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				for (int k = 0; k < acc_schema_size; k++)
					toKeep[k] = k;
				for (int k = 0; k < schemaSizes[j]; k++)
					toKeep[acc_schema_size + k] = k;

				Record *right_rec = &records[j][results[pairs_no][i][j]];
				newRec->MergeRecords(toMerge, right_rec, acc_schema_size, schemaSizes[j], toKeep, acc_schema_size + schemaSizes[j], acc_schema_size);

				acc_schema_size += schemaSizes[j];

				delete toMerge;
				toMerge = newRec;

				delete [] toKeep;
			}

/*
			Schema old_schema;
			old_schema.CopyFrom(&cur_schema);
			cur_schema.CopyFrom(&old_schema, schemas[j]);
			toMerge->Print(&cur_schema);
*/

		}


	        // LUIS: WHAT IS THIS? CAN WE MEMOIZE/AVOID THIS STEP?

		if (exact_hash->Compare(toMerge, extra_lit, extra_cond))
		{
			int intResult;
			double doubleResult;
	
			Type t = func->Apply(toMerge, intResult, doubleResult);
	
			if (t == Int)
				output[new_count] = (double)intResult;
			else if (t == Double)
				output[new_count] = doubleResult;


			join_res[new_count] = new (std::nothrow) int[relations_no];
			if (join_res[new_count] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			for (int j = 0; j < relations_no; j++)
			{
				join_res[new_count][j] = colaps_hash[j]->Compute_Hash(&records[j][results[pairs_no][i][j]], &colaps_om[j]);
			}

			new_count += 1;
		}

//		printf("Aggregate: %20.4f\n", p_output[i]);

		delete toMerge;
	}

	count = new_count;


	for (int i = 0; i < pairs_no + 1; i++)
	{
		for (int j = 0; j < max_res; j++)
			delete [] results[i][j];
		delete [] results[i];
	}
	delete [] results;


//	cout << "InMemoryJoinTuples: " << count << "\n";

	return count;
}



// does a DFS to find the order in which to process stuff, beginning with a given relation k.
int InMemoryJoin::GetDFSOrder(int starting_node, Pair *&output)
{
	int idx = 0;

	output = new (std::nothrow) Pair[(relations_no * (relations_no - 1)) / 2];
	if (output == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	stack<int> Q;
	Q.push(starting_node);


	int **visited = new (std::nothrow) int*[relations_no];
	if (visited == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < relations_no; i++)
	{
		visited[i] = new (std::nothrow) int[relations_no];
		if (visited[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (int j = 0; j < relations_no; j++)
			visited[i][j] = 0;
	}


	while (!Q.empty())
	{
		int k = Q.top();
		Q.pop();

		for (int i = 0; i < relations_no; i++)
		{
			if (!visited[i][k] && mask[i][k])
			{			
				output[idx] = (Pair){k,i};
				idx += 1;
				visited[i][k] = 1;
				Q.push(i);
			}

			if (!visited[k][i] && mask[k][i])
			{
				output[idx] = (Pair){k,i};
				idx += 1;
				visited[k][i] = 1;
				Q.push(i);
			}
		}
	}


	for (int i = 0; i < relations_no; i++)
		delete [] visited[i];
	delete [] visited;


	return idx;
}


