#include "InMemoryJoinWrapper.h"

#include <iostream>
#include <stack>
#include <stdlib.h>

#include<vector>


using namespace std;



InMemoryJoinWrapper::InMemoryJoinWrapper(int numRels, int **mask, Schema **relSchemas, struct AndList *myCondition, struct FuncOperator *funcParseTree, double softBound, double hardBound, struct AttrGroupbyList *myGroupBy)
{
	this->relations_no = numRels;

	recs_no = new (std::nothrow) int[this->relations_no];
	if (recs_no == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	records = new (std::nothrow) Record* [this->relations_no];
	if (records == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	for (int i = 0; i < this->relations_no; i++)
	{
		records[i] = NULL;
	}

	perm_order = new (std::nothrow) int[this->relations_no];
	if (perm_order == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	imj = new (std::nothrow) NewInMemoryJoinWrapper(numRels, mask, relSchemas, myCondition, funcParseTree, softBound, hardBound, myGroupBy);
	if (imj == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	imj->InitialConstRel(0, NULL);

	first_join = 0;

	Result.numRes = 0;
	Result.numDel = 0;
}


InMemoryJoinWrapper::~InMemoryJoinWrapper()
{
	Result.Print();

	int i;

	for (i = 0; i < relations_no; i++)
	{
		delete [] records[i];
		records[i] = NULL;
	}

	delete [] records;
	records = NULL;

	delete [] recs_no;
	recs_no = NULL;

	delete [] perm_order;
	perm_order = NULL;

	delete imj;
	imj = NULL;
}




void InMemoryJoinWrapper::Swap(Record *&input, Record *&output, int which_relation, int &output_no)
{
	if (records[which_relation] != NULL)
	{
		// move previous stuff to the output array.
		output = new (std::nothrow) Record[recs_no[which_relation]];
		if (output == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		for (int i = 0; i < recs_no[which_relation]; i++)
			output[i].Consume(&records[which_relation][i]);

		// destroy the old array.
		delete [] records[which_relation];
		records[which_relation] = NULL;
	}

	// we can put a 0 as a parameter if we only want to get the last records out of the join
	if (output_no > 0)
	{
		// build new array
		records[which_relation] = new (std::nothrow) Record[output_no];
		if (records[which_relation] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		// move everything in the structure.
		for (int x = 0; x < output_no; x++) 
			records[which_relation][x].Consume(&input[x]);

		delete [] input;
		input = NULL;
	}

	// exchange array sizes and stuff
	int size = output_no;
	output_no = recs_no[which_relation];
	recs_no[which_relation] = size;
}



// join whatever is in memory
SearchResults* InMemoryJoinWrapper::Join(double *threshold, int grouping)
{
	SearchResults *results = new (std::nothrow) SearchResults();
	if (results == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	vector<SearchResults*> *partial_results = new (std::nothrow) vector<SearchResults*>;
	if (partial_results == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	partial_results->clear();

	int results_no = 0;

	if (first_join == 0)
	{
		for (int i = 0; i < relations_no; i++)
		{
			for (int j = 0; j < recs_no[i]; j++)
			{
				Record *rec_tmp = new (std::nothrow) Record();
				if (rec_tmp == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
				rec_tmp->Copy(&records[i][j]);

				imj->AddStartTuples(*rec_tmp, i, i);

				delete rec_tmp;
			}
		}

		SearchResults *myResult = imj->ChoosePermOrder(threshold);
		if (myResult != NULL)
		{
			delete myResult;
		}

		imj->GetPermOrder(perm_order);

		imj->Erase();

		for (int i = 0; i < relations_no; i++)
		{
			for (int j = 0; j < recs_no[perm_order[i]]; j++)
			{
				Record *rec_tmp = new (std::nothrow) Record();
				if (rec_tmp == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
				rec_tmp->Copy(&records[perm_order[i]][j]);

				SearchResults *myResult = imj->ExternInsertRec(*rec_tmp, perm_order[i], rec_tmp->hash);
				if (myResult != NULL)
				{
					Result.numRes += myResult->numRes;
					Result.numDel += myResult->numDel;

					results_no += myResult->numRes;

					partial_results->push_back(myResult);
				}

				delete rec_tmp;
			}
		}

		first_join = 1;
	}
	else
	{
		imj->Erase();

		for (int i = 0; i < relations_no; i++)
		{
			for (int j = 0; j < recs_no[perm_order[i]]; j++)
			{
				Record *rec_tmp = new (std::nothrow) Record();
				if (rec_tmp == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
				rec_tmp->Copy(&records[perm_order[i]][j]);

				SearchResults *myResult = imj->ExternInsertRec(*rec_tmp, perm_order[i], rec_tmp->hash);
				if (myResult != NULL)
				{
					Result.numRes += myResult->numRes;
					Result.numDel += myResult->numDel;

					results_no += myResult->numRes;

					partial_results->push_back(myResult);
				}

				delete rec_tmp;
			}
		}
	}

	results->numRes = results_no;

	cout << "IMJ Results = " << results_no << "\n";

	results->resTuples = new (std::nothrow) int*[results_no];
	if (results->resTuples == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	results->funcRes = new (std::nothrow) double[results_no];
	if (results->funcRes == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	if (grouping != -1)
	{
		results->resGroups = new (std::nothrow) Record[results_no];
		if (results->resGroups == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	}

	int k = 0;

	int size = partial_results->size();
	for (int i = 0; i < size; i++)
	{
		SearchResults *r = (*partial_results)[i];
		for (int j = 0; j < r->numRes; j++)
		{
			results->resTuples[k] = new (std::nothrow) int[relations_no];
			if (results->resTuples[k] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			memcpy(results->resTuples[k], r->resTuples[j], sizeof(int) * relations_no);


/*
			for (int jj = 0; jj < relations_no; jj++)
			{
				printf("%d\t", results->resTuples[k][jj]);
				printf("%d\t", r->resTuples[j][jj]);
			}
			cout << endl;
*/

			results->funcRes[k] = r->funcRes[j];

			if (grouping != -1)
			{
				results->resGroups[k].Consume(&r->resGroups[j]);
			}

			k += 1;
		}

		delete r;
		(*partial_results)[i] = NULL;
	}

	partial_results->clear();
	delete partial_results;

	return results;
}

