#include "Schema.h"
#include "Comparison.h"
#include "Record.h"
#include "Pipe.h"
#include "RelOpLevel.h"
#include "Levelwise.h"


#include <vector>
#include <iostream>


LevelwiseStep::LevelwiseStep(int base_rels_no, int *relation_size)
{
	operators.clear();
	out_pipe_sizes.clear();

	this->base_rels_no = base_rels_no;
	this->relation_size = relation_size;

}

LevelwiseStep::~LevelwiseStep()
{
	int i, n;

	n = operators.size();
	for (i = 0; i < n; i++)
	{
		RelationalOpLevel *e = operators[i];
		delete e;
		e = NULL;
	}
	operators.clear();

	out_pipe_sizes.clear();
}


int LevelwiseStep::GetNextBatch(int which_pipe, double p, Record *output, int max_output_size, int &output_size)
{
	int res = operators[which_pipe]->GetNextBatch(p, output, max_output_size, output_size);
	return res;
}


void LevelwiseStep::StartUp()
{
	int i, n;

	n = operators.size();
	for (i = 0; i < n; i++)
	{
		operators[i]->StartUp();
	}
}


double LevelwiseStep::HowMuchIsLeft(int which_pipe)
{
	double res = operators[which_pipe]->HowMuchIsLeft();
	return res;
}


void LevelwiseStep::Append(const char *file_location, Pipe *out_pipe, int out_pipe_size, CNF *sel_op, Record *literal, Schema *schema, int *atts, int atts_no, int tuples_no)
{
	//R-STORE VS C-STORE

	RelationalOpLevel *t = new (std::nothrow) TableScanLevel(file_location, out_pipe, sel_op, literal, schema, atts, atts_no, tuples_no);
//	RelationalOpLevel *t = new (std::nothrow) TableScanLevelC(file_location, out_pipe, sel_op, literal, schema, atts, atts_no, tuples_no);
	if (t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	operators.push_back(t);

	out_pipe_sizes.push_back(out_pipe_size);
}


void LevelwiseStep::Append(char **file_location, Pipe *out_pipe, int out_pipe_size, Schema **schema, int **proj_atts, int *proj_atts_no, int *proj_atts_out, int proj_atts_out_no, AndList *conditions, int *tuples_no, int relations_no)
{
	RelationalOpLevel *t = new (std::nothrow) TableScanPreJoinLevel(file_location, out_pipe, schema, proj_atts, proj_atts_no, proj_atts_out, proj_atts_out_no, conditions, tuples_no, relations_no);
	if (t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	operators.push_back(t);

	out_pipe_sizes.push_back(out_pipe_size);
}


void LevelwiseStep::Append (Pipe *in_pipe_left, int pipe_left_index, Pipe *in_pipe_right, int pipe_right_index, Pipe *out_pipe, int out_pipe_size, CNF *sel_op, Record *literal, Schema *schema_left, Schema *schema_right, int *atts, int atts_no, int buffer_size)
{
	RelationalOpLevel *t = new (std::nothrow) SortJoinLevel(in_pipe_left, pipe_left_index, in_pipe_right, pipe_right_index, out_pipe, sel_op, literal, schema_left, schema_right, atts, atts_no, buffer_size);
	if (t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	operators.push_back(t);

	out_pipe_sizes.push_back(out_pipe_size);
}


void LevelwiseStep::Append(Pipe *in_pipe, int pipe_index, Pipe *out_pipe, int out_pipe_size, Schema *schema, int buffer_size)
{
	RelationalOpLevel *t = new (std::nothrow) RehashLevel(in_pipe, pipe_index, out_pipe, schema, buffer_size);
	if (t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	operators.push_back(t);

	out_pipe_sizes.push_back(out_pipe_size);
}


void LevelwiseStep::Append(Pipe *input_pipe, int pipe_index, Schema *schema, Function *func, int *join_atts, int join_atts_no)
{
	RelationalOpLevel *t = new (std::nothrow) WriteOutLevel(input_pipe, pipe_index, schema, func, join_atts, join_atts_no);
	if (t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	operators.push_back(t);
}


void LevelwiseStep::Append(Pipe *input_pipe, int pipe_index, Schema *schema, Function *func, int *join_atts, int join_atts_no, int *group_atts, int group_atts_no, int buffer_size)
{
	RelationalOpLevel *t = new (std::nothrow) GroupAggregateLevel(input_pipe, pipe_index, schema, func, join_atts, join_atts_no, group_atts, group_atts_no, buffer_size);
	if (t == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	operators.push_back(t);
}




int LevelwiseStep::GetSize()
{
	int n = operators.size();
	return n;
}


double LevelwiseStep::GetTuplesNo(int which_pipe)
{
	double res = operators[which_pipe]->GetTuplesNo();
	return res;
}


double LevelwiseStep::GetAggregate(int which_pipe)
{
	double res = operators[which_pipe]->GetAggregate();
	return res;
}


Schema* LevelwiseStep::GetSchema(int which_pipe)
{
	Schema* res = operators[which_pipe]->GetSchema();
	return res;
}


Function* LevelwiseStep::GetFunction(int which_pipe)
{
	Function* res = operators[which_pipe]->GetFunction();
	return res;
}


void LevelwiseStep::GetJoinAtts(int which_pipe, int *&join_atts, int &join_atts_no)
{
	operators[which_pipe]->GetJoinAtts(join_atts, join_atts_no);
	return;
}


void LevelwiseStep::GetGroupAtts(int which_pipe, int *&group_atts, int &group_atts_no)
{
	operators[which_pipe]->GetGroupAtts(group_atts, group_atts_no);
	return;
}


void LevelwiseStep::GetGroups(int which_pipe, vector<Record*> &desc, vector<double> &aggregate, vector<double> &tuples, int &groups_no)
{
	operators[which_pipe]->GetGroups(desc, aggregate, tuples, groups_no);
	return;
}

