#ifndef LEVELWISE_H
#define LEVELWISE_H

#include "Schema.h"
#include "Comparison.h"
#include "Record.h"
#include "Pipe.h"
#include "Function.h"
#include "RelOpLevel.h"

#include <vector>


using namespace std;


class LevelwiseStep {

public:
	LevelwiseStep(int base_rels_no, int *relation_size);
	virtual ~LevelwiseStep();

	// This asks the relational operation associated with the i^th output
	// pipe to put copies of the next p% of the tuples that the operation
	// will produce into the array output.  The call to GetNextBatch is 
	// blocking, but the LevelwiseStep class always does as much work as
	// it can after StartUp is called, but before GetNextBatch is 
	// actually called

	// For example, in the
	// case of a levelwise step comprised of joins, before GetNextBatch is
	// ever called the levelwise step will suck in all of the records from
	// all of the input streams, break them into sorted runs, and write the
	// runs to disk.  Then it will set up the merges, and stop and wait until
	// GetNextBatch is called in order to run the merge
	//
	// If p is negative, then the output stream is written to output in its
	// entirety, no matter how much of the output stream it contains
	virtual int GetNextBatch(int which_pipe, double p, Record *output, int max_output_size, int &output_size);

	// This tells the levelwise step that it is time to start up; this function
	// should be called after all of the operations have been added to the
	// step and before GetNextBatch is ever called
	virtual void StartUp();

	// This asks the i^th output pipe what percentage of its records still
	// remain to be processed
	virtual double HowMuchIsLeft(int which_pipe);


	int GetSize();

	double GetTuplesNo(int which_pipe);
	double GetAggregate(int which_pipe);
	Schema* GetSchema(int which_pipe);
	Function* GetFunction(int which_pipe);
	void GetJoinAtts(int which_pipe, int *&join_atts, int &join_atts_no);
	void GetGroupAtts(int which_pipe, int *&group_atts, int &group_atts_no);

	void GetGroups(int which_pipe, vector<Record*> &desc, vector<double> &aggregate, vector<double> &tuples, int &groups_no);


        void Append(const char *file_location, Pipe *out_pipe, int out_pipe_size, CNF *sel_op, Record *literal, Schema *schema, int *atts, int atts_no, int tuples_no);

	void Append(char **file_location, Pipe *out_pipe, int out_pipe_size, Schema **schema, int **proj_atts, int *proj_atts_no, int *proj_atts_out, int proj_atts_out_no, AndList *conditions, int *tuples_no, int relations_no);

	void Append(Pipe *in_pipe_left, int pipe_left_index, Pipe *in_pipe_right, int pipe_right_index, Pipe *out_pipe, int out_pipe_size, CNF *sel_op, Record *literal, Schema *schema_left, Schema *schema_right, int *atts, int atts_no, int buffer_size);

	void Append(Pipe *in_pipe, int pipe_index, Pipe *out_pipe, int out_pipe_size, Schema *schema, int buffer_size);

        void Append(Pipe *input_pipe, int pipe_index, Schema *schema, Function *func, int *join_atts, int join_atts_no);

        void Append(Pipe *input_pipe, int pipe_index, Schema *schema, Function *func, int *join_atts, int join_atts_no, int *group_atts, int group_atts_no, int buffer_size);


	int base_rels_no;
	int *relation_size;

	vector<RelationalOpLevel*> operators;
	vector<int> out_pipe_sizes;
};

#endif
