/***********************************************************************
 * SECTION 1
 ***********************************************************************/
/* The code in %{ %} is included as it is in lex.yy.c file
 * it has C global variables, prototypes, and comments 
 */


%{

#include <string.h> // For strdup()
#include <stdlib.h> // For malloc()

#include "ParseTree.h"
#include "y.tab.h"

int lineno = 1;
void yyerror(char*s);

%}


/******************************************************************************
 * SECTION 2
 ******************************************************************************/
/* This is the DEFINITION section which contains substitutions, code, and
 * start stats; will be copied into lex.yy.c
 */

/******************************************************************************
 * SECTION 3
 ******************************************************************************/
/* This is the RULES section which defines how to "scan" and what action
 * to take for each token
 */

%%

"("			return('(');

")"    	        	return(')');

"<"    		      	return('<');

">"    		 	return('>');

"=" 			return('=');

","			return(',');

"+"    	        	return('+');

"-"    	        	return('-');

"/"    	        	return('/');

"*"    	        	return('*');

			
-?[0-9]+ 		{
				yylval.actualChars = strdup(yytext);
  				return (Int);
		        }

-?[0-9]+\.[0-9]*	{
				yylval.actualChars = strdup(yytext);
				return (Float);
			}

"OR"      		return(OR);

"AND"			return(AND);

"BEGIN"			return(Begin);

"END"			return(END);

"TABLE_SCAN"		return(TABLE_SCAN);

"RELATION"		return(RELATION);

"OUTPUT_PIPE"		return(OUTPUT_PIPE);

"INMEMORYJOINBUFFERSIZE"	return(INMEMORYJOINBUFFERSIZE);

"RENAME_ATTRIBUTES"	return(RENAME_ATTRIBUTES);

"AS"			return(AS);

"ACCEPT"		return(ACCEPT);

"SORT_JOIN"		return(SORT_JOIN);

"INPUT_PIPE"		return(INPUT_PIPE);

"MEMORY"		return(MEMORY);

"TUPLES"		return (TUPLES);

"LEVELWISE"		return (LEVELWISE);

"STEP"			return  (STEP);

"AGGREGATE"		return (AGGREGATE);

"REHASHING"		return (REHASHING);

"GROUPBY"		return (GROUPBY);

"NUMGROUPSTODISPLAY"	return (NUMGROUPSTODISPLAY);

"NUMTUPLEFORSTATS"	return (NUMTUPLEFORSTATS);

"SOFTBOUND"		return (SOFTBOUND);

"HARDBOUND"		return (HARDBOUND);

"QUEUESIZE"		return (QUEUESIZE);

"MULTIRELATION"		return (MULTIRELATION);

\'[^'\n]*\'		{/* take care of ' in a string */
	 			if (yytext[yyleng - 2] == '\\') {
					yymore();
				}
				else {
					char *ss = (char*) malloc(sizeof(char) * yyleng - 1);
					strncpy(ss, yytext + sizeof(char), yyleng - 2);
					ss[yyleng - 2] = '\0';

					yylval.actualChars = strdup(ss);
					free(ss);

					return (String);
   				}
			}


["/""."A-Za-z]["/""."A-Za-z0-9_-]*	{
						yylval.actualChars = strdup(yytext);
						return (Name);
					}

\n                	lineno++;

[ \t]             	;

.                 	yyerror("LEX_ERROR: invalid character");

%%

void yyerror(char *s) {
  printf("%d: %s at %s\n", lineno, s, yytext);
}

int yywrap(){
  return 1;
}


