#include "MMIndex.h"
#include "File.h"

#include <string.h>
#include <iostream.h>



MMIndex :: MMIndex(OrderMaker *order, int recs_no) {
	this->hash_size = (recs_no >> 5) > 0 ? recs_no >> 5 : 1;
	this->hash_table = new (std::nothrow) TwoWayList<Record>[this->hash_size];
	if (this->hash_table == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->order = order;

	this->hash_func = new (std::nothrow) CWHash(this->order->numAtts + 1, this->hash_size);
	if (this->hash_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	this->exact_hash = new (std::nothrow) Hash();
	if (this->exact_hash == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}


MMIndex :: ~MMIndex() {
	delete [] hash_table;

	delete hash_func;
	delete exact_hash;
}



/*
Version never used
void MMIndex :: BuildIndex(char *run_name) {
	File run_file;
	run_file.Open(1, run_name);

	Record *temp_rec = new (std::nothrow) Record();
	if (temp_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	while (run_file.NextRec(temp_rec)) {
		int hash_index = hash_func->Compute_Hash(temp_rec, order);
		hash_table[hash_index].Insert(temp_rec);
	}
	delete temp_rec;

	run_file.Close();
}
*/



void MMIndex :: BuildIndex(Record *rec_array, int recs_no) {
	for (int i = 0; i < recs_no; i++) {
		int hash_index = hash_func->Compute_Hash(&rec_array[i], order);
		hash_table[hash_index].Insert(&rec_array[i]);
	}
}



int MMIndex :: Find(Record *rec, OrderMaker *rec_order, TwoWayList<Record> *match_list) {
	int res_size = 0;

	int hash_index = hash_func->Compute_Hash(rec, rec_order);

	hash_table[hash_index].MoveToStart();
	while (hash_table[hash_index].RightLength() > 0) {

		Record *cur_rec = hash_table[hash_index].Current(0);

		if (exact_hash->Compare(cur_rec, order, rec, rec_order) == 0) {
			Record *new_rec = new (std::nothrow) Record();
			if (new_rec == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			new_rec->Copy(cur_rec);
			match_list->Insert(new_rec);
			delete new_rec;

			res_size += 1;
		}

		hash_table[hash_index].Advance();
	}

	return res_size;
}

