#include <set>
#include <cstdlib>

#include "NewInMemoryJoin.h"
#include "TwoWayList.cc"


NewInMemoryJoin :: NewInMemoryJoin(int numRels, int **mask, Schema **relSchemas, struct AndList *myCondition, struct FuncOperator *funcParseTree, double softBound, double hardBound,struct AttrGroupbyList *myGroupBy)
{
    
    this->myGroupBy = myGroupBy;
    groupAtts = NULL;
    numGroupAtts = -1;	

    unsigned short int timet =(unsigned short) time(0);
    seed48(&timet);
    srand(time(0));

    this->myStatModule = NULL;

    this->numRels = numRels;
	
	this->mask = new int* [numRels];

    if(this->mask == NULL)
	{
		cerr<<"Memory Allocation error in NewInMemoryJoin :: NewInMemoryJoin. \n";
		exit(1);	
	}

    for(int i=0; i< numRels;++i)
	{
		this->mask[i] = new int[numRels];

        if(this->mask[i] == NULL)
		{
			cerr<<"Memory Allocation error in NewInMemoryJoin :: NewInMemoryJoin. \n";
			exit(1);	
		}		
	}

	for(int i=0; i<numRels; ++i)
	{
	   for(int j=0; j<numRels; ++j)
	   {
		if (i == j)
			this->mask[i][j] = 0;
		else
			this->mask[i][j] = mask[i][j];
	   }
	}
	
	this-> relSchemas = new Schema[numRels];

    if(this->relSchemas == NULL)
	{
		cerr<<"Memory Allocation error in NewInMemoryJoin :: NewInMemoryJoin. \n";
		exit(1);	
	}

    for(int i=0; i<numRels; ++i)
		this->relSchemas[i].CopyFrom(relSchemas[i]);

    this->myCondition = myCondition;

    this->funcParseTree = funcParseTree;	

    PermutationOrder = NULL;

    myPartialResults = NULL;

    constRels = NULL;

    numConstRels = 0;

    myPartialResults = NULL;

    IndextoPerm.clear();

    consttoPerm.clear();

    myStartTuples = new RecordArray[numRels];

    allStartTuples = new RecordArray;

    relations = new vector<int>;

    relations->clear();

    this->softBound = softBound;

    this->hardBound = hardBound;

    insertedTimeStamps = new vector <int*>;
    insertedTimeStamps->clear();

    insertedAggregates = new vector<double>;
    insertedAggregates->clear();

    insertedGroups = new vector<Record*>;
    insertedGroups->clear();	

    deletedTimeStamps = new vector <int*>;
    deletedTimeStamps->clear();

    deletedAggregates = new vector <double>;
    deletedAggregates->clear();

    deletedGroups = new vector<Record*>;
    deletedGroups->clear();

    myFunc = new Function();
    constMask = new double[numRels];
    for(int i=0;i<numRels;++i)
    constMask[i] = 0.0;
   
    numTuplesReceived = new int [numRels];
    for(int i =0; i<numRels; ++i)
	numTuplesReceived[i] = 0;

    myConstRelRecs = NULL;
}

void NewInMemoryJoin :: SetPermOrder(int * orders)
{
	PermutationOrder = new int[numRels];

    for(int i=0; i<numRels;++i)
		PermutationOrder[i]=orders[i];
}

void NewInMemoryJoin :: InitialConstRel(int numConstRels, int *constRels)
{
	this->numConstRels = numConstRels;

    if(numConstRels == 0) 
		return;

    this->constRels = new int[numConstRels];

    if(this->constRels == NULL)
	{
		cerr<<"Memory Allocation error in NewInMemoryJoin :: InitialConstRel. \n";
		exit(1);	
	}

    for(int i=0; i<numConstRels; ++i)
    {	
	this->constRels[i] = constRels[i];
	this->constMask[constRels[i]] = 1.0;
    }
    for(int i=0; i<numConstRels; ++i)
		this->consttoPerm[constRels[i]] = i;


    myConstRelRecs = new ConstRelPair[numConstRels];

    	    
}


NewInMemoryJoin :: ~NewInMemoryJoin()
{
    if(groupAtts!=NULL)
      delete[] groupAtts;
  
   delete [] relSchemas;

    for(int i=0; i<numRels;++i)
        delete [] mask[i];

    delete [] mask;

    delete []constRels;
	//we also need to delete the partialJoinResult

    delete [] myPartialResults;

    delete [] PermutationOrder;

    delete [] myStartTuples;

    delete  allStartTuples;


    relations->clear();

    delete relations;

    insertedTimeStamps->clear();
    delete insertedTimeStamps;

    insertedAggregates->clear();
    delete insertedAggregates;

    insertedGroups->clear();
    delete insertedGroups;

    deletedTimeStamps->clear();
    delete deletedTimeStamps;

    deletedAggregates->clear();
    delete deletedAggregates;

    deletedGroups->clear();
    delete deletedGroups;

    delete myFunc;
    delete constMask;


/*
    for (int i=0; i<numRels; ++i)
    cout<<"relation"<<i<<" receives "<<numTuplesReceived[i] <<" of records.\n";	
*/


    delete [] numTuplesReceived;

    delete [] myConstRelRecs;
}

NewInMemoryJoin :: PartialJoinResults :: ~PartialJoinResults()
{
	delete myRecs;

    delete myExtendedHash;

    delete outputSchema;

    delete resultHashingOrderMaker;

    delete inputHashingOrderMaker;	

    delete resultCNF;

    delete resultLit;

    delete myInputHash;//this is used for hash input records

}

NewInMemoryJoin :: ConstRelPair :: ~ConstRelPair()
{
	delete myRecs;

    delete myExtendedHash;
}

NewInMemoryJoin :: ConstRelPair :: ConstRelPair()
{
	myRecs = NULL;

    	myExtendedHash = NULL;
}

//a key function for In-memory-join to work
void NewInMemoryJoin :: PrepareParameters()
{
	
//some of the code should be modified when the start phase is added	
	
	//add PermutationOrder here for testing purpose
	//cout<<"numRels "<<numRels<<endl;
	for(int i=0; i<numRels; ++i)
	{
		//cout<<"I am in 123\n";
		IndextoPerm[PermutationOrder[i]] = i;
	
	}
	
	
	int tmpsize=numRels;
	myPartialResults = new struct PartialJoinResults[tmpsize];
		
	for(int i =0 ; i<numRels; ++i)
	{
		//cout<<"I am in 134 \n";
		myPartialResults[i].isConst = false;
        myPartialResults[i].ratio = 1.0;

        myPartialResults[i].currentTimeStamp = -1;
        myPartialResults[i].currentCoinFlip = -1.0;
    }

	for(int i=0; i<numConstRels; ++i)
	{
		
		myPartialResults[IndextoPerm[constRels[i]]].isConst = true;
	}

	
	AndListWrapper myWrapper(myCondition);
	
	for(int i =0; i<tmpsize;++i)
	{
	
	myPartialResults[i].myRecs = new RecordArray(i+1);
        myPartialResults[i].outputSchema = new Schema;
        myPartialResults[i].resultCNF = new CNF;
        myPartialResults[i].resultLit = new Record;	
        myPartialResults[i].resultHashingOrderMaker = new OrderMaker;
        myPartialResults[i].inputHashingOrderMaker = new OrderMaker;
        myPartialResults[i].myInputHash = NULL;
        myPartialResults[i].myExtendedHash = NULL;
	
        if(i ==0 )
		{
			myPartialResults[i].outputSchema->CopyFrom(&relSchemas[PermutationOrder[i]]);
						
			//NOTICE,WE DO not need to check the CNF for the first 
			//relation because each record sent from the controller should be correct and the record only needs to be inserted directly and hashed directly
		}
					
		//notice for the first relation in the permutation, the only thing we need 
		//to do is to insert the record, so except the outputSchema and RecordArray, we 
		//are not required to do anything, though we need to obtain resultHashingOrderMaker
		//which is always done in the next iteration of this for loop

        else 
		{
			
			myPartialResults[i].outputSchema->
			CopyFrom(myPartialResults[i-1].outputSchema, &relSchemas[PermutationOrder[i]]);	
			//we should prepare several parameters
			//notice the CNF requires two schemas when applying the compair function.

            myWrapper.BuildCNF(*(myPartialResults[i-1].outputSchema), relSchemas[PermutationOrder[i]], *myPartialResults[i].resultCNF, *myPartialResults[i].resultLit);

            myPartialResults[i].resultCNF->GetSortOrders(*myPartialResults[i-1].resultHashingOrderMaker, *myPartialResults[i].inputHashingOrderMaker);

            myPartialResults[i].myInputHash = new CWHash(myPartialResults[i].inputHashingOrderMaker->numAtts+1,MAXBUCKET);

            myPartialResults[i-1].myExtendedHash = new ExtendedHash(*myPartialResults[i-1].resultHashingOrderMaker, myPartialResults[i].myInputHash);
			
		}
	}

    if(funcParseTree != NULL)
        myFunc->GrowFromParseTree(funcParseTree, (myPartialResults[tmpsize - 1].outputSchema));
	
    //for groupby
      if(myGroupBy!=NULL)
	{
		
		vector<int>tempGroupAtts;
		int tempIndexGroupby = numRels -1;
		for(struct AttrGroupbyList* current = myGroupBy; current!=NULL;current=current->NextAttr)
		{
			int w=  myPartialResults[tempIndexGroupby].outputSchema->Find(current->CurrentAttr);
			if(w==-1)
			{
				cout<<"Could not find GROUPBY attribute \""<<current->CurrentAttr<<"\" in InMemoryJoinOutSchema\n";
				exit(1);
			}
			tempGroupAtts.push_back(w);
		}
		numGroupAtts = tempGroupAtts.size();
		groupAtts = new int[numGroupAtts];
		for(int i=0;i<numGroupAtts;++i)
			groupAtts[i]=tempGroupAtts[i];
	}
    	
}
//this is called before insert a rec
//this function hashes the constant recs + put all start tuples into the system
//based on their timestamp + if the first relation is a constant, then put 
//recs from this record in

void NewInMemoryJoin :: PreInsertRec()
{
	//also need to add in the code to create the extended hash object for 
	//each constant relation and hash all records in each constant relation
	for(int i=0; i<numConstRels; ++i)
	{
		int t =IndextoPerm[myConstRelRecs[i].relIndex];
		//the case that the first relation is constant is special
		//we do not need to hash it.

        if(t!=0)
		{
			myConstRelRecs[i].myExtendedHash = new ExtendedHash(
				*(myPartialResults[t].inputHashingOrderMaker), myPartialResults[t].myInputHash);
			//we want to hash each record to build index

            for(int j = 0, size = myConstRelRecs[i].myRecs->GetNumOfRecs(); j < size; ++j)
			{
				myConstRelRecs[i].myExtendedHash->AddRecIndex(
				*(myConstRelRecs[i].myRecs->allRecords[j]), j);
			}
		}
	}
	//we also need to start joining tuples when the first several relaitons
	//in the permutation are constant relations
	//We only need to add recs from the first relation if it is a constant
	//since we will directly check the next relation is a constant or not in 
	//AddRec(), for any constant relation after the first one, it should be 
	//automatically taken care of.
	if(myPartialResults[0].isConst==true)
	{
		int w = consttoPerm[PermutationOrder[0]];

        Record  tmpRec;

        for(int t=0, tmpsize = myConstRelRecs[w].myRecs->GetNumOfRecs(); t < tmpsize;++t)
		{

            tmpRec.Copy(myConstRelRecs[w].myRecs->allRecords[t]);

            InsertRec (tmpRec, PermutationOrder[0], t, myConstRelRecs[w].myRecs->ratios[t][0]);
		}
	}

    //now we insert all records obtained from the start phase.
    for(int t =0, tmpsize = allStartTuples->GetNumOfRecs(); t < tmpsize; ++t)
	{

        InsertRec(*(allStartTuples->allRecords[t]), (*relations)[t],  allStartTuples->timeStamps[t][0]);
	
	}

	delete allStartTuples;
	allStartTuples = NULL;
	
}

void NewInMemoryJoin :: Print()
{
	for(int i =0; i<numRels;++i)
	{
		
		cout<<"Permutation:"<<PermutationOrder[i]<<endl;

        if(i == 0)
		{
			myPartialResults[0].outputSchema->Print();
			myPartialResults[0].resultHashingOrderMaker->Print();
			//cout<<"now print the extended hash\n";
			//myPartialResults[0].myExtendedHash->myOrderMaker.Print();
			//myPartialResults[i].myRecs->Print(0, *myPartialResults[i].outputSchema);
		}

		else
		{

            myPartialResults[i].outputSchema->Print();

            myPartialResults[i].resultCNF->Print();

            myPartialResults[i].inputHashingOrderMaker->Print();

            myPartialResults[i].resultHashingOrderMaker->Print();
		}
	}
}



void NewInMemoryJoin :: InsertRec (Record& rec, int whichRel, int timeStamp)
{

    int t = IndextoPerm[whichRel];

    
    if(myPartialResults[t].currentTimeStamp != timeStamp)
    {
        myPartialResults[t].currentTimeStamp = timeStamp;
        myPartialResults[t].currentCoinFlip = drand48();
    }
    InsertRec(rec, whichRel, timeStamp, myPartialResults[t].currentCoinFlip);

}


//this is just a version that contains ratio, used for constant relation insertion
//note anytime when a constant relation tuple is inserted(happened in function PreInsertion
// and these two copies of InsertRec in recursively calling insertRec
void NewInMemoryJoin :: InsertRec (Record& rec, int whichRel, int timeStamp, double ratio)
{
   
	if(timeStamp <0)
	{
		cout<<"This is a negative timeStamp: "<<timeStamp<<endl;
		exit(1);
	}
    //first flip a coin to generate a ratio used for dropping later on
    double dp=ratio;
  
    //there is a problem for a constant relation.
    //in such a case, the random stuff should be fixed.
	int t = IndextoPerm[whichRel];

    //if the ratio is smaller than dp, we need to throw the record and do nothing
    if (myPartialResults[t].ratio < dp) 
        return;

    //if it is the first relation
	//we need only to insert this record and hash it
    if(t ==0)
	{
  		
		//if this is the first relation, we need to insert and hash it
		//this code should be modified once the dropping is added
       	myPartialResults[t].myExtendedHash->AddRecIndex(rec, myPartialResults[t].myRecs->GetNumOfRecs());

        myPartialResults[t].myRecs->AddRecord(rec, &timeStamp, &dp);
	}

	else
	//we need to following the regular way, do the join, check the result
	//we merge two records to create a new one when we only need to do that.
	{
		
		//we first hash the input record
		unsigned int hashvalue = myPartialResults[t].myInputHash->
			Compute_Hash(&rec, myPartialResults[t].inputHashingOrderMaker);

        //we then need to retrieve all indices from relation t-1 using the hash value
		TwoWayList<IntWrapper> * pIndex = NULL;
		int res=myPartialResults[t-1].myExtendedHash->Retrieve(hashvalue, &pIndex);
		if(res== -1)
			return;

        	else
		{	
            		int tmpInt; 

            		double tmpDouble;

    		//we need to merge the two records we found
			int sizeLeft = myPartialResults[t-1].outputSchema->GetNumAtts();

            		int sizeRight = relSchemas[PermutationOrder[t]].GetNumAtts();

           		 pIndex->MoveToStart();
			
			while(pIndex->RightLength()!=0)
			{	
				Record *newRec = new Record();

               			 if(newRec == NULL)
				{
					cerr << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

                		IntWrapper *tmp =pIndex->Current(0);
			
				//we should check whether the cnf is satisfied before merging
				if(myCompareHash.Compare(myPartialResults[t-1].myRecs->allRecords[tmp->value], 
				&rec, myPartialResults[t].resultLit, myPartialResults[t].resultCNF)!=0)
				{
					
					newRec->AppendRecords(myPartialResults[t-1].myRecs->allRecords[tmp->value], &rec,sizeLeft, sizeRight);

                    //now we need to insert and hash the new record
					//however, when this is the last relation, we do not need
					//to hash it.
					if(t < numRels -1)
					{
						//this code should be modified when the dropping is added
						myPartialResults[t].myExtendedHash->AddRecIndex(*newRec, myPartialResults[t].myRecs->GetNumOfRecs());
					}
                   
                    myPartialResults[t].myRecs->AddRecord(*newRec, 
							myPartialResults[t-1].myRecs->timeStamps[tmp->value], timeStamp,myPartialResults[t-1].myRecs->ratios[tmp->value],dp);
			
                    if(t == numRels -1 && funcParseTree != NULL)
                    {
                        //we need to put the timestamps and the aggregate value into the vector
			
                        if( Int == myFunc->Apply (myPartialResults[t].myRecs->allRecords[myPartialResults[t].myRecs->GetNumOfRecs()-1] ,tmpInt, tmpDouble))
                            insertedAggregates->push_back((double)tmpInt);
                        else
                            insertedAggregates->push_back(tmpDouble);
                        insertedTimeStamps->push_back(myPartialResults[t].myRecs->timeStamps[myPartialResults[t].myRecs->GetNumOfRecs()-1]);
                        if(myGroupBy!=NULL)
			{
				Record *p = new Record;
				p->Copy(myPartialResults[t].myRecs->allRecords[myPartialResults[t].myRecs->GetNumOfRecs()-1]);
				p->Project(groupAtts, numGroupAtts, myPartialResults[numRels-1].outputSchema->GetNumAtts());
				insertedGroups->push_back(p);    
			}
                    }

                    delete newRec;	
					

				}

				pIndex->Advance();
			}
			
			
		}		
	}
	//after adding all join results, we need to consider whether the next one is a
	//constant relation. if it is, automatically join the next one
	//if the relation is not the last one in the permutation, we need to do this check
	//we do this using the following three steps:
	//step1: find all tuples in the constant relation that will join with recs we inserted
	//step2: insert these tuples;
	//step3: clean the current join result, because we do not need to keep them.
	//remember we have to clean both the join result and the hash table
	if(t!= numRels -1 && myPartialResults[t+1].isConst == true)//not last rel + next is const rel
	{
		
		set<int> myIndices;

        myIndices.clear();

        //first find all involved constant recs in the join, we are doing so by
		//adding all indices from the constant relation that join with any of the tuples
		//from this join result

        for(int i =0, tmpsize = myPartialResults[t].myRecs->GetNumOfRecs(); i<tmpsize;++i)
		{

            unsigned int hashvalue = myPartialResults[t+1].myInputHash->
			Compute_Hash(myPartialResults[t].myRecs->allRecords[i], myPartialResults[t].resultHashingOrderMaker);

            TwoWayList<IntWrapper> * pIndex = NULL;
			
			int res=myConstRelRecs[consttoPerm[PermutationOrder[t+1]]].myExtendedHash->Retrieve(hashvalue, &pIndex);
		
			if(res!= -1)
			{
				pIndex->MoveToStart();

                while(pIndex->RightLength()!=0)
				{
					IntWrapper *tmp =pIndex->Current(0);

                    myIndices.insert(tmp->value);

                    pIndex->Advance();
				}
			}	
		}

		//now we have gathered every indices
		//we need to insert these recs in the constant relation into the in-memory-join
		int ind= consttoPerm[PermutationOrder[t+1]];

        int rel=PermutationOrder[t+1];

        Record * tmpRec = new Record;

        for(set<int> :: iterator ibegin = myIndices.begin(), iend = myIndices.end();
			ibegin!=iend; ++ibegin)
		{
			int constInd = *ibegin;

            tmpRec->Copy(myConstRelRecs[ind].myRecs->allRecords[constInd]);

            InsertRec(*tmpRec, rel, constInd,myConstRelRecs[ind].myRecs->ratios[constInd][0]);
		}

		delete tmpRec;

        //finally we need to clear current recs and hash tables;
		
		RecordArray * tmpArray = new RecordArray(*(myPartialResults[t].myRecs));
		delete myPartialResults[t].myRecs;
		myPartialResults[t].myRecs = tmpArray;

        delete myPartialResults[t].myExtendedHash;
		myPartialResults[t].myExtendedHash = new ExtendedHash(*myPartialResults[t].resultHashingOrderMaker, myPartialResults[t+1].myInputHash);
	}
}



//this is used to add non-constant relation tuples
//they must be added one by one in a sorted order by the timestamp
void NewInMemoryJoin :: AddStartTuples (Record &tuple, int timeStamp,  int whichRel)
{
   numTuplesReceived[whichRel] ++;	

    double p =drand48();
   
    	
    Record *tmpRec = new Record;

    tmpRec->Copy(&tuple);

    allStartTuples->AddRecord(*tmpRec, &timeStamp, &p);

    relations->push_back(whichRel);

    myStartTuples[whichRel].AddRecord(tuple, &timeStamp, &p);
		
	delete tmpRec;
}


//this is used to add all const relation tuples;
//we need to add records both in the constant relation data structure
//and the start tuple data structure
//we maintain a counter for the constant relations as the timestamp
//this is only used when the dropping happens
void NewInMemoryJoin :: AddConstTuples (Record* tuples, int numTuples, int whichRel)
{
    numTuplesReceived[whichRel] += numTuples;	

    double p;

    int t = consttoPerm[whichRel];

    myConstRelRecs[t].relIndex = whichRel;

    myConstRelRecs[t].myRecs = new RecordArray;

    Record *tmp = new Record;

    for(int  i = 0; i< numTuples; ++i)
	{
		p=drand48();
		tmp->Copy(&(tuples[i]));

        myConstRelRecs[t].myRecs->AddRecord(*tmp, &i, &p);

        //we also need to add this into the start tuple data structure
		myStartTuples[whichRel].AddRecord(tuples[i], &i, &p);
	}

	delete tmp;
}

inline double NewInMemoryJoin :: min (double a , double b)
{
	return (a<b)?a:b;
}

inline double NewInMemoryJoin :: max (double a , double b)
{
	return (a>b)?a:b;
}

//Get Total number of Bytes occpuied by the join result
double NewInMemoryJoin :: GetJoinSize()
{
	return myPartialResults[numRels -1].myRecs->GetNumOfBytes();
}



//this function pulls out all join result tuples (from myPartialResults[numRels -1]
//the Schema is also returned
void NewInMemoryJoin :: PullOutAllJoinResults(Schema & outputSchema, vector<Record*>& myRecs)
{
	int t = numRels - 1;

    outputSchema.CopyFrom(myPartialResults[t].outputSchema);
	
	for(int i =0, size = myPartialResults[t].myRecs->GetNumOfRecs(); i<size;++i)
	{
		myRecs.push_back(myPartialResults[t].myRecs->allRecords[i]);

        myPartialResults[t].myRecs->allRecords[i] = NULL;	
	}
}


//this is used to determine the permutation of the starting phase
//furthermore all start tuples are cleaned after this phase
SearchResults* NewInMemoryJoin :: ChoosePermOrder()
{
    int first = -1, second = -1;
    double size = -1.0;

    int tmpnumRels = 2;

    int ** tmpmask = new int*[2];

    tmpmask[0] =new int [2];

    tmpmask[1] = new int [2];

    tmpmask[0][0] = tmpmask[1][1] = 0;

    tmpmask[0][1] = tmpmask [1][0] =1;

    Schema ** tmpSchema;

    tmpSchema = new Schema*[2];

    tmpSchema[0] = new Schema;

    tmpSchema[1] = new Schema;	
	
     
	
	for(int i=0; i<numRels; ++i)
		for(int j = i+1; j<numRels; ++j)
	{

        //if there is a join condition
		if(mask[i][j] == 1)
		{
			tmpSchema[0]->CopyFrom(&relSchemas[i]);

            tmpSchema[1]->CopyFrom(&relSchemas[j]);

            NewInMemoryJoin * tmpJoin = new NewInMemoryJoin(2, tmpmask, tmpSchema, myCondition, NULL, 0.0, 0.0, NULL);

            //now we need to set the constant relation
			tmpJoin->InitialConstRel(0, NULL);

            //set the permutation order
			int permorder[] = {0,1};

            tmpJoin->SetPermOrder(permorder);

            //prepare all parameters
			tmpJoin->PrepareParameters();

            Record *tmpRec = new Record;

            for(int k=0, ksize=myStartTuples[i].GetNumOfRecs(); k< ksize; ++k)
			{

                tmpRec->Copy(myStartTuples[i].allRecords[k]);

                tmpJoin->InsertRec(*tmpRec,0,k, 0.0);
			}

            double tmpMinSize =  min(myStartTuples[i].GetNumOfBytes(),myStartTuples[j].GetNumOfBytes());
			for(int k=0, ksize=myStartTuples[j].GetNumOfRecs(); k< ksize; ++k)
			{

                tmpRec->Copy(myStartTuples[j].allRecords[k]);

                tmpJoin->InsertRec(*tmpRec,1,k, 0.0);
                if(size != -1 && tmpMinSize + tmpJoin->GetJoinSize() > size)
                    break;
                
            }

			delete tmpRec;

            double tmpsize = tmpMinSize + tmpJoin->GetJoinSize();

            if(size < 0 || tmpsize < size)
			{

                first = i;

                second = j;

                size = tmpsize;
			}

			delete tmpJoin;
		}
		
	}

	if(myStartTuples[first].GetNumOfBytes() > myStartTuples[second].GetNumOfBytes())
	{
		int tmpfirst = second;  second = first; first = tmpfirst;
	}
	//now we need to get the join result and pull them out
	
	tmpSchema[0]->CopyFrom(&relSchemas[first]);

    tmpSchema[1]->CopyFrom(&relSchemas[second]);

    NewInMemoryJoin * tmpJoin = new NewInMemoryJoin(2, tmpmask, tmpSchema, myCondition, NULL, 0.0 ,0.0, NULL);

    //now we need to set the constant relation
	tmpJoin->InitialConstRel(0, NULL);

    //set the permutation order
	int permorder[] = {0,1};

    tmpJoin->SetPermOrder(permorder);

    //prepare all parameters
	tmpJoin->PrepareParameters();

    Record *tmpRec = new Record;

    for(int k=0, ksize=myStartTuples[first].GetNumOfRecs(); k< ksize; ++k)
	{

        tmpRec->Copy(myStartTuples[first].allRecords[k]);

        tmpJoin->InsertRec(*tmpRec,0,k, 0.0);
	}
    
	for(int k=0, ksize=myStartTuples[second].GetNumOfRecs(); k< ksize; ++k)
	{
		tmpRec->Copy(myStartTuples[second].allRecords[k]);

        tmpJoin->InsertRec(*tmpRec,1,k,0.0);

	}

	delete tmpRec;

    int index = -1;

    vector <Record*> * myRecs = new vector<Record*>;

    myRecs->clear();

    Schema outputSchema;

    tmpJoin->PullOutAllJoinResults(outputSchema, *myRecs);

//    outputSchema.Print();

    delete tmpJoin;

    vector<int> firstPart, secondPart;

    firstPart.push_back(first);

    firstPart.push_back(second);

    for(int i=0; i<numRels;++i)
	{
		if(i!=first &&i!=second)
            secondPart.push_back(i);
	}

	while(secondPart.size()!=0 )
	{
        size = -1.0;

        int secondSize = secondPart.size();

        int firstSize = firstPart.size();

        //try for each relation, find the smallest one
		for(int i = 0; i<secondSize;++i)
		{
			
            //search to see whether the join is valid
			bool canjoin = false;

            for(int j =0; j<firstSize;++j)	
			{
            	if(mask[secondPart[i]][firstPart[j]] ==1)
				{
					canjoin = true;
				}
			}

            if(canjoin == true)
			{

                tmpSchema[0]->CopyFrom(&outputSchema);

                tmpSchema[1]->CopyFrom(&relSchemas[secondPart[i]]);

                NewInMemoryJoin * tmpJoin = new NewInMemoryJoin(2, tmpmask, tmpSchema, myCondition, NULL, 0.0, 0.0, NULL);

                //now we need to set the constant relation
				tmpJoin->InitialConstRel(0, NULL);

                //set the permutation order
				int permorder[] = {0,1};

                tmpJoin->SetPermOrder(permorder);

                //prepare all parameters
				tmpJoin->PrepareParameters();

                Record *tmpRec = new Record;

                for(int k=0, ksize=myRecs->size(); k< ksize; ++k)
				{
					tmpRec->Copy((*myRecs)[k]);
                    tmpJoin->InsertRec(*tmpRec,0,k ,0.0);
				}

				for(int k=0, ksize=myStartTuples[secondPart[i]].GetNumOfRecs(); k< ksize; ++k)
				{
                    tmpRec->Copy(myStartTuples[secondPart[i]].allRecords[k]);
                    tmpJoin->InsertRec(*tmpRec,1,k, 0.0);

                    if(size != -1 && tmpJoin->GetJoinSize() > size)
                    break;
                }

				delete tmpRec;

                double tmpsize = tmpJoin->GetJoinSize();

                if(size < 0 || tmpJoin->GetJoinSize() < size)
				{
					index =i;

                    size = tmpsize;
				}

				delete tmpJoin;

			}

        }
		//now we have obtained all necessary information 
		//we know the next one that should be added is secondPart[i]
		int tmpindex = secondPart[index];

        firstPart.push_back(tmpindex);

        secondPart[index]= secondPart[0];

        secondPart.erase(secondPart.begin());

        //we need to pull out all join results.
		tmpSchema[0]->CopyFrom(&outputSchema);

        tmpSchema[1]->CopyFrom(&relSchemas[tmpindex]);

        NewInMemoryJoin * tmpJoin = new NewInMemoryJoin(2, tmpmask, tmpSchema, myCondition, NULL, 0.0, 0.0, NULL);

        //now we need to set the constant relation
		tmpJoin->InitialConstRel(0, NULL);

        //set the permutation order
		int permorder[] = {0,1};
        tmpJoin->SetPermOrder(permorder);

        //prepare all parameters
		tmpJoin->PrepareParameters();

        Record *tmpRec = new Record;

        for(int k=0, ksize=myRecs->size(); k< ksize; ++k)
		{
			
			tmpJoin->InsertRec(*(*myRecs)[k],0,k, 0.0);

            delete (*myRecs)[k];

        }

        for(int k=0, ksize=myStartTuples[tmpindex].GetNumOfRecs(); k< ksize; ++k)
		{	
            tmpRec->Copy(myStartTuples[tmpindex].allRecords[k]);

            tmpJoin->InsertRec(*tmpRec,1,k, 0.0);
		}

		delete tmpRec;

        delete myRecs;

        myRecs = new vector<Record*>;

        tmpJoin->PullOutAllJoinResults(outputSchema, *myRecs);

        delete tmpJoin;
	}
	
	//everything is ready
	//some clean work;
	for(int i=0;i<myRecs->size(); ++i)
        delete (*myRecs)[i];

    delete myRecs;

    delete [] tmpmask[0];

    delete [] tmpmask[1];

    delete [] tmpmask;

    delete  tmpSchema[0];

    delete  tmpSchema[1];

    delete [] tmpSchema;

    //we need to setup the Permutation Order;
    PermutationOrder = new int[numRels];

    for(int i=0; i<numRels;++i)
    {
        cout<<firstPart[i]<<endl;
        PermutationOrder[i]=firstPart[i];
    }

    //clean the storage for inserted and deleted resulting tuples
    insertedTimeStamps->clear();
    insertedAggregates->clear();
    insertedGroups->clear();
    deletedAggregates->clear();
    deletedTimeStamps->clear();
    deletedGroups->clear();

    SearchResults* myResults = NULL;

    //after setting up the permutation order we need to prepare the parameters
    //that are needed for inserting a record
   
    PrepareParameters();
   
    PreInsertRec();

    //after PreInsertRec(), return the resulting tuples
    //nothing changed
   
    if(deletedAggregates->size() > 0)
    {
        cout<<"No delete should happen in the starting phase, ERROR!!!\n";
        exit(1);
    }

    if(insertedAggregates->size()==0)
        return NULL;
    //we need to build the structure
    else
    {
        myResults = new  SearchResults;
        myResults->constRels = new int[numRels];
        myResults->ratios = new double[numRels];
        myResults->numConst = numRels;
        for(int i = 0; i < numRels; ++i)
        {
           myResults->ratios[i] = myPartialResults[i].ratio;
           if(myPartialResults[i].isConst == true)
               myResults->constRels[i] = 1;
           else
               myResults->constRels[i] = 0;
        }
        //now we need to build the array for inserted tuples
        myResults->numRes = insertedAggregates->size();
        myResults->funcRes = new double[myResults->numRes];
        myResults->resTuples = new int* [myResults->numRes];


        for(int i=0; i<myResults->numRes; ++i)
            myResults->resTuples[i] = new int [numRels];
            
        for(int i = 0; i < myResults->numRes; ++i)
        {
            for(int j = 0; j< numRels; ++j)
                myResults->resTuples[i][j] = (*insertedTimeStamps)[i][j];

                myResults->funcRes[i] = (*insertedAggregates)[i];
        }

	if(myGroupBy!=NULL)
	{
		myResults->resGroups = new Record[myResults->numRes];
		for(int i=0; i<myResults->numRes;++i)
		{
			myResults->resGroups[i].Consume((*insertedGroups)[i]);
			delete (*insertedGroups)[i];
			(*insertedGroups)[i] = NULL;
		}
	}	
	else
		myResults->resGroups = NULL;

        myResults->numDel = 0;
        myResults->funcDel = NULL;
        myResults->delTuples = NULL;
  	myResults->delGroups = NULL;
   }
	
	//myResults->Print();
    return myResults;
        

}


//this is used to determine the permutation of the starting phase
//furthermore all start tuples are cleaned after this phase
SearchResults* NewInMemoryJoin :: ChoosePermOrder( double* samplingRatio)
{
    const double sampleNumber = 1000;

    int first = -1, second = -1;
    double size = -1.0;

    int tmpnumRels = 2;

    int ** tmpmask = new int*[2];

    tmpmask[0] =new int [2];

    tmpmask[1] = new int [2];

    tmpmask[0][0] = tmpmask[1][1] = 0;

    tmpmask[0][1] = tmpmask [1][0] =1;

    Schema ** tmpSchema;

    tmpSchema = new Schema*[2];

    tmpSchema[0] = new Schema;

    tmpSchema[1] = new Schema;	
	
     cout<<"Join Mask\n";
     for(int i=0;i<numRels;++i)
     {	
	for(int j=0; j<numRels;++j)
          cout<<mask[i][j]<<" ";
	cout<<endl;

     } 
    // while(1);	
	

	for(int i=0; i<numRels; ++i)
		for(int j = i+1; j<numRels; ++j)
	{

        //if there is a join condition
		if(mask[i][j] == 1)
		{
			tmpSchema[0]->CopyFrom(&relSchemas[i]);

           	 	tmpSchema[1]->CopyFrom(&relSchemas[j]);

            		NewInMemoryJoin * tmpJoin = new NewInMemoryJoin(2, tmpmask, tmpSchema, myCondition, NULL, 0.0, 0.0, NULL);

            		//now we need to set the constant relation
			tmpJoin->InitialConstRel(0, NULL);

            		//set the permutation order
			int permorder[] = {0,1};

            		tmpJoin->SetPermOrder(permorder);

            		//prepare all parameters
			tmpJoin->PrepareParameters();
			 Record *tmpRec = new Record;
	    int count1 = 0, count2 = 0;
	    double kratio1;
	    int ksize1 = myStartTuples[i].GetNumOfRecs();
	    if(ksize1 <= sampleNumber)
		kratio1 = 1.0;
	    else
	    {	
      		kratio1 = sampleNumber / ksize1; 
           	ksize1 = 1000;
	    }	
	    for(int k=0; k< ksize1; ++k)
	    {
//		if(drand48()<kratio1)
		{
			count1++;
                	tmpRec->Copy(myStartTuples[i].allRecords[k]);
                	tmpJoin->InsertRec(*tmpRec,0,k, 0.0);
	   	} 
	   }
	   
	    double kratio2;
	    int ksize2 = myStartTuples[j].GetNumOfRecs();
	    if(ksize2 <= sampleNumber)
		kratio2 = 1.0;
	    else
	    {
		kratio2 = sampleNumber / ksize2; 
	        ksize2=1000;
            }
	    double ratio1 = samplingRatio[i];
	    double ratio2 = samplingRatio[j];	
	    double tmpMinSize =  min(myStartTuples[i].GetNumOfBytes()/ratio1, myStartTuples[j].GetNumOfBytes()/ratio2);	
	    double wratio = ratio1*ratio2*kratio1*kratio2;
            for(int k=0; k< ksize2; ++k)
	   {
//		if(drand48()<kratio2)
		{
			count2++;
                	tmpRec->Copy(myStartTuples[j].allRecords[k]);
	                tmpJoin->InsertRec(*tmpRec,1,k, 0.0);
        		if(size >0 && tmpMinSize + tmpJoin->GetJoinSize()/wratio >size)
			break;
		
		}        
                
            }
	   
	 delete tmpRec;

	    
	    
            double tmpsize = tmpMinSize + tmpJoin->GetJoinSize()/wratio;
	
	    if(size < 0 || tmpsize < size)
		{

                first = i;

                second = j;

                size = tmpsize;
		}

			delete tmpJoin;
		}
		
	}

	if(myStartTuples[first].GetNumOfBytes() > myStartTuples[second].GetNumOfBytes())
	{
		int tmpfirst = second;  second = first; first = tmpfirst;
	}
	//now we need to get the join result and pull them out
	
	tmpSchema[0]->CopyFrom(&relSchemas[first]);

    tmpSchema[1]->CopyFrom(&relSchemas[second]);

    NewInMemoryJoin * tmpJoin = new NewInMemoryJoin(2, tmpmask, tmpSchema, myCondition, NULL, 0.0 ,0.0, NULL);

    //now we need to set the constant relation
	tmpJoin->InitialConstRel(0, NULL);

    //set the permutation order
	int permorder[] = {0,1};

    tmpJoin->SetPermOrder(permorder);

    //prepare all parameters
	tmpJoin->PrepareParameters();

    Record *tmpRec = new Record;

	double kratio1;
	int ksize1 = myStartTuples[first].GetNumOfRecs();
        if(ksize1 <= sampleNumber)
	   kratio1 = 1.0;
	else
	{
	   kratio1 = sampleNumber / ksize1; 
           ksize1 = 1000;
        }
        for(int k=0; k< ksize1; ++k)
	{
//	   if(drand48()<kratio1)
	   {
                	tmpRec->Copy(myStartTuples[first].allRecords[k]);
                	tmpJoin->InsertRec(*tmpRec,0,k, 0.0);
	   } 
	}

	double kratio2;
	int ksize2 = myStartTuples[second].GetNumOfRecs();
	if(ksize2 <= sampleNumber)
	   kratio2 = 1.0;
	else
	{
	   kratio2 = sampleNumber / ksize2; 
	   ksize2 = 1000;	
	}
        for(int k=0; k< ksize2; ++k)
	{
//	   if(drand48()<kratio2)
	   {
                	tmpRec->Copy(myStartTuples[second].allRecords[k]);
	                tmpJoin->InsertRec(*tmpRec,1,k, 0.0);
           }        
                
       }

       delete tmpRec;   


    int index = -1;

    vector <Record*> * myRecs = new vector<Record*>;

    myRecs->clear();

    Schema outputSchema;

    tmpJoin->PullOutAllJoinResults(outputSchema, *myRecs);

//    outputSchema.Print();

    delete tmpJoin;

    vector<int> firstPart, secondPart;

    firstPart.push_back(first);

    firstPart.push_back(second);

    for(int i=0; i<numRels;++i)
   {
 	if(i!=first &&i!=second)
            secondPart.push_back(i);
   }

	while(secondPart.size()!=0 )
	{
        size = -1.0;

        int secondSize = secondPart.size();

        int firstSize = firstPart.size();

        //try for each relation, find the smallest one
		for(int i = 0; i<secondSize;++i)
		{
			
            //search to see whether the join is valid
			bool canjoin = false;

   	         for(int j =0; j<firstSize;++j)	
		{
        	    	if(mask[secondPart[i]][firstPart[j]] ==1)
			{
				canjoin = true;
			}
		}

            if(canjoin == true)
		{
	
        	        tmpSchema[0]->CopyFrom(&outputSchema);
	
        	        tmpSchema[1]->CopyFrom(&relSchemas[secondPart[i]]);

                	NewInMemoryJoin * tmpJoin = new NewInMemoryJoin(2, tmpmask, tmpSchema, myCondition, NULL, 0.0, 0.0, NULL);

                	//now we need to set the constant relation
			tmpJoin->InitialConstRel(0, NULL);

                	//set the permutation order
			int permorder[] = {0,1};

                	tmpJoin->SetPermOrder(permorder);

                	//prepare all parameters
			tmpJoin->PrepareParameters();

                	Record *tmpRec = new Record;

                	for(int k=0, ksize=myRecs->size(); k< ksize; ++k)
			{
				tmpRec->Copy((*myRecs)[k]);
                    		tmpJoin->InsertRec(*tmpRec,0,k ,0.0);
			}

			double kratio;
			int ksize = myStartTuples[secondPart[i]].GetNumOfRecs();
			if(ksize <= sampleNumber)
	   			kratio = 1.0;
			else
			{
	   			kratio = sampleNumber / ksize; 
				ksize = 1000;
			}	
			double ratio = samplingRatio[secondPart[i]];
       			double wratio= ratio*kratio;
			for(int k=0; k< ksize; ++k)
			{
			//	if(drand48()<kratio)
	   			{		
                    			tmpRec->Copy(myStartTuples[secondPart[i]].allRecords[k]);
                    			tmpJoin->InsertRec(*tmpRec,1,k, 0.0);
					
                			if(size > 0 && tmpJoin->GetJoinSize()/wratio > size)
					  break;
				}
			}

				delete tmpRec;
			
                	double tmpsize = tmpJoin->GetJoinSize()/wratio;

                	if(size < 0 || tmpsize < size)
			{
				index =i;
                    		size = tmpsize;
			}
			delete tmpJoin;

		}

        }
		//now we have obtained all necessary information 
		//we know the next one that should be added is secondPart[i]
		int tmpindex = secondPart[index];

        firstPart.push_back(tmpindex);

        secondPart[index]= secondPart[0];

        secondPart.erase(secondPart.begin());

        //we need to pull out all join results.
		tmpSchema[0]->CopyFrom(&outputSchema);

        tmpSchema[1]->CopyFrom(&relSchemas[tmpindex]);

        NewInMemoryJoin * tmpJoin = new NewInMemoryJoin(2, tmpmask, tmpSchema, myCondition, NULL, 0.0, 0.0, NULL);

        //now we need to set the constant relation
		tmpJoin->InitialConstRel(0, NULL);

        //set the permutation order
		int permorder[] = {0,1};
        tmpJoin->SetPermOrder(permorder);

        //prepare all parameters
		tmpJoin->PrepareParameters();

        Record *tmpRec = new Record;

        for(int k=0, ksize=myRecs->size(); k< ksize; ++k)
	{
			
			tmpJoin->InsertRec(*(*myRecs)[k],0,k, 0.0);

            delete (*myRecs)[k];

        }
	
	double kratio;
	int ksize = myStartTuples[tmpindex].GetNumOfRecs();
	
	if(ksize <= sampleNumber)
		kratio = 1.0;
	else
	{
		kratio = sampleNumber / ksize;
		ksize = 1000; 
	}
	for(int k=0; k< ksize; ++k)
	{	
      //  	if(drand48()<kratio)
		{
			tmpRec->Copy(myStartTuples[tmpindex].allRecords[k]);
		        tmpJoin->InsertRec(*tmpRec,1,k, 0.0);
		}
	    
	}

		delete tmpRec;

        delete myRecs;

        myRecs = new vector<Record*>;

        tmpJoin->PullOutAllJoinResults(outputSchema, *myRecs);

        delete tmpJoin;
	}
	
	//everything is ready
	//some clean work;
	for(int i=0;i<myRecs->size(); ++i)
        delete (*myRecs)[i];

    delete myRecs;

    delete [] tmpmask[0];

    delete [] tmpmask[1];

    delete [] tmpmask;

    delete  tmpSchema[0];

    delete  tmpSchema[1];

    delete [] tmpSchema;

    //we need to setup the Permutation Order;
    PermutationOrder = new int[numRels];

    for(int i=0; i<numRels;++i)
    {
        cout<<firstPart[i]<<endl;
        PermutationOrder[i]=firstPart[i];
    }

    //clean the storage for inserted and deleted resulting tuples
    insertedTimeStamps->clear();
    insertedAggregates->clear();
    deletedAggregates->clear();
    deletedTimeStamps->clear();
    SearchResults* myResults = NULL;

    //after setting up the permutation order we need to prepare the parameters
    //that are needed for inserting a record
   
    PrepareParameters();
    //Print();
    PreInsertRec();

    //after PreInsertRec(), return the resulting tuples
    //nothing changed
   
    if(deletedAggregates->size() > 0)
    {
        cout<<"No delete should happen in the starting phase, ERROR!!!\n";
        exit(1);
    }

    if(insertedAggregates->size()==0)
        return NULL;
    //we need to build the structure
    else
    {
        myResults = new  SearchResults;
        myResults->constRels = new int[numRels];
        myResults->ratios = new double[numRels];
        myResults->numConst = numRels;
        for(int i = 0; i < numRels; ++i)
        {
           myResults->ratios[i] = myPartialResults[i].ratio;
           if(myPartialResults[i].isConst == true)
               myResults->constRels[i] = 1;
           else
               myResults->constRels[i] = 0;
        }
        //now we need to build the array for inserted tuples
        myResults->numRes = insertedAggregates->size();
        myResults->funcRes = new double[myResults->numRes];
        myResults->resTuples = new int* [myResults->numRes];

        for(int i=0; i<myResults->numRes; ++i)
            myResults->resTuples[i] = new int [numRels];
            
        for(int i = 0; i < myResults->numRes; ++i)
        {
            for(int j = 0; j< numRels; ++j)
                myResults->resTuples[i][j] = (*insertedTimeStamps)[i][j];

                myResults->funcRes[i] = (*insertedAggregates)[i];
        }

	if(myGroupBy!=NULL)
	{
		myResults->resGroups = new Record[myResults->numRes];
		for(int i=0; i<myResults->numRes;++i)
		{
			myResults->resGroups[i].Consume((*insertedGroups)[i]);
			delete (*insertedGroups)[i];
			(*insertedGroups)[i] = NULL;
		}
	}	
	else
		myResults->resGroups = NULL;

        myResults->numDel = 0;
        myResults->funcDel = NULL;
        myResults->delTuples = NULL;
  	myResults->delGroups = NULL;
    }

//    myResults->Print();	


    return myResults;
        

}

void NewInMemoryJoin :: CheckExceed()
{

    double size = 0;

    for(int i =0; i<numRels;++i)
    {
        size+=myPartialResults[i].myRecs->GetNumOfBytes();
    }
    
    if(size <= hardBound)
        return;

    //now we need to find which relation needs to be dropped and what should be the new sample ratio
    //cout<<"we start dropping\n";	
    vector<int> *validIndices =new vector<int>;

    vector<double> * newRatio = new vector <double>;

    double distance = hardBound - softBound;

    size = 0;

//    cerr << "BEFORE DROPPING\n";
    for(int i = numRels -1; i >= 0; --i)
    {
        size += myPartialResults[i].myRecs->GetNumOfBytes();

//        cerr << "\tBytes for rel " << i << ": " << myPartialResults[i].myRecs->GetNumOfBytes() << "\n";
//        cerr << "\tRecs in rel " << i << ": " << myPartialResults[i].myRecs->GetNumOfRecs() << "\n";

        if(size > distance)
        {
            validIndices->push_back(i);
            newRatio->push_back((1- distance/size)*myPartialResults[i].ratio);
        }

    }

    //check all the stuff and find the one that gives the least variance
    double * myRatio = new double[numRels];
    double minVar ; 
    int droppingIndex;
    for(int i=0; i<numRels;++i)
        myRatio[i]=myPartialResults[i].ratio;

    minVar = 9e199;
    droppingIndex = -1;

    for(int i=0, t= validIndices->size(); i<t; ++i)
    {
        double tmp = myRatio[(*validIndices)[i]];

        myRatio[(*validIndices)[i]]= (*newRatio)[i];
        double tmpVar = myStatModule->WhatIfVariance(myRatio);
//	cerr << "If ratio is " << myRatio[(*validIndices)[i]] << " for relation " << (*validIndices)[i] << " then variance is " << tmpVar << "\n";

        if( minVar > tmpVar)
        {
            minVar = tmpVar;
            droppingIndex = i;
        }
        myRatio[(*validIndices)[i]] = tmp;
    }
//   	cerr << "Automatically choose to drop " << (*validIndices)[droppingIndex] << "\n";
   
       Dropping ((*validIndices)[droppingIndex], (*newRatio)[droppingIndex]);
  
     delete myRatio;
     //call the procedure to drop samples
   
//	cerr << "AFTER DROPPING\n";
    size  = 0 ;
    for(int i =numRels - 1; i>= 0; --i)
    {
        size+=myPartialResults[i].myRecs->GetNumOfBytes();
//	cerr << "\tBytes for rel " << i << ": " << myPartialResults[i].myRecs->GetNumOfBytes() << "\n";
//	cerr << "\tRecs in rel " << i << ": " << myPartialResults[i].myRecs->GetNumOfRecs() << "\n";
    }
//	cerr << "\n\n";
   // cout<<"after dropping mysize is "<<size<<endl;
    delete validIndices;
    delete newRatio;
}


void NewInMemoryJoin :: Dropping(int relIndex, double newRatio)
{
//    cout << "Dropping is on!!!\n";

    myPartialResults[relIndex].ratio = newRatio;

    for(int i= relIndex; i<numRels - 1; ++i)
    {
        //we want to update ratio of partial result i
        //and drop all records from this to the end of the partial result chain
        int size = myPartialResults[i].myRecs->GetNumOfRecs();
     
        if(size!=0)
        {
	    RecordArray* tmpArray = new RecordArray(*(myPartialResults[i].myRecs));

            ExtendedHash * tmpHash = new ExtendedHash(*(myPartialResults[i].myExtendedHash));

            int tmpCount = 0;
             
//            cout<<"first ratio="<<myPartialResults[i].myRecs->ratios[0][relIndex]<<endl;
//            cout<<"second ratio="<<myPartialResults[i].myRecs->ratios[1][relIndex]<<endl;
//	    cout<<"third ratio="<<myPartialResults[i].myRecs->ratios[3][relIndex]<<endl;
            for(int j=0; j< size;++j)
            {
                //we need to drop this record
                if(myPartialResults[i].myRecs->ratios[j][relIndex] >= newRatio)
                {

                    delete [] myPartialResults[i].myRecs->ratios[j];
                    myPartialResults[i].myRecs->ratios[j] = NULL;
	            delete myPartialResults[i].myRecs->allRecords[j];
                    myPartialResults[i].myRecs->allRecords[j] = NULL;
	            delete [] myPartialResults[i].myRecs->timeStamps[j];
                    myPartialResults[i].myRecs->timeStamps[j] = NULL;
                }

                else //add to the new Array and ExtendedHash
                {
		   tmpHash->AddRecIndex(*(myPartialResults[i].myRecs->allRecords[j]), tmpCount);

                    tmpArray->AddRecord(*(myPartialResults[i].myRecs->allRecords[j]),  
                                        myPartialResults[i].myRecs->timeStamps[j],myPartialResults[i].myRecs->ratios[j]);
                    
                    tmpCount++;

                    myPartialResults[i].myRecs->ratios[j] = NULL;

                    myPartialResults[i].myRecs->allRecords[j] = NULL;

                    myPartialResults[i].myRecs->timeStamps[j] = NULL;
                }

            }
		
            delete myPartialResults[i].myRecs;
            myPartialResults[i].myRecs = tmpArray;
		
            delete myPartialResults[i].myExtendedHash;
            myPartialResults[i].myExtendedHash = tmpHash;
        }
	
    }

    //we take the dropping of the last relation out because this relation has something special
    //first this relation we do not need to hash ,there is no ExtendedHash object
    // second, we need to gather those records that are dropped so that we can return to the controller
    // the needed information

    int tmpInt;
    double tmpDouble;

    int i= numRels - 1;
    {
	//we want to update ratio of partial result i
        //and drop all records from this to the end of the partial result chain
        int size = myPartialResults[i].myRecs->GetNumOfRecs();

        if(size!=0)
        {
	    RecordArray* tmpArray = new RecordArray(*(myPartialResults[i].myRecs));
            int tmpCount = 0;

            for(int j=0; j< size;++j)
            {
                //we need to drop this record
                if(myPartialResults[i].myRecs->ratios[j][relIndex] >= newRatio)
                {
		 //add in stuff to gather the information needed for return.
                    //we store the times and the aggregate value into deletedAggregates and deletedTimeStamps
                    delete [] myPartialResults[i].myRecs->ratios[j];
                    myPartialResults[i].myRecs->ratios[j] = NULL;
                    
                    if( Int == myFunc->Apply (myPartialResults[i].myRecs->allRecords[j] ,tmpInt, tmpDouble))
                            deletedAggregates->push_back((double)tmpInt);
                    else
                            deletedAggregates->push_back(tmpDouble);
		    if(myGroupBy!=NULL)
		    {
		 	Record *p = new Record;
		    	p->Consume(myPartialResults[i].myRecs->allRecords[j]);
                        p->Project(groupAtts, numGroupAtts, myPartialResults[numRels-1].outputSchema->GetNumAtts());
			deletedGroups->push_back(p);
		    }
		    delete myPartialResults[i].myRecs->allRecords[j];
                    myPartialResults[i].myRecs->allRecords[j] = NULL;

		    deletedTimeStamps->push_back(myPartialResults[i].myRecs->timeStamps[j]);
                    myPartialResults[i].myRecs->timeStamps[j] = NULL;
                    
                }
		
                else //add to the new Array, no extendedhash needed
                {
                    tmpArray->AddRecord(*(myPartialResults[i].myRecs->allRecords[j]),  
                                        myPartialResults[i].myRecs->timeStamps[j],myPartialResults[i].myRecs->ratios[j]);
                    tmpCount++;

                    myPartialResults[i].myRecs->ratios[j] = NULL;

                    myPartialResults[i].myRecs->allRecords[j] = NULL;

                    myPartialResults[i].myRecs->timeStamps[j] = NULL;
                }

            }
		
            delete myPartialResults[i].myRecs;
            myPartialResults[i].myRecs = tmpArray;
            
        }
		
    }
    
		
    // finally, if this is a constant relation, we also need to drop tuples from the relation
   if(  myPartialResults[relIndex].isConst == true)
   {
       int t = consttoPerm[PermutationOrder[relIndex]];
       int size = myConstRelRecs[t].myRecs->GetNumOfRecs();

       RecordArray* tmpArray = new RecordArray;
       
       	
       	ExtendedHash * tmpHash = NULL;
	if(relIndex!=0)
		tmpHash = new ExtendedHash(*(myConstRelRecs[t].myExtendedHash));

       int tmpCount = 0;
	
       for(int j=0; j< size;++j)
       {
	   //we need to drop this record
           if(myConstRelRecs[t].myRecs->ratios[j][0] >= newRatio)
           {
	       delete [] myConstRelRecs[t].myRecs->ratios[j];
               myConstRelRecs[t].myRecs->ratios[j] = NULL;

               delete myConstRelRecs[t].myRecs->allRecords[j];
               myConstRelRecs[t].myRecs->allRecords[j] = NULL;

               delete [] myConstRelRecs[t].myRecs->timeStamps[j];
               myConstRelRecs[t].myRecs->timeStamps[j] = NULL;
           }

           else //add to the new Array and ExtendedHash
           {
		if(relIndex!=0)
                	tmpHash->AddRecIndex(*(myConstRelRecs[t].myRecs->allRecords[j]), tmpCount);
                tmpArray->AddRecord(*(myConstRelRecs[t].myRecs->allRecords[j]),  
                                         myConstRelRecs[t].myRecs->timeStamps[j],myConstRelRecs[t].myRecs->ratios[j]);
               
                tmpCount++;

                myConstRelRecs[t].myRecs->ratios[j] = NULL;

                myConstRelRecs[t].myRecs->allRecords[j] = NULL;

                myConstRelRecs[t].myRecs->timeStamps[j] = NULL;
            }
	
       }
       delete myConstRelRecs[t].myRecs;
       myConstRelRecs[t].myRecs = tmpArray;
	
       if(relIndex!=0)
       {	
		delete myConstRelRecs[t].myExtendedHash;
       		myConstRelRecs[t].myExtendedHash = tmpHash;
	}
	
    }
	
}

void NewInMemoryJoin :: obtainSamplingRatio(double *&ratios)
{
    ratios = new double[numRels];

    for(int i=0; i<numRels;++i)
        ratios[i]= myPartialResults[IndextoPerm[i]].ratio;
}

void NewInMemoryJoin :: outputResults(int**&hashes, double*& aggregates, int& numRecs)
{
    numRecs = myPartialResults[numRels-1].myRecs->GetNumOfRecs();

    aggregates = new double[numRecs];

    hashes = new int*[numRecs];

    vector <int> *tmpVector = new vector<int>;

    for(int i=0;i<numRels;++i)
      if(myPartialResults[IndextoPerm[i]].isConst != true)
          tmpVector->push_back(IndextoPerm[i]);

    int w = numRels -1;

    int size = tmpVector->size();

    int tmpInt; 

    double tmpDouble;

    for(int i=0; i<numRecs;++i)
    {
        if(myFunc->Apply( myPartialResults[w].myRecs->allRecords[i], tmpInt, tmpDouble) == Int)
            aggregates[i] = tmpInt;

        else
            aggregates[i] = tmpDouble;

        hashes[i] = new int[size];

        for(int j = 0; j < size; ++j)
            hashes[i][j]= myPartialResults[w].myRecs->timeStamps[i][(*tmpVector)[j]];
    }

    delete tmpVector;
    
}

SearchResults* NewInMemoryJoin :: ExternInsertRec(Record& rec, int whichRel, int timeStamp)
{
   static int tmpCount = 0;    
   tmpCount++;

   numTuplesReceived[whichRel] ++; 

    insertedTimeStamps->clear();
    insertedAggregates->clear();
    insertedGroups->clear();
  
    deletedAggregates->clear();
    deletedTimeStamps->clear();
    deletedGroups->clear();
    SearchResults* myResults = NULL;
     	
    InsertRec (rec, whichRel, timeStamp);
   
    CheckExceed();

/*
   if(tmpCount%100000==0) 
   {   
    cout<<"After an Insertion\n";
   for(int i=0; i<numRels;++i)
   cout<<"Partial result"<<i<<" has "<<myPartialResults[i].myRecs->GetNumOfRecs()<<endl;		
   }
*/

    //nothing changed
    if(insertedAggregates->size()==0 && deletedAggregates->size() == 0)
        return NULL;
    //we need to build the structure
    else
    {
        myResults = new  SearchResults;
        myResults->constRels = new int[numRels];
        myResults->ratios = new double[numRels];
        myResults->numConst = numRels;
        for(int i = 0; i < numRels; ++i)
        {
           myResults->ratios[i] = myPartialResults[i].ratio;
           if(myPartialResults[i].isConst == true)
               myResults->constRels[i] = 1;
           else
               myResults->constRels[i] = 0;
        }
        //now we need to build the array for inserted tuples
        if(0==insertedAggregates->size())
        {
            myResults->numRes = 0;
            myResults->funcRes = NULL;
            myResults->resTuples = NULL;
	    myResults->resGroups = NULL;
        }
        else
        {
            myResults->numRes = insertedAggregates->size();
            myResults->funcRes = new double[myResults->numRes];
            myResults->resTuples = new int* [myResults->numRes];

            for(int i=0; i<myResults->numRes; ++i)
                myResults->resTuples[i] = new int [numRels];
            
            for(int i = 0; i < myResults->numRes; ++i)
            {
                for(int j = 0; j< numRels; ++j)
                    myResults->resTuples[i][j] = (*insertedTimeStamps)[i][j];

                myResults->funcRes[i] = (*insertedAggregates)[i];
            }
	   if(myGroupBy!=NULL)
	   {
		myResults->resGroups = new Record[myResults->numRes];
		for(int i=0; i<myResults->numRes;++i)
		{
			myResults->resGroups[i].Consume((*insertedGroups)[i]);
			delete (*insertedGroups)[i];
			(*insertedGroups)[i] = NULL;
		}
	   }		
           else 
	      myResults->resGroups = NULL;
         }

        //now we need to build the array for deleted tuples
        if(0==deletedAggregates->size())
        {
            myResults->numDel = 0;
            myResults->funcDel = NULL;
            myResults->delTuples = NULL;
            myResults->delGroups = NULL;
        }
        else
        {
            myResults->numDel = deletedAggregates->size();
            myResults->funcDel = new double[myResults->numDel];
            myResults->delTuples = new int* [myResults->numDel];

            for(int i=0; i<myResults->numDel; ++i)
                myResults->delTuples[i] = new int [numRels];
            
            for(int i = 0; i < myResults->numDel; ++i)
            {
                for(int j = 0; j< numRels; ++j)
                    myResults->delTuples[i][j] = (*deletedTimeStamps)[i][j];

                myResults->funcDel[i] = (*deletedAggregates)[i];
                delete (*deletedTimeStamps)[i];
                (*deletedTimeStamps)[i] = NULL;
            }
	    if(myGroupBy!=NULL)
	    {
		myResults->delGroups = new Record[myResults->numDel];
		for(int i=0; i<myResults->numDel;++i)
		{
			myResults->delGroups[i].Consume((*deletedGroups)[i]);
			delete (*deletedGroups)[i];
			(*deletedGroups)[i] = NULL;
		}
	    }		
           else 
	      myResults->delGroups = NULL;
         }
    }
    //myResults->Print();
    return myResults;
    
}



int NewInMemoryJoin :: GetResultTupleNumber()
{
	return myPartialResults[numRels -1].myRecs->GetNumOfRecs();
}

void NewInMemoryJoin :: SetStatModule(Estimation* fromMe)
{
	myStatModule = fromMe;
}


