#include "ParseTree.h"

#include <stdio.h>
#include <stdlib.h>
#include <iostream>


using namespace std;



Operand* CopyOperand(Operand *p)
{
	if (p == NULL)
		return NULL;

	Operand *returnvalue = new (std::nothrow) Operand;
	if (returnvalue == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	returnvalue->code = p->code;
	returnvalue->value = strdup(p->value);
	return returnvalue;

}

ComparisonOp* CopyComparisonOp(ComparisonOp *p)
{
	if(p == NULL)
		return NULL;

	ComparisonOp *returnvalue = new (std::nothrow) ComparisonOp;
	if (returnvalue == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	returnvalue->left = CopyOperand(p->left);
	returnvalue->right = CopyOperand(p->right);
	returnvalue->code = p->code;
	return returnvalue;
}

OrList* CopyOrList(OrList *p)
{
	if(p == NULL)
		return NULL;

	OrList *returnvalue = new (std::nothrow) OrList;
	if (returnvalue == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	returnvalue->left = CopyComparisonOp(p->left);
	returnvalue->rightOr = CopyOrList(p->rightOr);	
	return returnvalue;
}

AndList* CopyAndList(AndList *p)
{
	if(p == NULL)
		return NULL;

	AndList* returnvalue = new (std::nothrow) AndList;
	if (returnvalue == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	returnvalue->left = CopyOrList(p->left);
	returnvalue->rightAnd = CopyAndList(p->rightAnd);
	return returnvalue;
}




void CleanComparisonOp(ComparisonOp *op)
{
	Operand *left = op->left;
	free(left->value);
	free(left);
	left = NULL;

	Operand *right = op->right;
	free(right->value);
	free(right);
	right = NULL;

	free(op);
	op = NULL;
}


void CleanOrList(OrList *list)
{
	while (list != NULL)
	{
		ComparisonOp *CurrentItem = list->left;
		CleanComparisonOp(CurrentItem);

		OrList *t = list->rightOr;

		free(list);
		list = t;
	}
}


void CleanAndList(AndList *list)
{
	while (list != NULL)
	{
		OrList *CurrentItem = list->left;
		CleanOrList(CurrentItem);

		AndList *t = list->rightAnd;

		free(list);
		list = t;
	}
}


void CleanAttrRenameList(AttrRenameList *list)
{
	while (list != NULL)
	{
		AttrRenameItem *CurrentItem = list->CurrentItem;
		free(CurrentItem->OldName);
		free(CurrentItem->NewName);
		free(CurrentItem);

		AttrRenameList *t = list->NextItem;

		free(list);
		list = t;
	}
}




void CleanTableScanList(TableScanList *list)
{
	while (list != NULL)
	{
		TableScanYac *CurrentItem = list->CurrentItem;
		free(CurrentItem->RelationName);
		CleanAttrRenameList(CurrentItem->RenameList);
		CleanAndList(CurrentItem->Condition);
		free(CurrentItem);

		TableScanList *t = list->NextItem;

		free(list);
		list = t;
	}
}


void CleanSortJoinList(SortJoinList *list)
{
	while (list != NULL)
	{
		SortJoinYac *CurrentItem = list->CurrentItem;
		CleanAndList(CurrentItem->Condition);
		free(CurrentItem);

		SortJoinList *t = list->NextItem;

		free(list);
		list = t;
	}
}

void CleanRehashingList(RehashingList *list)
{
	while (list != NULL)
	{
		RehashingYac *CurrentItem = list->CurrentItem;
		free(CurrentItem);

		RehashingList *t = list->NextItem;

		free(list);
		list = t;
	}
}

void CleanLevelWiseList(LevelWiseList *list)
{
	while (list != NULL)
	{
		SortJoinList *CurrentItem = list->CurrentJoinList;
		CleanSortJoinList(CurrentItem);
		RehashingList *CurrentRehashingList = list->CurrentRehashingList;
		CleanRehashingList(CurrentRehashingList);
		LevelWiseList *t = list->NextLevel;

		free(list);
		list = t;
	}
}


void CleanFuncOperand(FuncOperand *op)
{
	if (op != NULL)
	{
		free(op->value);
		free(op);
	}
}


void CleanFuncOperator(FuncOperator *list)
{
	if (list != NULL)
	{
		CleanFuncOperator(list->leftOperator);
		CleanFuncOperator(list->right);
		CleanFuncOperand(list->leftOperand);

		free(list);
		list = NULL;
	}
}

void CleanAggregateStep(AggregateStep *list)
{
	CleanFuncOperator(list->Operator);
        
	free(list);
	list = NULL;
}


void CleanOperationList(OperationList *list)
{
	CleanTableScanList(list->CurrentTableScanList);
	CleanLevelWiseList(list->CurrentLevelWiseList);
	CleanAggregateStep(list->CurrentAggregationStep);

	free(list);
	list = NULL;
}

