#ifndef PARSETREE_H
#define PARSETREE_H


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


//Compariosn operators
#define LESS_THAN 1
#define GREATER_THAN 2
#define EQUALS 3

//Comparison operand types
#define DOUBLE 1
#define INT 2
#define STRING 3
#define NAME 4

//types of operations in a query plan
#define TYPETABLESCAN 1
#define TYPESORTJOIN 2
#define TYPEAGGREGATE 3
#define TYPEREHASHING 4
#define TYPEMULTISCAN 5


//Operand in a Compariosn expression
struct Operand {

	// this tells us the type of the operand: FLOAT, INT, STRING...
	int code;

	// this is the actual operand
	char *value;
};

//Comparison operation
struct ComparisonOp {

	// this corresponds to one of the codes describing what type
	// of literal value we have in this nodes: LESS_THAN, EQUALS...
	int code;

	// these are the operands on the left and on the right
	struct Operand *left;
	struct Operand *right;
};

//Complex comparison expression containing multiple simple comparisons connected through ORs
struct OrList {

	// this is the comparison to the left of the OR
	struct ComparisonOp *left;

	// this is the OrList to the right of the OR; again,
	// this might be NULL if the left is a simple comparison
	struct OrList *rightOr;
};

//Multiple OR comparison expressions connected by ANDs
struct AndList {

	// this is the disjunction to the left of the AND
	struct OrList *left;

	// this is the AndList to the right of the AND
	// note that this can be NULL if the left is a simple disjunction
	struct AndList *rightAnd;

};


//Operand in an arithmetic expression defining the aggregate function
struct FuncOperand {
	// this tells us the type of the operand: FLOAT, INT, STRING...
	int code;

	// this is the actual operand
	char *value;
};

//Operator in an arithmetic expression defining the aggregate function
struct FuncOperator {
	// this tells us which operator to use: '+', '-', ...
	int code;

	// these are the operators on the left and on the right
	struct FuncOperator *leftOperator;
	struct FuncOperand *leftOperand;
	struct FuncOperator *right;

};


struct AttrRenameItem
{
  char *OldName;
  char *NewName;
};

struct AttrRenameList
{
   struct AttrRenameItem *CurrentItem;
   struct AttrRenameList *NextItem;
};


struct TableScanYac
{
   //indicate the type of operation,which is TYPETABLESCAN (=1) here;
   int type;
   char *RelationName;// used to store the relation it scans. its' type is Name
   int OutPutPipe;// used to store the output pipe, it's type is Int
   struct AttrRenameList *RenameList; //struct Attribute RenameList
   struct AndList *Condition; //store the condition to be accepted
   int InMemoryJoinBufferSize;
};

struct TableScanList
{
  struct TableScanYac *CurrentItem;
  struct TableScanList *NextItem;
};


struct SortJoinYac
{
  //indicate the type of operation,which is TYPESORTJOIN (=2)here;
  int type;
  int InPutPipe1;
  int InPutPipe2;
  int OutPutPipe;
  int MemorySize;
  struct AndList *Condition;
  int InMemoryJoinBufferSize;	
};

struct SortJoinList
{
   struct SortJoinYac *CurrentItem;
   struct SortJoinList *NextItem;
};


struct AttrGroupbyList
{
  char* CurrentAttr;
  struct AttrGroupbyList *NextAttr;
};


struct RehashingYac
{
  int type;
  int InPutPipe;
  int OutPutPipe;
  int MemorySize;	
  int InMemoryJoinBufferSize;
};


struct RehashingList
{
  struct RehashingYac *CurrentItem;
  struct RehashingList *NextItem;
};

struct LevelWiseList
{
  struct SortJoinList *CurrentJoinList;
  struct RehashingList *CurrentRehashingList;
  int CurrentLevelNumber;
  int softBound;
  int hardBound;
  struct LevelWiseList *NextLevel;
};



//for aggregate in the parser.
struct AggregateStep
{
  int InPutPipe;
  struct AttrGroupbyList *myGroupBy;
  int numGroups;
  int MemorySize;
  int numTuplesforStats;
  int queueSize;
  struct FuncOperator *Operator;
};

struct MultiScanYac
{
   //indicate the type of operation,which is TYPETABLESCAN (=1) here;
   int type;
   struct AttrGroupbyList *Relations;// used to store the relation it scans. its' type is Name
   int OutPutPipe;// used to store the output pipe, it's type is Int
   struct AttrRenameList *RenameList; //struct Attribute RenameList
   struct AndList *Condition; //store the condition to be accepted
   int InMemoryJoinBufferSize;
};

struct MultiScanList
{
  struct MultiScanYac *CurrentItem;
  struct MultiScanList *NextItem;
};


struct OperationList
{
  struct TableScanList *CurrentTableScanList;
  struct MultiScanList *CurrentMultiScanList;
  int softBound;
  int hardBound;
  struct LevelWiseList *CurrentLevelWiseList;
  struct AggregateStep *CurrentAggregationStep;
};

struct Operand* CopyOperand(struct Operand* p);
struct ComparisonOp* CopyComparisonOp(struct ComparisonOp* p);
struct OrList* CopyOrList(struct OrList* p);
struct AndList* CopyAndList(struct AndList* p);



/*
Functions to deallocate the memory occupied by the query tree built by the compiler
Memory is recursively deallocated for all the structures, starting with the top-most level,
and decending to the bottom-most level
*/
void CleanComparisonOp(struct ComparisonOp *op);
void CleanOrList(struct OrList *list);
void CleanAndList(struct AndList *list);
void CleanAttrRenameList(struct AttrRenameList *list);
void CleanTableScanList(struct TableScanList *list);
void CleanSortJoinList(struct SortJoinList *list);
void CleanLevelWiseList(struct LevelWiseList *list);
void CleanFuncOperand(struct FuncOperand *op);
void CleanFuncOperator(struct FuncOperator *list);
void CleanAggregateStep(struct AggregateStep *list);
void CleanOperationList(struct OperationList *list);
void CleanRehashingList(struct RehashingList *list);

#endif


