#include "ParseTreeWrapper.h"
#include <iostream>

using namespace std;

AndListWrapper :: AndListWrapper ( struct AndList* parseTree)
{
	myParseTree = CopyAndList(parseTree);
}

AndListWrapper :: ~AndListWrapper ()
{
	myParseTree = NULL;
}

int AndListWrapper :: checkOrConditions(struct OrList *currentOrList,Schema& leftSchema, Schema& rightSchema)
{
	for (struct OrList *current = currentOrList; current != NULL; current = current->rightOr)
	{
		if (current->left->left->code == NAME)
		{
			if (leftSchema.Find(current->left->left->value) == -1 && rightSchema.Find(current->left->left->value) == -1)
				return -1;
		}

		if (current->left->right->code == NAME)
		{
			if (leftSchema.Find(current->left->right->value) == -1 && rightSchema.Find(current->left->right->value) == -1)
				return -1;
		}
	}

	return 0;
}


int AndListWrapper :: checkOrConditionsSelect(struct OrList *currentOrList, Schema &schema)
{
	for(struct OrList *current = currentOrList; current != NULL; current = current->rightOr)
	{
		if (current->left->left->code == NAME)
		{
			if (schema.Find(current->left->left->value) == -1)
				return -1;
		}

		if (current->left->right->code == NAME)
		{
			if (schema.Find(current->left->right->value) == -1)
				return -1;
		}
	}

	return 0;
}



int AndListWrapper :: BuildCNF(Schema &leftSchema, Schema &rightSchema, CNF &cnf, Record &recLiteral)
{
	if(myParseTree == NULL)
	{
		Schema* temp = new Schema;

        recLiteral.GrowFromParseTree(myParseTree, &leftSchema, &rightSchema, cnf, temp);

        delete temp;

        return 0;
	}

	//we would like to search through every and node and keep this node when
	//all attributes are from the two schemas
	//This is done by moving the current AndNode and attache it into the new
	//tree. After the new tree is built, a call to GrowFromParseTree is performed
	//We then free the tree and return the cnf and recliteral

    struct AndList* tempTree = NULL, *currentNodeForTempTree = NULL;
    struct AndList *first = myParseTree, *second = NULL;

    //we first move nodes that at the beginning of the tree

	while(first != NULL)
	{
		// this condition is true when the first several nodes
		//are the nodes we need to move.
		if( checkOrConditions(first->left, leftSchema, rightSchema)==0)
		{
			//we need to first attach tempTree to this node
			if(tempTree == NULL)
			{
				tempTree = first;

                currentNodeForTempTree = first;
			}

		//we need to change currentNodeForTempTree pointer
		//notice these nodes are already chained, we dont need to 
		//chain them again.

            else 
            {
				currentNodeForTempTree = first;
			}

		//now move to the next Node to see whether it is the node 
		//we want to move
			first = first ->rightAnd;
			
		}

		//now we find the first node we dont need to move
		else
		{
	          //if some node is attached, we need to set the rightAnd = NULL
			if(currentNodeForTempTree != NULL) 
			{
				currentNodeForTempTree -> rightAnd =NULL;
			}

            break;
		}
	}

	//we need to set the head of the tree to the new first
	myParseTree = first;
	

    //if first is already NULL, we no longer need to continue check
	if(first != NULL)
	{

    //we have solved the problem that the head needs to move
	//now first points to some nodes that we dont need to move	
		for(second = first ->rightAnd; second !=NULL; )
		{

            if(checkOrConditions(second->left, leftSchema, rightSchema)==0)
			{
				if(tempTree == NULL)
				{
					tempTree = second;

                    currentNodeForTempTree = second;
				}

				else
				{
					currentNodeForTempTree->rightAnd = second;

                    currentNodeForTempTree = second;
				}

				first -> rightAnd = second ->rightAnd;

                second = second -> rightAnd;

                currentNodeForTempTree -> rightAnd = NULL;
			}

			else
			{
				first = second;	

                second = second ->rightAnd;
			}
		}
	}
	
	Schema* temp = new Schema;

    recLiteral.GrowFromParseTree(tempTree, &leftSchema, &rightSchema, cnf, temp);

    CleanAndList(tempTree);

    delete temp;

    if (myParseTree == NULL)
	return 0;

	return -1;
}



int AndListWrapper :: BuildCNFSelect(Schema &schema, CNF &cnf, Record &recLiteral)
{
	if (myParseTree == NULL)
	{
		Schema *temp = new (std::nothrow) Schema();
		if (temp == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		recLiteral.GrowFromParseTree(myParseTree, &schema, cnf, temp);
	
		delete temp;
	
		return 0;
	}

	//we would like to search through every and node and keep this node when
	//all attributes are from the two schemas
	//This is done by moving the current AndNode and attache it into the new
	//tree. After the new tree is built, a call to GrowFromParseTree is performed
	//We then free the tree and return the cnf and recliteral
	struct AndList *tempTree = NULL, *currentNodeForTempTree = NULL;
	struct AndList *first = myParseTree, *second = NULL;


	//we first move nodes that at the beginning of the tree
	while (first != NULL)
	{
		//this condition is true when the first several nodes
		//are the nodes we need to move.
		if (checkOrConditionsSelect(first->left, schema) == 0)
		{
			//we need to first attach tempTree to this node
			if (tempTree == NULL)
			{
				tempTree = first;
				currentNodeForTempTree = first;
			}

			//we need to change currentNodeForTempTree pointer
			//notice these nodes are already chained, we dont need to
			//chain them again.
			else
			{
				currentNodeForTempTree = first;
			}

			//now move to the next Node to see whether it is the node 
			//we want to move
			first = first->rightAnd;
			
		}

		//now we find the first node we dont need to move
		else
		{
			//if some node is attached, we need to set the rightAnd = NULL
			if (currentNodeForTempTree != NULL)
			{
				currentNodeForTempTree->rightAnd = NULL;
			}

			break;
		}
	}

	//we need to set the head of the tree to the new first
	myParseTree = first;
	

	//if first is already NULL, we no longer need to continue check
	if (first != NULL)
	{
		//we have solved the problem that the head needs to move
		//now first points to some nodes that we dont need to move
		for (second = first->rightAnd; second != NULL;)
		{
			if (checkOrConditionsSelect(second->left, schema) == 0)
			{
				if (tempTree == NULL)
				{
					tempTree = second;
					currentNodeForTempTree = second;
				}
				else
				{
					currentNodeForTempTree->rightAnd = second;
					currentNodeForTempTree = second;
				}

				first->rightAnd = second->rightAnd;
				second = second->rightAnd;

				currentNodeForTempTree->rightAnd = NULL;
			}
			else
			{
				first = second;
				second = second ->rightAnd;
			}
		}
	}
	
	Schema *temp = new (std::nothrow) Schema();
	if (temp == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	recLiteral.GrowFromParseTree(tempTree, &schema, cnf, temp);

	CleanAndList(tempTree);

	delete temp;

	if (myParseTree == NULL)
		return 0;

	return -1;
}



