#ifndef PARSETREEWRAPPER_H
#define PARSETREEWRAPPER_H
#include "Record.h"
#include "Schema.h"
#include "ParseTree.h"
//The Wrapper class takes an AndList parse tree and provides 
//the functions to produce necessary CNFs
//CNFs can then been used to produce ordermakers for hashing

class AndListWrapper{
private:
	//This stores the parse tree to be used to generate the CNFs	
	struct AndList *myParseTree;

	//This is a function called by BuildCNF to check whether
	//an and node should copy or not. It returns 0 when the check is passed and -1 otherwise
	//It basically scan each attribute and find whether this attributes in one of
	//the two schemas or not.
	int checkOrConditions(struct OrList *currentOrList, Schema &leftSchema, Schema &rightSchema);

	//as above but for selection conditions
	int checkOrConditionsSelect(struct OrList *currentOrList, Schema &schema);
public:
	//constructor, the only parameter is an andlist parsetree built by the parser
	AndListWrapper (struct AndList *parseTree);
	virtual ~AndListWrapper();

	//This returns the necessary cnf and the record literal given
	//the left Schema and right Schema by skipping those branches
	//in the parse tree that contain atts outside of the two schemas
	//parameters: leftSchema and rightSchema are schemas for two relations with join conditions
	//cnf and recLiteral contains the schema built
	//the program returns 0 if it finishes normally and -1 otherwise
	int BuildCNF(Schema &leftSchema, Schema &rightSchema, CNF &cnf, Record &recLiteral);

	//builds the selection conditions for a given relation (schema)
	int BuildCNFSelect(Schema &schema, CNF &cnf, Record &recLiteral);
};
#endif
