%{
	#include "ParseTree.h"

	#include <math.h>
	#include <stdio.h>
	#include <string.h>
	#include <stdlib.h>
	#include <iostream>

	extern "C" int yylex();
	extern "C" int yyparse();
	extern "C" void yyerror(char *s);


	// this is the final parse tree that is returned
	struct OperationList *finalope;

%}

// this stores all of the types returned by production rules
%union {
 	struct Operand *myOperand;
	struct ComparisonOp *myComparison;
  	struct OrList *myOrList;
  	struct AndList *myAndList;

	char *actualChars;

	struct OperationList *myOperationList;
	struct TableScanList *myTableScanList;
 	struct TableScanYac  *myTableScanYac;
	struct AttrRenameList *myRenameList;
	struct AttrRenameItem *myRenameItem;
	struct SortJoinYac *mySortJoinYac;
	struct SortJoinList *mySortJoinList;
	struct LevelWiseList *myLevelWiseList;
	struct AggregateStep *myAggregateStep;
        struct FuncOperand *myFunOperand;
	struct FuncOperator *myOperator;
	struct RehashingYac *myRehashingYac;
	struct RehashingList *myRehashingList;
	struct AttrGroupbyList *myGroupBy; 
	struct MultiScanList *myMultiScanList;
	struct MultiScanYac *myMultiScanYac;
	char whichOne;
}


%token LEVELWISE
%token STEP
%token OR
%token AND
%token Begin
%token AGGREGATE
%token END
%token TABLE_SCAN
%token RELATION
%token OUTPUT_PIPE
%token RENAME_ATTRIBUTES
%token AS
%token ACCEPT
%token SORT_JOIN
%token INPUT_PIPE
%token MEMORY
%token TUPLES
%token REHASHING
%token INMEMORYJOINBUFFERSIZE
%token GROUPBY
%token NUMGROUPSTODISPLAY
%token NUMTUPLEFORSTATS
%token SOFTBOUND
%token HARDBOUND
%token QUEUESIZE
%token MULTIRELATION

%token <actualChars> Name
%token <actualChars> String
%token <actualChars> Float
%token <actualChars> Int


%type <myOrList> OrList
%type <myAndList> AndList 
%type <myComparison> Condition 
%type <myOperand> Literal 
%type <myComparison> Op 
%type <myOperationList> OperationList
%type <myTableScanList> TableScanList
%type <mySortJoinYac> SortJoin
%type <mySortJoinList> SortJoinList
%type <myLevelWiseList> LevelWiseList
%type <myTableScanYac> TableScan
%type <myRenameList> RenameList
%type <myRenameItem> RenameItem
%type <myFunOperand> SimpleExp
%type <myOperator> CompoundExp
%type <whichOne> OpFun 
%type <myAggregateStep> AggregateStep
%type <myRehashingYac> Rehashing
%type <myRehashingList> RehashingList
%type <myGroupBy> GroupAtts
%type <myMultiScanList> MultiScanList
%type <myMultiScanYac> MultiScan

%start OperationList 


//******************************************************************************
// SECTION 3
//******************************************************************************
/* This is the PRODUCTION RULES section which defines how to "understand" the 
 * input language and what action to take for each "statment"
 */

%%
OperationList: TABLE_SCAN TableScanList SOFTBOUND Int HARDBOUND Int END LevelWiseList AggregateStep
{
		//here we need to prepend the operation to the existing OperationList
		// first we allocate the space for this node
		$$=(struct OperationList *) malloc(sizeof (struct OperationList));
	        finalope = $$;
		
		//prepend all tablescanoperations
		$$->CurrentTableScanList = $2;
		$$->softBound = atoi($4);
		free($4);
		$$->hardBound = atoi($6);
		free($6);
		//prepend all levewise operations
		$$->CurrentLevelWiseList = $8;
		//prepend the aggregation function
		$$->CurrentAggregationStep = $9;
		$$->CurrentMultiScanList = NULL;
}
| TABLE_SCAN TableScanList SOFTBOUND Int HARDBOUND Int END AggregateStep
{
		//here we need to prepend the operation to the existing OperationList
		// first we allocate the space for this node
		$$=(struct OperationList *) malloc(sizeof (struct OperationList));
	        finalope = $$;
		
		//prepend all tablescanoperations
		$$->CurrentTableScanList = $2;

		$$->softBound = atoi($4);
		free($4);
		$$->hardBound = atoi($6);
		free($6);
	
		//prepend all levewise operations
		$$->CurrentLevelWiseList = NULL;
		//prepend the aggregation function
		$$->CurrentAggregationStep = $8;
		$$->CurrentMultiScanList = NULL;	
}

|TABLE_SCAN TableScanList MultiScanList SOFTBOUND Int HARDBOUND Int END LevelWiseList AggregateStep
{
		//here we need to prepend the operation to the existing OperationList
		// first we allocate the space for this node
		$$=(struct OperationList *) malloc(sizeof (struct OperationList));
	        finalope = $$;
		
		//prepend all tablescanoperations
		$$->CurrentTableScanList = $2;
		$$->CurrentMultiScanList = $3;
		$$->softBound = atoi($5);
		free($5);
		$$->hardBound = atoi($7);
		free($7);
		//prepend all levewise operations
		$$->CurrentLevelWiseList = $9;
		//prepend the aggregation function
		$$->CurrentAggregationStep = $10;
}
| TABLE_SCAN TableScanList MultiScanList SOFTBOUND Int HARDBOUND Int END AggregateStep
{
		//here we need to prepend the operation to the existing OperationList
		// first we allocate the space for this node
		$$=(struct OperationList *) malloc(sizeof (struct OperationList));
	        finalope = $$;
		
		//prepend all tablescanoperations
		$$->CurrentTableScanList = $2;
		$$->CurrentMultiScanList = $3;
		$$->softBound = atoi($5);
		free($5);
		$$->hardBound = atoi($7);
		free($7);
	
		//prepend all levewise operations
		$$->CurrentLevelWiseList = NULL;
		//prepend the aggregation function
		$$->CurrentAggregationStep = $9;
}
;

MultiScanList : MultiScan MultiScanList
{
    $$=(struct MultiScanList*) malloc (sizeof (struct MultiScanList));
    $$->CurrentItem = $1;
    $$->NextItem = $2;
}
|MultiScan
{
    $$=(struct MultiScanList*) malloc (sizeof (struct MultiScanList));
    $$->CurrentItem = $1;
    $$->NextItem = NULL;
};

MultiScan :  MULTIRELATION GroupAtts OUTPUT_PIPE Int ACCEPT AndList INMEMORYJOINBUFFERSIZE Int TUPLES END
{
	   //no renaming no condition
	   //allocate space
  	   $$=(struct MultiScanYac *) malloc(sizeof (struct MultiScanYac));
           //save relation name
	   $$->Relations=$2;
	  //save outputpipe
           $$->OutPutPipe = atoi($4);
	   free($4);
	  //rename list 
           $$->RenameList = NULL;
          //condition
           $$->Condition = $6;
	   $$->InMemoryJoinBufferSize = atoi($8);
	   free($8);
}
| MULTIRELATION GroupAtts OUTPUT_PIPE Int RENAME_ATTRIBUTES RenameList ACCEPT AndList INMEMORYJOINBUFFERSIZE Int TUPLES END
{
	   //no renaming no condition
	   //allocate space
  	   $$=(struct MultiScanYac *) malloc(sizeof (struct MultiScanYac));
           //save relation name
	   $$->Relations=$2;
	  //save outputpipe
           $$->OutPutPipe = atoi($4);
	   free($4);
	  //rename list 
           $$->RenameList = $6;
          //condition
           $$->Condition = $8;
	   $$->InMemoryJoinBufferSize = atoi($10);
	   free($10);
};


TableScanList : TableScan TableScanList
{
    $$=(struct TableScanList*) malloc (sizeof (struct TableScanList));
    $$->CurrentItem = $1;
    $$->NextItem = $2;
}
| TableScan
{
    $$=(struct TableScanList*) malloc (sizeof (struct TableScanList));
    $$->CurrentItem = $1;
    $$->NextItem = NULL;
};


TableScan: RELATION Name OUTPUT_PIPE Int INMEMORYJOINBUFFERSIZE Int TUPLES END
{
	   //no renaming no condition
	   //allocate space
  	   $$=(struct TableScanYac *) malloc(sizeof (struct TableScanYac));
           //save relation name
	   $$->RelationName=$2;
	  //save outputpipe
           $$->OutPutPipe = atoi($4);
	   free($4);
	  //rename list 
           $$->RenameList = NULL;
          //condition
           $$->Condition = NULL;
	   $$->InMemoryJoinBufferSize = atoi($6);
	   free($6);	
}
| RELATION Name OUTPUT_PIPE Int RENAME_ATTRIBUTES RenameList INMEMORYJOINBUFFERSIZE Int TUPLES END
{
	    //no renaming no condition
	   //allocate space
  	   $$=(struct TableScanYac *) malloc(sizeof (struct TableScanYac));
           //save relation name
	   $$->RelationName=$2;
	  //save outputpipe
           $$->OutPutPipe = atoi($4);
	   free($4);
	  //rename list 
           $$->RenameList = $6;
          //condition
           $$->Condition = NULL;
	   $$->InMemoryJoinBufferSize = atoi($8);
	   free($8);
}
| RELATION Name OUTPUT_PIPE Int ACCEPT AndList INMEMORYJOINBUFFERSIZE Int TUPLES END
{
	   //no renaming no condition
	   //allocate space
  	   $$=(struct TableScanYac *) malloc(sizeof (struct TableScanYac));
           //save relation name
	   $$->RelationName=$2;
	  //save outputpipe
           $$->OutPutPipe = atoi($4);
	   free($4);
	  //rename list 
           $$->RenameList = NULL;
          //condition
           $$->Condition = $6;
	   $$->InMemoryJoinBufferSize = atoi($8);
	   free($8);
}
| RELATION Name OUTPUT_PIPE Int RENAME_ATTRIBUTES RenameList ACCEPT  AndList INMEMORYJOINBUFFERSIZE Int TUPLES END
{
	    //no renaming no condition
	   //allocate space
  	   $$=(struct TableScanYac *) malloc(sizeof (struct TableScanYac));
           //save relation name
	   $$->RelationName=$2;
	  //save outputpipe
           $$->OutPutPipe = atoi($4);
	   free($4);
	  //rename list 
           $$->RenameList = $6;
          //condition
           $$->Condition = $8;
	   $$->InMemoryJoinBufferSize = atoi($10);
	   free($10);
}
;


RenameList: RenameItem ',' RenameList
{
    //Attribute Rename List
    //first allocate space
    $$=(struct AttrRenameList*) malloc (sizeof (struct AttrRenameList));
    //han the current RenameItem to the list
    $$->CurrentItem = $1;
    //link this to the rest of the renamelist;
    $$->NextItem = $3;
}
|RenameItem
{
   //Attribute Rename List
    //first allocate space
    $$=(struct AttrRenameList*) malloc (sizeof (struct AttrRenameList));
    //han the current RenameItem to the list
    $$->CurrentItem = $1;
    //Here the rest is NULL;
    $$->NextItem = NULL;
}
;

RenameItem: Name AS Name
{
    // allocate space and save the two names
    $$=(struct AttrRenameItem*) malloc (sizeof (struct AttrRenameItem));
    $$->OldName = $1;
    $$->NewName = $3;
}
;

LevelWiseList: LEVELWISE STEP Int SortJoinList RehashingList SOFTBOUND Int HARDBOUND Int END LevelWiseList 
{
    $$=(struct LevelWiseList*) malloc(sizeof (struct LevelWiseList) );
    $$->CurrentJoinList = $4;
    $$->CurrentRehashingList = $5;
    $$->CurrentLevelNumber = atoi($3);
    free($3);
    $$->softBound= atoi($7);
    free($7);
    $$->hardBound = atoi($9);
    free($9);	
    $$->NextLevel = $11;
}
|LEVELWISE STEP Int SortJoinList RehashingList SOFTBOUND Int HARDBOUND Int END
{
    $$=(struct LevelWiseList*) malloc(sizeof (struct LevelWiseList) );
    $$->CurrentJoinList = $4;
    $$->CurrentRehashingList = $5;	

    $$->CurrentLevelNumber = atoi($3);
    free($3);

    $$->softBound= atoi($7);
    free($7);
    $$->hardBound = atoi($9);
    free($9);		

    $$->NextLevel = NULL;
}
|LEVELWISE STEP Int SortJoinList SOFTBOUND Int HARDBOUND Int END LevelWiseList
{
    $$=(struct LevelWiseList*) malloc(sizeof (struct LevelWiseList) );
    $$->CurrentJoinList = $4;
    $$->CurrentRehashingList = NULL;
    $$->CurrentLevelNumber = atoi($3);
    free($3);
     $$->softBound= atoi($6);
    free($6);
    $$->hardBound = atoi($8);
    free($8);	 

    $$->NextLevel = $10;
}
| LEVELWISE STEP Int SortJoinList SOFTBOUND Int HARDBOUND Int END
{
    $$=(struct LevelWiseList*) malloc(sizeof (struct LevelWiseList) );
    $$->CurrentJoinList = $4;
    $$->CurrentRehashingList = NULL;	

    $$->CurrentLevelNumber = atoi($3);
    free($3);
    
    $$->softBound= atoi($6);
    free($6);
    $$->hardBound = atoi($8);
    free($8);

    $$->NextLevel = NULL;
}

;

RehashingList : Rehashing RehashingList
{
  $$=(struct RehashingList*) malloc (sizeof (struct RehashingList));
  $$->CurrentItem = $1;
  $$->NextItem = $2;
}
| Rehashing
{
  $$=(struct RehashingList*) malloc (sizeof (struct RehashingList));
  $$->CurrentItem = $1;
  $$->NextItem = NULL;
};

Rehashing : REHASHING INPUT_PIPE Int OUTPUT_PIPE Int MEMORY Int TUPLES INMEMORYJOINBUFFERSIZE Int TUPLES END
{
  $$=(struct RehashingYac*) malloc (sizeof (struct RehashingYac));
  $$->InPutPipe = atoi ($3);
  free($3);
  $$->OutPutPipe = atoi ($5);
  free ($5);
  $$->MemorySize = atoi ($7);
  free ($7);
  $$->InMemoryJoinBufferSize = atoi ($10);
  free ($10);
};

SortJoinList : SortJoin SortJoinList
{
  $$=(struct SortJoinList*) malloc (sizeof (struct SortJoinList));
  $$->CurrentItem=$1;
  $$->NextItem=$2;
}
| SortJoin
{
  $$=(struct SortJoinList*) malloc (sizeof (struct SortJoinList));
  $$->CurrentItem=$1;
  $$->NextItem=NULL;
};

SortJoin:SORT_JOIN INPUT_PIPE Int INPUT_PIPE Int OUTPUT_PIPE Int MEMORY Int TUPLES INMEMORYJOINBUFFERSIZE Int TUPLES END
{
	   //Allocate space
	   $$=(struct SortJoinYac*) malloc(sizeof (struct SortJoinYac));

           $$->InPutPipe1=atoi($3);
	   free($3);

	   $$->InPutPipe2 = atoi($5);
	   free($5);

	   $$->OutPutPipe = atoi($7);
	   free($7);

	   $$->MemorySize = atoi($9);
	   free($9);

           $$->Condition =NULL;

	   $$->InMemoryJoinBufferSize = atoi($12);
	   free($12);
}
	|SORT_JOIN INPUT_PIPE Int INPUT_PIPE Int OUTPUT_PIPE Int MEMORY Int TUPLES ACCEPT AndList INMEMORYJOINBUFFERSIZE Int TUPLES END
{
	   //Allocate space
	   $$=(struct SortJoinYac*) malloc(sizeof (struct SortJoinYac));

           $$->InPutPipe1=atoi($3);
	   free($3);

	   $$->InPutPipe2 = atoi($5);
	   free($5);

	   $$->OutPutPipe = atoi($7);
	   free($7);

	   $$->MemorySize = atoi($9);
	   free($9);

           $$->Condition = $12;
	   $$->InMemoryJoinBufferSize = atoi ($14);
           free($14);	
}
;

AndList: '(' OrList ')' AND AndList
{
	// here we need to pre-pend the OrList to the AndList
	// first we allocate space for this node
	$$ = (struct AndList *) malloc (sizeof (struct AndList));
	// hang the OrList off of the left
	$$->left = $2;

	// hang the AndList off of the right
	$$->rightAnd = $5;

}

| '(' OrList ')'
{
	// just return the OrList!
	$$ = (struct AndList *) malloc (sizeof (struct AndList));
	$$->left = $2;
	$$->rightAnd = NULL;
}
;



OrList: Condition OR OrList
{ 
	// here we have to hang the condition off the left of the OrList
	$$ = (struct OrList *) malloc (sizeof (struct OrList));	
	$$->left = $1;
	$$->rightOr = $3;
}

| Condition
{
	// nothing to hang off of the right
	$$ = (struct OrList *) malloc (sizeof (struct OrList));
	$$->left = $1;
	$$->rightOr = NULL;
}
;

Condition: Literal Op Literal 
{
	// in this case we have a simple literal/variable comparison
	$$ = $2;
	$$->left = $1;
	$$->right = $3;
}
;

Op: '<' 
{
	// construct and send up the comparison
	$$ = (struct ComparisonOp *) malloc (sizeof (struct ComparisonOp));
	$$->code = LESS_THAN;
}  

| '>'
{
	// construct and send up the comparison
	$$ = (struct ComparisonOp *) malloc (sizeof (struct ComparisonOp));
	$$->code = GREATER_THAN;
}  

| '='
{
	// construct and send up the comparison
	$$ = (struct ComparisonOp *) malloc (sizeof (struct ComparisonOp));
	$$->code = EQUALS;
}  
;

Literal : String 
{
	// construct and send up the operand containing the string
	$$ = (struct Operand *) malloc (sizeof (struct Operand));
	$$->code = STRING;
	$$->value = $1;
} 

| Float
{
	// construct and send up the operand containing the FP number
	$$ = (struct Operand *) malloc (sizeof (struct Operand));
	$$->code = DOUBLE;
	$$->value = $1;
} 

| Int
{
	// construct and send up the operand containing the integer
	$$ = (struct Operand *) malloc (sizeof (struct Operand));
	$$->code = INT;
	$$->value = $1;
} 

| Name
{
	// construct and send up the operand containing the name 
	$$ = (struct Operand *) malloc (sizeof (struct Operand));
	$$->code = NAME;
	$$->value = $1;
}
;


AggregateStep : AGGREGATE INPUT_PIPE Int NUMTUPLEFORSTATS Int QUEUESIZE Int CompoundExp END
{
   $$= (struct AggregateStep *) malloc (sizeof (struct AggregateStep));

   $$->InPutPipe=atoi($3);
   free($3);
   
   $$->myGroupBy = NULL;
   $$->numGroups = -1;
   $$->MemorySize = -1;	
   $$->numTuplesforStats = atoi($5);  
   free($5);
   $$->queueSize = atoi($7);
   free($7);
   $$->Operator= $8;
}
| AGGREGATE INPUT_PIPE Int GROUPBY GroupAtts NUMGROUPSTODISPLAY Int MEMORY Int TUPLES NUMTUPLEFORSTATS Int  QUEUESIZE Int CompoundExp END
{
   $$= (struct AggregateStep *) malloc (sizeof (struct AggregateStep));

   $$->InPutPipe=atoi($3);
   free($3);

   $$->myGroupBy = $5;

   $$->numGroups = atoi($7);
   free($7);

   $$->MemorySize = atoi($9);
   free($9);
	
   $$->numTuplesforStats = atoi($12);
   free($12);

   $$->queueSize = atoi($14);
   free($14);
	
   $$->Operator= $15;
};

GroupAtts: Name
{
      $$=(struct AttrGroupbyList*) malloc (sizeof(struct AttrGroupbyList)); 
      $$->CurrentAttr = $1;
      $$->NextAttr = NULL;	 
}
| Name ',' GroupAtts
{
	$$=(struct AttrGroupbyList*) malloc (sizeof(struct AttrGroupbyList)); 
      $$->CurrentAttr = $1;
      $$->NextAttr = $3;
}
;

CompoundExp: SimpleExp OpFun CompoundExp
{
	$$ = (struct FuncOperator *) malloc (sizeof (struct FuncOperator));	
	$$->leftOperator = (struct FuncOperator *) malloc (sizeof (struct FuncOperator));
	$$->leftOperator->leftOperator = NULL;
	$$->leftOperator->leftOperand = $1;
	$$->leftOperator->right = NULL;
	$$->leftOperand = NULL;
	$$->right = $3;
	$$->code = $2;	
}

| '(' CompoundExp ')' OpFun CompoundExp
{
	$$ = (struct FuncOperator *) malloc (sizeof (struct FuncOperator));	
	$$->leftOperator = $2;
	$$->leftOperand = NULL;
	$$->right = $5;
	$$->code = $4;	
}

| '(' CompoundExp ')'
{
	$$ = $2;
}

| SimpleExp
{
	$$ = (struct FuncOperator *) malloc (sizeof (struct FuncOperator));	
	$$->leftOperator = NULL;
	$$->leftOperand = $1;
	$$->right = NULL;	
}

| '-' CompoundExp
{
	$$ = (struct FuncOperator *) malloc (sizeof (struct FuncOperator));	
	$$->leftOperator = $2;
	$$->leftOperand = NULL;
	$$->right = NULL;	
	$$->code = '-';
}
;

OpFun: '-'
{
	$$ = '-';
}

| '+'
{
	$$ = '+';
}

| '*'
{
	$$ = '*';
}

| '/'
{
	$$ = '/';
}
;


SimpleExp: 

Float
{
        // construct and send up the operand containing the FP number
        $$ = (struct FuncOperand *) malloc (sizeof (struct FuncOperand));
        $$->code = DOUBLE;
        $$->value = $1;
} 

| Int
{
        // construct and send up the operand containing the integer
        $$ = (struct FuncOperand *) malloc (sizeof (struct FuncOperand));
        $$->code = INT;
        $$->value = $1;
} 

| Name
{
        // construct and send up the operand containing the name
        $$ = (struct FuncOperand *) malloc (sizeof (struct FuncOperand));
        $$->code = NAME;
        $$->value = $1;
}
;

%%

