#ifndef _PRIORITY_QUEUE
#define _PRIORITY_QUEUE

#include "Record.h"
#include "Comparison.h"
#include "Hashes.h"
#include "File.h"


// stores part of a run in memory.
// It is refilled each time the records are used up, until the run is exhausted.
struct InMemoryRun {
	// an array holding 'size' records in memory
	Record *buffer;

	// size is the size of the in memory run
	// curIndex points to the next unread record
	int buffer_size;

	int cur_size;
	int cur_index;
};

struct ExternalNode {
	// holds a pointer to a record
	// actual record in InMemory run
	Record *rec;

	//the number of the run corresponding to this node
	int run_no;

	// holds the records of this run currenly in main memory
	InMemoryRun *mem_run;
};

struct InternalNode {
	// holds a pointer to a record
	// actual record in InMemory run
	Record *rec;

	// the run from which the record stored here comes from.
	int which_external;
};



// A min heap to merge multiple sorted runs into a single run
class PriorityQueue {

private:
	char *run_name;
	int runs_no;

	OrderMaker *order;
	Hash *sort_func;
	Schema *schema;

	// this is where the runs are stored on disk
	Run_File_Interface *run_file;


	// The heap nodes
	// number of external (or internal) nodes in the heap
	// total num of nodes in heap = 2 * numHeapNodes
	int nodes_no;

	// An internal stores a record pointer
	InternalNode *internal_heap;
	// stores a record and a run to refill the node
	ExternalNode *external_heap;

	// the total memory that can be used by this heap
	int buffer_size;


	void InitializeHeap();

	void ReadDataIntoRun(InMemoryRun *run_mem, int run_no);

	void UpdateFromExternal(int parent_index, int child_index);
	void UpdateFromInternal(int parent_index, int child_index);

public:
	PriorityQueue(char *run_name, int runs_no, OrderMaker *order, Hash *sort_func, Schema *schema,  int buffer_size);
	virtual ~PriorityQueue();

	// return 1 if a record exists, 0 otherwise
	int GetNext(Record *record);

	// compare the top of the priority queue
	int CompareToTop(Record *record);
};

#endif
