#include <iostream>
#include <pthread.h>
#include <limits.h>
#include <sys/time.h>
#include <time.h>

#include "Record.h"
#include "SMOrder.h"
#include "Pipe.h"


using namespace std;


File_Interface *fileIn;
File_Interface *fileOut;
Schema *schema;
Pipe *myPipe;
int *atts;
int atts_no;

void *producer(void *arg)
{
	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	while (fileIn->NextRec(rec))
		myPipe->Insert(rec);

	myPipe->ShutDown();

	delete rec;
}

void *consumer(void *arg)
{
	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	OrderMaker sortOrder;

	sortOrder.numAtts = atts_no;
	for (int i = 0; i < atts_no; i++)
	{
		sortOrder.whichAtts[i] = atts[i];
		sortOrder.whichTypes[i] = schema->GetAtts()[atts[i]].myType;
	}

	Hash *hash_f = new (std::nothrow) CWHash(sortOrder.numAtts + 1, HASH_DOMAIN);
	if (hash_f == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Order orderMe(1000000, &sortOrder, hash_f, schema, "mx");

	while (myPipe->Remove(rec))
		orderMe.Insert(rec);

	orderMe.BuildQueue();

	while (orderMe.GetNextRec(rec))
	{
		fileOut->Append(rec);
	}

	delete rec;
}

int main(int argc, char **argv)
{
	cerr << "Usage: " << argv[0] << " [input file] [output file] [schema file] [schema name] [atts_no] [atts_list]\n";

	srand((unsigned int)time((time_t*)NULL));

	atts_no = atoi(argv[5]);
	atts = new (std::nothrow) int[atts_no];
	if (atts == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	for (int i = 0; i < atts_no; i++)
	{
		atts[i] = atoi(argv[6 + i]);
	}

	schema = new (std::nothrow) Schema(argv[3], argv[4]);
	if (schema == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	fileIn = new (std::nothrow) File();
	if (fileIn == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	fileIn->Open(1, argv[1], FT_RECORD);

	fileOut = new (std::nothrow) File();
	if (fileOut == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	fileOut->Open(0, argv[2], FT_RECORD);


	myPipe = new (std::nothrow) Pipe(10000);
	if (myPipe == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	pthread_t producerThread, consumerThread;

	if (pthread_create(&producerThread, NULL, producer, NULL) == 0)
		if (pthread_create(&consumerThread, NULL, consumer, NULL) == 0)
		{
			pthread_join(producerThread, NULL);
			pthread_join(consumerThread, NULL);
		}
		else
			cerr << "Could not create consumer thread...\n";
	else
		cerr << "Could not create producer thread...\n";


	fileIn->Close();
	delete fileIn;

	fileOut->Close();
	delete fileOut;

	delete schema;

	delete myPipe;

	delete [] atts;

	return(0);
}

