#include <iostream>
#include <pthread.h>
#include <limits.h>
#include <sys/time.h>
#include <time.h>

#include "Record.h"
#include "SMOrder.h"
#include "Pipe.h"


using namespace std;


File_Interface *fileIn;
File_Interface **fileOut;
Schema *schema;
Schema **schemas;
Pipe *myPipe;

void *producer(void *arg)
{
	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	while (fileIn->NextRec(rec))
		myPipe->Insert(rec);

	myPipe->ShutDown();

	delete rec;
}

void *consumer(void *arg)
{
	int sch_atts_no = schema->GetNumAtts();

	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Record *out_rec = new (std::nothrow) Record();
	if (out_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	OrderMaker sortOrder;

	sortOrder.numAtts = schema->GetNumAtts();
	for (int i = 0; i < sortOrder.numAtts; i++)
	{
		sortOrder.whichAtts[i] = i;
		sortOrder.whichTypes[i] = schema->GetAtts()[i].myType;
	}

	Hash *hash_f = new (std::nothrow) CWHash(sortOrder.numAtts + 1, HASH_DOMAIN);
	if (hash_f == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Order orderMe(1000000, &sortOrder, hash_f, schema, "mx");

	while (myPipe->Remove(rec))
		orderMe.Insert(rec);

	orderMe.BuildQueue();

	while (orderMe.GetNextRec(rec))
	{
		for (int i = 0; i < sch_atts_no; i++)
		{
			out_rec->Copy(rec);
			out_rec->Project(&i, 1, sch_atts_no);

			fileOut[i]->Append(out_rec);
		}
	}

	delete rec;
	delete out_rec;
}

int main(int argc, char **argv)
{
	if (argc == 5)
	{
		srand((unsigned int)time((time_t*)NULL));


		schema = new (std::nothrow) Schema(argv[3], argv[4]);
		if (schema == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		int sch_atts_no = schema->GetNumAtts();
		Attribute *sch_atts = schema->GetAtts();

		schemas = new (std::nothrow) Schema*[sch_atts_no];
		if (schemas == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}


		fileIn = new (std::nothrow) File();
		if (fileIn == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
		fileIn->Open(1, argv[1], FT_RECORD);

		fileOut = new (std::nothrow) File_Interface*[sch_atts_no];
		if (fileOut == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}


		for (int i = 0; i < sch_atts_no; i++)
		{
			//STRIPE
//			fileOut[i] = new (std::nothrow) File();
			fileOut[i] = new (std::nothrow) Stripe_File();
			if (fileOut[i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			int file_type;
			switch(sch_atts[i].myType)
			{
				case Int:
					file_type = FT_INT;
					break;
				case Double:
					file_type = FT_REAL;
					break;
				case String:
					file_type = FT_STRING;
					break;
			}

			//STRIPE
/*
			char fc_name[200];
			sprintf(fc_name, "data/%s", sch_atts[i].name);
*/

			fileOut[i]->Open(0, sch_atts[i].name, file_type);


			schemas[i] = new (std::nothrow) Schema();
			if (schemas[i] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			schemas[i]->CopyFrom(schema);
			schemas[i]->Project(&i, 1);
		}


		myPipe = new (std::nothrow) Pipe(10000);
		if (myPipe == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}


		pthread_t producerThread, consumerThread;

		if (pthread_create(&producerThread, NULL, producer, NULL) == 0)
			if (pthread_create(&consumerThread, NULL, consumer, NULL) == 0)
			{
				pthread_join(producerThread, NULL);
				pthread_join(consumerThread, NULL);
			}
			else
				cerr << "Could not create consumer thread...\n";
		else
			cerr << "Could not create producer thread...\n";


		fileIn->Close();
		delete fileIn;

		for (int i = 0; i < sch_atts_no; i++)
		{
			fileOut[i]->Close();

			delete fileOut[i];
			fileOut[i] = NULL;

			delete schemas[i];
		}

		delete [] fileOut;

		delete [] schemas;
		delete schema;

		delete myPipe;


	} else
		cerr << "Usage: " << argv[0] << " [input file] [output file] [schema file] [schema name]\n";

	return(0);
}
