#include <iostream>
#include <pthread.h>
#include <limits.h>
#include <sys/time.h>
#include <time.h>

#include "Record.h"
#include "SMOrder.h"
#include "Pipe.h"


using namespace std;


FILE *fileIn;
File fileOut;
Schema *schema;
Pipe *myPipe;

void *producer(void *arg)
{
	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	while (rec->SuckNextRecord(schema, fileIn) == 1)
		myPipe->Insert(rec);

	myPipe->ShutDown();
}

void *consumer(void *arg)
{
	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	OrderMaker sortOrder;

	sortOrder.numAtts = schema->GetNumAtts();
	for (int i = 0; i < sortOrder.numAtts; i++)
	{
		sortOrder.whichAtts[i] = i;
		sortOrder.whichTypes[i] = schema->GetAtts()[i].myType;
	}

	Hash *hash_f = new (std::nothrow) CWHash(sortOrder.numAtts + 1, HASH_DOMAIN);
	if (hash_f == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Order orderMe(1000000, &sortOrder, hash_f, schema, "mx");

	while (myPipe->Remove(rec))
		orderMe.Insert(rec);

	orderMe.BuildQueue();


	int n = 0;
	while (orderMe.GetNextRec(rec))
	{
/*
		n += 1;
		printf("%10d\n", n);

		Record temp_rec;
		temp_rec.Copy(rec);
		temp_rec.Print(schema);
*/

		fileOut.Append(rec);
	}



}

int main(int argc, char **argv)
{
	if (argc == 5)
	{
		srand((unsigned int)time((time_t*)NULL));

		//cerr << "opening fIn... " << argv[1] << "\n";

		fileIn = fopen(argv[1], "r");
		myPipe = new (std::nothrow) Pipe(10000);
		if (myPipe == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		if (fileIn != NULL)
		{
			//cerr << "opening fOut... " << argv[2] << "\n";

			fileOut.Open(0, argv[2], FT_RECORD);

			schema = new (std::nothrow) Schema(argv[3], argv[4]);
			if (schema == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			pthread_t producerThread, consumerThread;

			if (pthread_create(&producerThread, NULL, producer, NULL) == 0)
				if (pthread_create(&consumerThread, NULL, consumer, NULL) == 0)
				{
					pthread_join(producerThread, NULL);
					pthread_join(consumerThread, NULL);

					fclose(fileIn);
					fileOut.Close();

					//cerr << "Successfully generated files...\n";
				}
					else
						cerr << "Could not create consumer thread...\n";
			else
				cerr << "Could not create producer thread...\n";
			

		} else
			cerr << "Could not open input file " << argv[1] << "\n";
		
		;
		
	} else
		cerr << "Usage: " << argv[0] << " [input file] [output file] [schema file] [schema name]\n";

	return(0);
}


