#include "RecPriorityQ.h"
#include "TwoWayList.cc"


void Holder :: Consume (Holder *me) {
	myRec->Consume (me->myRec);
	timeStamp = me->timeStamp;
};

Holder :: Holder () {
	myRec = new Record();
	timeStamp = 0;
}

Holder :: ~Holder () {
	delete myRec;
}

RecPriorityQ :: ~RecPriorityQ () {

/*
	for (int i = 0; i < numLists; i++) {
		cerr << i << "\t" << allLists[i].RightLength () << endl;
	}
*/

	delete [] allLists;
	delete [] whichAreDone;
}


RecPriorityQ :: RecPriorityQ (int numInputRels, int *whichAreDoneIn) {

	// create all of the individual queues
	allLists = new TwoWayList <Holder>[numInputRels];

	// this records which ones we should not worry about
	whichAreDone = new int[numInputRels];
	for (int i = 0; i < numInputRels; i++) {
		whichAreDone[i] = whichAreDoneIn[i];
	}
	numLists = numInputRels;
}

void RecPriorityQ :: Insert (Record *insertMe, int whichRel, int timeStamp) {

	// first, make sure we are not inserting into a done queue
	if (whichAreDone[whichRel] == 1) {
		cerr << "You are trying to insert into the queue for a relation that is done!!!\n";
		exit (1);
	}

	Holder *tempHolder = new Holder;
	tempHolder->myRec->Consume (insertMe);
	tempHolder->timeStamp = timeStamp;

	allLists[whichRel].MoveToFinish ();
	allLists[whichRel].Insert (tempHolder);
	allLists[whichRel].MoveToStart ();

	delete tempHolder;
}

void RecPriorityQ :: ThisGuyIsDone (int whichRel) {

	whichAreDone[whichRel] = 1;

	// make sure we're not marking one with data as being done
	if (allLists[whichRel].RightLength ()) {
		cerr << "You are trying to kill a relation that I still have more data for!!\n";
		exit (1);
	}
}

int RecPriorityQ :: Remove (Record *removeMe, int &timeStamp, int &whichRel) {

	// find the lowest timestamp
	int lowest = -1;
	int lowestIndex = -1;

	for (int i = 0; i < numLists; i++) {

		// see if this particulare relation is done with
		if (whichAreDone[i])
			continue;

		// it is not, so deal with it
		if (allLists[i].RightLength () == 0) {
			cerr << "Bad!!  There is an active relation with no data in the queue.\n";
			cerr << "Either add data to this queue, or else let me know it's inactive.\n";
			exit (1);
		}

		// in the first case, this is the first one to check
		if (lowestIndex < 0) {
			lowestIndex = i;
			lowest = allLists[i].Current(0)->timeStamp;

//			cout << lowest << "\n";

		// in the second case, see if it is the least one
		} else {
			if (allLists[i].Current(0)->timeStamp < lowest) {
				lowestIndex = i;
				lowest = allLists[i].Current(0)->timeStamp;
			}
		}

//		cout << "Found active list.\n";
	}

	// now, return the lowest one and exit
	if (lowest < 0) {
		cerr << "Bad! Somehow, no more relations were active in Remove.\n";
		exit (1);
	}

	// set up the return params
	Holder *foo = new Holder;
	allLists[lowestIndex].Remove (foo);
	removeMe->Consume (foo->myRec);
	timeStamp = foo->timeStamp;
	whichRel = lowestIndex;
	delete foo;

	// return a one if this relation is not exhausted
	if (allLists[lowestIndex].RightLength ())
		return 1;


/*
	for (int i = 0; i < numLists; i++) {

		// see if this particulare relation is done with
		if (whichAreDone[i])
			continue;

		// it is not, so deal with it
		if (allLists[i].RightLength () > 0) {
			cerr << i << "\t" << allLists[i].Current(0)->timeStamp << "\t" << allLists[i].RightLength() << endl;
		}
	}
*/


	// return a zero otherwise
	return 0;
}

