#include "RelOpLevel.h"
#include "TwoWayList.cc"
#include "DEFS.h"
#include "Hashes.h"

#include <string.h>
#include <stdlib.h>
#include <iostream>
#include <vector>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>



RelationalOpLevel::~RelationalOpLevel()
{
}


double RelationalOpLevel::GetTuplesNo()
{
	return 0.0;
}


double RelationalOpLevel::GetAggregate()
{
	return 0.0;
}


Schema* RelationalOpLevel::GetSchema()
{
	return NULL;
}


Function* RelationalOpLevel::GetFunction()
{
	return NULL;
}


void RelationalOpLevel::GetJoinAtts(int *&join_atts, int &join_atts_no)
{
	return;
}


void RelationalOpLevel::GetGroupAtts(int *&group_atts, int &group_atts_no)
{
	return;
}


void RelationalOpLevel::GetGroups(vector<Record*> &desc, vector<double> &aggregate, vector<double> &tuples, int &groups_no)
{
	return;
}




TableScanLevel::TableScanLevel()
{
}


TableScanLevel::TableScanLevel(const char *file_location, Pipe *out_pipe, CNF *sel_op, Record *literal, Schema *schema, int *atts, int atts_no, int tuples_no)
{
//	printf("TableScanLevel:Constructor:In\n");

	this->file_location = strdup(file_location);

	this->pipe = out_pipe;

	this->op = sel_op;
	this->lit = literal;
	this->schema = schema;

	this->atts = atts;
	this->atts_no = atts_no;

	this->tuples_no = tuples_no;
	this->tuples_proc = 0;

	this->exact_hash = new (std::nothrow) Hash();
	if (this->exact_hash == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->hash_time = new (std::nothrow) CWHash(2, HASH_DOMAIN / this->tuples_no);
	if (this->hash_time == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	gsl_rng_env_setup();
	
	T = gsl_rng_default;
	r = gsl_rng_alloc (T);

	gsl_rng_set(r, Random_Generate());

	last_time_stamp = 0.0;
	last_tuple_id = -1.0;

	//STRIPE
//	this->f = new (std::nothrow) Stripe_File();
	this->f = new (std::nothrow) File();
	if (this->f == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	pthread_mutex_init (&start_mutex, NULL);

//	printf("TableScanLevel:Constructor:Out\n");
}


TableScanLevel::~TableScanLevel()
{
//	printf("TableScanLevel:Destructor:In\n");

	free(file_location);

	lit = NULL;
	op = NULL;
	schema = NULL;

	atts = NULL;
	atts_no = 0;

	if (tuples_proc < tuples_no)
	{
		if (f != NULL)
			f->Close();

		pipe->ShutDown();
	}

	pipe = NULL;

	tuples_no = 0;
	tuples_proc = 0;

	delete exact_hash;
	exact_hash = NULL;

	delete hash_time;
	hash_time = NULL;

	gsl_rng_free (r);

	delete f;
	f = NULL;

	pthread_mutex_destroy (&start_mutex);

//	printf("TableScanLevel:Destructor:Out\n");
}



int TableScanLevel::GenerateTimeStamp(int tuple_id)
{

//This version generates pseudo-random values based on the domain and hashing.
/*
	int base = tuple_id * (HASH_DOMAIN / tuples_no);
	base += hash_time->Compute_Hash(&tuple_id, 1);

	return base;
*/



//This version uses the Beta distribution numbers as described in the paper
	double beta_v = gsl_ran_beta (r, (double)tuple_id - last_tuple_id, tuples_no - tuple_id);
	last_time_stamp = last_time_stamp + beta_v * (1.0 - last_time_stamp);
	last_tuple_id = tuple_id;

	int base = (int)(last_time_stamp * HASH_DOMAIN);

	return base;


/*
//This version is a combination of the two above
	double beta_v = gsl_ran_beta (r, 1.0, HASH_DOMAIN / tuples_no);

	int base = tuple_id * (HASH_DOMAIN / tuples_no);
	base += (int)(beta_v * HASH_DOMAIN / tuples_no);

	return base;
*/
}



int TableScanLevel::GetNextBatch (double p, Record *output, int max_output_size, int &output_size)
{
	int ret_val = 0;

	output_size = 0;

	pthread_mutex_lock(&start_mutex);

//	printf("TableScanLevel:GetNextBatch:In\n");

	if ((tuples_proc >= tuples_no) || (p == 0.0))
	{
//		printf("TableScanLevel:GetNextBatch:Out\n");

		pthread_mutex_unlock(&start_mutex);

		return ret_val;
	}

	ret_val = 1;
	if (p < 0.0)
	{
		int schema_atts_no = schema->GetNumAtts();

		Record *in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		while (f->NextRec(in_rec))
		{
			if (exact_hash->Compare(in_rec, lit, op))
			{
				in_rec->Project(atts, atts_no, schema_atts_no);

				in_rec->hash = GenerateTimeStamp(tuples_proc);
				output[output_size].Copy(in_rec);
				output_size += 1;
				
				in_rec->hash = -1;
				pipe->Insert(in_rec);
			}

			tuples_proc += 1;
			if (tuples_proc >= tuples_no)
				break;

			if (output_size == max_output_size)
			{
				ret_val = -1;
				break;
			}
		}
		delete in_rec;

		if (tuples_proc >= tuples_no)
		{
			f->Close();
			pipe->ShutDown();
		}
	}

	else
	{
		int tuples_proc_now = (int)(p * tuples_no);
		tuples_proc_now = tuples_proc_now > 0 ? tuples_proc_now : 1;

		int i = 0;

		int schema_atts_no = schema->GetNumAtts();
	
		Record *in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		while (i < tuples_proc_now)
		{
			if (f->NextRec(in_rec))
			{
				if (exact_hash->Compare(in_rec, lit, op))
				{
					in_rec->Project(atts, atts_no, schema_atts_no);
	
					in_rec->hash = GenerateTimeStamp(tuples_proc + i);
					output[output_size].Copy(in_rec);
					output_size += 1;
					
					in_rec->hash = -1;
					pipe->Insert(in_rec);
				}

				i += 1;

				if (output_size == max_output_size)
				{
					ret_val = -1;
					break;
				}
			}
			else
			{
				break;
			}
		}
		delete in_rec;
	
		tuples_proc += i;
		if (tuples_proc >= tuples_no)
		{
			f->Close();
			pipe->ShutDown();
		}
	}

//	printf("TableScanLevel:GetNextBatch:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return ret_val;
}


void TableScanLevel::StartUp ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("TableScanLevel:StartUp:In\n");

	pthread_create(&thr, NULL, TableScanLevel::DoStartUp, (void *)this);
	pthread_join(thr, NULL);

//	printf("TableScanLevel:StartUp:Out\n");

	pthread_mutex_unlock(&start_mutex);
}


double TableScanLevel::HowMuchIsLeft ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("TableScanLevel:HowMuchIsLeft:In\n");

	double res = (double)(tuples_no - tuples_proc) / (double)tuples_no;

//	printf("TableScanLevel:HowMuchIsLeft:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return res;
}


void *TableScanLevel::DoStartUp(void *arg)
{
//	printf("TableScanLevel:DoStartUp:In\n");

	TableScanLevel *ts = (TableScanLevel *)arg;
	ts->f->Open(1, ts->file_location, FT_RECORD);

//	printf("TableScanLevel:DoStartUp:Out\n");
}




TableScanLevelC::TableScanLevelC(const char *file_location, Pipe *out_pipe, CNF *sel_op, Record *literal, Schema *schema, int *atts, int atts_no, int tuples_no)
{
//	printf("TableScanLevelC:Constructor:In\n");

	this->file_location = strdup(file_location);

	this->pipe = out_pipe;

	this->op = sel_op;
	this->lit = literal;
	this->schema = schema;

	this->atts = atts;
	this->atts_no = atts_no;

	this->tuples_no = tuples_no;
	this->tuples_proc = 0;


	this->read_atts_no = this->atts_no;
	this->read_atts = new (std::nothrow) int[this->schema->GetNumAtts()];
	if (this->read_atts == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	//the attributes in the output schema have to be read (at least)
	memcpy(this->read_atts, this->atts, this->atts_no * sizeof(int));

	//next step is to add the attributes with predicates and not in the output schema
	//these attributes appear in the CNF
	//the indices in the CNF are modified at the same time
	this->op->AddCNFAtts(this->read_atts, this->read_atts_no);

	//modify the attributes in the output schema
	//they are the first in read_atts
	for (int i = 0; i < this->atts_no; i++)
		this->atts[i] = i;


	this->exact_hash = new (std::nothrow) Hash();
	if (this->exact_hash == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->hash_time = new (std::nothrow) CWHash(2, HASH_DOMAIN / this->tuples_no);
	if (this->hash_time == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	gsl_rng_env_setup();
	
	T = gsl_rng_default;
	r = gsl_rng_alloc (T);

	last_time_stamp = 0.0;
	last_tuple_id = -1.0;


	this->f = NULL;


	this->fc = new (std::nothrow) File_Interface*[this->read_atts_no];
	if (this->fc == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->f_pipe = new (std::nothrow) Pipe*[this->read_atts_no];
	if (this->f_pipe == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->thr_params = new (std::nothrow) ThreadTableScanParam*[this->read_atts_no];
	if (this->thr_params == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->myThreads = new (std::nothrow) pthread_t[this->read_atts_no];
	if (this->myThreads == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	for (int i = 0; i < this->read_atts_no; i++)
	{
		//STRIPE
//		fc[i] = new (std::nothrow) Stripe_File();
		fc[i] = new (std::nothrow) File();

		if (fc[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		f_pipe[i] = new (std::nothrow) Pipe(PIPE_BUFFERSIZE);
		if (f_pipe[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		thr_params[i] = new (std::nothrow) ThreadTableScanParam;
		if (thr_params[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		thr_params[i]->f = fc[i];
		thr_params[i]->out_pipe = f_pipe[i];
	}


	pthread_mutex_init (&start_mutex, NULL);

//	printf("TableScanLevelC:Constructor:Out\n");
}


TableScanLevelC::~TableScanLevelC()
{
//	printf("TableScanLevelC:Destructor:In\n");

	for (int i = 0; i < this->read_atts_no; i++)
	{
		delete fc[i];
		fc[i] = NULL;

		delete f_pipe[i];
		f_pipe[i] = NULL;

		thr_params[i]->f = NULL;
		thr_params[i]->out_pipe = NULL;
		delete thr_params[i];
		thr_params[i] = NULL;
	}

	delete [] fc;
	delete [] f_pipe;
	delete [] thr_params;
	delete [] myThreads;

	delete [] read_atts;

	pthread_mutex_destroy (&start_mutex);

//	printf("TableScanLevelC:Destructor:Out\n");
}



int TableScanLevelC::GetNextBatch (double p, Record *output, int max_output_size, int &output_size)
{
	int ret_val = 0;

	output_size = 0;

	pthread_mutex_lock(&start_mutex);

//	printf("TableScanLevel:GetNextBatch:In\n");

	if ((tuples_proc >= tuples_no) || (p == 0.0))
	{
//		printf("TableScanLevel:GetNextBatch:Out\n");

		pthread_mutex_unlock(&start_mutex);

		return ret_val;
	}

	ret_val = 1;
	if (p < 0.0)
	{
		Record *in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Record *in_rec1 = new (std::nothrow) Record();
		if (in_rec1 == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Record *in_rec2 = new (std::nothrow) Record();
		if (in_rec2 == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		while (f_pipe[0]->Remove(in_rec))
		{
			for (int i = 1; i < read_atts_no; i++)
			{
				f_pipe[i]->Remove(in_rec1);

				in_rec2->AppendRecords(in_rec, in_rec1, i, 1);

				in_rec->Consume(in_rec2);
			}

			if (exact_hash->Compare(in_rec, lit, op))
			{
				if (atts_no < read_atts_no)
					in_rec->Project(atts, atts_no, read_atts_no);

				in_rec->hash = GenerateTimeStamp(tuples_proc);
				output[output_size].Copy(in_rec);
				output_size += 1;
				
				in_rec->hash = -1;
				pipe->Insert(in_rec);
			}

			tuples_proc += 1;
			if (tuples_proc >= tuples_no)
				break;

			if (output_size == max_output_size)
			{
				ret_val = -1;
				break;
			}
		}

		delete in_rec;
		delete in_rec1;
		delete in_rec2;

		if (tuples_proc >= tuples_no)
		{
			pipe->ShutDown();
		}
	}

	else
	{
		int tuples_proc_now = (int)(p * tuples_no);
		tuples_proc_now = tuples_proc_now > 0 ? tuples_proc_now : 1;

		int i = 0;

		Record *in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Record *in_rec1 = new (std::nothrow) Record();
		if (in_rec1 == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		Record *in_rec2 = new (std::nothrow) Record();
		if (in_rec2 == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		while (i < tuples_proc_now)
		{
			if (f_pipe[0]->Remove(in_rec))
			{
				for (int j = 1; j < read_atts_no; j++)
				{
					f_pipe[j]->Remove(in_rec1);
	
					in_rec2->AppendRecords(in_rec, in_rec1, j, 1);
	
					in_rec->Consume(in_rec2);
				}

				if (exact_hash->Compare(in_rec, lit, op))
				{
					if (atts_no < read_atts_no)
						in_rec->Project(atts, atts_no, read_atts_no);

					in_rec->hash = GenerateTimeStamp(tuples_proc + i);
					output[output_size].Copy(in_rec);
					output_size += 1;
					
					in_rec->hash = -1;
					pipe->Insert(in_rec);
				}

				i += 1;

				if (output_size == max_output_size)
				{
					ret_val = -1;
					break;
				}
			}
			else
			{
				break;
			}
		}

		delete in_rec;
		delete in_rec1;
		delete in_rec2;
	
		tuples_proc += i;
		if (tuples_proc >= tuples_no)
		{
			pipe->ShutDown();
		}
	}

//	printf("TableScanLevel:GetNextBatch:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return ret_val;
}


void TableScanLevelC::StartUp ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("TableScanLevel:StartUp:In\n");

	pthread_create(&thr, NULL, TableScanLevelC::DoStartUp, (void *)this);
	pthread_join(thr, NULL);

//	printf("TableScanLevel:StartUp:Out\n");

	pthread_mutex_unlock(&start_mutex);
}


void *TableScanLevelC::DoStartUp(void *arg)
{
//	printf("TableScanLevel:DoStartUp:In\n");

	TableScanLevelC *ts = (TableScanLevelC *)arg;

	Attribute *sch_atts = ts->schema->GetAtts();
	for (int i = 0; i < ts->read_atts_no; i++)
	{
		int file_type;
		switch(sch_atts[ts->read_atts[i]].myType)
		{
			case Int:
				file_type = FT_INT;
				break;
			case Double:
				file_type = FT_REAL;
				break;
			case String:
				file_type = FT_STRING;
				break;
		}

		//STRIPE
		char fc_name[200];
		sprintf(fc_name, "%s%s", ts->file_location, sch_atts[ts->read_atts[i]].name);

//		ts->fc[i]->Open(1, sch_atts[ts->read_atts[i]].name, file_type);
		ts->fc[i]->Open(1, fc_name, file_type);

		// and let the corresponding thread go
		pthread_attr_t t_attr;
		int ret;
	
		ret = pthread_attr_init(&t_attr);
		ret = pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_DETACHED);
		ret = pthread_create(&(ts->myThreads[i]), &t_attr, TableScan_Thread, (void *) (ts->thr_params[i]));
		if (ret) {
			cerr << "ERROR; return code from pthread_create() is " << ret << "\n";
			exit(-1);
		}
	}

//	printf("TableScanLevel:DoStartUp:Out\n");
}





TableScanPreJoinLevel::TableScanPreJoinLevel(char **file_location, Pipe *out_pipe, Schema **schema, int **proj_atts, int *proj_atts_no, int *proj_atts_out, int proj_atts_out_no, AndList *conditions, int *tuples_no, int relations_no)
{
//	printf("TableScanLevel:Constructor:In\n");

	this->small_relations_no = relations_no - 1;

	this->big_file_location = strdup(file_location[relations_no - 1]);

	this->file_location = new (std::nothrow) char*[this->small_relations_no];
	if (this->file_location == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	for (int i = 0; i < this->small_relations_no; i++)
	{
		this->file_location[i] = strdup(file_location[i]);
	}

	//STRIPE
//	this->big_file = new (std::nothrow) Stripe_File();
	this->big_file = new (std::nothrow) File();
	if (this->big_file == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->f = new (std::nothrow) File_Interface*[this->small_relations_no];
	if (this->f == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	for (int i = 0; i < this->small_relations_no; i++)
	{
		//STRIPE
//		this->f[i] = new (std::nothrow) Stripe_File();
		this->f[i] = new (std::nothrow) File();
	}

	this->pipe = out_pipe;

	this->big_file_schema = schema[relations_no - 1];
	this->schema = schema;

	this->proj_atts_no = proj_atts_no;
	this->proj_atts = proj_atts;

	this->small_atts_no = 0;
	for (int i = 0; i < this->small_relations_no; i++)
	{
		this->small_atts_no += this->proj_atts_no[i];
	}
	this->total_atts_no = this->small_atts_no + this->proj_atts_no[this->small_relations_no];

	this->proj_atts_out_no = proj_atts_out_no;
	this->proj_atts_out = proj_atts_out;

	this->conditions = new (std::nothrow) AndListWrapper(conditions);
	if (this->conditions == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->recLiteral = new (std::nothrow) Record();
	if (this->recLiteral == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->cnf = new (std::nothrow) CNF();
	if (this->cnf == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->conditions->BuildCNFSelect(*big_file_schema, *cnf, *recLiteral);

	this->big_file_tuples_no = tuples_no[relations_no - 1];
	this->tuples_proc = 0;

	this->tuples_no = tuples_no;

	this->exact_hash = new (std::nothrow) Hash();
	if (this->exact_hash == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->hash_table = NULL;

	gsl_rng_env_setup();
	
	T = gsl_rng_default;
	r = gsl_rng_alloc (T);

	gsl_rng_set(r, Random_Generate());

	last_time_stamp = 0.0;
	last_tuple_id = -1.0;


	pthread_mutex_init (&start_mutex, NULL);

//	printf("TableScanLevel:Constructor:Out\n");
}


TableScanPreJoinLevel::~TableScanPreJoinLevel()
{
//	printf("TableScanLevel:Destructor:In\n");

	free(big_file_location);

	for (int i = 0; i < small_relations_no; i++)
	{
		free(file_location[i]);
	}
	delete [] file_location;

	for (int i = 0; i < small_relations_no; i++)
	{
		delete f[i];
		f[i] = NULL;
	}
	delete [] f;

	big_file_schema = NULL;
	schema = NULL;

	proj_atts_no = NULL;
	proj_atts = NULL;

	proj_atts_out_no = 0;
	proj_atts_out = NULL;

	delete conditions;
	conditions = NULL;

	delete recLiteral;
	recLiteral = NULL;

	delete cnf;
	cnf = NULL;

	proj_atts_no = NULL;
	proj_atts = NULL;

	proj_atts_out_no = 0;
	proj_atts_out = NULL;

	tuples_no = NULL;

	if (tuples_proc < big_file_tuples_no)
	{
		if (big_file != NULL)
			big_file->Close();

		pipe->ShutDown();
	}

	delete big_file;
	big_file = NULL;

	pipe = NULL;

	big_file_tuples_no = 0;
	tuples_proc = 0;

	delete exact_hash;
	exact_hash = NULL;

	delete hash_table;
	hash_table = NULL;

	gsl_rng_free (r);

	pthread_mutex_destroy (&start_mutex);

//	printf("TableScanLevel:Destructor:Out\n");
}



int TableScanPreJoinLevel::GenerateTimeStamp(int tuple_id)
{

//This version generates pseudo-random values based on the domain and hashing.
/*
	int base = tuple_id * (HASH_DOMAIN / tuples_no);
	base += hash_time->Compute_Hash(&tuple_id, 1);

	return base;
*/



//This version uses the Beta distribution numbers as described in the paper
	double beta_v = gsl_ran_beta (r, (double)tuple_id - last_tuple_id, big_file_tuples_no - tuple_id);
	last_time_stamp = last_time_stamp + beta_v * (1.0 - last_time_stamp);
	last_tuple_id = tuple_id;

	int base = (int)(last_time_stamp * HASH_DOMAIN);

	return base;


/*
//This version is a combination of the two above
	double beta_v = gsl_ran_beta (r, 1.0, HASH_DOMAIN / tuples_no);

	int base = tuple_id * (HASH_DOMAIN / tuples_no);
	base += (int)(beta_v * HASH_DOMAIN / tuples_no);

	return base;
*/
}



int TableScanPreJoinLevel::GetNextBatch (double p, Record *output, int max_output_size, int &output_size)
{
	int ret_val = 0;

	output_size = 0;

	pthread_mutex_lock(&start_mutex);

//	printf("TableScanLevel:GetNextBatch:In\n");

	if ((tuples_proc >= big_file_tuples_no) || (p == 0.0))
	{
//		printf("TableScanLevel:GetNextBatch:Out\n");

		pthread_mutex_unlock(&start_mutex);

		return ret_val;
	}

	ret_val = 1;
	if (p < 0.0)
	{
		int schema_atts_no = big_file_schema->GetNumAtts();

		Record *in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		while (big_file->NextRec(in_rec))
		{
			if (exact_hash->Compare(in_rec, recLiteral, cnf))
			{
				in_rec->Project(proj_atts[small_relations_no], proj_atts_no[small_relations_no], schema_atts_no);

				TwoWayList<Record> *match_list = new (std::nothrow) TwoWayList<Record>;
				if (match_list == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				int ret_res = hash_table->Find(in_rec, &big_order, match_list);
				if (ret_res > 0)
				{
					//there exist matching records
					Record *m_rec = new (std::nothrow) Record();
					if (m_rec == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					match_list->MoveToStart();
					while (match_list->RightLength() > 0)
					{
						Record *l_rec = match_list->Current(0);

						if (exact_hash->Compare(in_rec, &big_order, l_rec, &small_order) == 0)
						{
							m_rec->AppendRecords(l_rec, in_rec, small_atts_no, proj_atts_no[small_relations_no]);
							m_rec->Project(proj_atts_out, proj_atts_out_no, total_atts_no);

							m_rec->hash = GenerateTimeStamp(tuples_proc);
							output[output_size].Copy(m_rec);
							output_size += 1;
							
							m_rec->hash = -1;
							pipe->Insert(m_rec);
						}

						match_list->Advance();
					}
					delete m_rec;
				}

				delete match_list;
			}

			tuples_proc += 1;

			if (tuples_proc >= big_file_tuples_no)
				break;

			if (output_size == max_output_size)
			{
				ret_val = -1;
				break;
			}
		}
		delete in_rec;

		if (tuples_proc >= big_file_tuples_no)
		{
			big_file->Close();
			pipe->ShutDown();
		}
	}

	else
	{
		int tuples_proc_now = (int)(p * big_file_tuples_no);
		tuples_proc_now = tuples_proc_now > 0 ? tuples_proc_now : 1;

		int i = 0;

		int schema_atts_no = big_file_schema->GetNumAtts();
	
		Record *in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		while (i < tuples_proc_now)
		{
			if (big_file->NextRec(in_rec))
			{
				if (exact_hash->Compare(in_rec, recLiteral, cnf))
				{
					in_rec->Project(proj_atts[small_relations_no], proj_atts_no[small_relations_no], schema_atts_no);
	
					TwoWayList<Record> *match_list = new (std::nothrow) TwoWayList<Record>;
					if (match_list == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
	
					int ret_res = hash_table->Find(in_rec, &big_order, match_list);
					if (ret_res > 0)
					{
						//there exist matching records
						Record *m_rec = new (std::nothrow) Record();
						if (m_rec == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
	
						match_list->MoveToStart();
						while (match_list->RightLength() > 0)
						{
							Record *l_rec = match_list->Current(0);
	
							if (exact_hash->Compare(in_rec, &big_order, l_rec, &small_order) == 0)
							{
								m_rec->AppendRecords(l_rec, in_rec, small_atts_no, proj_atts_no[small_relations_no]);
								m_rec->Project(proj_atts_out, proj_atts_out_no, total_atts_no);
	
								m_rec->hash = GenerateTimeStamp(tuples_proc + i);
								output[output_size].Copy(m_rec);
								output_size += 1;
								
								m_rec->hash = -1;
								pipe->Insert(m_rec);	
							}
	
							match_list->Advance();
						}
						delete m_rec;
					}
	
					delete match_list;
				}

				if (output_size == max_output_size)
				{
					ret_val = -1;
					break;
				}

				i += 1;
			}
			else
			{
				break;
			}
		}
		delete in_rec;
	
		tuples_proc += i;
		if (tuples_proc >= big_file_tuples_no)
		{
			big_file->Close();
			pipe->ShutDown();
		}
	}

//	printf("TableScanLevel:GetNextBatch:Out\n");

	pthread_mutex_unlock(&start_mutex);


	return ret_val;
}


void TableScanPreJoinLevel::StartUp ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("TableScanLevel:StartUp:In\n");

	pthread_create(&thr, NULL, TableScanPreJoinLevel::DoStartUp, (void *)this);
	pthread_join(thr, NULL);

//	printf("TableScanLevel:StartUp:Out\n");

	pthread_mutex_unlock(&start_mutex);
}


double TableScanPreJoinLevel::HowMuchIsLeft ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("TableScanLevel:HowMuchIsLeft:In\n");

	double res = (double)(big_file_tuples_no - tuples_proc) / (double)big_file_tuples_no;

//	printf("TableScanLevel:HowMuchIsLeft:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return res;
}


void *TableScanPreJoinLevel::DoStartUp(void *arg)
{
//	printf("TableScanLevel:DoStartUp:In\n");

	TableScanPreJoinLevel *ts = (TableScanPreJoinLevel *)arg;


	//read the records from the small relations in memory
	//apply all the selection predicates and project while reading

	Record **rels = new (std::nothrow) Record*[ts->small_relations_no];
	if (rels == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int *rels_no = new (std::nothrow) int[ts->small_relations_no];
	if (rels_no == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int max_join_size = 1;

	for (int i = 0; i < ts->small_relations_no; i++)
	{
		rels[i] = new (std::nothrow) Record[ts->tuples_no[i]];
		if (rels[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		int schema_atts_no = ts->schema[i]->GetNumAtts();

		Record recLiteral;
		CNF cnf;
		ts->conditions->BuildCNFSelect(*ts->schema[i], cnf, recLiteral);

		Record *in_rec = new (std::nothrow) Record();
		if (in_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		int k = 0;

		ts->f[i]->Open(1, ts->file_location[i], FT_RECORD);
		while (ts->f[i]->NextRec(in_rec))
		{
			if (ts->exact_hash->Compare(in_rec, &recLiteral, &cnf))
			{
				in_rec->Project(ts->proj_atts[i], ts->proj_atts_no[i], schema_atts_no);
				in_rec->hash = -1;
				rels[i][k].Consume(in_rec);

				k += 1;
			}

		}
		ts->f[i]->Close();

		rels_no[i] = k;
		max_join_size = max_join_size * k;

		delete in_rec;

		//generate the reduced schema (only the attributes used are kept)
		ts->schema[i]->Project(ts->proj_atts[i], ts->proj_atts_no[i]);
	}

	//also project the unneeded attributes from the big file schema
	ts->big_file_schema->Project(ts->proj_atts[ts->small_relations_no], ts->proj_atts_no[ts->small_relations_no]);


	//size of join at the current moment
	int k = 0;

	//join schema at the current moment
	Schema *join_schema = new (std::nothrow) Schema();
	if (join_schema == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	join_schema->CopyFrom(ts->schema[0]);

	//join attributes number
	int join_atts_no = 0;

	Record *join_res = new (std::nothrow) Record[max_join_size];
	if (join_res == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	//the join contains initially all the records from the first relation
	for (int i = 0; i < rels_no[0]; i++)
	{
		join_res[i].Copy(&rels[0][i]);
	}
	k = rels_no[0];
	join_atts_no = ts->proj_atts_no[0];


	for (int i = 1; i < ts->small_relations_no; i++)
	{
		Schema *right_schema = ts->schema[i];
		int right_atts_no = ts->proj_atts_no[i];

		//find the join conditions
		Record recLiteral;
		CNF cnf;
		ts->conditions->BuildCNF(*join_schema, *right_schema, cnf, recLiteral);

		OrderMaker left_ord_join, right_ord_join;
		cnf.GetSortOrders(left_ord_join, right_ord_join);

		Record *aux_join_res = new (std::nothrow) Record[max_join_size];
		if (aux_join_res == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		//here comes the nested loop join
		int aux_k = 0;
		for (int r1 = 0; r1 < k; r1++)
		{
			for (int r2 = 0; r2 < rels_no[i]; r2++)
			{
				//verify the join condition and do the actual join
				if ((ts->exact_hash->Compare(&join_res[r1], &left_ord_join, &rels[i][r2], &right_ord_join) == 0) && (ts->exact_hash->Compare(&join_res[r1], &rels[i][r2], &recLiteral, &cnf) == 1))
				{
					aux_join_res[aux_k].AppendRecords(&join_res[r1], &rels[i][r2], join_atts_no, right_atts_no);
					aux_join_res[aux_k].hash = -1;

					aux_k += 1;
				}
			}
		}

		k = aux_k;
		join_atts_no += right_atts_no;

		Schema *temp = new (std::nothrow) Schema();
		if (temp == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
		temp->CopyFrom(join_schema, ts->schema[i]);

		delete join_schema;
		join_schema = temp;

		delete [] join_res;
		join_res = new (std::nothrow) Record[max_join_size];
		if (join_res == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		//the join contains initially all the records from the first relation
		for (int j = 0; j < aux_k; j++)
		{
			join_res[j].Copy(&aux_join_res[j]);
		}

		delete [] aux_join_res;
	}


	//build the hash table for the small relations

	//identify the joining attributes
	Record recLiteral;
	CNF cnf;
	ts->conditions->BuildCNF(*join_schema, *ts->big_file_schema, cnf, recLiteral);
	cnf.GetSortOrders(ts->small_order, ts->big_order);

	ts->hash_table = new (std::nothrow) MMIndex(&ts->small_order, max_join_size);
	if (ts->hash_table == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	ts->hash_table->BuildIndex(join_res, k);

	delete join_schema;
	delete [] join_res;


	delete rels_no;

	for (int i = 0; i < ts->small_relations_no; i++)
	{
		delete [] rels[i];
		rels[i] = NULL;
	}
	delete [] rels;


	ts->big_file->Open(1, ts->big_file_location, FT_RECORD);

//	printf("TableScanLevel:DoStartUp:Out\n");
}







SortJoinLevel::SortJoinLevel(Pipe *in_pipe_left, int pipe_left_index, Pipe *in_pipe_right, int pipe_right_index, Pipe *out_pipe, CNF *sel_op, Record *literal, Schema *schema_left, Schema *schema_right, int *atts, int atts_no, int buffer_size)
{
//	printf("SortJoinLevel:Constructor:In\n");

	this->buffer_size = buffer_size;
	this->domain_proc = 0;

	this->in_pipe_left = in_pipe_left;
	this->in_pipe_right = in_pipe_right;
	this->out_pipe = out_pipe;

	this->op = sel_op;
	this->extra_cond = new (std::nothrow) CNF();
	if (this->extra_cond == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->extra_cond->EliminateEquiJoinCond(this->op);

	this->lit = literal;

	this->schema_left = schema_left;
	this->schema_right = schema_right;
	
	this->atts_no = atts_no;
	this->atts = new (std::nothrow) int[this->atts_no];
	if (this->atts == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int atts_left = schema_left->GetNumAtts();
        for(right_start = 0; right_start < atts_no; right_start++) {
                if(atts[right_start] >= atts_left) {
                        break;
                }
        }

        for(int i = 0; i < this->atts_no; i++) {
		if (i >= right_start)
                	this->atts[i] = atts[i] - atts_left;
		else
			this->atts[i] = atts[i];
        }


	this->om_left = new (std::nothrow) OrderMaker();
	if (om_left == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->om_right = new (std::nothrow) OrderMaker();
	if (om_right == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->op->GetSortOrders(*om_left, *om_right);


	this->sort_func = new (std::nothrow) CWHash(om_left->numAtts + 1, HASH_DOMAIN);
	if (sort_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->exact_hash = new (std::nothrow) Hash();
	if (exact_hash == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	char run_name[200];

	sprintf(run_name, "run_%d", pipe_left_index);
	this->order_left = new (std::nothrow) Order(this->buffer_size / 2, om_left, sort_func, this->schema_left, run_name);
	if (this->order_left == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	sprintf(run_name, "run_%d", pipe_right_index);
	this->order_right = new (std::nothrow) Order(this->buffer_size / 2, om_right, sort_func, this->schema_right, run_name);
	if (this->order_right == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->which_index = 0;

	this->is_done = 0;

	this->cur_left = new (std::nothrow) TwoWayList<Record>;
	if (this->cur_left == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->cur_right = new (std::nothrow) TwoWayList<Record>;
	if (this->cur_right == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	pthread_mutex_init (&start_mutex, NULL);

//	printf("SortJoinLevel:Constructor:Out\n");
}


SortJoinLevel::~SortJoinLevel()
{
//	printf("SortJoinLevel:Destructor:In\n");

	delete in_pipe_left;
	in_pipe_left = NULL;
	delete in_pipe_right;
	in_pipe_right = NULL;

	delete order_left;
	delete order_right;
	order_left = NULL;
	order_right = NULL;

	delete om_left;
	delete om_right;
	om_left = NULL;
	om_right = NULL;

	delete sort_func;
	sort_func = NULL;
	
	delete exact_hash;
	exact_hash = NULL;

	in_pipe_left = NULL;
	in_pipe_right = NULL;

	out_pipe->ShutDown();
	out_pipe = NULL;

	delete extra_cond;
	extra_cond = NULL;
	op = NULL;

	lit = NULL;

	schema_left = NULL;
	schema_right = NULL;

	atts_no = 0;
	delete [] atts;
	atts = NULL;

	domain_proc = 0;
	buffer_size = 0;

	which_index = 0;

	is_done = 0;

	delete cur_left;
	cur_left = NULL;

	delete cur_right;
	cur_right = NULL;

	pthread_mutex_destroy (&start_mutex);

//	printf("SortJoinLevel:Destructor:Out\n");
}




int SortJoinLevel::GetNextBatch (double p, Record *output, int max_output_size, int &output_size)
{
	int ret_val = 0;

	output_size = 0;

	pthread_mutex_lock(&start_mutex);

//	printf("SortJoinLevel:GetNextBatch:In\n");

	if ((domain_proc >= HASH_DOMAIN) || (p == 0.0) || (is_done == 1))
	{
//		printf("SortJoinLevel:GetNextBatch:Out\n");

		pthread_mutex_unlock(&start_mutex);

		return ret_val;
	}

	ret_val = 1;
	if (p < 0.0)
	{
		int la_no = schema_left->GetNumAtts();
		int ra_no = schema_right->GetNumAtts();


//		FILE *f1 = fopen("LEFT_hash", "at+");
//		FILE *f2 = fopen("RIGHT_hash", "at+");



		switch (which_index) {

		case 0: {
			while (ret_val != -1)
			{
				Record *r_rec = cur_right->Current(0);
				Record *l_rec = cur_left->Current(0);

//				r_rec->PrintToFile(schema_right, "RIGHT_text");
//				l_rec->PrintToFile(schema_left, "LEFT_text");

//				fprintf(f1, "%20d\n", sort_func->Compute_Hash(l_rec, om_left));
//				fprintf(f2, "%20d\n", sort_func->Compute_Hash(r_rec, om_right));


				int c_res = sort_func->Compare(l_rec, om_left, r_rec, om_right);
				if (c_res == -1)
				{
					delete cur_left;
					cur_left = new (std::nothrow) TwoWayList<Record>;
					if (cur_left == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_left->ExtractQueueTop(cur_left);
					if (ret_res == 0)
						break;
	
					cur_left->MoveToStart();
				}

				else if (c_res == 1)
				{
					delete cur_right;
					cur_right = new (std::nothrow) TwoWayList<Record>;
					if (cur_right == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_right->ExtractQueueTop(cur_right);
					if (ret_res == 0)
						break;
	
					cur_right->MoveToStart();
				}

				else if (c_res == 0)
				{
					int r_len = cur_right->RightLength();
					int l_len = cur_left->RightLength();

					if (output_size + r_len * l_len > max_output_size)
					{
						domain_proc = (l_rec->hash > 0) ? l_rec->hash - 1 : 0;

						ret_val = -1;
						break;
					}

					Record *m_rec = new (std::nothrow) Record();
					if (m_rec == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					while (cur_right->RightLength() > 0)
					{
						r_rec = cur_right->Current(0);

						while (cur_left->RightLength() > 0)
						{
							l_rec = cur_left->Current(0);
	
							if ((exact_hash->Compare(l_rec, om_left, r_rec, om_right) == 0) && (exact_hash->Compare(l_rec, r_rec, lit, extra_cond) == 1))
							{
								m_rec->MergeRecords(l_rec, r_rec, la_no, ra_no, atts, atts_no, right_start);

								m_rec->hash = l_rec->hash;
								output[output_size].Copy(m_rec);
								output_size += 1;

								m_rec->hash = -1;
								out_pipe->Insert(m_rec);
							}
	
							cur_left->Advance();
						}
		
						cur_left->MoveToStart();
				
						cur_right->Advance();
					}
					delete m_rec;


					delete cur_left;
					cur_left = new (std::nothrow) TwoWayList<Record>;
					if (cur_left == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_left->ExtractQueueTop(cur_left);
					if (ret_res == 0)
						break;
	
					cur_left->MoveToStart();


					delete cur_right;
					cur_right = new (std::nothrow) TwoWayList<Record>;
					if (cur_right == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					ret_res = order_right->ExtractQueueTop(cur_right);
					if (ret_res == 0)
						break;
	
					cur_right->MoveToStart();
				}
			}

			break;
		}

		case 1: {
			Record *m_rec = new (std::nothrow) Record();
			if (m_rec == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			while (ret_val != -1)
			{
				int r_len = cur_right->RightLength();
				while (cur_right->RightLength() > 0)
				{
					Record *r_rec = cur_right->Current(0);
					if (order_left->FindIndex(r_rec, om_right, cur_left) > 0)
					{
						cur_left->MoveToStart();

						int l_len = cur_left->RightLength();
						if (output_size + r_len * l_len > max_output_size)
						{
							domain_proc = sort_func->Compute_Hash(r_rec, om_right);
							domain_proc = (domain_proc > 0) ? domain_proc - 1 : 0;

							ret_val = -1;

							delete cur_left;
							cur_left = new (std::nothrow) TwoWayList<Record>;
							if (cur_left == NULL)
							{
								cout << "ERROR : Not enough memory. EXIT !!!\n";
								exit(1);
							}

							break;
						}

						while (cur_left->RightLength() > 0)
						{
							Record *l_rec = cur_left->Current(0);

							if (exact_hash->Compare(l_rec, r_rec, lit, extra_cond) == 1)
							{
								m_rec->MergeRecords(l_rec, r_rec, la_no, ra_no, atts, atts_no, right_start);
		
								m_rec->hash = r_rec->hash;
								output[output_size].Copy(m_rec);
								output_size += 1;
	
								m_rec->hash = -1;
								out_pipe->Insert(m_rec);
							}
	
							cur_left->Advance();
						}
	
						delete cur_left;
						cur_left = new (std::nothrow) TwoWayList<Record>;
						if (cur_left == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
					}
			
					cur_right->Advance();
				}
	
				if (ret_val != -1)
				{
					delete cur_right;
					cur_right = new (std::nothrow) TwoWayList<Record>;
					if (cur_right == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_right->ExtractQueueTop(cur_right);
					if (ret_res == 0)
						break;
	
					cur_right->MoveToStart();
				}
			}

			delete m_rec;

			break;
		}

		case 2: {
			Record *m_rec = new (std::nothrow) Record();
			if (m_rec == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			while (ret_val != -1)
			{
				int l_len = cur_left->RightLength();
				while (cur_left->RightLength() > 0)
				{
					Record *l_rec = cur_left->Current(0);
					if (order_right->FindIndex(l_rec, om_left, cur_right) > 0)
					{
						cur_right->MoveToStart();

						int r_len = cur_right->RightLength();
						if (output_size + r_len * l_len > max_output_size)
						{
							domain_proc = sort_func->Compute_Hash(l_rec, om_left);
							domain_proc = (domain_proc > 0) ? domain_proc - 1 : 0;

							ret_val = -1;

							delete cur_right;
							cur_right = new (std::nothrow) TwoWayList<Record>;
							if (cur_right == NULL)
							{
								cout << "ERROR : Not enough memory. EXIT !!!\n";
								exit(1);
							}

							break;
						}

						while (cur_right->RightLength() > 0)
						{
							Record *r_rec = cur_right->Current(0);
	
							if (exact_hash->Compare(l_rec, r_rec, lit, extra_cond) == 1)
							{
								m_rec->MergeRecords(l_rec, r_rec, la_no, ra_no, atts, atts_no, right_start);
	
								m_rec->hash = l_rec->hash;
								output[output_size].Copy(m_rec);
								output_size += 1;
	
								m_rec->hash = -1;
								out_pipe->Insert(m_rec);
							}
	
							cur_right->Advance();
						}
	
						delete cur_right;
						cur_right = new (std::nothrow) TwoWayList<Record>;
						if (cur_right == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
					}
			
					cur_left->Advance();
				}
	
				if (ret_val != -1)
				{
					delete cur_left;
					cur_left = new (std::nothrow) TwoWayList<Record>;
					if (cur_left == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_left->ExtractQueueTop(cur_left);
					if (ret_res == 0)
						break;
	
					cur_left->MoveToStart();
				}
			}

			delete m_rec;

			break;
		}

		default: {
			break;
		}

		}


		if (ret_val != -1)
		{
			domain_proc = HASH_DOMAIN;
			out_pipe->ShutDown();
		}


//		fclose(f1);
//		fclose(f2);
	}

	else
	{
		int domain_proc_now = (int)(p * HASH_DOMAIN);
		domain_proc_now = domain_proc_now > 0 ? domain_proc_now : 1;
		domain_proc += domain_proc_now;


		int la_no = schema_left->GetNumAtts();
		int ra_no = schema_right->GetNumAtts();

		switch (which_index) {

		case 0: {
			while (ret_val != -1)
			{
				Record *r_rec = cur_right->Current(0);
				Record *l_rec = cur_left->Current(0);

				unsigned int r_val = sort_func->Compute_Hash(r_rec, om_right);
				unsigned int l_val = sort_func->Compute_Hash(l_rec, om_left);
				if ((r_val > domain_proc) || (l_val > domain_proc))
					break;

				int c_res = l_val < r_val ? -1 : (l_val == r_val ? 0 : 1);
				if (c_res == -1)
				{
					delete cur_left;
					cur_left = new (std::nothrow) TwoWayList<Record>;
					if (cur_left == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_left->ExtractQueueTop(cur_left);
					if (ret_res == 0)
					{
						is_done = 1;
						break;
					}
	
					cur_left->MoveToStart();
				}

				else if (c_res == 1)
				{
					delete cur_right;
					cur_right = new (std::nothrow) TwoWayList<Record>;
					if (cur_right == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_right->ExtractQueueTop(cur_right);
					if (ret_res == 0)
					{
						is_done = 1;
						break;
					}
	
					cur_right->MoveToStart();
				}

				else if (c_res == 0)
				{
					int r_len = cur_right->RightLength();
					int l_len = cur_left->RightLength();

					if (output_size + r_len * l_len > max_output_size)
					{
						domain_proc = (l_val > 0) ? l_val - 1 : 0;

						ret_val = -1;
						break;
					}

					Record *m_rec = new (std::nothrow) Record();
					if (m_rec == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					while (cur_right->RightLength() > 0)
					{
						r_rec = cur_right->Current(0);

						while (cur_left->RightLength() > 0)
						{
							l_rec = cur_left->Current(0);
	
							if ((exact_hash->Compare(l_rec, om_left, r_rec, om_right) == 0) && (exact_hash->Compare(l_rec, r_rec, lit, extra_cond) == 1))
							{
								m_rec->MergeRecords(l_rec, r_rec, la_no, ra_no, atts, atts_no, right_start);

								m_rec->hash = l_rec->hash;
								output[output_size].Copy(m_rec);
								output_size += 1;

								m_rec->hash = -1;
								out_pipe->Insert(m_rec);
							}
	
							cur_left->Advance();
						}
		
						cur_left->MoveToStart();
				
						cur_right->Advance();
					}
					delete m_rec;

					delete cur_left;
					cur_left = new (std::nothrow) TwoWayList<Record>;
					if (cur_left == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_left->ExtractQueueTop(cur_left);
					if (ret_res == 0)
					{
						is_done = 1;
						break;
					}
	
					cur_left->MoveToStart();


					delete cur_right;
					cur_right = new (std::nothrow) TwoWayList<Record>;
					if (cur_right == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					ret_res = order_right->ExtractQueueTop(cur_right);
					if (ret_res == 0)
					{
						is_done = 1;
						break;
					}
	
					cur_right->MoveToStart();
				}
			}

			break;
		}

		case 1: {
			Record *m_rec = new (std::nothrow) Record();
			if (m_rec == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			while (ret_val != -1)
			{
				Record *rec = cur_right->Current(0);
				unsigned int r_val = sort_func->Compute_Hash(rec, om_right);
				if (r_val > domain_proc)
					break;

				int r_len = cur_right->RightLength();
				while (cur_right->RightLength() > 0)
				{
					Record *r_rec = cur_right->Current(0);
					if (order_left->FindIndex(r_rec, om_right, cur_left) > 0)
					{
						cur_left->MoveToStart();

						int l_len = cur_left->RightLength();
						if (output_size + r_len * l_len > max_output_size)
						{
							domain_proc = (r_val > 0) ? r_val - 1 : 0;

							ret_val = -1;

							delete cur_left;
							cur_left = new (std::nothrow) TwoWayList<Record>;
							if (cur_left == NULL)
							{
								cout << "ERROR : Not enough memory. EXIT !!!\n";
								exit(1);
							}

							break;
						}

						while (cur_left->RightLength() > 0)
						{
							Record *l_rec = cur_left->Current(0);

							if (exact_hash->Compare(l_rec, r_rec, lit, extra_cond) == 1)
							{
								m_rec->MergeRecords(l_rec, r_rec, la_no, ra_no, atts, atts_no, right_start);
	
								m_rec->hash = r_rec->hash;
								output[output_size].Copy(m_rec);
								output_size += 1;
	
								m_rec->hash = -1;
								out_pipe->Insert(m_rec);
							}
	
							cur_left->Advance();
						}
	
						delete cur_left;
						cur_left = new (std::nothrow) TwoWayList<Record>;
						if (cur_left == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
					}
			
					cur_right->Advance();
				}
	
				if (ret_val != -1)
				{
					delete cur_right;
					cur_right = new (std::nothrow) TwoWayList<Record>;
					if (cur_right == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_right->ExtractQueueTop(cur_right);
					if (ret_res == 0)
					{
						is_done = 1;
						break;
					}
	
					cur_right->MoveToStart();
				}
			}

			delete m_rec;

			break;
		}

		case 2: {
			Record *m_rec = new (std::nothrow) Record();
			if (m_rec == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			while (ret_val != -1)
			{
				Record *rec = cur_left->Current(0);
				unsigned int l_val = sort_func->Compute_Hash(rec, om_left);
				if (l_val > domain_proc)
					break;

				int l_len = cur_left->RightLength();
				while (cur_left->RightLength() > 0)
				{
					Record *l_rec = cur_left->Current(0);
					if (order_right->FindIndex(l_rec, om_left, cur_right) > 0)
					{
						cur_right->MoveToStart();

						int r_len = cur_right->RightLength();
						if (output_size + r_len * l_len > max_output_size)
						{
							domain_proc = (l_val > 0) ? l_val - 1 : 0;

							ret_val = -1;

							delete cur_right;
							cur_right = new (std::nothrow) TwoWayList<Record>;
							if (cur_right == NULL)
							{
								cout << "ERROR : Not enough memory. EXIT !!!\n";
								exit(1);
							}

							break;
						}

						while (cur_right->RightLength() > 0)
						{
							Record *r_rec = cur_right->Current(0);
	
							if (exact_hash->Compare(l_rec, r_rec, lit, extra_cond) == 1)
							{
								m_rec->MergeRecords(l_rec, r_rec, la_no, ra_no, atts, atts_no, right_start);
	
								m_rec->hash = l_rec->hash;
								output[output_size].Copy(m_rec);
								output_size += 1;
	
								m_rec->hash = -1;
								out_pipe->Insert(m_rec);
							}
	
							cur_right->Advance();
						}
	
						delete cur_right;
						cur_right = new (std::nothrow) TwoWayList<Record>;
						if (cur_right == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
					}
			
					cur_left->Advance();
				}
	
				if (ret_val != -1)
				{
					delete cur_left;
					cur_left = new (std::nothrow) TwoWayList<Record>;
					if (cur_left == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}

					int ret_res = order_left->ExtractQueueTop(cur_left);
					if (ret_res == 0)
					{
						is_done = 1;
						break;
					}
	
					cur_left->MoveToStart();
				}
			}

			delete m_rec;

			break;
		}

		default: {
			break;
		}

		}


		if ((domain_proc >= HASH_DOMAIN) || (is_done == 1))
		{
			domain_proc = HASH_DOMAIN;
			out_pipe->ShutDown();
		}
	}

//	printf("SortJoinLevel:GetNextBatch:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return ret_val;
}


void SortJoinLevel::StartUp ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("SortJoinLevel:StartUp:In\n");

	pthread_attr_t t_attr;
	int ret;

	ret = pthread_attr_init(&t_attr);
	ret = pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_DETACHED);
	ret = pthread_create(&thread3, &t_attr, SortJoinLevel::DoSort, (void *)this);

//	printf("SortJoinLevel:StartUp:Out\n");
}


double SortJoinLevel::HowMuchIsLeft ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("SortJoinLevel:HowMuchIsLeft:In\n");

	double res = (double)(HASH_DOMAIN - domain_proc) / (double)HASH_DOMAIN;

//	printf("SortJoinLevel:HowMuchIsLeft:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return res;
}




void* SortJoinLevel::ReadRight(void *in_right)
{
	Record *temp_rec = new (std::nothrow) Record();
	if (temp_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	SortJoinLevel *sj = (SortJoinLevel *)in_right;


	// insert records in left and right, until pipe ends
	while(sj->in_pipe_right->Remove(temp_rec))
	{
		sj->order_right->Insert(temp_rec);
	}

	delete temp_rec;
	
	return NULL;
}


void* SortJoinLevel::ReadLeft(void *in_left)
{
	Record *temp_rec = new (std::nothrow) Record();
	if (temp_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	SortJoinLevel *sj = (SortJoinLevel *)in_left;


	// insert records in left and right, until pipe ends
	while(sj->in_pipe_left->Remove(temp_rec))
	{
		sj->order_left->Insert(temp_rec);
	}
	
	delete temp_rec;

	return NULL;
}



void* SortJoinLevel :: DoSort(void *args)
{
	SortJoinLevel *sj = (SortJoinLevel *)args;

//	printf("SortJoinLevel:DoSort:In\n");


	pthread_create(&sj->thread1, NULL, SortJoinLevel::ReadLeft, (void *)sj);
	pthread_create(&sj->thread2, NULL, SortJoinLevel::ReadRight, (void *)sj);

	pthread_join(sj->thread1, NULL);
	pthread_join(sj->thread2, NULL);



	int left_runs_no = sj->order_left->GetRunsNo();
	int right_runs_no = sj->order_right->GetRunsNo();
	
	if (left_runs_no == 0)
		sj->which_index = 1;
	else if (right_runs_no == 0)
		sj->which_index = 2;


	if (sj->which_index == 1)
	{
		sj->order_left->BuildIndex();
		sj->order_right->BuildQueue();

		int res = sj->order_right->ExtractQueueTop(sj->cur_right);
		if (res == 0)
			sj->is_done = 1;

		sj->cur_right->MoveToStart();
	}

	else if (sj->which_index == 2)
	{
		sj->order_left->BuildQueue();
		sj->order_right->BuildIndex();

		int res = sj->order_left->ExtractQueueTop(sj->cur_left);
		if (res == 0)
			sj->is_done = 1;

		sj->cur_left->MoveToStart();
	}

	else if (sj->which_index == 0)
	{
		sj->order_left->BuildQueue();
		sj->order_right->BuildQueue();

		int res = sj->order_left->ExtractQueueTop(sj->cur_left);
		if (res == 0)
			sj->is_done = 1;

		sj->cur_left->MoveToStart();

		res = sj->order_right->ExtractQueueTop(sj->cur_right);
		if (res == 0)
			sj->is_done = 1;

		sj->cur_right->MoveToStart();
	}


//	printf("SortJoinLevel:DoSort:Out\n");

	pthread_mutex_unlock(&sj->start_mutex);
}




RehashLevel::RehashLevel(Pipe *in_pipe, int pipe_index, Pipe *out_pipe, Schema *schema, int buffer_size)
{
//	printf("RehashLevel:Constructor:In\n");

	this->buffer_size = buffer_size;
	this->domain_proc = 0;

	this->in_pipe = in_pipe;
	this->out_pipe = out_pipe;

	this->schema = schema;
	
	this->om = new (std::nothrow) OrderMaker(schema);
	if (this->om == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->sort_func = new (std::nothrow) CWHash(om->numAtts + 1, HASH_DOMAIN);
	if (this->sort_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	char run_name[200];
	sprintf(run_name, "run_%d", pipe_index);

	this->order = new (std::nothrow) Order(this->buffer_size, om, sort_func, this->schema, run_name);
	if (this->order == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->is_done = 0;

	this->cur_rec = new (std::nothrow) TwoWayList<Record>;
	if (this->cur_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	pthread_mutex_init (&start_mutex, NULL);

//	printf("RehashLevel:Constructor:Out\n");
}


RehashLevel::~RehashLevel()
{
//	printf("RehashLevel:Destructor:In\n");

	delete in_pipe;
	in_pipe = NULL;

	delete om;
	om = NULL;

	delete order;
	order = NULL;

	delete sort_func;
	sort_func = NULL;
	
	out_pipe->ShutDown();
	out_pipe = NULL;

	schema = NULL;

	domain_proc = 0;
	buffer_size = 0;

	is_done = 0;

	delete cur_rec;
	cur_rec = NULL;

	pthread_mutex_destroy (&start_mutex);

//	printf("RehashLevel:Destructor:Out\n");
}



int RehashLevel::GetNextBatch (double p, Record *output, int max_output_size, int &output_size)
{
	int ret_val = 0;

	output_size = 0;

	pthread_mutex_lock(&start_mutex);

//	printf("RehashLevel:GetNextBatch:In\n");

	if ((domain_proc >= HASH_DOMAIN) || (p == 0.0) || (is_done == 1))
	{
//		printf("RehashLevel:GetNextBatch:Out\n");

		pthread_mutex_unlock(&start_mutex);

		return ret_val;
	}

	ret_val = 1;
	if (p < 0.0)
	{

		while (ret_val != -1)
		{
			Record *rec = cur_rec->Current(0);

			if (rec->hash < 0 )
			{
				rec->Print(schema);
				cout << "Negative hash value!!!\n";
				exit (1);
			}

			int len = cur_rec->RightLength();
			if (output_size + len > max_output_size)
			{
				unsigned int hash_val = sort_func->Compute_Hash(rec, om);
				domain_proc = (hash_val > 0) ? hash_val - 1 : 0;

				ret_val = -1;
				break;
			}

			while (cur_rec->RightLength() > 0)
			{
				rec = cur_rec->Current(0);


				if (rec->hash < 0 )
				{
					rec->Print(schema);
					cout << "Negative hash value!!!\n";
					exit (1);
				}


				output[output_size].Copy(rec);
				output_size += 1;

				rec->hash = -1;
				out_pipe->Insert(rec);

				cur_rec->Advance();
			}

			delete cur_rec;
			cur_rec = new (std::nothrow) TwoWayList<Record>;
			if (cur_rec == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			int ret_res = order->ExtractQueueTop(cur_rec);
			if (ret_res == 0)
				break;

			cur_rec->MoveToStart();
		}

		if (ret_val != -1)
		{
			domain_proc = HASH_DOMAIN;
			out_pipe->ShutDown();
		}
	}

	else
	{
		int domain_proc_now = (int)(p * HASH_DOMAIN);
		domain_proc_now = domain_proc_now > 0 ? domain_proc_now : 1;
		domain_proc += domain_proc_now;

		while (ret_val != -1)
		{
			Record *rec = cur_rec->Current(0);

			if (rec->hash < 0 )
			{
				rec->Print(schema);
				cout << "Negative hash value!!!\n";
				exit (1);
			}

			unsigned int val = sort_func->Compute_Hash(rec, om);
			if (val > domain_proc)
				break;

			int len = cur_rec->RightLength();
			if (output_size + len > max_output_size)
			{
				domain_proc = (val > 0) ? val - 1 : 0;

				ret_val = -1;
				break;
			}

			while (cur_rec->RightLength() > 0)
			{
				rec = cur_rec->Current(0);

				if (rec->hash < 0 )
				{
					rec->Print(schema);
					cout << "Negative hash value!!!\n";
					exit (1);
				}

				output[output_size].Copy(rec);
				output_size += 1;

				rec->hash = -1;
				out_pipe->Insert(rec);

				cur_rec->Advance();
			}

			delete cur_rec;
			cur_rec = new (std::nothrow) TwoWayList<Record>;
			if (cur_rec == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			int ret_res = order->ExtractQueueTop(cur_rec);
			if (ret_res == 0)
			{
				is_done = 1;
				break;
			}

			cur_rec->MoveToStart();
		}

		if ((domain_proc >= HASH_DOMAIN) || (is_done == 1))
		{
			domain_proc = HASH_DOMAIN;
			out_pipe->ShutDown();
		}
	}

//	printf("RehashLevel:GetNextBatch:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return ret_val;
}


void RehashLevel::StartUp ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("RehashLevel:StartUp:In\n");

	pthread_attr_t t_attr;
	int ret;

	ret = pthread_attr_init(&t_attr);
	ret = pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_DETACHED);
	ret = pthread_create(&thread1, &t_attr, RehashLevel::DoSort, (void *)this);

//	printf("RehashLevel:StartUp:Out\n");
}


double RehashLevel::HowMuchIsLeft ()
{
	pthread_mutex_lock(&start_mutex);

//	printf("RehashLevel:HowMuchIsLeft:In\n");

	double res = (double)(HASH_DOMAIN - domain_proc) / (double)HASH_DOMAIN;

//	printf("RehashLevel:HowMuchIsLeft:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return res;
}



void* RehashLevel :: DoSort(void *args)
{
	RehashLevel *sj = (RehashLevel *)args;

//	printf("RehashLevel:DoSort:In\n");


	Record *temp_rec = new (std::nothrow) Record();
	if (temp_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	while(sj->in_pipe->Remove(temp_rec))
	{
		sj->order->Insert(temp_rec);
	}
	delete temp_rec;


	sj->order->BuildQueue();

	int res = sj->order->ExtractQueueTop(sj->cur_rec);
	if (res == 0)
		sj->is_done = 1;

	sj->cur_rec->MoveToStart();

//	printf("RehashLevel:DoSort:Out\n");

	pthread_mutex_unlock(&sj->start_mutex);
}







WriteOutLevel::WriteOutLevel(Pipe *pipe, int pipe_index, Schema *schema, Function *func, int *join_atts, int join_atts_no)
{
//	printf("WriteOutLevel:Constructor:In\n");

	this->pipe = pipe;
	this->schema = schema;
	this->func = func;

	tuples_no = 0.0;
	tuples_proc = 0.0;

	aggregate = 0.0;

	this->join_atts = join_atts;
	this->join_atts_no = join_atts_no;

	//STRIPE
//	f = new (std::nothrow) Stripe_File();
	f = new (std::nothrow) File();

	if (f == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	sprintf(file_name, "%d_", pipe_index);

	pthread_mutex_init (&start_mutex, NULL);

//	printf("WriteOutLevel:Constructor:Out\n");
}


WriteOutLevel::~WriteOutLevel()
{
//	printf("WriteOutLevel:Destructor:In\n");

	delete pipe;
	pipe = NULL;

	schema = NULL;
	func = NULL;

//	f->Kill();
	delete f;
	f = NULL;

	tuples_no = 0.0;
	tuples_proc = 0.0;

	aggregate = 0.0;

	join_atts = NULL;
	join_atts_no = 0;

	pthread_mutex_destroy (&start_mutex);

//	printf("WriteOutLevel:Destructor:Out\n");
}


void WriteOutLevel::StartUp()
{
	pthread_mutex_lock(&start_mutex);

//	printf("WriteOutLevel:StartUp:In\n");

	pthread_attr_t t_attr;
	int ret;

	ret = pthread_attr_init(&t_attr);
	ret = pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_DETACHED);
	ret = pthread_create(&thr, &t_attr, WriteOutLevel::DoStartUp, (void *)this);

//	printf("WriteOutLevel:StartUp:Out\n");
}


double WriteOutLevel::HowMuchIsLeft()
{
	pthread_mutex_lock(&start_mutex);

//	printf("WriteOutLevel:HowMuchIsLeft:In\n");

	double res = (double)(tuples_no - tuples_proc) / (double)tuples_no;

//	printf("WriteOutLevel:HowMuchIsLeft:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return res;
}


int WriteOutLevel::GetNextBatch (double p, Record *output, int max_output_size, int &output_size)
{
	pthread_mutex_lock(&start_mutex);

//	printf("WriteOutLevel:GetNextBatch:In\n");

/*
	if ((tuples_proc >= tuples_no) || (p == 0.0))
	{
//		printf("WriteOutLevel:GetNextBatch:Out\n");

		pthread_mutex_unlock(&start_mutex);

		return 0;
	}

	if (p < 0.0)
	{
		Record *rec = new (std::nothrow) Record();
		while (f->NextRec(rec) == 1)
		{
//			rec->Print(schema);

			int int_result;
			double double_result;

			Type t = func->Apply(rec, int_result, double_result);
			if (t == Int)
				aggregate += int_result;
			else if (t == Double)
				aggregate += double_result;

//			output->Insert(rec);

			tuples_proc += 1;
			if (tuples_proc >= tuples_no)
				break;
		}
		delete rec;
	
		tuples_proc = tuples_no;
	}

	else
	{
		int tuples_proc_now = (int)(p * tuples_no);
		tuples_proc_now = tuples_proc_now > 0 ? tuples_proc_now : 1;

		Record *rec = new (std::nothrow) Record();

		int i = 0;
		while (i < tuples_proc_now)
		{
			if (f->NextRec(rec) == 1)
			{
				int int_result;
				double double_result;
	
				Type t = func->Apply(rec, int_result, double_result);
				if (t == Int)
					aggregate += int_result;
				else if (t == Double)
					aggregate += double_result;

//				output->Insert(rec);

				i += 1;
			}
			else
			{
				break;
			}
		}
	
		delete rec;

		tuples_proc += i;
	}

//	printf("WriteOutLevel:GetNextBatch:Out\n");
*/

	pthread_mutex_unlock(&start_mutex);

	return 1;
}



void *WriteOutLevel::DoStartUp(void *args)
{
	WriteOutLevel *wo = (WriteOutLevel *)args;

//	printf("WriteOutLevel:DoStartUp:In\n");

//	File f;
//	f.Open(0, wo->file_name);

	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	while (wo->pipe->Remove(rec))
	{
//		rec->Print(wo->schema);
//		f.Append(rec);

		int int_result;
		double double_result;

		Type t = wo->func->Apply(rec, int_result, double_result);
		if (t == Int)
			wo->aggregate += int_result;
		else if (t == Double)
			wo->aggregate += double_result;

		wo->tuples_no += 1.0;
	}
	delete rec;

//	f.Close();

//	wo->f->Open(1, wo->file_name);

//	printf("WriteOutLevel:DoStartUp:Out\n");

	pthread_mutex_unlock(&wo->start_mutex);
}



double WriteOutLevel::GetTuplesNo()
{
	return tuples_no;
}


double WriteOutLevel::GetAggregate()
{
	return aggregate;
}


Schema* WriteOutLevel::GetSchema()
{
	return schema;
}


Function* WriteOutLevel::GetFunction()
{
	return func;
}


void WriteOutLevel::GetJoinAtts(int *&join_atts, int &join_atts_no)
{
	join_atts = this->join_atts;
	join_atts_no = this->join_atts_no;
}







GroupAggregateLevel::GroupAggregateLevel(Pipe *pipe, int pipe_index, Schema *schema, Function *func, int *join_atts, int join_atts_no, int *group_atts, int group_atts_no, int buffer_size)
{
//	printf("GroupAggregateLevel:Constructor:In\n");

	this->buffer_size = buffer_size;

	this->pipe = pipe;
	this->schema = schema;
	this->func = func;

	tuples_no = 0.0;
	tuples_proc = 0.0;

	this->join_atts = join_atts;
	this->join_atts_no = join_atts_no;

	this->group_atts = group_atts;
	this->group_atts_no = group_atts_no;

	//create the schema for pre-aggregation
	store_schema = new (std::nothrow) Schema();
	if (store_schema == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	store_schema->CopyFrom(this->schema);
	store_schema->Project(this->group_atts, this->group_atts_no);

	//create the OrderMaker for storing pre-aggregated tuples
	store_order = new (std::nothrow) OrderMaker(store_schema);
	if (store_order == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	//add the extra two attributes: the number of tuples in the pre-aggregated group and the function
	Attribute *extraAtts = new (std::nothrow) Attribute[2];
	if (extraAtts == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	extraAtts[0].name = strdup("func");
	extraAtts[1].name = strdup("tuples");

	extraAtts[0].myType = Double;
	extraAtts[1].myType = Double;

	store_schema->AppendAtts(extraAtts, 2);

	//delete the extraAtts structure
	free(extraAtts[0].name);
	extraAtts[0].name = NULL;
	free(extraAtts[1].name);
	extraAtts[1].name = NULL;

	delete [] extraAtts;


	group_order = new (std::nothrow) OrderMaker(schema, this->group_atts, this->group_atts_no);
	if (group_order == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->sort_func = new (std::nothrow) Hash();
	if (this->sort_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	char run_name[200];
	sprintf(run_name, "run_%d", pipe_index);

	this->order = new (std::nothrow) Order(this->buffer_size, store_order, sort_func, store_schema, run_name);
	if (this->order == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->cur_rec = new (std::nothrow) TwoWayList<Record>;
	if (this->cur_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	groups.clear();
	group_aggs.clear();
	group_tuples.clear();


	pthread_mutex_init (&start_mutex, NULL);

//	printf("GroupAggregateLevel:Constructor:Out\n");
}


GroupAggregateLevel::~GroupAggregateLevel()
{
//	printf("GroupAggregateLevel:Destructor:In\n");

	delete pipe;
	pipe = NULL;

	schema = NULL;
	func = NULL;

	tuples_no = 0.0;
	tuples_proc = 0.0;

	join_atts = NULL;
	join_atts_no = 0;

	group_atts = NULL;
	group_atts_no = 0;

	delete group_order;
	group_order = NULL;

	delete order;
	order = NULL;

	delete sort_func;
	sort_func = NULL;

	delete store_schema;
	store_schema = NULL;

	delete store_order;
	store_order = NULL;

	delete cur_rec;
	cur_rec = NULL;


	for (int i = 0; i < groups.size(); i++)
	{
		Record *rec = groups[i];
		delete rec;
		rec = NULL;

		groups[i] = NULL;
	}

	groups.clear();
	group_aggs.clear();
	group_tuples.clear();

	pthread_mutex_destroy (&start_mutex);

//	printf("GroupAggregateLevel:Destructor:Out\n");
}


void GroupAggregateLevel::StartUp()
{
	pthread_mutex_lock(&start_mutex);

//	printf("GroupAggregateLevel:StartUp:In\n");

	pthread_attr_t t_attr;
	int ret;

	ret = pthread_attr_init(&t_attr);
	ret = pthread_attr_setdetachstate(&t_attr, PTHREAD_CREATE_DETACHED);
	ret = pthread_create(&thr, &t_attr, GroupAggregateLevel::DoStartUp, (void *)this);

//	printf("GroupAggregateLevel:StartUp:Out\n");
}


double GroupAggregateLevel::HowMuchIsLeft()
{
	pthread_mutex_lock(&start_mutex);

//	printf("GroupAggregateLevel:HowMuchIsLeft:In\n");

	double res = (double)(tuples_no - tuples_proc) / (double)tuples_no;

//	printf("GroupAggregateLevel:HowMuchIsLeft:Out\n");

	pthread_mutex_unlock(&start_mutex);

	return res;
}


int GroupAggregateLevel::GetNextBatch (double p, Record *output, int max_output_size, int &output_size)
{
	pthread_mutex_lock(&start_mutex);

	pthread_mutex_unlock(&start_mutex);

	return 1;
}



void *GroupAggregateLevel::DoStartUp(void *args)
{
	GroupAggregateLevel *wo = (GroupAggregateLevel *)args;

//	printf("GroupAggregateLevel:DoStartUp:In\n");

	//structures needed for hashing
	Record **colaps_array = new (std::nothrow) Record*[wo->buffer_size];
	if (colaps_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *colaps_func = new (std::nothrow) double[wo->buffer_size];
	if (colaps_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *colaps_tuples = new (std::nothrow) double[wo->buffer_size];
	if (colaps_tuples == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < wo->buffer_size; i++)
	{
		colaps_array[i] = NULL;
		colaps_func[i] = 0.0;
		colaps_tuples[i] = 0.0;
	}

	//the hash function
	Hash *hash_func = new (std::nothrow) CWHash(wo->store_order->numAtts + 1, wo->buffer_size);
	if (hash_func == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}


	int colaps_count = 0;

	//extract the records from the pipe and hash them
	while(wo->pipe->Remove(rec))
	{
		double aggregate;

		//first compute the aggregate function
		int int_result;
		double double_result;

		Type t = wo->func->Apply(rec, int_result, double_result);
		if (t == Int)
			aggregate = int_result;
		else if (t == Double)
			aggregate = double_result;

		//project out the non-grouping attributes
		rec->Project(wo->group_atts, wo->group_atts_no, wo->schema->GetNumAtts());

		//compute the position in the hash table
		unsigned int index = hash_func->Compute_Hash(rec, wo->store_order);

		//check to see if there is space in the hash table
		if (colaps_array[index] == NULL)
		{
			//in this case the position is free, so just add everything in the hash tables
			colaps_array[index] = new (std::nothrow) Record();
			if (colaps_array[index] == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
			colaps_array[index]->Consume(rec);

			colaps_func[index] += aggregate;
			colaps_tuples[index] += 1;

			colaps_count += 1;
		}
		else
		{
			//in this case there already exists a record at the current position

			//first verify that the two records are indeed identical
			int are_identical = wo->sort_func->Compare(rec, colaps_array[index], wo->store_order);
			if (are_identical == 0)
			{
				//the records are identical, so collapse the aggregate values
				colaps_func[index] += aggregate;
				colaps_tuples[index] += 1;
			}
			else
			{
				//the records are different; use linear hashing and find the next free position where to insert the current record
				int j;
				for (j = (index + 1) % wo->buffer_size; (colaps_array[j] != NULL) && (j != index); j = (j + 1) % wo->buffer_size)
				{
					//verify equality with each of these since it is possible that this is not the first record with these values that met a conflict
					are_identical = wo->sort_func->Compare(rec, colaps_array[j], wo->store_order);
					if (are_identical == 0)
						break;
				}


				if (are_identical == 0)
				{
					//we found a record with the same values that was previously rehashed at a different position
					//the records are identical, so collapse the aggregate values
					colaps_func[j] += aggregate;
					colaps_tuples[j] += 1;
				}
				else
				{
					if (colaps_array[j] == NULL)
					{
						//in this case the position is free, so just add everything in the hash tables
						colaps_array[j] = new (std::nothrow) Record();
						if (colaps_array[j] == NULL)
						{
							cout << "ERROR : Not enough memory. EXIT !!!\n";
							exit(1);
						}
						colaps_array[j]->Consume(rec);
			
						colaps_func[j] += aggregate;
						colaps_tuples[j] += 1;
			
						colaps_count += 1;
					}
					else
					{
						cout << "This should be an error. No place was found in the hash array for record: " << colaps_count << "\n";
						exit(1);
					}
				}
			}
		}

		//at this point we inserted our stuff into the hash tables
		//check to see if the hash tables are full
		//if they are full, it is neccessary to flush them on the disk
		if (colaps_count == wo->buffer_size)
		{
			for (int i = 0; i < colaps_count; i++)
			{
				if (colaps_array[i] == NULL)
				{
					cout << "ERROR : the HASH should be full and actually it is not !!!\n";
					exit(1);
				}

				//for each slot in the hash tables create the extended record that is stored
				//first add a record with the function
				char *first = new (std::nothrow) char[(sizeof(int) << 1) + sizeof(double)];
				if (first == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
			
				((int*)first)[0] = (sizeof(int) << 1) + sizeof(double);
				((int*)first)[1] = sizeof(int) << 1;
				memcpy(first + (sizeof(int) << 1), &colaps_func[i], sizeof(double));

				Record *append_rec = new (std::nothrow) Record();
				if (append_rec == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				rec->Consume(first);
				append_rec->AppendRecords(colaps_array[i], rec, wo->group_atts_no, 1);

				//second add a record with the number of tuples
				first = new (std::nothrow) char[(sizeof(int) << 1) + sizeof(double)];
				if (first == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
			
				((int*)first)[0] = (sizeof(int) << 1) + sizeof(double);
				((int*)first)[1] = sizeof(int) << 1;
				memcpy(first + (sizeof(int) << 1), &colaps_tuples[i], sizeof(double));

				rec->Consume(first);
				colaps_array[i]->AppendRecords(append_rec, rec, wo->group_atts_no + 1, 1);

				//and now insert into the SMOrder for sorting and writing into a run
				wo->order->Insert(colaps_array[i]);

				//clean the hash tables structures
				delete colaps_array[i];
				colaps_array[i] = NULL;

				colaps_func[i] = 0.0;
				colaps_tuples[i] = 0.0;

				delete append_rec;
			}

			colaps_count = 0;
		}
	}



	//we have to do this one more time for sure (we need at least one run)
	for (int i = 0; i < wo->buffer_size; i++)
	{
		if (colaps_array[i] == NULL)
		{
			//this position in the hash table was not occupied
			continue;
		}

		//for each slot in the hash tables create the extended record that is stored
		//first add a record with the function
		char *first = new (std::nothrow) char[(sizeof(int) << 1) + sizeof(double)];
		if (first == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	
		((int*)first)[0] = (sizeof(int) << 1) + sizeof(double);
		((int*)first)[1] = sizeof(int) << 1;
		memcpy(first + (sizeof(int) << 1), &colaps_func[i], sizeof(double));

		Record *append_rec = new (std::nothrow) Record();
		if (append_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		rec->Consume(first);
		append_rec->AppendRecords(colaps_array[i], rec, wo->group_atts_no, 1);

		//second add a record with the number of tuples
		first = new (std::nothrow) char[(sizeof(int) << 1) + sizeof(double)];
		if (first == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	
		((int*)first)[0] = (sizeof(int) << 1) + sizeof(double);
		((int*)first)[1] = sizeof(int) << 1;
		memcpy(first + (sizeof(int) << 1), &colaps_tuples[i], sizeof(double));

		rec->Consume(first);
		colaps_array[i]->AppendRecords(append_rec, rec, wo->group_atts_no + 1, 1);

		//and now insert into the SMOrder for sorting and writing into a run
		wo->order->Insert(colaps_array[i]);

		//clean the hash tables structures
		delete colaps_array[i];
		colaps_array[i] = NULL;

		colaps_func[i] = 0.0;
		colaps_tuples[i] = 0.0;

		delete append_rec;
	}

	colaps_count = 0;

	delete rec;
	rec = NULL;

	//clean the space used by the hashing structures
	delete [] colaps_array;
	delete [] colaps_func;
	delete [] colaps_tuples;

	delete hash_func;


	//build the priority queue for extracting the pre-aggregated records in order
	wo->order->BuildQueue();


	while (wo->order->ExtractQueueTop(wo->cur_rec) > 0)
	{
		wo->cur_rec->MoveToStart();

		//build the grouping attributes for the pre-aggregated records
		int *group_atts = new (std::nothrow) int[wo->group_atts_no];
		if (group_atts == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
		for (int i = 0; i < wo->group_atts_no; i++)
		{
			group_atts[i] = i;
		}


		//add the data for the description of the group
		rec = wo->cur_rec->Current(0);

		Record *group_rec = new (std::nothrow) Record();
		if (group_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
		group_rec->Copy(rec);
		group_rec->Project(group_atts, wo->group_atts_no, wo->store_schema->GetNumAtts());

		wo->groups.push_back(group_rec);

		delete [] group_atts;


		//do the final aggregation from the pre-aggregated parts
		double aggregate = 0.0;
		double group_tuples_no = 0.0;

		while (wo->cur_rec->RightLength() > 0)
		{
			rec = wo->cur_rec->Current(0);

			char *func = rec->GetColumn(wo->group_atts_no);
			aggregate += *(double*)func;

			char *tuples = rec->GetColumn(wo->group_atts_no + 1);
			wo->tuples_no += *(double*)tuples;
			group_tuples_no += *(double*)tuples;

			wo->cur_rec->Advance();
		}

		//add the data for the group aggregate and the number of tuples in the group
		wo->group_aggs.push_back(aggregate);
		wo->group_tuples.push_back(group_tuples_no);


		delete wo->cur_rec;
		wo->cur_rec = new (std::nothrow) TwoWayList<Record>;
		if (wo->cur_rec == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}
	}


//	printf("GroupAggregateLevel:DoStartUp:Out\n");

	pthread_mutex_unlock(&wo->start_mutex);
}



double GroupAggregateLevel::GetTuplesNo()
{
	return tuples_no;
}


Schema* GroupAggregateLevel::GetSchema()
{
	return schema;
}


Function* GroupAggregateLevel::GetFunction()
{
	return func;
}


void GroupAggregateLevel::GetJoinAtts(int *&join_atts, int &join_atts_no)
{
	join_atts = this->join_atts;
	join_atts_no = this->join_atts_no;
}

void GroupAggregateLevel::GetGroupAtts(int *&group_atts, int &group_atts_no)
{
	group_atts = this->group_atts;
	group_atts_no = this->group_atts_no;
}


void GroupAggregateLevel::GetGroups(vector<Record*> &desc, vector<double> &aggregate, vector<double> &tuples, int &groups_no)
{
	desc = groups;
	aggregate = group_aggs;
	tuples = group_tuples;

	groups_no = groups.size();
}





// table scan C-STORE thread function
// reads all the records from the given column file and puts them into the corresponding pipe
void *TableScan_Thread (void *param) {
	// remember all of our params
	File_Interface *f = ((ThreadTableScanParam*) param)->f;
	Pipe *out_pipe = ((ThreadTableScanParam*) param)->out_pipe;

	//read all the records from the file and insert them into the pipe
	Record *in_rec = new (std::nothrow) Record();
	if (in_rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	while (f->NextRec(in_rec))
	{
		out_pipe->Insert(in_rec);
	}

	delete in_rec;

	out_pipe->ShutDown();
	f->Close();
}


