#include "SMOrder.h"
#include "File.h"

#include <string.h>
#include <stdlib.h>
#include <iostream>


Order::Order(int buffer_size, OrderMaker *order, Hash *sort_func, Schema *schema, char *run_name)
{
	this->buffer_size = buffer_size;
	buffer = new (std::nothrow) SortRecord[buffer_size];
	if (buffer == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	this->recs_no = 0;
	this->cur_index = 0;

	this->run_name = strdup(run_name);
	this->runs_no = 0;

	this->order = order;
	this->sort_func = sort_func;

	this->schema = schema;

	for (int i = 0; i < buffer_size; i++) {
		buffer[i].record = new (std::nothrow) Record();
		if (buffer[i].record == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		buffer[i].order = this->order;
		buffer[i].hash = this->sort_func;
	}

	this->queue = NULL;
	this->index = NULL;

	this->run_file = NULL;
}

Order::~Order()
{
	for (int i = 0; i < buffer_size; i++) {
		delete buffer[i].record;
		buffer[i].record = NULL;
		buffer[i].order = NULL;
		buffer[i].hash = NULL;
	}

	delete [] buffer;
	buffer = NULL;

	free(run_name);
	run_name = NULL;

	sort_func = NULL;

	schema = NULL;

	delete queue;
	queue = NULL;

	delete index;
	index = NULL;

	delete run_file;
	run_file = NULL;
}


int Order::GetRecsNo()
{
	return recs_no;
}

int Order::GetRunsNo()
{
	return runs_no;
}


int Order::Insert(Record *rec)
{
	recs_no += 1;
/*
	rec->hash = sort_func->Compute_Hash(rec, order);
	if (rec->hash < 0)
	{
		cout << "ORDER: Negative hash value!!!\n";
		exit (1);
	}
*/

	if (cur_index < buffer_size)
	{
		//insert into current run
		buffer[cur_index].record->Consume(rec);
		cur_index += 1;

		return 0;
	}

	else
	{
		//sort run, write run to disk, and create a new run
		SortRun();

		//this is the first run; so we need to create the file for the runs and to open it
		if (runs_no == 0)
		{
			//STRIPE
//			run_file = new (std::nothrow) Run_Stripe_File();
			run_file = new (std::nothrow) Run_File();

			if (run_file == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

			char f_name[200];
			sprintf(f_name, "%s", run_name);

			run_file->Open(0, f_name);
		}

		WriteRun();

		buffer[0].record->Consume(rec);
		cur_index = 1;

		runs_no += 1;

		return 1;
	}
	
}



void Order::SortRun()
{
	qsort(buffer, cur_index, sizeof(SortRecord), Compare);
}



void Order::WriteRun()
{
/*
	char f_name[200];
	sprintf(f_name, "%s%d", run_name, runs_no);


//	char f_text[200];
//	sprintf(f_text, "%s%d_text", run_name, runs_no);

//	char f_hash[200];
//	sprintf(f_hash, "%s%d_hash", run_name, runs_no);

//	FILE *f = fopen(f_hash, "wt");


	File new_file;
	new_file.Open(0, f_name);

	for(int i = 0; i < cur_index; i++) {
//		buffer[i].record->PrintToFile(schema, f_text);
//		fprintf(f, "%20d\n", sort_func->Compute_Hash(buffer[i].record, order));


		new_file.Append(buffer[i].record);
	}

	new_file.Close();


//	fclose(f);
*/

	run_file->Append_Run(buffer, cur_index);
}



void Order::BuildQueue()
{
	//write the last run
	SortRun();

	if (runs_no == 0)
	{
		//STRIPE
//		run_file = new (std::nothrow) Run_Stripe_File();
		run_file = new (std::nothrow) Run_File();

		if (run_file == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		char f_name[200];
		sprintf(f_name, "%s", run_name);

		run_file->Open(0, f_name);
	}

	WriteRun();

	run_file->Close(1);


	cur_index = 0;
	runs_no += 1;

	queue = new (std::nothrow) PriorityQueue(run_name, runs_no, order, sort_func, schema, buffer_size);
	if (queue == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}



void Order::BuildIndex()
{
	Record *rec_array = new (std::nothrow) Record[cur_index];
	if (rec_array == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < cur_index; i++)
	{
		rec_array[i].Consume(buffer[i].record);
	}

	index = new (std::nothrow) MMIndex(order, cur_index);
	if (index == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	index->BuildIndex(rec_array, cur_index);

	delete [] rec_array;
}


int Order::ExtractQueueTop(TwoWayList<Record> *top_list)
{
	int records_no = 0;

	Record *rec = new (std::nothrow) Record();
	if (rec == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	if (queue->GetNext(rec) == 1) {
		Record *temp = new (std::nothrow) Record();
		if (temp == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		temp->Copy(rec);

		top_list->Insert(rec);
		records_no += 1;

		while (queue->CompareToTop(temp) == 0) {
			queue->GetNext(rec);
			top_list->Insert(rec);
			records_no += 1;
		}

		delete temp;
	}
	delete rec;

	return records_no;
}


int Order::GetNextRec(Record *rec)
{
	int res = 0;

	if (queue->GetNext(rec) == 1) {
		res = 1;
	}

	return res;
}



int Order::FindIndex(Record *rec, OrderMaker *rec_order, TwoWayList<Record> *match_list)
{
	int res = index->Find(rec, rec_order, match_list);
	return res;
}




int Compare(const void *r1, const void *r2)
{
	SortRecord *a, *b;

	a = (SortRecord*)r1;
	b = (SortRecord*)r2;

	int res = a->hash->Compare(a->record, b->record, a->order);
	return res;
}


