#include "Schema.h"

#include <string.h>
#include <stdio.h>
#include <iostream.h>


Schema :: Schema() {
	numAtts = 0;
	myAtts = NULL;
	
	fileName = NULL;

	recs_no = 0;
}

Schema :: Schema (char *fName, char *relName) {
	FILE *foo = fopen (fName, "r");
		
	// this is enough space to hold any tokens
	char space[200];

	fscanf (foo, "%s", space);
	int totscans = 1;

	// see if the file starts with the correct keyword
	if (strcmp (space, "BEGIN")) {
		cout << "Unfortunately, this does not seem to be a schema file.\n";
		exit (1);
	}	
		
	while (1) {

		// check to see if this is the one we want
		fscanf (foo, "%s", space);
		totscans++;
		if (strcmp (space, relName)) {

			// it is not, so suck up everything to past the BEGIN
			while (1) {

				// suck up another token
				if (fscanf (foo, "%s", space) == EOF) {
					cerr << "Could not find the schema for the specified relation.\n";
					exit (1);
				}

				totscans++;
				if (!strcmp (space, "BEGIN")) {
					break;
				}
			}

		// otherwise, got the correct file!!
		} else {
			break;
		}
	}

	// suck in the file name
	fscanf (foo, "%s", space);
	totscans++;
	fileName = strdup (space);

	// suck in the number of records
	fscanf (foo, "%s", space);
	totscans++;
	recs_no = atoi(space);


	// count the number of attributes specified
	numAtts = 0;
	while (1) {
		fscanf (foo, "%s", space);
		if (!strcmp (space, "END")) {
			break;		
		} else {
			fscanf (foo, "%s", space);
			numAtts++;
		}
	}

	// now actually load up the schema
	fclose (foo);
	foo = fopen (fName, "r");

	// go past any un-needed info
	for (int i = 0; i < totscans; i++) {
		fscanf (foo, "%s", space);
	}

	// and load up the schema
	myAtts = new (std::nothrow) Attribute[numAtts];
	if (myAtts == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < numAtts; i++ ) {

		// read in the attribute name
		fscanf (foo, "%s", space);	
		myAtts[i].name = strdup (space);

		// read in the attribute type
		fscanf (foo, "%s", space);
		if (!strcmp (space, "Int")) {
			myAtts[i].myType = Int;
		} else if (!strcmp (space, "Double")) {
			myAtts[i].myType = Double;
		} else if (!strcmp (space, "String")) {
			myAtts[i].myType = String;
		} else {
			cout << "Bad attribute type for " << myAtts[i].name << "\n";
			exit (1);
		}
	}
	
	fclose (foo);
}

Schema :: ~Schema () {
	free(fileName);

	for(int i=0;i<numAtts;i++)
	{
		free(myAtts[i].name);
		myAtts[i].name = NULL;
	}

	delete [] myAtts;
	myAtts = NULL;
}


int Schema :: GetNumAtts() {
	return numAtts;
}

Attribute* Schema :: GetAtts() {
	return myAtts;
}

//Append the list of attributes to the current attributes in the schema
int Schema :: AppendAtts(Attribute *new_atts, int new_atts_no) {
	int new_numAtts = numAtts + new_atts_no;

	Attribute *myNewAtts = new (std::nothrow) Attribute[new_numAtts];
	if (myNewAtts == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	//our attributes go first
	for(int i = 0; i < numAtts; i++)
	{
		myNewAtts[i].name = strdup(myAtts[i].name);
		myNewAtts[i].myType = myAtts[i].myType;

		free(myAtts[i].name);
		myAtts[i].name = NULL;
	}

	//the extra ones go at the end
	for(int i = 0; i < new_atts_no; i++)
	{
		myNewAtts[i + numAtts].name = strdup(new_atts[i].name);
		myNewAtts[i + numAtts].myType = new_atts[i].myType;
	}

	delete [] myAtts;
	myAtts = myNewAtts;

	numAtts = new_numAtts;

	return numAtts;
}



int Schema :: GetRecsNo() {
	return recs_no;
}

char* Schema :: GetName() {
	return fileName;
}


//this copies the content in fromMe to current Schema
//return -1 if fromMe has no attributes, the current
//schema is not changed in this case
//return 0 otherwise
int Schema :: CopyFrom(Schema *fromMe) {
	if(fromMe->GetNumAtts()==0) 
	{
		cerr<<"Schema fromMe is empty\n";
		return -1;
	}

	//free memory if necessary
	for(int i=0;i<numAtts;i++)
	{
		free(myAtts[i].name);
	}
	
	delete [] myAtts;

	free(fileName);
	fileName = NULL;


	//get number of attributes
	numAtts= fromMe->GetNumAtts();

	//prepare space
	myAtts = new (std::nothrow) Attribute[numAtts];
	if (myAtts == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Attribute* fromAtt=fromMe->GetAtts();
	for(int i=0;i<numAtts;i++)
	{
		myAtts[i].name = strdup(fromAtt[i].name);
		myAtts[i].myType = fromAtt[i].myType;
	}

	recs_no = fromMe->GetRecsNo();

	if(fromMe->fileName!=NULL)
		fileName = strdup(fromMe->fileName);

	return 0;
}

//this copy of CopyFrom copies from two schemas instead of one, the second is attached
//at the end of the first
// return -1 and does not change the schema if both are empty, 0 otherwise
int Schema :: CopyFrom(Schema *first, Schema *second) {
	//if both schemas are empty, return -1;
	if(first->GetNumAtts()==0 && second->GetNumAtts()==0)
	{
		cerr<<"At least one of the schemas is empty\n";
		return -1;
	}

	//free memory if necessary
	for(int i=0;i<numAtts;i++)
	{
		free(myAtts[i].name);
	}
	
	delete [] myAtts;

	free(fileName);
	fileName = NULL;


	int num1=first->GetNumAtts(), num2=second->GetNumAtts();
	numAtts = num1+num2;
	myAtts=new (std::nothrow) Attribute[numAtts];
	if (myAtts == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	Attribute* Att1 = first->GetAtts();
	Attribute* Att2 = second->GetAtts();

	//from first
	for(int i=0;i<num1;i++)
	{
		myAtts[i].name=strdup(Att1[i].name);
		myAtts[i].myType=Att1[i].myType;
	}

	// from second
	for(int i=0;i<num2;i++)
	{
		myAtts[i+num1].name=strdup(Att2[i].name);
		myAtts[i+num1].myType=Att2[i].myType;
	}

	return 0;
}


//Rename an Attribute, return -1 if the attribute is not there, 0 otherwise
int Schema :: RenameAtt(char *oldName, char *newName) {
	int attPos = Find(oldName);
	if(attPos==-1) 
		return -1;

	free(myAtts[attPos].name);
	myAtts[attPos].name=strdup(newName);

	return 0;
}


//Project the attributes of a schema
//only those attributes indexed in the array are left after projection
//index begins from 0.
//return -1 if failure 0 otherwise
//add code to deal with the Probability Atts
int Schema:: Project(int *attsToKeep, int numAttsToKeep) {
	if(numAttsToKeep>numAtts)
	{
		cerr<<"Number of attributes to project is larger than the total number of atts.\n";
		return -1;
	}

	// store those projected atts in a tempory space
	Attribute * tempAttr = new (std::nothrow) Attribute[numAttsToKeep];
	if (tempAttr == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for(int i=0; i<numAttsToKeep;i++)
	{
		int index=attsToKeep[i];
		if(index>=numAtts)
		{
			for(int j=0;j<i;j++)
				free(tempAttr[j].name);
		
			delete [] tempAttr;

			cerr<<"Index out of range\n";
			return -1;
		}

		tempAttr[i].name = strdup(myAtts[index].name);
		tempAttr[i].myType = myAtts[index].myType;
	}

	//free original memory
	for(int i=0;i<numAtts;i++)
	{
		free(myAtts[i].name);
		myAtts[i].name=NULL;
	}

	delete [] myAtts;

	//set to new position
	myAtts=tempAttr;
	numAtts=numAttsToKeep;

	return 0;
}


int Schema :: Find (char *attName) {
	for (int i = 0; i < numAtts; i++) {
		if (!strcmp (attName, myAtts[i].name)) {
			return i;
		}
	}

	// if we made it here, the attribute was not found
	return -1;
}

Type Schema :: FindType (char *attName) {
	for (int i = 0; i < numAtts; i++) {
		if (!strcmp (attName, myAtts[i].name)) {
			return myAtts[i].myType;
		}
	}

	// if we made it here, the attribute was not found
	return Int;
}



//print the schema
void Schema:: Print() {
	if (numAtts<=0)
		cout<<"This is an empty schema\n";
	else
	{
		if (fileName != NULL)
		{
			cout << "Filename = " << fileName << "\n";
			cout << "Number of records = " << recs_no << "\n";
		}

		for(int i=0;i<numAtts;i++)
		{
			cout<<myAtts[i].name << '\t';

			switch(myAtts[i].myType)
			{
				case Int:
					cout << "INT";
					break;
				case Double:
					cout << "DOUBLE";
					break;
				case String:
					cout << "STRING";
					break;
				default:
					cout << "UNKNOWN";
					break;
			}

			cout << "\n";
		}
	}
}


