#ifndef SCHEMA_H
#define SCHEMA_H

#include "DEFS.h"


/*
Attribute that can be part of a schema
- name of the attribute
- type of the attribute
*/
struct Attribute {
	char *name;
	Type myType;
};



/*
Manage schema of a relation
- materialized on disk
- intermediary result during query execution
*/
class Schema {
private:
	// gives the physical location of the binary file storing the relation
	char *fileName;

	// gives the attributes in the schema: array with numAtts elements
	int numAtts;
	Attribute *myAtts;

	// gives the number of records in the relation
	// this is known only for relations materialized on disk
	int recs_no;

public:
	Schema ();

	// this reads the specification for the schema from a file
	// fName is the name of the file storing all the schemas
	// relName is the name of the relation for which the schema is needed
	Schema (char *fName, char *relName);

	virtual ~Schema ();

	int GetNumAtts();
	Attribute* GetAtts();

	//Append Attributes
	int AppendAtts(Attribute *new_atts, int new_atts_no);


	int GetRecsNo();

	char* GetName();

	//this copies the content in fromMe to current Schema
	//return -1 if fromMe has no attributes, the current schema is not changed in this case
	//return 0 otherwise
	int CopyFrom(Schema *fromMe);

	//this copies from two schemas instead of one, the second is attached
	//at the end of the first
	// return -1 and does not change the schema if both are empty, 0 otherwise
	int CopyFrom(Schema *first, Schema *second);

	//Rename an Attribute
	int RenameAtt(char* oldName, char* newName);

	//Project the attributes of a schema
	//only those attributes indexed in the passed array are left after projection
	//index begins from 0.
	//return -1 if failure 0 otherwise
	int Project(int *attsToKeep, int numAttsToKeep);

	// this finds the position of the specified attribute in the schema
	// returns a -1 if the attribute is not present in the schema
	int Find(char *attName);

	// this finds the type of the given attribute
	// returns arbitrary type if the passed attribute is not present
	// this method should be called only after Find method is called
	Type FindType(char *attName);

	//This print out the Schema()
	void Print();

};

#endif
