#ifndef _STAT_STRUCT_H_
#define _STAT_STRUCT_H_


#include "Hashes.h"

#include <vector>


using namespace std;


int Compare_Tuples(const void *r1, const void *r2);


class TupleRepository {
public:
	int relations_no;
	int tuples_no;

	int *tuples;
	double *func;


	TupleRepository(int relations_no, int tuples_no, int **tuples, double *func);
	virtual ~TupleRepository();

	void Sort(int order);
	double ComputeOrderCoef();
};



//class that stores a factor in a variance term
class VarianceFactor {
public:
	//the size of the factor (how many relations)
	int n;

	//the actual tuple ids
	int *v;

	//the value of the function for this factor
	double func;


	//creates the actual factor
	//1 in mask says that the relation is in
	//n is the size of the vector v and of the bits in mask
	VarianceFactor(int mask, int *v, int n);

	virtual ~VarianceFactor();

	//compare current factor with a given one
	//returns 0 for equal and 1 for different
	int Compare(VarianceFactor *f);
};



//class that stores a term in the variance
class VarianceTerm {
public:
	//the set of the term
	int mask;

	//the size of the term (and factor)
	int n;

	//the estimate for the term
	double estimate;

	//hash function
	Hash *hash_func;

	//hash repository
	vector<VarianceFactor*> *repository;
	int repository_size;


	VarianceTerm(int mask, int n, int repository_size);
	virtual ~VarianceTerm();


	//adds a tuple for this term
	void AddTuple(int *tuple, int n, double func);

	//removes an existing tuple from this term
	void RemoveTuple(int *tuple, int n, double func);

	double GetEstimate();
};


#endif

