#include "SamplingEstimate.h"
#include "StatisticsLevel.h"

#include <vector>
#include <iostream>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>


using namespace std;


StatisticsLevel::StatisticsLevel(int relations_no)
{
	this->relations_no = relations_no;
}


StatisticsLevel::~StatisticsLevel()
{
	relations_no = 0;
}



double StatisticsLevel::GetExpectation(double p)
{
	return 0.0;
}


double StatisticsLevel::GetVariance(double p)
{
	return 0.0;
}


double StatisticsLevel::GetExpectation()
{
	return 0.0;
}


double StatisticsLevel::GetVariance()
{
	return 0.0;
}



int StatisticsLevel::ProcessTuples(int **tuples, double *funcVals, int numTuples, SimpleSamplingData *data)
{
	return 0;
}

int StatisticsLevel::ProcessTuples(double *funcVals, int numTuples, SimpleSamplingData *data)
{
	return 0;
}

int StatisticsLevel::ProcessTuples(SearchResults *newResults, double p)
{
	return 0;
}


double StatisticsLevel::WhatIfVariance(double *proposedSamplingRatios)
{
	return 0.0;
}


void StatisticsLevel::FreezeCoefficients()
{
}


void StatisticsLevel::Print()
{
}




StatisticsLowLevel::StatisticsLowLevel(int relations_no, int var_estimates_needed)
: StatisticsLevel(relations_no)
{
	this->var_estimates_needed = var_estimates_needed;

	variance_estimates = 0;

	estimates_no = 0;

	e = new (std::nothrow) vector<double>;
	if (e == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	var = new (std::nothrow) vector<double>;
	if (var == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	sampling_params = new (std::nothrow) vector<SimpleSamplingData*>;
	if (sampling_params == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	sampling_estimator = new (std::nothrow) SamplingEstimate(relations_no);
	if (sampling_estimator == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
}


StatisticsLowLevel::~StatisticsLowLevel()
{
	e->clear();
	delete e;
	e = NULL;

	var->clear();
	delete var;
	var = NULL;

	for (int i = 0; i < estimates_no; i++)
	{
		delete (*sampling_params)[i];
		(*sampling_params)[i] = NULL;
	}
	sampling_params->clear();
	delete sampling_params;
	sampling_params = NULL;

	delete sampling_estimator;
	sampling_estimator = NULL;

	estimates_no = 0;

	variance_estimates = 0;
}



double StatisticsLowLevel::GetExpectation()
{
	int n = estimates_no;


	double expectation;
	double variance;


	int ret = Solve_Optimization(e, var, n, expectation, variance);
	if (ret == 0)
	{
		expectation = 0.0;
		for (int i = 0; i < n; i++)
		{
			expectation += (*e)[i];
		}
		expectation = expectation / n;
	}


/*
	for (int i = 0; i < n; i++)
	{
		expectation += (*e)[i];
	}
	expectation = expectation / n;
*/

	return expectation;
}


double StatisticsLowLevel::GetVariance()
{
	int n = estimates_no;


	double expectation;
	double variance;


	int ret = Solve_Optimization(e, var, n, expectation, variance);
	if (ret == 0)
	{
		variance = 0.0;
		for (int i = 0; i < n; i++)
		{
			variance += (*var)[i];
		}
		variance = variance / n;
	}


/*
	for (int i = 0; i < n; i++)
	{
		variance += (*var)[i];
	}
	variance = variance / n;
*/

	return variance;
}


int StatisticsLowLevel::ProcessTuples(int **tuples, double *funcVals, int numTuples, SimpleSamplingData *data)
{
	double aux_e = 0.0;
	double aux_v = 0.0;

	SimpleSamplingData *aux_d = new (std::nothrow) SimpleSamplingData(*data);
	if (aux_d == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	if (numTuples > 0)
	{
		//generate more statistics only if we do not have enough estimates for the variance
		if (variance_estimates < var_estimates_needed)
		{
			sampling_estimator->GenerateStatistics(numTuples, tuples, funcVals);
			sampling_estimator->ComputeUnbiasedCoefs(data);

/*
			for (int i = 0; i < estimates_no; i++)
			{
				//recompute the variance only for those blocks that returned tuples
				if (((*e)[i] != 0.0) || ((*var)[i] != 0.0))
				{
					(*var)[i] = sampling_estimator->ComputeVariance((*sampling_params)[i]);
				}
			}
*/

			variance_estimates += 1;
		}
	
		aux_e = sampling_estimator->ComputeExpectation(data, numTuples, funcVals);
		aux_v = sampling_estimator->ComputeVariance(data);
	}


	e->push_back(aux_e);
	var->push_back(aux_v);
	sampling_params->push_back(aux_d);

	estimates_no += 1;

	return 0;
}



void StatisticsLowLevel::Print()
{
	for (int i = 0; i < estimates_no; i++)
	{
		cout << "E = " << (*e)[i] << "\n";
		cout << "VAR = " << (*var)[i] << "\n";
		cout << "Parameters:\n";
		(*sampling_params)[i]->Print();
	}

	cout << "Estimator:\n";
	sampling_estimator->PrintCoefficients();
}





StatisticsUpLevel::StatisticsUpLevel(int relations_no)
: StatisticsLevel(relations_no)
{
	f = 0.0;
	f2 = 0.0;

	proc_data = NULL;
}


StatisticsUpLevel::~StatisticsUpLevel()
{
	f = 0.0;
	f2 = 0.0;

	delete proc_data;
	proc_data = NULL;
}


double StatisticsUpLevel::GetExpectation()
{
//	printf("E %10.5f : %10.5f %30.4f\n", proc_data->e[0], 1.0 / proc_data->e[0], f);


	double res = (1.0 / proc_data->e[0]) * f;
	return res;
}


double StatisticsUpLevel::GetVariance()
{
//	printf("V %10.5f : %10.5f %30.4f\n", proc_data->e[0], (1.0 / proc_data->e[0]) * (1.0 / proc_data->e[0] - 1.0), f2);

	double res = (1.0 / proc_data->e[0]) * (1.0 / proc_data->e[0] - 1.0) * f2;

/*
	if (res < 0.0)
	{
		cout << "Variance is negative. EXIT !!!\n";
		exit(1);
	}
*/

	return res;
}


int StatisticsUpLevel::ProcessTuples(double *funcVals, int numTuples, SimpleSamplingData *data)
{
	for (int i = 0; i < numTuples; i++)
	{
		f += funcVals[i];
		f2 += (funcVals[i] * funcVals[i]);
	}

	delete proc_data;
	proc_data = new SimpleSamplingData(*data);

	return 0;
}


void StatisticsUpLevel::Print()
{
}





StatisticsGeneralLevel::StatisticsGeneralLevel(int relations_no, int tuples_stable)
: StatisticsLevel(relations_no)
{
	sampling_estimator = new (std::nothrow) GeneralizedSamplingEstimate(relations_no, tuples_stable);
	if (sampling_estimator == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	bernoulli_p = new (std::nothrow) double[relations_no];
	if (bernoulli_p == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < relations_no; i++)
		bernoulli_p[i] = 1.0;

	threshold = 0.0;

	constant_no = 0;
	constant_mask = 0;

	this->tuples_stable = tuples_stable;
}


StatisticsGeneralLevel::~StatisticsGeneralLevel()
{
	delete sampling_estimator;
	sampling_estimator = NULL;

	delete bernoulli_p;
	bernoulli_p = NULL;
}



double StatisticsGeneralLevel::GetExpectation(double p)
{
	threshold = p;

	GeneralizedSamplingData sampling_data(relations_no, constant_no, constant_mask, threshold, bernoulli_p);
//	sampling_data.Print();

	double e = sampling_estimator->ComputeExpectation(&sampling_data);
	return e;
}


double StatisticsGeneralLevel::GetVariance(double p)
{
	threshold = p;

	GeneralizedSamplingData sampling_data(relations_no, constant_no, constant_mask, threshold, bernoulli_p);

	double var = sampling_estimator->ComputeVariance(&sampling_data);

/*
	if (var < 0.0)
	{
		cout << "Variance is negative. EXIT !!!\n";
		exit(1);
	}
*/

	return var;
}


int StatisticsGeneralLevel::ProcessTuples(SearchResults *newResults, double p)
{
	threshold = p;

	memcpy(bernoulli_p, newResults->ratios, sizeof(double) * relations_no);

	constant_no = 0;
	constant_mask = 0;

	for (int i = 0; i < relations_no; i++)
	{
		if (newResults->constRels[i] == 1)
		{
			constant_no += 1;
			constant_mask = constant_mask | (1 << i);
		}
	}

	GeneralizedSamplingData sampling_data(relations_no, constant_no, constant_mask, threshold, bernoulli_p);

	sampling_estimator->UpdateStatistics(newResults->numRes, newResults->resTuples, newResults->funcRes, newResults->numDel, newResults->delTuples, newResults->funcDel, &sampling_data);

	return 0;
}


double StatisticsGeneralLevel::WhatIfVariance(double *proposedSamplingRatios)
{
	GeneralizedSamplingData sampling_data(relations_no, constant_no, constant_mask, threshold, proposedSamplingRatios);

	double var = sampling_estimator->ComputeVariance(&sampling_data);

/*
	if (var < 0.0)
	{
		cout << "Variance is negative. EXIT !!!\n";
		exit(1);
	}
*/


	cout << "WhatIf ???\n";
	for (int i = 0; i < relations_no; i++)
	{
		cout << proposedSamplingRatios[i] << "\t";
	}
	cout << "\n";

	cout << var << "\n";


	return var;
}


void StatisticsGeneralLevel::FreezeCoefficients()
{
	sampling_estimator->FreezeCoefficients();
}


void StatisticsGeneralLevel::Print()
{
}




StatisticsGeneralLevelCov::StatisticsGeneralLevelCov(int relations_no, int tuples_stable)
: StatisticsLevel(relations_no)
{
	sampling_estimator = new (std::nothrow) GeneralizedSamplingEstimate(relations_no, tuples_stable);
	if (sampling_estimator == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	vector<double> bernoulli;
	bernoulli.clear();
	for (int i = 0; i < relations_no; i++)
	{
		bernoulli.push_back(1.0);
	}

	bernoulli_p.clear();
	bernoulli_p.push_back(bernoulli);

	threshold.clear();
	threshold.push_back(0.0);

	expectation.clear();
	expectation.push_back(0.0);

	vector<double> cov;
	cov.clear();
	cov.push_back(0.0);

	cov_table.clear();
	cov_table.push_back(cov);

	constant_no = 0;
	constant_mask = 0;

	this->tuples_stable = tuples_stable;
}


StatisticsGeneralLevelCov::~StatisticsGeneralLevelCov()
{
	delete sampling_estimator;
	sampling_estimator = NULL;

	threshold.clear();

	int n = bernoulli_p.size();
	for (int i = 0; i < n; i++)
	{
		bernoulli_p[i].clear();
	}
	bernoulli_p.clear();

	expectation.clear();

	n = cov_table.size();
	for (int i = 0; i < n; i++)
	{
		cov_table[i].clear();
	}
	cov_table.clear();
}



double StatisticsGeneralLevelCov::GetExpectation(double p)
{
	int last = threshold.size();
	threshold[last - 1] = p;

	GeneralizedSamplingData *sampling_data = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[last - 1], &bernoulli_p[last - 1]);
	if (sampling_data == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	//get the last expectation estimate for the current region
	last = expectation.size();
	double e = sampling_estimator->ComputeExpectation(sampling_data);
	expectation[last - 1] = e;

	delete sampling_data;

	//compute the updated covariance row
	last = cov_table.size();
	for (int i = 0; i < last; i++)
	{
		for (int j = 0; j <= i; j++)
		{
/*
			if (i == j)
			{
				GeneralizedSamplingData *sampling_data = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[i], &bernoulli_p[i]);
				if (sampling_data == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				//get the last expectation estimate for the current region
				double var = sampling_estimator->ComputeVariance(sampling_data);
//				printf("Variance = %30.10f\n", var);
				if (var > 0)
				{
					sampling_data->Print();
					printf("Rel No = %10d\n", relations_no);
					printf("Const No = %10d\n", constant_no);
					printf("Const Mask = %10d\n", constant_mask);
				}
			
				cov_table[i][j] = var;

				delete sampling_data;
			}
			else
			{
				GeneralizedSamplingDataCov *cov_data = new (std::nothrow) GeneralizedSamplingDataCov(relations_no, constant_no, constant_mask, threshold[j], threshold[i], &bernoulli_p[j], &bernoulli_p[i]);
				if (cov_data == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
		
//				cov_data->Print();
	
				cov_table[i][j] = sampling_estimator->ComputeCovariance(cov_data);
		
				delete cov_data;
			}
*/

			GeneralizedSamplingDataCov *cov_data = new (std::nothrow) GeneralizedSamplingDataCov(relations_no, constant_no, constant_mask, threshold[j], threshold[i], &bernoulli_p[j], &bernoulli_p[i]);
			if (cov_data == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}

/*	
			printf("Rel No = %10d\n", relations_no);
			printf("Const No = %10d\n", constant_no);
			printf("Const Mask = %10d\n", constant_mask);
			cov_data->Print();
*/

			cov_table[i][j] = sampling_estimator->ComputeCovariance(cov_data);
	
			delete cov_data;
		}
	}

//	Print();

	double opt_e, opt_var;
	Solve_Optimization_Cov(&expectation, &cov_table, last, opt_e, opt_var);

	return opt_e;
}


double StatisticsGeneralLevelCov::GetVariance(double p)
{
	int last = threshold.size();
	threshold[last - 1] = p;

	GeneralizedSamplingData *sampling_data = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[last - 1], &bernoulli_p[last - 1]);
	if (sampling_data == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	//get the last expectation estimate for the current region
	last = expectation.size();
	double e = sampling_estimator->ComputeExpectation(sampling_data);
	expectation[last - 1] = e;

	delete sampling_data;

	//compute the updated covariance row
	last = cov_table.size();
	for (int i = 0; i < last; i++)
	{
		for (int j = 0; j <= i; j++)
		{
/*
			if (i == j)
			{
				GeneralizedSamplingData *sampling_data = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[i], &bernoulli_p[i]);
				if (sampling_data == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}

				//get the last expectation estimate for the current region
				double var = sampling_estimator->ComputeVariance(sampling_data);
//				printf("Variance = %30.10f\n", var);
			
				cov_table[i][j] = var;

				delete sampling_data;
			}
			else
			{
				GeneralizedSamplingDataCov *cov_data = new (std::nothrow) GeneralizedSamplingDataCov(relations_no, constant_no, constant_mask, threshold[j], threshold[i], &bernoulli_p[j], &bernoulli_p[i]);
				if (cov_data == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
		
//				cov_data->Print();
	
				cov_table[i][j] = sampling_estimator->ComputeCovariance(cov_data);
		
				delete cov_data;
			}
*/

			GeneralizedSamplingDataCov *cov_data = new (std::nothrow) GeneralizedSamplingDataCov(relations_no, constant_no, constant_mask, threshold[j], threshold[i], &bernoulli_p[j], &bernoulli_p[i]);
			if (cov_data == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
	
//			cov_data->Print();

			cov_table[i][j] = sampling_estimator->ComputeCovariance(cov_data);
	
			delete cov_data;
		}
	}

//	Print();

	double opt_e, opt_var;
	Solve_Optimization_Cov(&expectation, &cov_table, last, opt_e, opt_var);

/*
	if (opt_var < 0.0)
	{
		cout << "Variance is negative. EXIT !!!\n";
		exit(1);
	}
*/

	return opt_var;
}


int StatisticsGeneralLevelCov::ProcessTuples(SearchResults *newResults, double p)
{
	constant_no = 0;
	constant_mask = 0;

	for (int i = 0; i < relations_no; i++)
	{
		if (newResults->constRels[i] == 1)
		{
			constant_no += 1;
			constant_mask = constant_mask | (1 << i);
		}
	}


	if (newResults->numDel > 0)
	{
		//tuples are deleted -- special action required
		int last = threshold.size();
		threshold[last - 1] = p;

		GeneralizedSamplingData *sampling_data = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[last - 1], &bernoulli_p[last - 1]);
		if (sampling_data == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}


		//get the last expectation estimate for the current region
		last = expectation.size();
		double e = sampling_estimator->ComputeExpectation(sampling_data);
		expectation[last - 1] = e;
		expectation.push_back(e);

		delete sampling_data;

		//compute the row in the covariance matrix
		last = cov_table.size();
		for (int i = 0; i < last; i++)
		{
			for (int j = 0; j <= i; j++)
			{
/*
				if (i == j)
				{
					GeneralizedSamplingData *sampling_data = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[i], &bernoulli_p[i]);
					if (sampling_data == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
	
					//get the last expectation estimate for the current region
					double var = sampling_estimator->ComputeVariance(sampling_data);
	//				printf("Variance = %30.10f\n", var);
				
					cov_table[i][j] = var;
	
					delete sampling_data;
				}
				else
				{
					GeneralizedSamplingDataCov *cov_data = new (std::nothrow) GeneralizedSamplingDataCov(relations_no, constant_no, constant_mask, threshold[j], threshold[i], &bernoulli_p[j], &bernoulli_p[i]);
					if (cov_data == NULL)
					{
						cout << "ERROR : Not enough memory. EXIT !!!\n";
						exit(1);
					}
			
	//				cov_data->Print();
		
					cov_table[i][j] = sampling_estimator->ComputeCovariance(cov_data);
			
					delete cov_data;
				}
*/

				GeneralizedSamplingDataCov *cov_data = new (std::nothrow) GeneralizedSamplingDataCov(relations_no, constant_no, constant_mask, threshold[j], threshold[i], &bernoulli_p[j], &bernoulli_p[i]);
				if (cov_data == NULL)
				{
					cout << "ERROR : Not enough memory. EXIT !!!\n";
					exit(1);
				}
		
	//			cov_data->Print();
	
				cov_table[i][j] = sampling_estimator->ComputeCovariance(cov_data);
		
				delete cov_data;

			}
		}

//		Print();

		//add the new parameters into the corresponding vectors
		threshold.push_back(p);

		vector<double> bernoulli;
		bernoulli.clear();
		for (int i = 0; i < relations_no; i++)
		{
			bernoulli.push_back(newResults->ratios[i]);
		}
		bernoulli_p.push_back(bernoulli);


		last = threshold.size();

		vector<double> cov;
		cov.clear();

		int nc = cov_table.size();
		for (int i = 0; i <= nc; i++)
		{
/*
			GeneralizedSamplingDataCov *cov_data = new (std::nothrow) GeneralizedSamplingDataCov(relations_no, constant_no, constant_mask, threshold[i], threshold[last - 1], &bernoulli_p[i], &bernoulli_p[last - 1]);
			if (cov_data == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
	
			cov.push_back(sampling_estimator->ComputeCovariance(cov_data));

			delete cov_data;
*/
			cov.push_back(0.0);
		}

		cov_table.push_back(cov);

//		Print();
	}
	else
	{
		int last = threshold.size();
		threshold[last - 1] = p;

		for (int i = 0; i < relations_no; i++)
		{
			bernoulli_p[last - 1][i] = newResults->ratios[i];
		}
	}

	int last = threshold.size();
	GeneralizedSamplingData *sdata = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[last - 1], &bernoulli_p[last - 1]);
	if (sdata == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	sampling_estimator->UpdateStatistics(newResults->numRes, newResults->resTuples, newResults->funcRes, newResults->numDel, newResults->delTuples, newResults->funcDel, sdata);

	delete sdata;

	return 0;
}


double StatisticsGeneralLevelCov::WhatIfVariance(double *proposedSamplingRatios)
{
/*
	int last  = threshold.size();

	GeneralizedSamplingData *sampling_data = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[last - 1], proposedSamplingRatios);
	if (sampling_data == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double var = sampling_estimator->ComputeVariance(sampling_data);

	delete sampling_data;
*/


	int last = threshold.size();

	GeneralizedSamplingData *sampling_data = new (std::nothrow) GeneralizedSamplingData(relations_no, constant_no, constant_mask, threshold[last - 1], proposedSamplingRatios);
	if (sampling_data == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	//get the last expectation estimate for the current region
	last = expectation.size();
	double e = sampling_estimator->ComputeExpectation(sampling_data);
	expectation[last - 1] = e;

	delete sampling_data;


	last = bernoulli_p.size();

	vector<double> old_bernoulli;
	old_bernoulli.clear();
	for (int i = 0; i < relations_no; i++)
	{
		old_bernoulli.push_back(bernoulli_p[last - 1][i]);
		bernoulli_p[last - 1][i] = proposedSamplingRatios[i];
	}


	//compute the updated covariance row
	last = cov_table.size();
	for (int i = 0; i < last; i++)
	{
		for (int j = 0; j <= i; j++)
		{
			GeneralizedSamplingDataCov *cov_data = new (std::nothrow) GeneralizedSamplingDataCov(relations_no, constant_no, constant_mask, threshold[j], threshold[i], &bernoulli_p[j], &bernoulli_p[i]);
			if (cov_data == NULL)
			{
				cout << "ERROR : Not enough memory. EXIT !!!\n";
				exit(1);
			}
	
			cov_table[i][j] = sampling_estimator->ComputeCovariance(cov_data);
	
			delete cov_data;
		}
	}

//	Print();

	double opt_e, opt_var;
	Solve_Optimization_Cov(&expectation, &cov_table, last, opt_e, opt_var);


	last = bernoulli_p.size();
	for (int i = 0; i < relations_no; i++)
	{
		bernoulli_p[last - 1][i] = old_bernoulli[i];
	}

	old_bernoulli.clear();



/*
	if (var < 0.0)
	{
		cout << "Variance is negative. EXIT !!!\n";
		exit(1);
	}
*/


	cout << "WhatIf ???\n";
	for (int i = 0; i < relations_no; i++)
	{
		cout << proposedSamplingRatios[i] << "\t";
	}
	cout << "\n";

	cout << opt_var << "\n";


	return opt_var;
}


void StatisticsGeneralLevelCov::FreezeCoefficients()
{
	sampling_estimator->FreezeCoefficients();
}


void StatisticsGeneralLevelCov::Print()
{
	int n = threshold.size();

	cout << "Threshold\n";
	for (int i = 0; i < n; i++)
	{
		cout << threshold[i] << "\t";
	}
	cout << "\n";

	cout << "Bernoulli\n";
	for (int i = 0; i < n; i++)
	{
		for (int j = 0; j < relations_no; j++)
		{
			cout << bernoulli_p[i][j] << "\t";
		}
		cout << "\n";
	}
	cout << "\n";

	cout << "Expectation\n";
	for (int i = 0; i < n; i++)
	{
		cout << expectation[i] << "\t";
	}
	cout << "\n";

	cout << "Covariance Table\n";
	for (int i = 0; i < n; i++)
	{
		for (int j = 0; j <= i; j++)
		{
			cout << cov_table[i][j] << "\t";
		}
		cout << "\n";
	}
	cout << "\n";
}




StatisticsGeneralLevelNoDiscard::StatisticsGeneralLevelNoDiscard(int relations_no, int tuples_stable)
: StatisticsLevel(relations_no)
{
	sampling_estimator = new (std::nothrow) GeneralizedSamplingEstimateNoDiscard(relations_no, tuples_stable);
	if (sampling_estimator == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	bernoulli_p = new (std::nothrow) vector<double>*[relations_no];
	if (bernoulli_p == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < relations_no; i++)
	{
		bernoulli_p[i] = new (std::nothrow) vector<double>;
		if (bernoulli_p[i] == NULL)
		{
			cout << "ERROR : Not enough memory. EXIT !!!\n";
			exit(1);
		}

		bernoulli_p[i]->clear();
		bernoulli_p[i]->push_back(1.0);
	}

	threshold = new (std::nothrow) vector<double>;
	if (threshold == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}
	threshold->clear();
	threshold->push_back(0.0);

	constant_no = 0;
	constant_mask = 0;

	this->tuples_stable = tuples_stable;
}


StatisticsGeneralLevelNoDiscard::~StatisticsGeneralLevelNoDiscard()
{
	delete sampling_estimator;
	sampling_estimator = NULL;

	threshold->clear();
	delete threshold;

	for (int i = 0; i < relations_no; i++)
	{
		bernoulli_p[i]->clear();

		delete bernoulli_p[i];
		bernoulli_p[i] = NULL;
	}

	delete bernoulli_p;
	bernoulli_p = NULL;
}



double StatisticsGeneralLevelNoDiscard::GetExpectation(double p)
{
	int last = threshold->size() - 1;
	(*threshold)[last] = p;

	GeneralizedSamplingDataNoDiscard sampling_data(relations_no, constant_no, constant_mask, threshold, bernoulli_p);
//	sampling_data.Print();

	double e = sampling_estimator->ComputeExpectation(&sampling_data);
	return e;
}


double StatisticsGeneralLevelNoDiscard::GetVariance(double p)
{
	int last = threshold->size() - 1;
	(*threshold)[last] = p;

	GeneralizedSamplingDataNoDiscard sampling_data(relations_no, constant_no, constant_mask, threshold, bernoulli_p);

	double var = sampling_estimator->ComputeVariance(&sampling_data);

/*
	if (var < 0.0)
	{
		cout << "Variance is negative. EXIT !!!\n";
		exit(1);
	}
*/

	return var;
}


int StatisticsGeneralLevelNoDiscard::ProcessTuples(SearchResults *newResults, double p)
{
	if (newResults->numDel > 0)
	{
		int last = threshold->size() - 1;
		(*threshold)[last] = p;

		threshold->push_back(p);

		for (int i = 0; i < relations_no; i++)
		{
			bernoulli_p[i]->push_back(newResults->ratios[i]);
		}
	}
	else if (newResults->numDel == 0)
	{
		int last = threshold->size() - 1;
		(*threshold)[last] = p;
	}

	constant_no = 0;
	constant_mask = 0;

	for (int i = 0; i < relations_no; i++)
	{
		if (newResults->constRels[i] == 1)
		{
			constant_no += 1;
			constant_mask = constant_mask | (1 << i);
		}
	}

	GeneralizedSamplingDataNoDiscard sampling_data(relations_no, constant_no, constant_mask, threshold, bernoulli_p);

	sampling_estimator->UpdateStatistics(newResults->numRes, newResults->resTuples, newResults->funcRes, newResults->numDel, newResults->delTuples, newResults->funcDel, &sampling_data);

	return 0;
}


double StatisticsGeneralLevelNoDiscard::WhatIfVariance(double *proposedSamplingRatios)
{
	int last = threshold->size() - 1;
	threshold->push_back((*threshold)[last]);

	for (int i = 0; i < relations_no; i++)
	{
		bernoulli_p[i]->push_back(proposedSamplingRatios[i]);
	}


	GeneralizedSamplingDataNoDiscard sampling_data(relations_no, constant_no, constant_mask, threshold, bernoulli_p);

	double var = sampling_estimator->ComputeVariance(&sampling_data);

/*
	if (var < 0.0)
	{
		cout << "Variance is negative. EXIT !!!\n";
		exit(1);
	}
*/


	cout << "WhatIf ???\n";
	for (int i = 0; i < relations_no; i++)
	{
		cout << proposedSamplingRatios[i] << "\t";
	}
	cout << "\n";

	cout << var << "\n";


	//undo the changes in the two vectors
	threshold->pop_back();

	for (int i = 0; i < relations_no; i++)
	{
		bernoulli_p[i]->pop_back();
	}
	

	return var;
}


void StatisticsGeneralLevelNoDiscard::FreezeCoefficients()
{
	sampling_estimator->FreezeCoefficients();
}


void StatisticsGeneralLevelNoDiscard::Print()
{
}




int Solve_Optimization(vector<double> *e, vector<double> *var, int n, double &opt_e, double &opt_var)
{
	opt_e = 0.0;
	opt_var = 0.0;

	for (int i = 0; i < n; i++)
	{
		if ((*var)[i] > 0.0)
		{
			opt_e += (*e)[i] / (*var)[i];
			opt_var += 1.0 / (*var)[i];
		}
		else
		{
			//call off optimization if any of the variances is not nice
			return 0;
		}
	}

	if (opt_var != 0.0)
	{
		opt_e = opt_e / opt_var;
		opt_var = 1.0 / opt_var;

		return 1;
	}

	return 0;
}




int Solve_Optimization_Cov(vector<double> *e, vector<vector<double> > *cov, int n, double &opt_e, double &opt_var)
{

	int nn = n;

	double *alpha = new (std::nothrow) double[n];
	if (alpha == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	double *alphaM = new (std::nothrow) double[n];
	if (alphaM == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	int *bitset = new (std::nothrow) int[n];
	if (bitset == NULL)
	{
		cout << "ERROR : Not enough memory. EXIT !!!\n";
		exit(1);
	}

	for (int i = 0; i < n; i++)
	{
		bitset[i] = 1;
	}

	gsl_set_error_handler_off();

	while (1)
	{
		if (nn == 1)
		{
			for (int i = 0; i < n; i++)
			{
				if (bitset[i] == 0)
				{
					alpha[i] = 0.0;
				}
				else
				{
					alpha[i] = 1.0;
				}
			}

			break;
		}

		gsl_matrix *A = gsl_matrix_alloc(nn, nn);
		gsl_vector *b = gsl_vector_alloc(nn);
		gsl_vector *x = gsl_vector_alloc(nn);

		int k1 = 0;
		for (int i = 0; i < n; i++)
		{
			//exclude the values set to 0 from the optimization
			if (bitset[i] == 0)
				continue;

			int k2 = 0;
			for (int j = 0; j <= i; j++)
			{
				//exclude the values set to 0 from the optimization
				if (bitset[j] == 0)
					continue;

				gsl_matrix_set(A, k1, k2, (*cov)[i][j]);
				gsl_matrix_set(A, k2, k1, (*cov)[i][j]);

				k2 += 1;
			}
	
			gsl_vector_set(b, k1, 1.0);

			k1 += 1;
		}


//		cout << "A" << endl;
//		gsl_matrix_fprintf(stdout, A, "%f");
//		cout << "b" << endl;
//		gsl_vector_fprintf(stdout, b, "%f");


		int err_code = gsl_linalg_cholesky_decomp(A);
		if (err_code == GSL_EDOM)
		{
			int s;
	
			gsl_permutation *p = gsl_permutation_alloc(nn);
			gsl_linalg_LU_decomp (A, p, &s);
			
			gsl_linalg_LU_solve (A, p, b, x);
	
			gsl_permutation_free (p);
		}
		else
		{
			gsl_linalg_cholesky_solve(A, b, x);
		}


//		cout << "x" << endl;
//		gsl_vector_fprintf(stdout, x, "%f");


		int nnn = nn;
		for (int i = 0; i < nn; i++)
		{
			double xx = gsl_vector_get(x, i);
			if (xx < 0.0)
			{
				bitset[i] = 0;
				nn -= 1;
			}
		}

		if ((nnn == nn) || (nn == 0))
		{
			//in this case everything is fine
			//exit
			int k2 = 0;
			for (int i = 0; i < n; i++)
			{
				if (bitset[i] == 0)
				{
					alpha[i] = 0.0;
				}
				else
				{
					alpha[i] = gsl_vector_get(x, k2);
					k2 += 1;
				}
			}

			gsl_vector_free(x);
			gsl_vector_free(b);
			gsl_matrix_free(A);

			break;
		}

		//some of the solutions are positive, some are negative
		//keep on looping and solving
		gsl_vector_free(x);
		gsl_vector_free(b);
		gsl_matrix_free(A);
	}

	//normalize the solutions alpha such that sum(alpha(i)) = 1
	double s_alpha = 0.0;
	for (int i = 0; i < n; i++)
	{
		s_alpha += alpha[i];

//		cout << "alpha[" << i << "] = " << alpha[i] << endl;
	}

	for (int i = 0; i < n; i++)
	{
		alpha[i] = alpha[i] / s_alpha;

//		cout << "alpha[" << i << "] = " << alpha[i] << endl;
	}

	//compute the optimal expectation
	opt_e = 0.0;
	for (int i = 0; i < n; i++)
	{
		opt_e += alpha[i] * (*e)[i];
	}

	//compute the optimal variance
	for (int i = 0; i < n; i++)
	{
		alphaM[i] = 0.0;
		for (int j = 0; j < n; j++)
		{
			if (j < i)
			{
				alphaM[i] += alpha[j] * (*cov)[i][j];
			}
			else
			{
				alphaM[i] += alpha[j] * (*cov)[j][i];
			}
		}
	}

	opt_var = 0.0;
	for (int i = 0; i < n; i++)
	{
		opt_var += alphaM[i] * alpha[i];
	}

	delete [] alphaM;
	delete [] alpha;
	delete [] bitset;


/*
	//this solution avoids the problems with the numeric stability
	//it returns the minimum variance available and the average of expected values

	opt_e = 0.0;

	int k = 0;
	for (int i = 0; i < n; i++)
	{
		if ((*e)[i] > 0)
		{
			opt_e += (*e)[i];
			k += 1;
		}
	}

	opt_e = opt_e / k;


	opt_var = (*cov)[0][0];
	for (int i = 1; i < n; i++)
	{
		if ((*cov)[i][i] < opt_var)
		{
			opt_var = (*cov)[i][i];
		}
	}
*/

	return 1;
}

