#ifndef _STAT_LEVEL_H_
#define _STAT_LEVEL_H_


#include "SamplingEstimate.h"
#include "DEFS.h"

#include <vector>


using namespace std;


int Solve_Optimization(vector<double> *e, vector<double> *var, int n, double &opt_e, double &opt_var);
int Solve_Optimization_Cov(vector<double> *e, vector<vector<double> > *cov, int n, double &opt_e, double &opt_var);


class StatisticsLevel {
public:
	int relations_no;


	StatisticsLevel(int relations_no);
	virtual ~StatisticsLevel();

	virtual double GetExpectation(double p);
	virtual double GetVariance(double p);

	virtual double GetExpectation();
	virtual double GetVariance();

	virtual int ProcessTuples(int **tuples, double *funcVals, int numTuples, SimpleSamplingData *data);
	virtual int ProcessTuples(double *funcVals, int numTuples, SimpleSamplingData *data);
	virtual int ProcessTuples(SearchResults *newResults, double p);

	//needed in the IMJ for determining the dropping
	virtual double WhatIfVariance(double *proposedSamplingRatios);

	virtual void FreezeCoefficients();

	virtual void Print();
};



class StatisticsLowLevel : public StatisticsLevel {
public:
	int variance_estimates;
	int var_estimates_needed;

	int estimates_no;

	vector<double> *e;
	vector<double> *var;

	vector<SimpleSamplingData*> *sampling_params;

	SamplingEstimate *sampling_estimator;


	StatisticsLowLevel(int relations_no, int var_estimates_needed);
	virtual ~StatisticsLowLevel();

	virtual double GetExpectation();
	virtual double GetVariance();

	virtual int ProcessTuples(int **tuples, double *funcVals, int numTuples, SimpleSamplingData *data);

	virtual void Print();
};



class StatisticsUpLevel : public StatisticsLevel {
public:
	double f;
	double f2;

	SimpleSamplingData *proc_data;


	StatisticsUpLevel(int relations_no);
	virtual ~StatisticsUpLevel();

	virtual double GetExpectation();
	virtual double GetVariance();

	virtual int ProcessTuples(double *funcVals, int numTuples, SimpleSamplingData *data);

	virtual void Print();
};




class StatisticsGeneralLevel : public StatisticsLevel {
public:
	GeneralizedSamplingEstimate *sampling_estimator;

	double *bernoulli_p;
	double threshold;

	int constant_no;
	int constant_mask;

	int tuples_stable;


	StatisticsGeneralLevel(int relations_no, int tuples_stable);
	virtual ~StatisticsGeneralLevel();


	virtual double GetExpectation(double p);
	virtual double GetVariance(double p);


	virtual int ProcessTuples(SearchResults *newResults, double p);

	//needed in the IMJ for determining the dropping
	virtual double WhatIfVariance(double *proposedSamplingRatios);

	virtual void FreezeCoefficients();

	virtual void Print();
};



class StatisticsGeneralLevelCov : public StatisticsLevel {
public:
	GeneralizedSamplingEstimate *sampling_estimator;

	vector<double> threshold;
	vector<vector<double> > bernoulli_p;

	vector<double> expectation;
	vector<vector<double> > cov_table;

	int constant_no;
	int constant_mask;

	int tuples_stable;


	StatisticsGeneralLevelCov(int relations_no, int tuples_stable);
	virtual ~StatisticsGeneralLevelCov();


	virtual double GetExpectation(double p);
	virtual double GetVariance(double p);


	virtual int ProcessTuples(SearchResults *newResults, double p);

	//needed in the IMJ for determining the dropping
	virtual double WhatIfVariance(double *proposedSamplingRatios);

	virtual void FreezeCoefficients();

	virtual void Print();
};




class StatisticsGeneralLevelNoDiscard : public StatisticsLevel {
public:
	GeneralizedSamplingEstimateNoDiscard *sampling_estimator;

	vector<double> **bernoulli_p;
	vector<double> *threshold;

	int constant_no;
	int constant_mask;

	int tuples_stable;


	StatisticsGeneralLevelNoDiscard(int relations_no, int tuples_stable);
	virtual ~StatisticsGeneralLevelNoDiscard();


	virtual double GetExpectation(double p);
	virtual double GetVariance(double p);


	virtual int ProcessTuples(SearchResults *newResults, double p);

	//needed in the IMJ for determining the dropping
	virtual double WhatIfVariance(double *proposedSamplingRatios);

	virtual void FreezeCoefficients();

	virtual void Print();
};


#endif

